/*
 * fsResource.c
 * FileSystem resource for teams, wraps the unix-jthreads2 fs object.
 *
 * Copyright (c) 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

#include "assert_pp.h"
#include "tags.h"
#include "memResource.h"
#include "fsResource.h"
#include "jteam.h"
#include "module.h"
#include "moduleList.h"
#include "logResource.h"
#include "soft.h"
#include "nativeClass.h"
#include "nativeClass-inline.h"

#if defined(UNIX_JTHREADS2)
#include "jthread.h"
#include "jthreadState.h"
#include "jthreadScheduler.h"

#include "edu_utah_janosvm_resources_FileSystemSpec.h"
#include "edu_utah_janosvm_resources_FileSystemResource.h"

#define JAVA_FS_SPEC "edu/utah/janosvm/resources/FileSystemSpec"

/** Module declarations **/
static int initFSResourceModule(int stage);

module_t fsResourceModule = {
	"fs",
	0,
	initFSResourceModule,
	0,
	0,
	0,
};

/** Resource declarations **/
static struct _resourceSpec *fsCreateJSISpec(struct _resource *res,
						jsiMessage *jm);
static void fsDeleteJSISpec(struct _resource *res,
			       struct _resourceSpec *rs);
static int fsSubscribe(struct _resource *res,
			  struct _resourceClient *rc,
			  struct _resourceSpec *rs);
static void fsUnsubscribe(struct _resource *res,
			     struct _resourceClient *rc);
static char *fsHandleJSIQuery(struct _resourceClient *rc,
				 jsiMessage *jm);

static resourceFunctions fsResourceFunctions = {
	0,
	0,
	0,
	0,
	fsCreateJSISpec,
	fsDeleteJSISpec,
	0,
	fsSubscribe,
	0,
	fsUnsubscribe,
	fsHandleJSIQuery,
};

/** fsResource variables **/

fsResource metalFSResource;

static char *realRootDirectory = "/";

fsResourceClient metalFSResourceClient = {
	{
		0,
		0,
		&metalFSResource.link,
		0
	},
};

int initFSResourceModule(int stage)
{
	int retval = 0;
	jsked *js;
	
	switch( stage )
	{
	case SID_FSRESOURCEMODULE_INITMETAL:
		installResource(&metalFSResource.link,
				RA_Name, "fs",
				RA_Flags, RF_INTERNAL,
				RA_Functions, &fsResourceFunctions,
				RA_ClientSize, sizeof(fsResourceClient),
				RA_SpecSize, sizeof(fsResourceSpec),
				RA_TAG_DONE);
		metalTeam->internalResources[metalFSResource.link.
					    internalIndex] =
			&metalFSResourceClient.client;
		metalFSResourceClient.client.owner = metalTeam;
		metalFSResourceClient.rootDirectory = realRootDirectory;
		retval = 1;
		break;
	case SID_FSRESOURCEMODULE_REINITMETAL:
		js = jskedCurrent();
		metalFSResourceClient.js = js;
		jioSetRootDirectory(&js->js_IORequestPort, realRootDirectory);
		retval = 1;
		break;
	default:
		installResource(&metalFSResource.link,
				RA_SpecNativeClass, JAVA_FS_SPEC,
				RA_TAG_DONE);
		retval = 1;
		break;
	}
	return( retval );
}

static
int changeFSSpec(fsResourceClient *frc, char *new_root)
{
	int retval = 0;
	int iLockRoot;

	require(frc != 0);
	require(new_root != 0);

	if( (new_root[0] == '\0') || (access(new_root, R_OK | X_OK) == 0) )
	{
		char *old_root;
		jsked *js;
		
		js = frc->js;
		lockMutex(frc);
		old_root = frc->rootDirectory;
		frc->rootDirectory = new_root;
		jioSetRootDirectory(&js->js_IORequestPort, new_root);
		unlockMutex(frc);
		if( old_root != realRootDirectory )
		{
			gc_free(old_root);
		}
		retval = 1;
	}
	return( retval );
}

resourceSpec *fsCreateJSISpec(struct _resource *res, jsiMessage *jm)
{
	resourceSpec *retval = 0;
	int count = 0;

	count = jsiMessageIdentifierCount(jm, "fs.root");
	if( count && (retval = createResourceSpec(res)) )
	{
		fsResourceSpec *frs;

		frs = (fsResourceSpec *)retval;
		jsiMessageGet(jm,
			      JMA_String, "fs.root", 0, &frs->rootDirectory,
			      JMA_TAG_DONE);
	}
	return( retval );
}

void fsDeleteJSISpec(struct _resource *res, struct _resourceSpec *rs)
{
	fsResourceSpec *frs;

	frs = (fsResourceSpec *)rs;
	deleteResourceSpec(rs);
}

char *fsHandleJSIQuery(struct _resourceClient *rc, jsiMessage *jm)
{
	fsResourceClient *frc = (fsResourceClient *)rc;
	char *spec_name, *spec_value, *retval = 0;
	int iLockRoot;

	switch( jm->what )
	{
	case JSI_REPLY:
		lockMutex(frc);
		jsiMessageAdd(jm,
			      JMA_String, "fs.root", frc->rootDirectory,
			      JMA_TAG_DONE);
		unlockMutex(frc);
		break;
	case JSI_SET_PROPERTY:
		if( jsiPopMessageSpecifier(jm, &spec_name, &spec_value) )
		{
			if( !strcmp(spec_name, "root") )
			{
				char *root;
				
				if( !jsiMessageGet(
					jm,
					JMA_String, "data", 0, &root,
					JMA_TAG_DONE) )
				{
					char *root_cp;
					
					if( (root_cp = gc_malloc(
						strlen(root) + 1,
						GC_ALLOC_FIXED)) )
					{
						strcpy(root_cp, root);
						if( changeFSSpec(frc,
								 root_cp) )
						{
							jsiClearMessage(jm);
							jsiMessageAdd(
								jm,
								JMA_What,
								JSI_REPLY,
								JMA_String,
								"result", "ok",
								JMA_TAG_DONE);
						}
						else
						{
							retval =
								"Invalid root";
						}
					}
					else
					{
						retval = "No memory";
					}
				}
				else
				{
					retval = "No data to set";
				}
			}
			else
			{
				retval = "Unknown property specifier";
			}
		}
		else
		{
			retval = "Missing fs property specifier";
		}
		break;
	default:
		retval = "Unsupported operation";
		break;
	}
	return( retval );
}

int fsSubscribe(struct _resource *res,
		struct _resourceClient *rc,
		struct _resourceSpec *rs)
{
	fsResourceClient *frc = (fsResourceClient *)rc;
	fsResourceSpec *frs = (fsResourceSpec *)rs;
	char *new_root;
	int retval = 0;
	jsked *js;
	
	js = jskedContext();
	frc->js = js;
	if( frs )
	{
		if( (frs->rootDirectory[0] == '\0') ||
		    (access(frs->rootDirectory, R_OK | X_OK) == 0) )
		{
			if( (new_root =
			     gc_malloc(strlen(frs->rootDirectory) + 1,
				       GC_ALLOC_FIXED)) )
			{
				strcpy(new_root, frs->rootDirectory);
				frc->rootDirectory = new_root;
				jioSetRootDirectory(&js->js_IORequestPort,
						    new_root);
				retval = 1;
			}
		}
		else
		{
			teamLogEntry(
				"fs",
				LEA_String,
				"Error", "Root directory is unaccessible.", -1,
				LEA_String, "root", frs->rootDirectory, -1,
				LEA_TAG_DONE);
		}
	}
	else
	{
		/* Default to '/', the real root directory. */
		frc->rootDirectory = realRootDirectory;
		jioSetRootDirectory(&js->js_IORequestPort, realRootDirectory);
		retval = 1;
	}
	return( retval );
}

void fsUnsubscribe(struct _resource *res, struct _resourceClient *rc)
{
	// fsResourceClient *frc = (fsResourceClient *)rc;
}

void edu_utah_janosvm_resources_FileSystemSpec_nativeConstructor(
	struct Hedu_utah_janosvm_resources_FileSystemSpec *jfrs,
	struct Hjava_lang_String *jroot)
{
	fsResourceSpec *frs = OBJECT2NATIVE(&metalFSResource.link.
					    specNativeClass,
					    (Hjava_lang_Object *)jfrs);

	require(jroot != 0);

	frs->spec.handler = &metalFSResource.link;
	frs->rootDirectory = checkPtr(stringJava2C(jroot));
}

struct Hjava_lang_String *
edu_utah_janosvm_resources_FileSystemSpec_getRoot(
	struct Hedu_utah_janosvm_resources_FileSystemSpec *jfrs)
{
	fsResourceSpec *frs = OBJECT2NATIVE(&metalFSResource.link.
					    specNativeClass,
					    (Hjava_lang_Object *)jfrs);
	struct Hjava_lang_String *retval = 0;

	retval = stringC2Java(frs->rootDirectory);
	return( retval );
}

void edu_utah_janosvm_resources_FileSystemSpec_nativeFinalizer(
	struct Hedu_utah_janosvm_resources_FileSystemSpec *jfrs)
{
	fsResourceSpec *frs = OBJECT2NATIVE(&metalFSResource.link.
					    specNativeClass,
					    (Hjava_lang_Object *)jfrs);

	if( !jfrs->found )
	{
		gc_free(frs->rootDirectory);
		frs->rootDirectory = 0;
	}
}

Hjava_lang_Object *
edu_utah_janosvm_resources_FileSystemSpec_clone(
	struct Hedu_utah_janosvm_resources_FileSystemSpec *jfrs)
{
	struct Hedu_utah_janosvm_resources_FileSystemSpec *retval = 0;
	fsResourceSpec *frs, *retval_frs;
	
	retval = (struct Hedu_utah_janosvm_resources_FileSystemSpec *)
		newObject(metalFSResource.link.specNativeClass.peer);
	retval_frs = OBJECT2NATIVE(&metalFSResource.link.specNativeClass,
				   (Hjava_lang_Object *)retval);
	frs = OBJECT2NATIVE(&metalFSResource.link.specNativeClass,
			    (Hjava_lang_Object *)jfrs);
	retval_frs->spec.handler = &metalFSResource.link;
	if( frs->rootDirectory )
	{
		retval_frs->rootDirectory =
			checkPtr(gc_malloc(strlen(frs->rootDirectory) + 1,
					   GC_ALLOC_FIXED));
		strcpy(retval_frs->rootDirectory, frs->rootDirectory);
	}
	return( (Hjava_lang_Object *)retval );
}

struct Hedu_utah_janosvm_resources_FileSystemSpec *
edu_utah_janosvm_resources_FileSystemSpec_current(void)
{
	struct Hedu_utah_janosvm_resources_FileSystemSpec *retval = 0;
	fsResourceClient *frc;
	fsResourceSpec *frs;
	int iLockRoot;

	frc = (fsResourceClient *)
		getTeamInternalResource(0, &metalFSResource.link);
	retval = (struct Hedu_utah_janosvm_resources_FileSystemSpec *)
		newObject(metalFSResource.link.specNativeClass.peer);
	frs = OBJECT2NATIVE(&metalFSResource.link.specNativeClass,
			    (Hjava_lang_Object *)retval);
	frs->spec.handler = &metalFSResource.link;
	lockMutex(frc);
	if( frc->rootDirectory )
	{
		frs->rootDirectory =
			checkPtr(gc_malloc(strlen(frc->rootDirectory) + 1,
					   GC_ALLOC_FIXED));
		strcpy(frs->rootDirectory, frc->rootDirectory);
	}
	unlockMutex(frc);
	return( retval );
}

void edu_utah_janosvm_resources_FileSystemResource_change(
	struct Hedu_utah_janosvm_resources_FileSystemSpec *jfrs)
{
	errorInfo einfo;
	int error = 0;

	if( jfrs )
	{
		fsResourceSpec *frs = OBJECT2NATIVE(&metalFSResource.link.
						    specNativeClass,
						    (Hjava_lang_Object *)jfrs);
		fsResourceClient *frc;
		char *new_root = 0;
		
		frc = (fsResourceClient *)
			getTeamInternalResource(0, &metalFSResource.link);
		new_root = checkPtr(gc_malloc(strlen(frs->rootDirectory) + 1,
					      GC_ALLOC_FIXED));
		strcpy(new_root, frs->rootDirectory);
		if( !changeFSSpec(frc, new_root) )
		{
			gc_free(new_root);
			postExceptionMessage(&einfo,
					     RESOURCE_SPEC_EXCEPTION_S,
					     "Invalid root directory: %s",
					     frs->rootDirectory);
			error = 1;
		}
	}
	else
	{
		postExceptionMessage(&einfo,
				     JAVA_LANG(NullPointerException),
				     "Null spec");
		error = 1;
	}
	if( error )
	{
		throwError(&einfo);
	}
}

#endif
