/*
 * hwHandlers.c
 *
 * Copyright (c) 2002, 2003 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

#include "config.h"

#include "gc.h"
#include "locks.h"
#include "module.h"
#include "registry.h"
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/socket.h>
#include <net/if.h>

#include <jsi/jsi.h>
#include <janosvm/scripting.h>
#include <janosvm/logResource.h>

/* XXX add to configure.in */
#if defined(__FreeBSD__) || defined(__APPLE__)
#include <net/if_dl.h>
#include <net/if_arp.h>
#include <net/if_types.h>
#endif

#if defined(linux)
#include <net/if_arp.h>
#include <linux/if_packet.h>
#endif

#include <ifaddrs.h>

static jsiHandlerFunctions hwNicFunctions;
static jsiNamedHandler hwNicHandler;

static char *hwHandleNicMessage(jsiHandlerContext *jhc,
				jsiHandler *jh,
				jsiMessage *jm);
static jsiHandler *hwResolveNicSpecifier(jsiHandler *jh,
					 jsiMessage *jm,
					 char *name,
					 char *data);
static int hwGetNicSuites(jsiHandler *jh, jsiMessage *jm);

static jsiHandlerFunctions hwNameFunctions;
static jsiNamedHandler hwNameHandler;

static char *hwHandleNameMessage(jsiHandlerContext *jhc,
				 jsiHandler *jh,
				 jsiMessage *jm);
static jsiHandler *hwResolveNameSpecifier(jsiHandler *jh,
					  jsiMessage *jm,
					  char *name,
					  char *data);
static int hwGetNameSuites(jsiHandler *jh, jsiMessage *jm);

static jsiHandlerFunctions hwCPUFunctions;
static jsiNamedHandler hwCPUHandler;

static char *hwHandleCPUMessage(jsiHandlerContext *jhc,
				jsiHandler *jh,
				jsiMessage *jm);
static jsiHandler *hwResolveCPUSpecifier(jsiHandler *jh,
					 jsiMessage *jm,
					 char *name,
					 char *data);
static int hwGetCPUSuites(jsiHandler *jh, jsiMessage *jm);

static jsiHandlerFunctions hwMemoryFunctions;
static jsiNamedHandler hwMemoryHandler;

static char *hwHandleMemoryMessage(jsiHandlerContext *jhc,
				   jsiHandler *jh,
				   jsiMessage *jm);
static jsiHandler *hwResolveMemorySpecifier(jsiHandler *jh,
					    jsiMessage *jm,
					    char *name,
					    char *data);
static int hwGetMemorySuites(jsiHandler *jh, jsiMessage *jm);

static jsiHandlerFunctions hwIOFunctions;
static jsiNamedHandler hwIOHandler;

static char *hwHandleIOMessage(jsiHandlerContext *jhc,
			       jsiHandler *jh,
			       jsiMessage *jm);
static jsiHandler *hwResolveIOSpecifier(jsiHandler *jh,
					jsiMessage *jm,
					char *name,
					char *data);
static int hwGetIOSuites(jsiHandler *jh, jsiMessage *jm);

static jsiHandlerFunctions hwUptimeFunctions;
static jsiNamedHandler hwUptimeHandler;

static char *hwHandleUptimeMessage(jsiHandlerContext *jhc,
				   jsiHandler *jh,
				   jsiMessage *jm);
static jsiHandler *hwResolveUptimeSpecifier(jsiHandler *jh,
					    jsiMessage *jm,
					    char *name,
					    char *data);
static int hwGetUptimeSuites(jsiHandler *jh, jsiMessage *jm);

static jsiHandlerFunctions hwRootFunctions;
static jsiNamedHandler hwRootHandler;

static jsiHandler *hwResolveRootSpecifier(jsiHandler *jh,
					  jsiMessage *jm,
					  char *name,
					  char *data);
static int hwGetRootSuites(jsiHandler *jh, jsiMessage *jm);

static struct ifaddrs *hwNIC;

/** Module declarations **/
static int initHWHandlersModule(int stage);
static char *configHWHandlersModule(void);

module_t hwHandlersModule = {
	"hwHandlers",
	0,
	initHWHandlersModule,
	0,
	0,
	0,
	configHWHandlersModule,
};

int initHWHandlersModule(int stage)
{
	int retval = 1;

	if( getifaddrs(&hwNIC) == 0 )
	{
		jsiInitNamedHandler(
			&hwRootHandler,
			JNA_Name, "machine",
			JNA_Functions, &hwRootFunctions,
			JNA_ResolveSpecifier, hwResolveRootSpecifier,
			JNA_GetSuites, hwGetRootSuites,
			JNA_Child, jsiInitNamedHandler(
				&hwNameHandler,
				JNA_Name, "name",
				JNA_Functions, &hwNameFunctions,
				JNA_HandleMessage, hwHandleNameMessage,
				JNA_ResolveSpecifier, hwResolveNameSpecifier,
				JNA_GetSuites, hwGetNameSuites,
				JNA_TAG_DONE),
			JNA_Child, jsiInitNamedHandler(
				&hwCPUHandler,
				JNA_Name, "cpu",
				JNA_Functions, &hwCPUFunctions,
				JNA_HandleMessage, hwHandleCPUMessage,
				JNA_ResolveSpecifier, hwResolveCPUSpecifier,
				JNA_GetSuites, hwGetCPUSuites,
				JNA_TAG_DONE),
			JNA_Child, jsiInitNamedHandler(
				&hwMemoryHandler,
				JNA_Name, "memory",
				JNA_Functions, &hwMemoryFunctions,
				JNA_HandleMessage, hwHandleMemoryMessage,
				JNA_ResolveSpecifier, hwResolveMemorySpecifier,
				JNA_GetSuites, hwGetMemorySuites,
				JNA_TAG_DONE),
			JNA_Child, jsiInitNamedHandler(
				&hwIOHandler,
				JNA_Name, "io",
				JNA_Functions, &hwIOFunctions,
				JNA_HandleMessage, hwHandleIOMessage,
				JNA_ResolveSpecifier, hwResolveIOSpecifier,
				JNA_GetSuites, hwGetIOSuites,
				JNA_TAG_DONE),
			JNA_Child, jsiInitNamedHandler(
				&hwUptimeHandler,
				JNA_Name, "uptime",
				JNA_Functions, &hwUptimeFunctions,
				JNA_HandleMessage, hwHandleUptimeMessage,
				JNA_ResolveSpecifier, hwResolveUptimeSpecifier,
				JNA_GetSuites, hwGetUptimeSuites,
				JNA_TAG_DONE),
			JNA_Child, jsiInitNamedHandler(
				&hwNicHandler,
				JNA_Name, "nic",
				JNA_Functions, &hwNicFunctions,
				JNA_HandleMessage, hwHandleNicMessage,
				JNA_ResolveSpecifier, hwResolveNicSpecifier,
				JNA_GetSuites, hwGetNicSuites,
				JNA_TAG_DONE),
			JNA_TAG_DONE);
		jsiInstallHandler(&jvmHandlerList, &hwRootHandler);
	}
#if 0
	else
	{
		teamLogEntry("hwHandlers",
			     LEA_String, "Warning", "getifaddrs() failed", -1,
			     LEA_String, "Errno", strerror(errno), -1,
			     LEA_TAG_DONE);
	}
#endif

	return( retval );
}

char *configHWHandlersModule(void)
{
	char *retval;

	retval = ""
#if defined(HAVE_GETIFADDRS)
		" (HAVE_GETIFADDRS)"
#elif defined(linux)
		" (Using Linux emulated getifaddrs)"
#endif
		;
	return( retval );
}

static struct ifaddrs *hwFindNIC(char *name)
{
	struct ifaddrs *ifa, *retval = 0;
	int lpc, id;
	
	ifa = hwNIC;
	if( sscanf(name, "%d", &id) == 1 )
	{
		for( lpc = 0; ifa && lpc < id; lpc++, ifa = ifa->ifa_next );
		if( lpc == id && ifa )
			retval = ifa;
	}
	else
	{
		while( ifa && !retval )
		{
			if( !strcmp(ifa->ifa_name, name) )
				retval = ifa;
			ifa = ifa->ifa_next;
		}
	}
	return( retval );
}

static char *hwType(struct ifaddrs *ifa)
{
	char *retval;

	if( ifa->ifa_addr )
	{
		
		switch( ifa->ifa_addr->sa_family )
		{
		case AF_INET:
			retval = "inet";
			break;
#if defined(AF_LINK)
		case AF_LINK:
			{
				struct sockaddr_dl *sdl;
				
				sdl = (struct sockaddr_dl *)ifa->ifa_addr;
				switch( sdl->sdl_type )
				{
				case IFT_ETHER:
					retval = "eth";
					break;
				case IFT_LOOP:
					retval = "loopback";
					break;
				default:
					retval = "link";
					break;
				}
			}
			break;
#endif
#if defined(PF_PACKET)
		case PF_PACKET:
			{
				struct sockaddr_ll *sll;
				
				sll = (struct sockaddr_ll *)ifa->ifa_addr;
				switch( sll->sll_hatype )
				{
				case ARPHRD_ETHER:
					retval = "eth";
					break;
				case ARPHRD_LOOPBACK:
					retval = "loopback";
					break;
				default:
					retval = "link";
					break;
				}
			}
			break;
#endif
		case AF_LOCAL:
			retval = "localhost";
			break;
		default:
			retval = "(unknown)";
			break;
		}
	}
	else
	{
		retval = "(unknown)";
	}
	return( retval );
}

static char *hwAddress(char *dest, struct sockaddr *sa)
{
	char *retval = dest;
	int lpc, hlen;
	char *haddr;
	
#if defined(AF_LINK)
	struct sockaddr_dl *sdl;

	sdl = (struct sockaddr_dl *)sa;
	hlen = sdl->sdl_alen;
	haddr = sdl->sdl_data + sdl->sdl_nlen;
#elif defined(PF_PACKET)
	struct sockaddr_ll *sll;

	sll = (struct sockaddr_ll *)sa;
	hlen = sll->sll_halen;
	haddr = sll->sll_addr;
#else
	hlen = 0;
#endif
	for( lpc = 0; lpc < hlen; lpc++ )
	{
		sprintf(dest,
			"%s%02x",
			(lpc > 0) ? ":" : "",
			haddr[lpc] & 0xff);
		dest += strlen(dest);
	}
	return( retval );
}

static char *ifAddress(char *dest, struct sockaddr *sa)
{
	if( sa )
	{
		switch( sa->sa_family )
		{
		case AF_INET:
			inet_ntop(sa->sa_family,
				  &((struct sockaddr_in *)sa)->sin_addr,
				  dest,
				  32);
			break;
#if defined(PF_PACKET)
		case AF_PACKET:
#endif
#if defined(AF_LINK)
		case AF_LINK:
#endif
			hwAddress(dest, sa);
			break;
		default:
			sprintf(dest, "...");
			break;
		}
	}
	else
	{
		sprintf(dest, "...");
	}
	return( dest );
}

static void hwAddNic(jsiMessage *jm, struct ifaddrs *ifa)
{
	char addrbuf[32], maskbuf[32];

	jsiMessageAdd(jm,
		      JMA_String, "name", ifa->ifa_name,
		      JMA_String, "type", hwType(ifa),
		      JMA_String, "address", ifAddress(addrbuf, ifa->ifa_addr),
		      JMA_String, "netmask", ifAddress(maskbuf,
						       ifa->ifa_netmask),
		      JMA_TAG_DONE);
}

/** BEGIN nic **/

jsiHandler *hwResolveNicSpecifier(jsiHandler *jh,
				  jsiMessage *jm,
				  char *name,
				  char *data)
{
	return jh;
}

static jsiPropertyInfo hwSubNicProperties[] = {
	{ "type", (1L << JSI_GET_PROPERTY), "",
	  "The network interface type.\n" },
	{ "id", (1L << JSI_GET_PROPERTY), "",
	  "The network interface unique id.\n" },
	{ "name", (1L << JSI_GET_PROPERTY), "",
	  "The physical name.\n" },
	{ "bandwdith", (1L << JSI_GET_PROPERTY), "",
	  "The total bandwidth.\n" },
	{ "address", (1L << JSI_GET_PROPERTY), "",
	  "The MAC address.\n" },
	{ 0 }
};

char *hwHandleNicMessage(jsiHandlerContext *jhc,
			 jsiHandler *jh,
			 jsiMessage *jm)
{
	char *name, *value, *retval = 0;
	struct ifaddrs *ifa = 0;
	
	switch( jm->what )
	{
	case JSI_LIST_PROPERTIES:
		jsiRestoreMessageSpecifier(jm);
		jsiPopMessageSpecifier(jm, &name, &value);
		if( (ifa = hwFindNIC(value)) )
		{
			jsiClearMessage(jm);
			jsiMessageAdd(jm,
				      JMA_What, JSI_REPLY,
				      JMA_String, "result", "type",
				      JMA_String, "result", "id",
				      JMA_String, "result", "name",
				      JMA_String, "result", "bandwidth",
				      JMA_String, "result", "address",
				      JMA_TAG_DONE);
		}
		else
		{
			jsiClearMessage(jm);
			jm->what = JSI_REPLY;
			for( ifa = hwNIC; ifa; ifa = ifa->ifa_next )
			{
				if( ifa->ifa_flags & IFF_UP )
					hwAddNic(jm, ifa);
			}
		}
		break;
	case JSI_GET_PROPERTY:
		jsiRestoreMessageSpecifier(jm);
		jsiPopMessageSpecifier(jm, &name, &value);
		if( (ifa = hwFindNIC(value)) )
		{
			if( jsiPopMessageSpecifier(jm, &name, &value) )
			{
				jsiClearMessage(jm);
				jm->what = JSI_REPLY;
				if( !strcmp(name, "name") )
				{
					jsiMessageAdd(
						jm,
						JMA_What, JSI_REPLY,
						JMA_String, "result", ifa->ifa_name,
						JMA_TAG_DONE);
				}
				else
				{
					retval = "Unknown nic property";
				}
			}
			else
			{
				jsiClearMessage(jm);
				hwAddNic(jm, ifa);
			}
		}
		else
		{
			retval = "Invalid NIC id";
		}
		break;
	case JSI_GET_SUITES:
		jsiClearMessage(jm);
		if( !jsiMessageAdd(
			jm,
			JMA_What, JSI_REPLY,
			JMA_PropertyList, "result", hwSubNicProperties,
			JMA_TAG_DONE) )
		{
			retval = "Out of memory";
		}
		break;
	default:
		retval = "Operation not supported";
		break;
	}
	return( retval );
}

static jsiPropertyInfo hwNicProperties[] = {
	{ "nic", (1L << JSI_LIST_PROPERTIES), "",
	  "List the Network Interface Cards in the machine.\n" },
	{ "nic", (1L << JSI_GET_PROPERTY), "nicID:int/string",
	  "Get info about a specific NIC.\n" },
	{ 0 }
};

int hwGetNicSuites(jsiHandler *jh, jsiMessage *jm)
{
	int retval;

	retval = jsiMessageAdd(jm,
			       JMA_PropertyList, "result", hwNicProperties,
			       JMA_TAG_DONE);
	return( retval );
}

/** END nic **/

/** BEGIN name **/

jsiHandler *hwResolveNameSpecifier(jsiHandler *jh,
				   jsiMessage *jm,
				   char *name,
				   char *data)
{
	return jh;
}

char *hwHandleNameMessage(jsiHandlerContext *jhc,
			 jsiHandler *jh,
			 jsiMessage *jm)
{
	char *retval = 0;
	
	switch( jm->what )
	{
	case JSI_GET_PROPERTY:
		{
			char hostnameBuffer[64];
			char *name;

			if( registryName )
			{
				name = registryName;
			}
			else
			{
				gethostname(hostnameBuffer, 64);
				name = hostnameBuffer;
			}
			jsiClearMessage(jm);
			jsiMessageAdd(jm,
				      JMA_What, JSI_REPLY,
				      JMA_String, "result", name,
				      JMA_TAG_DONE);
		}
		break;
	default:
		retval = "Operation not supported";
		break;
	}
	return( retval );
}

static jsiPropertyInfo hwNameProperties[] = {
	{ "name", (1L << JSI_GET_PROPERTY), "",
	  "Get the name of this virtual machine or the host it is executing "
	  "on.\n" },
	{ 0 }
};

int hwGetNameSuites(jsiHandler *jh, jsiMessage *jm)
{
	int retval;

	retval = jsiMessageAdd(jm,
			       JMA_PropertyList, "result", hwNameProperties,
			       JMA_TAG_DONE);
	return( retval );
}

/** END name **/

/** BEGIN cpu **/

jsiHandler *hwResolveCPUSpecifier(jsiHandler *jh,
				  jsiMessage *jm,
				  char *name,
				  char *data)
{
	return jh;
}

char *hwHandleCPUMessage(jsiHandlerContext *jhc,
			 jsiHandler *jh,
			 jsiMessage *jm)
{
	char *retval = 0;
	
	switch( jm->what )
	{
	case JSI_GET_PROPERTY:
#if defined(UNIX_JTHREADS2)
		jsiClearMessage(jm);
		jsiMessageAdd(jm,
			      JMA_What, JSI_REPLY,
			      JMA_LongLongUnsignedInteger, "used", jsked_state.jss_CPUAccount.jca_TotalUsed,
			      JMA_LongLongUnsignedInteger, "idle", jcpuAccountingIdle(&jsked_state.jss_CPUAccount),
			      JMA_TAG_DONE);
#else
		retval = "Operation not supported in this configuration";
#endif
		break;
	default:
		retval = "Operation not supported";
		break;
	}
	return( retval );
}

static jsiPropertyInfo hwCPUProperties[] = {
	{ "cpu", (1L << JSI_GET_PROPERTY), "",
	  "Get the used and idle CPU time.\n" },
	{ 0 }
};

int hwGetCPUSuites(jsiHandler *jh, jsiMessage *jm)
{
	int retval;

	retval = jsiMessageAdd(jm,
			       JMA_PropertyList, "result", hwCPUProperties,
			       JMA_TAG_DONE);
	return( retval );
}

/** END cpu **/

/** BEGIN memory **/

jsiHandler *hwResolveMemorySpecifier(jsiHandler *jh,
				     jsiMessage *jm,
				     char *name,
				     char *data)
{
	return jh;
}

char *hwHandleMemoryMessage(jsiHandlerContext *jhc,
			    jsiHandler *jh,
			    jsiMessage *jm)
{
	extern size_t gc_heap_total;
	extern size_t gc_heap_limit;

	char *retval = 0;
	
	switch( jm->what )
	{
	case JSI_GET_PROPERTY:
		jsiClearMessage(jm);
		jsiMessageAdd(jm,
			      JMA_What, JSI_REPLY,
			      JMA_UnsignedInteger, "used", gc_heap_total,
			      JMA_UnsignedInteger, "max", gc_heap_limit,
			      JMA_TAG_DONE);
		break;
	default:
		retval = "Operation not supported";
		break;
	}
	return( retval );
}

static jsiPropertyInfo hwMemoryProperties[] = {
	{ "memory", (1L << JSI_GET_PROPERTY), "",
	  "Get the maximum and used amount of memory.\n" },
	{ 0 }
};

int hwGetMemorySuites(jsiHandler *jh, jsiMessage *jm)
{
	int retval;

	retval = jsiMessageAdd(jm,
			       JMA_PropertyList, "result", hwMemoryProperties,
			       JMA_TAG_DONE);
	return( retval );
}

/** END memory **/

/** BEGIN io **/

jsiHandler *hwResolveIOSpecifier(jsiHandler *jh,
				 jsiMessage *jm,
				 char *name,
				 char *data)
{
	return jh;
}

char *hwHandleIOMessage(jsiHandlerContext *jhc,
			jsiHandler *jh,
			jsiMessage *jm)
{
	char *retval = 0;
	
	switch( jm->what )
	{
	case JSI_GET_PROPERTY:
#if defined(UNIX_JTHREADS2)
		jsiClearMessage(jm);
		jsiMessageAdd(jm,
			      JMA_What, JSI_REPLY,
			      JMA_LongLongUnsignedInteger, "ipc.in", jsked_state.jss_IOAccount[JSKED_IOACCOUNT_IPC].jia_Accumulator[JIO_ACCOUNT_IN],
			      JMA_LongLongUnsignedInteger, "ipc.out", jsked_state.jss_IOAccount[JSKED_IOACCOUNT_IPC].jia_Accumulator[JIO_ACCOUNT_OUT],
			      JMA_LongLongUnsignedInteger, "disk.in", jsked_state.jss_IOAccount[JSKED_IOACCOUNT_DISK].jia_Accumulator[JIO_ACCOUNT_IN],
			      JMA_LongLongUnsignedInteger, "disk.out", jsked_state.jss_IOAccount[JSKED_IOACCOUNT_DISK].jia_Accumulator[JIO_ACCOUNT_OUT],
			      JMA_LongLongUnsignedInteger, "network.in", jsked_state.jss_IOAccount[JSKED_IOACCOUNT_NETWORK].jia_Accumulator[JIO_ACCOUNT_IN],
			      JMA_LongLongUnsignedInteger, "network.out", jsked_state.jss_IOAccount[JSKED_IOACCOUNT_NETWORK].jia_Accumulator[JIO_ACCOUNT_OUT],
			      JMA_TAG_DONE);
#else
		retval = "Operation not supported in this configuration";
#endif
		break;
	default:
		retval = "Operation not supported";
		break;
	}
	return( retval );
}

static jsiPropertyInfo hwIOProperties[] = {
	{ "io", (1L << JSI_GET_PROPERTY), "",
	  "Get the I/O stats for this virtual machine.\n" },
	{ 0 }
};

int hwGetIOSuites(jsiHandler *jh, jsiMessage *jm)
{
	int retval;

	retval = jsiMessageAdd(jm,
			       JMA_PropertyList, "result", hwIOProperties,
			       JMA_TAG_DONE);
	return( retval );
}

/** END io **/

/** BEGIN uptime **/

jsiHandler *hwResolveUptimeSpecifier(jsiHandler *jh,
				     jsiMessage *jm,
				     char *name,
				     char *data)
{
	return jh;
}

char *hwHandleUptimeMessage(jsiHandlerContext *jhc,
			 jsiHandler *jh,
			 jsiMessage *jm)
{
	char *retval = 0;
	
	switch( jm->what )
	{
	case JSI_GET_PROPERTY:
#if defined(UNIX_JTHREADS2)
		jsiClearMessage(jm);
		jsiMessageAdd(jm,
			      JMA_What, JSI_REPLY,
			      JMA_UnsignedInteger, "result", jcpuTime(),
			      JMA_TAG_DONE);
#else
		retval = "Operation not supported in this configuration";
#endif
		break;
	default:
		retval = "Operation not supported";
		break;
	}
	return( retval );
}

static jsiPropertyInfo hwUptimeProperties[] = {
	{ "uptime", (1L << JSI_GET_PROPERTY), "",
	  "Get the uptime, in milliseconds, for this virtual machine.\n" },
	{ 0 }
};

int hwGetUptimeSuites(jsiHandler *jh, jsiMessage *jm)
{
	int retval;

	retval = jsiMessageAdd(jm,
			       JMA_PropertyList, "result", hwUptimeProperties,
			       JMA_TAG_DONE);
	return( retval );
}

/** END uptime **/

jsiHandler *hwResolveRootSpecifier(jsiHandler *jh,
				   jsiMessage *jm,
				   char *name,
				   char *data)
{
	jsiNamedHandler *jnh = (jsiNamedHandler *)jh;
	jsiHandler *retval;

	if( jsiPopMessageSpecifier(jm, &name, &data) )
	{
		if( (retval = (jsiHandler *)
		     jsiFindHandler(&jnh->children, name)) )
		{
			retval = retval->funcs->resolveSpecifier(retval,
								 jm,
								 name,
								 data);
		}
	}
	else
	{
		retval = jh;
	}
	return retval;
}

static jsiPropertyInfo hwRootProperties[] = {
	{ "machine", (1L << JSI_LIST_PROPERTIES), "",
	  "List the builtin properties of the machine.\n" },
	{ 0 }
};

int hwGetRootSuites(jsiHandler *jh, jsiMessage *jm)
{
	int retval;

	retval = jsiMessageAdd(jm,
			       JMA_PropertyList, "result", hwRootProperties,
			       JMA_TAG_DONE);
	return( retval );
}
