/*
 * jteam.c
 * Maintains separation of code running in the jvm.
 *
 * Copyright (c) 2000, 2001, 2002, 2003 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <assert.h>

#include "assert_pp.h"

#include "gtypes.h"
#include "threadData.h"
#include "thread.h"
#include "jthread.h"
#include "stringSupport.h"
#include "object.h"
#include "locks.h"
#include "jteam.h"
#include "lookup.h"
#include "memResource.h"
#include "classPath.h"
#include "logResource.h"
#include "module.h"
#include "nameSpace.h"
#include "resource-inline.h"
#include "configFile.h"
#include "moduleList.h"
#include "scripting.h"
#include "jthread.h"
#include "support.h"
#include "soft.h"
#include "lock-impl.h"

#include "jskedResource.h"
#include "jskedResource-inline.h"

#include "java_lang_ThreadObjectHolder.h"

#include "threadData-inline.h"
#include "jteamReservation.h"
#include "nativeClass-inline.h"
#include "threadData-inline.h"
#include "locks-inline.h"

#if defined(TRANSLATOR) 
#if defined(JIT3)
#include "jit3/machine.h"
#else
#include "jit/machine.h"
#endif
#endif

/** jteam variables **/
sparseArray *teamArray;
struct lnMinList teamList;
iLock *teamArrayLock;

nativeClass jteamNativeClass;
Hjava_lang_Class *TeamClass;
Hjava_lang_Class *TeamBackEndClass;
Hjava_lang_Class *TeamReservationClass;
Hjava_lang_Class *ListClass;
Hedu_utah_janosvm_sys_TeamBackEnd *kernelTeamBackEnd;
jthread_t reaperThread;
Method *FinishTeamEventMethod;
Method *DeadTeamEventMethod;
Method *GetTeamIDMethod;
Method *AddTailMethod;
Method *FindTeamMethod;

struct _metalTeamAndResources metalTeamAndResources;
jteam *metalTeam = &metalTeamAndResources.team;
iLock **kernelLock;
jteam *kernelTeam;
static char *defaultTeamName = "anonymous";

resourceSpec *defaultInternalSpec = (resourceSpec *)~0;

struct Hedu_utah_janosvm_sys_EventQueue *KernelEventQueue;
struct Hjava_lang_Thread *KernelEventThread;

static void terminateNativeTeam(jteam *jt);
static void scheduleCollectTeam(jteam *jt);

char *teamHandleMessage(jsiHandlerContext *hc, jsiHandler *jh, jsiMessage *jm);
jsiHandler *teamResolveSpecifier(jsiHandler *jh, jsiMessage *jm,
				 char *name, char *data);

static jsiHandlerFunctions teamFunctions = {
	teamResolveSpecifier,
	teamHandleMessage
};

static jsiHandlerFunctions teamsRootFunctions;
static jsiHandlerFunctions teamRootFunctions;
static jsiNamedHandler teamsRootHandler;
static jsiNamedHandler teamRootHandler;

char *teamHandleRootMessage(jsiHandlerContext *hc,
			    jsiHandler *jh, jsiMessage *jm);
jsiHandler *teamResolveRootSpecifier(jsiHandler *jh, jsiMessage *jm,
				     char *name, char *data);
int teamGetRootSuites(jsiHandler *jh, jsiMessage *jm);

/** Module declarations **/
static int initJTeamModule(int stage);

module_t jteamModule = {
	"jteam",
	0,
	initJTeamModule,
	0,
	0
};

/** Module functions **/

/* Create the kernel team object and setup the global pointers */
static
int createKernelTeam(void)
{
	int retval = 0;

	if( (kernelTeamBackEnd = (Hedu_utah_janosvm_sys_TeamBackEnd *)
	     execute_java_constructor(0, 0, TeamBackEndClass, "()V")) )
	{
		kernelLock = &kernelTeamBackEnd->base.lock;
		kernelTeam = OBJECT2NATIVE(&jteamNativeClass,
					   &kernelTeamBackEnd->base);
		retval = 1;
	}
	return( retval );
}

static
void installJTeamScripting(void)
{
	jsiInitNamedHandler(&teamRootHandler,
			    JNA_Name, "team",
			    JNA_Functions, &teamRootFunctions,
			    JNA_HandleMessage, teamHandleRootMessage,
			    JNA_ResolveSpecifier, teamResolveRootSpecifier,
			    JNA_TAG_DONE);
	jsiInstallHandler(&jvmHandlerList, &teamRootHandler);
	jsiInitNamedHandler(&teamsRootHandler,
			    JNA_Name, "teams",
			    JNA_Functions, &teamsRootFunctions,
			    JNA_HandleMessage, teamHandleRootMessage,
			    JNA_ResolveSpecifier, teamResolveRootSpecifier,
			    JNA_GetSuites, teamGetRootSuites,
			    JNA_TAG_DONE);
	jsiInstallHandler(&jvmHandlerList, &teamsRootHandler);
}

int initJTeamModule(int stage)
{
	int retval = 1;
	
	switch( stage )
	{
	case SID_JTEAMMODULE_INITMETAL:
		metalTeam->name = "metal";
		metalTeam->state = JTS_ALIVE;
		break;
	case SID_JTEAMMODULE:
		{
			errorInfo einfo;
			Utf8Const *nm;

			/*
			 * Setup the jteamNativeClass to include not only
			 * the size of the jteam structure, but also an array
			 * for the internal resources.
			 */
			jteamNativeClass.size = sizeof(jteam) +
				(sizeof(resourceClient *) *
				 internalResourceCount);
			/*
			 * Try to get the TeamBackEnd class, making sure its
			 * the real JanosVM.jar one and not the emulated one.
			 * Then make the kernel team object.
			 */
			TeamClass = lookupClass(
				"edu/utah/janosvm/sys/Team",
				NULL, &einfo);
			if( (nm = utf8ConstNew("_versionName", -1)) &&
			    (TeamBackEndClass =
			     lookupClass("edu/utah/janosvm/sys/TeamBackEnd",
					 NULL, &einfo)) &&
			    (ListClass =
			     lookupClass("edu/utah/janosvm/sys/List",
					 NULL, &einfo)) &&
			    /* Make sure its the "real thing" */
			    lookupClassField(TeamBackEndClass,
					     nm,
					     true,
					     &einfo) &&
			    (TeamReservationClass =
			     lookupClass("edu/utah/janosvm/sys/TeamReservation",
					 NULL, &einfo)) )
			{
				FindTeamMethod = lookupClassMethod(
					TeamClass,
					"find",
					"(Ledu/utah/janosvm/sys/TeamHandle;"
					"Ledu/utah/janosvm/sys/TeamBackEnd;)"
					"Ledu/utah/janosvm/sys/TeamHandle;",
					&einfo);
				translate(FindTeamMethod, &einfo);
				DeadTeamEventMethod = lookupClassMethod(
					TeamBackEndClass,
					"deadTeamEvent",
					"()V",
					&einfo);
				translate(DeadTeamEventMethod, &einfo);
				GetTeamIDMethod = lookupClassMethod(
					TeamClass,
					"getID",
					"()I",
					&einfo);
				translate(GetTeamIDMethod, &einfo);
				FinishTeamEventMethod = lookupClassMethod(
					TeamBackEndClass,
					"finishTeamEvent",
					"()V",
					&einfo);
				translate(FinishTeamEventMethod, &einfo);
				AddTailMethod = lookupClassMethod(
					ListClass,
					"addTail",
					"(Ledu/utah/janosvm/sys/ListNode;)V",
					&einfo);
				translate(AddTailMethod, &einfo);
				
				/*
				 * lookupClass will put the class in
				 * CSTATE_COMPLETE, so we're `safe' to bind
				 * the native data here.
				 */
				retval = bindNativeClass(&jteamNativeClass,
							 TeamBackEndClass,
							 &einfo);
				if( retval &&
				    (teamArray =
				     createSparseArray(SAA_TAG_DONE)) )
				{
					lnNewList(&teamList);
					if( (retval = createKernelTeam()) )
					{
						installJTeamScripting();
					}
					else
					{
					}
				}
				else
				{
					retval = 0;
				}
			}
		}
		break;
	}
	return( retval );
}

/* static inline */
void _lockTeam(jteam *jt, int *iLockRoot)
{
	require(jt != 0);
	require(jt != metalTeam);

	jthread_disable_stop();
	_lockMutex(&NATIVE2OBJECT(&jteamNativeClass, jt)->lock, iLockRoot);
}

void _unlockTeam(jteam *jt, int *iLockRoot)
{
	require(jt != 0);
	require(jt != metalTeam);
	
	_unlockMutex(&NATIVE2OBJECT(&jteamNativeClass, jt)->lock, iLockRoot);
	jthread_enable_stop();
}

void setTeamDescription(jteam *jt, char *str)
{
	if( switchToTeam(jt, 0) )
	{
		char *str_cp;

		if( (str_cp = gc_malloc(strlen(str) + 1, GC_ALLOC_JTEAM)) )
		{
			char *old_str;
			int iLockRoot;

			strcpy(str_cp, str);
			lockStaticMutex(&teamArrayLock);
			lockTeam(jt); // XXX Is this needed?
			lockMutex(&jt->internalLock);
			old_str = jt->description;
			jt->description = str_cp;
			unlockMutex(&jt->internalLock);
			unlockTeam(jt);
			unlockStaticMutex(&teamArrayLock);
			gc_free(old_str);
		}
		returnFromTeam();
	}
}

#if !defined(NO_PRECONDITIONS) || !defined(NO_POSTCONDITIONS)
static
int jteamInvariant(jteam *jt)
{
	require(jt != 0);
	require(jt->state >= JTS_NEW);
	require(jt->state <= JTS_BURIED);

	require(JTS_NEW < JTS_ALIVE);
	require(JTS_ALIVE < JTS_DYING);
	require(JTS_DYING < JTS_DEAD);
	require(JTS_DEAD < JTS_BURIED);
	
	switch(jt->state)
	{
	case JTS_NEW:
		/*
		 * NEW state only exists (only briefly) during the
		 * nativeConstructor() call.  Not much you can say
		 * about a team in this state.
		 */
		require(jt->elemID == 0);
		require(jt->zombieCount == 0);
		break;
	case JTS_ALIVE:
		// require(jt->elemID != 0);
		require(jt->name != 0);
		require(jt->zombieCount == 0);
		require(jt->liveThreads >= 0);
		require(jt->daemonThreads >= 0);
		require(jt->liveThreads >= jt->daemonThreads);
		require(jt->guestThreads >= 0);
		require(jt->absentThreads >= 0);
		break;
	case JTS_REVOKING:
		require(jt->name != 0);
		require(jt->zombieCount == 0);
		require(jt->liveThreads >= 0);
		require(jt->daemonThreads >= 0);
		require(jt->liveThreads >= jt->daemonThreads);
		require(jt->guestThreads >= 0);
		require(jt->absentThreads >= 0);
		break;
	case JTS_DYING:
		require(jt->name != 0);
		require(jt->zombieCount == 0);
		require(jt->liveThreads >= 0);
		require(jt->daemonThreads >= 0);
		require(jt->liveThreads >= jt->daemonThreads);
		require(jt->guestThreads >= 0);
		require(jt->absentThreads >= 0);
		break;
	case JTS_DEAD:
		require(jt->zombieCount >= 0);
		require(jt->liveThreads >= 0);
		require(jt->daemonThreads >= 0);
		require(jt->liveThreads >= jt->daemonThreads);
		require((jt->guestThreads == 0) ||
			(jt->guestThreads == 1) /* For the reaper thread. */);
		require(jt->absentThreads == 0);
		break;
	case JTS_BURIED:
		// require(jt->elemID != 0);
		require(jt->name == 0);
		require(jt->zombieCount == 0);
		require(jt->liveThreads == 0);
		require(jt->daemonThreads == 0);
		require(jt->guestThreads == 0);
		require(jt->absentThreads == 0);
		break;
	}

	return( 1 );
}
#endif

static iLock *dependencyLock;

int addTeamDependency(jteam *jt, jteam *dep)
{
	int retval = 0;
	int iLockRoot;

	require(currentTeam() == kernelTeam);
	require(jt != 0);
	require(dep != 0);

	lockStaticMutex(&dependencyLock);
	if( jt == dep )
	{
		/* ... */
	}
	else if( hasTeamDependency(dep, jt) )
	{
		/* Can't have cycles... */
		// assert(0);
	}
	else if( hasTeamDependency(jt, dep) )
	{
		/* Already have the dependency */
		retval = 1;
	}
	else
	{
		int valid = 0;
		
		lockTeam(dep);
		/* Make sure the thing is still alive */
		if( dep->state == JTS_ALIVE )
		{
			dep->dependents++;
			valid = 1;
		}
		unlockTeam(dep);
		
		if( valid )
		{
			jteamDependency *jd;
			int lpc;
			
			jd = &jt->deps;
			while( jd && !retval )
			{
				for( lpc = 0;
				     (lpc < JTEAM_DEPENDENCY_POOL) && !retval;
				     lpc++ )
				{
					if( jd->deps[lpc] == 0 )
					{
						jd->deps[lpc] = dep;
						retval = 1;
					}
				}
				jd = jd->next;
			}
			if( !retval )
			{
				if( (jd = gc_malloc(sizeof(jteamDependency),
						    GC_ALLOC_JTEAM)) )
				{
					jd->next = jt->deps.next;
					jd->deps[0] = dep;
					jt->deps.next = jd;
					retval = 1;
				}
			}
		}
	}
	unlockStaticMutex(&dependencyLock);
	return( retval );
}

/*
 * Drop any dependencies on other teams.  If we're they've been terminated and
 * are waiting for the last dependents to finish then this function is
 * responsible for scheduling the collection.
 */
static
void dropTeamDependencies(jteam *jt)
{
	jteamDependency *jd, *jdnext;
	int lpc;
	int iLockRoot;
	
	require(currentTeam() == kernelTeam);
	
	jd = &jt->deps;
	while( jd )
	{
		for( lpc = 0; lpc < JTEAM_DEPENDENCY_POOL; lpc++ )
		{
			if( jd->deps[lpc] )
			{
				jteam *dep = jd->deps[lpc];
				int collect = 0;
				
				lockTeam(dep);
				dep->dependents--;
				if( (dep->state == JTS_DYING) &&
				    (dep->guestThreads == 0) &&
				    (dep->absentThreads == 0) &&
				    (dep->dependents == 0) )
				{
					dep->state = JTS_DEAD;
					collect = 1;
				}
				unlockTeam(dep);
				if( collect )
					scheduleCollectTeam(dep);
			}
		}
		jdnext = jd->next;
		if( jd != &jt->deps )
			gc_free(jd);
		jd = jdnext;
	}
}

/*
 * Terminate any teams that depend on the given one.
 */
static
void terminateTeamDependencies(jteam *jt)
{
	jteam *curr;
	int iLockRoot;
	
	require(currentTeam() == kernelTeam);
	
	lockStaticMutex(&teamArrayLock);
	curr = (jteam *)teamList.lh_Head;
	while( curr->link.ln_Succ )
	{
		if( curr->state == JTS_ALIVE )
		{
			if( hasTeamDependency(curr, jt) )
			{
				unlockStaticMutex(&teamArrayLock);
				terminateTeam(curr);
				lockStaticMutex(&teamArrayLock);
				curr = (jteam *)&teamList;
			}
		}
		curr = (jteam *)curr->link.ln_Succ;
	}
	unlockStaticMutex(&teamArrayLock);
}

/*
 * Distribute the java specifications into an array of pointers to the native
 * spec structures.  In doing so the specs are sorted so that we will
 * subscribe to them in the correct order.  Only the internal specs are
 * included in the output array.  Internal specs are ignored for now.
 *
 * rsArray is an output array.  Assumed to have internalResourceCount
 * entries.
 */
static
int sortSpecs(resourceSpec *rsArray[],
	      HArrayOfObject *jSpecs,
	      resourceSpec **nSpecs)
{
	int lpc, retval = 1;
	
	require(rsArray != 0);

	for( lpc = 0; retval && jSpecs && (lpc < ARRAY_SIZE(jSpecs)); lpc++ )
	{
		struct Hedu_utah_janosvm_resources_ResourceSpec *curr;
		nativeClass *type;

		curr = ((struct Hedu_utah_janosvm_resources_ResourceSpec **)
			ARRAY_DATA(jSpecs))[lpc];

		/*
		 * Only include specs that have a native class object
		 * defined and that are internal specs in the
		 * sorted array.
		 */
		if( curr && (type = CLASS2NATIVE(curr->base.dtable->class)) )
		{
			resourceSpec *rs;
			
			rs = OBJECT2NATIVE(type, &curr->base);
			
			require(rs != 0);
			require(rs->handler != 0);
			
			if( rs->handler->internalIndex >= 0 )
			{
				if( rsArray[rs->handler->internalIndex] == 0 )
				{
					rsArray[rs->handler->
					       internalIndex] = rs;
				}
				else
				{
					/* Don't allow overwrites */
					retval = 0;
				}
			}
		}
		else if( curr )
		{
			teamLogEntry("jteam",
				     LEA_String, "Error",
				     "No native data in java spec", -1,
				     LEA_Pointer, "obj", curr,
				     LEA_TAG_DONE);
		}
		else
		{
		    /* NULL slot in the array, don't care */
		}
	}
	for( lpc = 0;
	     retval && nSpecs && (lpc < internalResourceCount);
	     lpc++ )
	{
		/* Only overwrite when the native spec is non-null */
		if( nSpecs[lpc] )
			rsArray[lpc] = nSpecs[lpc];
	}
	return( retval );
}

static void
unsubscribeAllInternalResources(jteam *jt)
{
	int lpc, did_return = 0;

	require(jt);
	
	for( lpc = internalResourceCount - 1; lpc >= 0; lpc-- )
	{
		resource *res;
		
		require(jt->internalResources[lpc]);
		// require((currentTeam() == jt));
		
		res = metalTeam->internalResources[lpc]->handler;
		if( (res->flags & RF_KERNEL_CLIENT) && !did_return )
		{
			quickChangeTeamReservation(kernelTeam);
			did_return = 1;
		}
		unsubscribeResource(jt->internalResources[lpc]);
		jt->internalResources[lpc] = NULL;

		// ensure(currentTeam() == jt);
	}
}

		
static void
unsubscribeAllExternalResources(jteam* jt)
{
	resourceClient *rc, *rc_next;
	
	rc = jt->externalResources;
	while (rc != NULL)
	{
		rc_next = rc->next;
		/*
		 * There is no counter-part to enableResource().
		 * Just do an unsubscribeResource
		 */
		unsubscribeResource(rc);
		rc = rc_next;
	}
}

/*
 * Subscribe the given team to the given sorted array of
 * resourceSpecs, if everything goes right the result is an team with
 * the requested set of resources.  For each internal resource that
 * is defined, either the given resourceSpec must succeed or the
 * default resourceSpec creation must succeed.  Otherwise 0 is
 * returned to indicate an error...
 *
 * All internal resources in the rsArray are subscribed and enabled
 * before initializing any external resources.
 */
static
resourceSpec *subscribeAllInternalResources(jteam *jt, resourceSpec *rsArray[])
{
	resourceClient **rcArray = (resourceClient **)rsArray;
	resourceSpec *retval = 0;
	int lpc, did_visit = 0;

	/*
	 * Note that rcArray *is* rsArray, just with a different type.
	 * The specs are replaced, in turn with actual resourceClient
	 * objects.  All specs are replaced.
	 */

	require(jt != 0);
	require(rsArray != 0);
	require(currentTeam() == kernelTeam);
	
	for( lpc = 0; !retval && (lpc < internalResourceCount); lpc++ )
	{
		resource *res;

		res = metalTeam->internalResources[lpc]->handler;
		if( !(res->flags & RF_KERNEL_CLIENT) && !did_visit )
		{
			jteamReservation *jr;
			void *rpc, *fp;
			
			/* Assume the worst. */
			retval = defaultInternalSpec;
			getCallerFrameAndPC(&fp, &rpc);
			if( pushTeamReservation_raw(jt,
						    rpc,
						    fp,
						    false,
						    &jr) == JRV_OK )
			{
				if( visitReservedTeam_raw(
					jthread_current(),
					jthread_get_data(jthread_current()),
					jr, rpc, fp, 0) )
				{
					did_visit = 1;
					retval = 0;
				}
				else
				{
					popTeamReservation(jr);
				}
			}
		}
		if( retval )
		{
			/* XXX refactor me. */
		}
		else if( rsArray[lpc] )
		{
			resourceSpec *rs = rsArray[lpc];

			require(rs->handler == res);
			
			/* A spec was provided, try to subscribe with it */
			if( (rcArray[lpc] = subscribeResource(jt, rs)) )
			{
			}
			else
			{
				retval = rs;
			}
		}
		else if( (rcArray[lpc] = subscribeResourceDefault(jt, res)) )
		{
			/*
			 * No spec, but we managed to subscribe with the
			 * default spec.
			 */
		}
		else
		{
			retval = defaultInternalSpec;

			ensure(retval != 0);
		}
		if( retval )
		{
			teamLogEntry("jteam",
				     LEA_String, "Error",
				     "Unable to subscribe to resource", -1,
				     LEA_String, "team", jt->name, -1,
				     LEA_String, "resource",
				     metalTeam->internalResources[lpc]->
				     handler->name, -1,
				     LEA_TAG_DONE);
		}
	}
	if( !retval )
	{
		/* Enable them in reverse order */
		for( lpc = (internalResourceCount - 1); lpc >= 0; lpc-- )
		{
			int rc;

			rc = enableResource(rcArray[lpc]);

			assert(rc); /* XXX */
		}
		returnFromTeam();
	}
	else
	{
		int did_return = 0;

		if( (lpc > 0) &&
		    (metalTeam->internalResources[lpc - 1]->
		     handler->flags & RF_KERNEL_CLIENT) )
		{
			did_return = 1;
		}
		for( lpc--; lpc >= 0; lpc-- )
		{
			if( !did_return &&
			    (metalTeam->internalResources[lpc]->
			     handler->flags & RF_KERNEL_CLIENT) )
			{
				returnFromTeam();
				did_return = 1;
			}
			/* Back out any subscriptions */
			unsubscribeResource(rcArray[lpc]);
		}
	}
	
	ensure(currentTeam() == kernelTeam);
	return( retval );
}

/*
 * Subscribe to only the external resources defined in the jSpec.
 * On success all resources are subscribed, installed, and enabled.
 * If there is any failure, all external resources will be
 * unsubscribed.
 */
static resourceSpec *
subscribeAllExternalResources(jteam *jt, HArrayOfObject *jSpecs)
{
	resourceSpec *retval = 0;
	int lpc;
	
	require(jt != NULL);
	require(jt->externalResources == NULL); /* Cleanup code assumes this. */
	
	switchToTeam(jt, 0);

	for( lpc = 0;
	     jSpecs && (!retval) && (lpc < ARRAY_SIZE(jSpecs));
	     lpc++ )
	{
		struct Hedu_utah_janosvm_resources_ResourceSpec *curr;
		nativeClass *type;

		curr = ((struct Hedu_utah_janosvm_resources_ResourceSpec **)
			ARRAY_DATA(jSpecs))[lpc];
		
		if( curr && (type = curr->base.dtable->class->nativeclass) )
		{
			resourceSpec *rs;
			rs = OBJECT2NATIVE(type, &curr->base);
			
			require(rs != 0);
			require(rs->handler != 0);

			if (rs->handler->internalIndex == -1)
			{
				/* Subscribe+enable the resource. */
				if (attachTeamResource(jt, rs))
				{
					/* yeah */
				}
				else
				{
					/* Subscribe failed.  Bail. */
					retval = rs;
				}
			}
		}
		else
		{
			/*
			 * No native spec, don't care...  We already spit out
			 * a warning.
			 */
		}
	}

	if( retval )
	{
		/*
		 * Must cleanup resources that were successfully installed,
		 * that's just the jt->externalResources list.
		 */
		unsubscribeAllExternalResources(jt);
	}

	returnFromTeam();

	return retval;
}

struct threadWalkerObject {
	jteam *team;
};

static void teamThreadMigrater(void *arg, jthread_t th)
{
	struct threadWalkerObject *two = arg;
	threadData *td;
	
	require(arg != 0);
	require(th != 0);

	td = jthread_get_data(th);
	if( (metalTeam == td->ownerTeam) )
	{
		td->boundary->currentReservation->entry.id = two->team;
		td->currentTeam = two->team;
		td->ownerTeam = two->team;
	}
}

/*
 * Native part of the java constructor
 * 
 * Returns -1 if things are okay.  Returns >0 as index of
 * bad resource if there was a bad resource.  Also can throw
 * exceptions.  Returns size of array (one off the end)
 * if there was a problem with an internal resource...
 */
jint
edu_utah_janosvm_sys_TeamBackEnd_nativeConstructor(
	struct Hedu_utah_janosvm_sys_TeamBackEnd *rt,
	struct Hjava_lang_String *jName,
	HArrayOfObject *jSpecs,
	struct Hkaffe_util_Ptr *nSpecs)
{
	jteam *jt = OBJECT2NATIVE(&jteamNativeClass, &rt->base);
	jbool error = false;
	sparseElement *se;
	jint retval = -1;
	errorInfo einfo;
	int iLockRoot;

	require(rt != 0);
	require((currentTeam() == kernelTeam) ||
		(currentTeam() == metalTeam));

	jmutex_initialise(&jt->ghouls.lock);
	jcondvar_initialise(&jt->ghouls.cvar);
	jt->handler.funcs = &teamFunctions;
	jt->state = JTS_NEW;

	/*
	 * Reserve a slot in the TeamArray.  Do not point the se to
	 * the jteam until the team initialization is more or less
	 * complete.
	 */
	lockStaticMutex(&teamArrayLock);
	se = getSparseElement(teamArray);
	/* XXX check this return value. And GC if its null. */
	assert(se != NULL);
	se->data.ptr = NULL;
	unlockStaticMutex(&teamArrayLock);

	/* The internal resource pointer array is at the end of the jteam */
	jt->externalResources = NULL;

	if( !jName )
	{
		jt->name = defaultTeamName;
	}
	else if( !(jt->name = stringJava2C(jName)) )
	{
		postOutOfMemory(&einfo);
		error = true;
	}
	if( !error )
	{
		/*
		 * We do some visits during initialization, so the team
		 * becomes "live" at this point.
		 */
		jt->state = JTS_ALIVE;

		/* We must do a special initialization for the kernel team */
		if( currentTeam() == metalTeam )
		{
			int lpc;
			
			/* Just steal the metal team's resources */
			for( lpc = 0; lpc < internalResourceCount; lpc++ )
			{
				jt->internalResources[lpc] =
					metalTeam->internalResources[lpc];
				jt->internalResources[lpc]->owner = jt;
			}
			/* Migrate threads to the kernel */
			{
				struct threadWalkerObject two;

				two.team = jt;
				STOPWORLD();
				jthread_walk_jthreads(&two,
						      teamThreadMigrater);
				RESUMEWORLD();
			}
			/*
			 * Emulate the switch when allocating resource clients
			 */
			switchToTeam(jt, 0);
			/* Turn the GC on */
			enableResource(getTeamInternalResource(jt,
				&metalMemResource.link));
			returnFromTeam();
			jt->liveThreads += 1;
			registerTeamThread(jt, jthread_current(), 0);
		}
		else
		{
			require(currentTeam() == kernelTeam);

			addTeamDependency(jt, kernelTeam);
			/* Sort specs in internalResources, temporarily. */
			if( sortSpecs((resourceSpec **)jt->internalResources,
				      jSpecs,
				      (resourceSpec **)nSpecs) )
			{
				resourceSpec *rs;
				
				/*
				 * fixup internal resources (spec -> resource
				 * transformation)
				 */
				if( !(rs = subscribeAllInternalResources(
					jt,
					(resourceSpec **)jt->internalResources)
					    ) )
				{
					if( !(rs =
					      subscribeAllExternalResources(
						jt,
						jSpecs)) )
					{
						/*
						 * Okay, both internal and
						 * external succeeded.
						 */
					}
					else
					{
						/*
						 * External resource failed.
						 * Cleanup internals.
						 */
						unsubscribeAllInternalResources(jt);
					}
				}
				if( rs == 0 )
				{
					// all is good
				}
				else if( rs == defaultInternalSpec )
				{
					if( jSpecs )
						retval = ARRAY_SIZE(jSpecs);
					else
						retval = 0;
				}
				else if( jSpecs )
				{
					int lpc;

					retval = ARRAY_SIZE(jSpecs);
					for( lpc = 0;
					     lpc < ARRAY_SIZE(jSpecs);
					     lpc++ )
					{
						if( NATIVE2OBJECT(
							&rs->handler->
							specNativeClass,
							rs) ==
						    ((Hjava_lang_Object **)
						     ARRAY_DATA(jSpecs))[lpc] )
						{
							retval = lpc;
						}
					}
				}
				else
				{
					retval = 0;
				}
			}
			else
			{
				postException(
					&einfo,
					"edu.utah.janosvm.resources."
					"DuplicateResourceSpecException");
				error = true;
			}
		}
	}
	if( !error && (retval == -1) )
	{
		require(se);
		require(jt->state == JTS_ALIVE);
		require(se->data.ptr == NULL);

		/*
		 * Point the sparse element at the team, now that the
		 * team is ready.  And, put the team in the teamList.
		 */
		lockStaticMutex(&teamArrayLock);
		jt->elemID = se->id;
		se->data.ptr = jt;
		lnAddTail(&teamList, &jt->link);
		unlockStaticMutex(&teamArrayLock);
	}
	else
	{
		require(se);
		require(se->data.ptr == NULL);
		require((jt->state == JTS_ALIVE) || (jt->state == JTS_NEW));

		if( jt->name != defaultTeamName )
			gc_free(jt->name);
		jt->elemID = 0;
		jt->name = NULL;
		jt->state = JTS_BURIED;

		dropTeamDependencies(jt);
		
		lockStaticMutex(&teamArrayLock);
		dropSparseElement(teamArray, se);
		unlockStaticMutex(&teamArrayLock);

		if( retval == -1 )
			throwError(&einfo);
	}
	ensure(jteamInvariant(jt));
	return( retval );
}

#if !defined(UNIX_JTHREADS2)
/*
 * Function called by jthread_walk_threads to kill any threads belonging
 * to a dead team.
 */
static void teamThreadKiller(void *arg, jthread_t th)
{
	struct threadWalkerObject *tko = arg;

	require(arg != 0);
	require(th != 0);
	
	if( (tko->team == jthread_get_data(th)->ownerTeam) )
	{
		require(th != jthread_current());
		require(jthread_get_data(th)->visits == 0);
		require(jteamInvariant(tko->team));

		tko->team->zombieCount += jthread_slaughter(th);
	}
}
#endif

void waitTeam(jteam *jt, teamwaitconditions_t condition)
{
	int done = 0;
	
	require(jt != 0);
	require((currentTeam() == kernelTeam) || (currentTeam() == jt));
	require((condition == WAIT_FOR_TEAM_THREADS)
		|| (condition == WAIT_FOR_TEAM_ZOMBIES)
		|| (condition == WAIT_FOR_TEAM_BURIAL));

	jmutex_lock(&jt->ghouls.lock);
	while( !done )
	{
		switch( condition )
		{
		case WAIT_FOR_TEAM_THREADS:
			if( (jt->liveThreads == jt->daemonThreads) ||
			    (jt->state >= JTS_DYING) )
			{
				done = 1;
			}
			break;
		case WAIT_FOR_TEAM_ZOMBIES:
			if( jt->zombieCount == 0 )
				done = 1;
			break;
		case WAIT_FOR_TEAM_BURIAL:
			if( jt->state == JTS_BURIED )
				done = 1; 
			break;
		}
		if( !done )
		{
			/* We wait with a timeout to resist bugs in our code */
			jcondvar_wait(&jt->ghouls.cvar, &jt->ghouls.lock, 100);
		}
	}
	jmutex_unlock(&jt->ghouls.lock);
}

void signalTeam(jteam *jt, int condition)
{
	require(jt != 0);
	require((condition == SIGNAL_THREAD_DEATH)
		|| (condition == SIGNAL_TEAM_BURIAL));
	
	jthread_disable_stop();
	switch( condition )
	{
	case SIGNAL_THREAD_DEATH:
		jmutex_lock(&jt->ghouls.lock);
		if( jthread_get_data(jthread_current())->flags & TDF_ZOMBIE )
		{
			require(jt->zombieCount > 0);
			
			jt->zombieCount--;
			if( jt->zombieCount == 0 )
			{
				jcondvar_signal(&jt->ghouls.cvar,
						&jt->ghouls.lock);
			}
		}
		else
		{
			jcondvar_signal(&jt->ghouls.cvar, &jt->ghouls.lock);
		}
		jmutex_unlock(&jt->ghouls.lock);
		break;
	case SIGNAL_TEAM_BURIAL:
		jmutex_lock(&jt->ghouls.lock);

		require(jt->state == JTS_DEAD);
		jt->state = JTS_BURIED;

		jcondvar_signal(&jt->ghouls.cvar, &jt->ghouls.lock);
		jmutex_unlock(&jt->ghouls.lock);
		break;
	}
	jthread_enable_stop();
	if( !(jthread_get_data(jthread_current())->flags & TDF_ZOMBIE) &&
	    (condition == SIGNAL_THREAD_DEATH) &&
	    (jt->flags & JTF_NOTIFY_ON_FINISH) )
	{
		if( switchToTeam(kernelTeam, 0) )
		{
			do_execute_java_method(
				NATIVE2OBJECT(&jteamNativeClass, jt),
				0, 0, FinishTeamEventMethod, 0);
			returnFromTeam();
		}
	}
}

int allowTeamThread(jteam *jt)
{
	int retval = 1, iLockRoot;

	require(jt != 0);

	if( jt != metalTeam )
	{
		lockTeam(jt);
		require(jteamInvariant(jt));
		
		if( jt->state >= JTS_DYING )
			retval = 0;
		else
			jt->liveThreads++;
		
		ensure(jteamInvariant(jt));
		unlockTeam(jt);
	}
	return( retval );
}

int registerTeamThread(jteam *jt, jthread_t tid, int daemon)
{
	if( jt != metalTeam )
	{
		int iLockRoot;
		
		lockTeam(jt);
		require(jteamInvariant(jt));
#if 0
		jthread_get_data(tid)->teamlk = jt->threadList;
		jt->threadList = tid;
#endif
		// jt->liveThreads++;
		if( daemon )
			jt->daemonThreads++;
		
		ensure(jteamInvariant(jt));
		unlockTeam(jt);
	}
	return( 1 );
}

void reregisterTeamThread(jthread_t tid, int daemon)
{
	jteam *jt = jthread_get_data(tid)->ownerTeam;
	
	if( jt != metalTeam )
	{
		int iLockRoot;

		lockTeam(jt);
		require(jteamInvariant(jt));
		
		if( daemon )
			jt->daemonThreads++;
		else
			jt->daemonThreads--;
		
		ensure(jteamInvariant(jt));
		unlockTeam(jt);
		if( jt->liveThreads == jt->daemonThreads )
		{
			signalTeam(jt, SIGNAL_THREAD_DEATH);
		}
	}
}

void unregisterTeamThread(jteam *jt, jthread_t tid, int daemon)
{
	if( jt != metalTeam )
	{
		int iLockRoot;
		
		lockTeam(jt);
		require(jteamInvariant(jt));
		require(jt->liveThreads > 0);
		
		jt->liveThreads--;
		if( daemon )
		{
			require(jt->daemonThreads > 0);
			
			jt->daemonThreads--;
		}
#if 0
		{
			jthread_t curr, *prev;
			
			prev = &jt->threadList;
			curr = jt->threadList;
			while( curr != tid )
			{
				prev = &jthread_get_data(curr)->teamlk;
				curr = jthread_get_data(curr)->teamlk;
			}
			if( curr )
			{
				*prev = jthread_get_data(curr)->teamlk;
			}
			else
			{
				ensure(0);
			}
		}
#endif
		
		ensure(jteamInvariant(jt));
		unlockTeam(jt);
		if( jt->liveThreads == jt->daemonThreads )
		{
			signalTeam(jt, SIGNAL_THREAD_DEATH);
		}
	}
}

static
jint teamGetID(jteam *jt)
{
	require(jt);
	require(jteamInvariant(jt));

	return ( jt->elemID );
}


/**
 * Schedule the given team to be collected by the kernel reaper
 * thread.  This will return once the team is scheduled (it does not
 * wait for the collection to complete).
 *
 * The caller is responsible for ensuring that this is invoked at most
 * once for each team.
 */
static void scheduleCollectTeam(jteam *jt)
{
	jteam *oldTeam;
	
	require(jt);
	require(jt->guestThreads == 0);
	require(jt->absentThreads == 0);
	require(jt->dependents == 0);
	require(jteamInvariant(jt));
	require(jt->state == JTS_DEAD);
	
	/* Temporarily switch to the kernel team to kill it */
	/* XXX Find a better way to do this... */
	oldTeam = quickChangeTeamReservation(kernelTeam);
	do_execute_java_method(NATIVE2OBJECT(&jteamNativeClass, jt),
			       0, 0, DeadTeamEventMethod, 0);
	oldTeam = quickChangeTeamReservation(oldTeam);
	
	ensure(oldTeam == kernelTeam);
}

/**
 * Invoked by the kernel reaper thread.  Performs the collection
 * of a dead-and-buried team.
 */
static void collectNativeTeam(jteam *jt)
{
	int iLockRoot;

	require(jt != 0);
	require(jt != kernelTeam);
	require(currentTeam() == kernelTeam);
	require(jt->guestThreads == 0);
	require(jt->absentThreads == 0);

	/* XXX Break this monster up */

#if defined(UNIX_JTHREADS2)
	jskedDisable(teamJsked(jt));
#else
	{
		struct threadWalkerObject tko;
		
		tko.team = jt;

		/* Kill all the team's threads */
		lockObject(NATIVE2OBJECT(&jteamNativeClass, jt));
		jmutex_lock(&jt->ghouls.lock);
		jthread_walk_jthreads(&tko, teamThreadKiller);
		jmutex_unlock(&jt->ghouls.lock);
		unlockObject(NATIVE2OBJECT(&jteamNativeClass, jt));
	}
	/* Wait for zombies to exit */
	waitTeam(jt, WAIT_FOR_TEAM_ZOMBIES);
#endif
	
	jt->liveThreads = 0;
	jt->daemonThreads = 0;
	reclaimTeamLocks(jt);
	/*
	 * Switch to the team and drop the resources.  NOTE: we do this
	 * backwards since there's an implicit dependency between
	 * the resources (e.g. everything depends on memory so we
	 * unsubscribe it last, which is in the zeroth slot).
	 */
	quickChangeTeamReservation(jt);
	{
		extern void gcUnlockMe(void *);

		gcUnlockMe(teamCollector(jt));
	}

	unsubscribeAllExternalResources(jt);

	unsubscribeAllInternalResources(jt);

	/* Switch back to the kernel and drop any memory */
	// quickChangeTeamReservation(kernelTeam);
	dropTeamDependencies(jt);
	if( jt->name != defaultTeamName )
	{
		gc_free(jt->name);
	}
	jt->name = 0;

	signalTeam(jt, SIGNAL_TEAM_BURIAL);

	lockStaticMutex(&teamArrayLock);
	if( jt->elemID )
	{
		sparseElement *se;

		se = findSparseElement(teamArray, jt->elemID);
		dropSparseElement(teamArray, se);
		lnRemove(&jt->link);
	}
	unlockStaticMutex(&teamArrayLock);

	ensure(currentTeam() == kernelTeam);
}

void terminateTeam(jteam *jt)
{
	/* XXX cache this method lookup */
	do_execute_java_method(NATIVE2OBJECT(&jteamNativeClass, jt),
			       "terminate", "()V", 0, 0, 0);
}

static void teamThreadInterrupter(void *arg, jthread_t th)
{
	struct threadWalkerObject *tko = arg;

	require(arg != 0);
	require(th != 0);
	
	if( (tko->team == jthread_get_data(th)->ownerTeam) )
	{
	}
	else if( (tko->team == jthread_get_data(th)->currentTeam) )
	{
		/* Need to make sure they're awake */
		jthread_interrupt(th);
	}
}

void edu_utah_janosvm_sys_TeamBackEnd_staticNativeInit(struct Hedu_utah_janosvm_sys_EventQueue *eq, struct Hjava_lang_Thread *th)
{
	KernelEventQueue = eq;
	KernelEventThread = th;
}

void edu_utah_janosvm_sys_TeamBackEnd_nativeTerminating(struct Hedu_utah_janosvm_sys_TeamBackEnd *rt)
{
	jteam *jt = OBJECT2NATIVE(&jteamNativeClass, &rt->base);

	require(rt != 0);
	require(jt->state != JTS_REVOKING);

	jt->state = JTS_REVOKING;
}

static
void terminateNativeTeam(jteam *jt)
{
	int killDeps = 0, collect = 0;
	int iLockRoot;
	
	require(jt != 0);
	require(currentTeam() == kernelTeam);

	lockTeam(jt);
	require(jteamInvariant(jt));
	if (jt->state == JTS_REVOKING)
	{
		killDeps = 1;
		jt->state = JTS_DYING;
		if( (jt->guestThreads == 0) &&
		    (jt->absentThreads == 0) &&
		    (jt->dependents == 0) )
		{
			jt->state = JTS_DEAD;
			collect = 1;
		}
	}
	ensure(jteamInvariant(jt));
	unlockTeam(jt);
	if( killDeps )
		terminateTeamDependencies(jt);
	{
		struct threadWalkerObject tko;
		tko.team = jt;

		STOPWORLD();
		jthread_walk_jthreads(&tko, teamThreadInterrupter);
		RESUMEWORLD();
	}
	if( collect )
	{
		/* No one is visiting/away kill it */
		scheduleCollectTeam(jt);
	}
}

void edu_utah_janosvm_sys_TeamBackEnd_nativeCage(struct Hedu_utah_janosvm_sys_TeamBackEnd *rt)
{
	jteam *jt = OBJECT2NATIVE(&jteamNativeClass, &rt->base);

	require(rt != 0);
	
	terminateNativeTeam(jt);
}


jbool edu_utah_janosvm_sys_TeamBackEnd_owns(
	struct Hedu_utah_janosvm_sys_TeamBackEnd *rt,
	struct Hjava_lang_Object *obj)
{
	jteam *jt;
	jbool retval = false;

	if (obj == 0)
		return true;

	jt = OBJECT2NATIVE(&jteamNativeClass, &rt->base);
	if( jt == gc_get_resource(obj)->client.owner )
		retval = true;
	return retval;
}

jint edu_utah_janosvm_sys_TeamBackEnd_getID(
	struct Hedu_utah_janosvm_sys_TeamBackEnd *rt)
{
	jteam *jt;
	jint retval;

	require(rt);

	jt = OBJECT2NATIVE(&jteamNativeClass, &rt->base);
	require(jt);

	retval = teamGetID(jt);
	return( retval );
}

jint edu_utah_janosvm_sys_TeamBackEnd_getThreadCount(
	struct Hedu_utah_janosvm_sys_TeamBackEnd *rt)
{
	jteam *jt;
	jint retval;

	require(rt);

	jt = OBJECT2NATIVE(&jteamNativeClass, &rt->base);
	require(jt);

	retval = jt->liveThreads;
	return( retval );
}

jint edu_utah_janosvm_sys_TeamBackEnd_getDaemonCount(
	struct Hedu_utah_janosvm_sys_TeamBackEnd *rt)
{
	jteam *jt;
	jint retval;

	require(rt);

	jt = OBJECT2NATIVE(&jteamNativeClass, &rt->base);
	require(jt);

	retval = jt->daemonThreads;
	return( retval );
}

void edu_utah_janosvm_sys_TeamBackEnd_setDescription(
	struct Hedu_utah_janosvm_sys_TeamBackEnd *rt,
	struct Hjava_lang_String *jdesc)
{
	char *desc;
	jteam *jt;

	require(rt);

	jt = OBJECT2NATIVE(&jteamNativeClass, &rt->base);
	require(jt);

	if( (desc = stringJava2C(jdesc)) )
	{
		setTeamDescription(jt, desc);
		gc_free(desc);
	}
}

void edu_utah_janosvm_sys_TeamBackEnd_anchorObject(
	struct Hedu_utah_janosvm_sys_TeamBackEnd *rt,
	struct Hjava_lang_Object *obj)
{
	jteam *jt = OBJECT2NATIVE(&jteamNativeClass, &rt->base);

	require(rt != 0);
	require(obj != 0);
	
	GC_add_ref((Collector *)teamCollector(jt), obj);
}

/*
 * Remove reservation on home team.
 */
static
void releaseHomeTeam(jteam *jt)
{
	jthread_t th = jthread_current();
	jteam *owner = jthread_get_data(th)->ownerTeam;

	require(jt != 0);

	/* Only need to do something if we're in the home team */
	if( jt == owner )
	{
		threadData *td = jthread_get_data(th);
		int iLockRoot, terminate = 0;
		
		lockTeam(jt);
		require(jteamInvariant(jt));
		require(jt->absentThreads > 0);
		jt->absentThreads--;
		/* Check if we need to kill the team */
		if( (jt->state == JTS_DYING) &&
		    (td->visits == 0) &&
		    (jt->absentThreads == 0) &&
		    (jt->guestThreads == 0) &&
		    (jt->dependents == 0) )
		{
			jt->state = JTS_DEAD;
			terminate = 1;
		}
		ensure(jteamInvariant(jt));
		unlockTeam(jt);
		if( terminate )
		{
			scheduleCollectTeam(jt);
		}
	}
}

team_error_t reserveTeam(jteam *jt)
{
	jthread_t th = jthread_current();
	jteam *curr = jthread_get_data(th)->currentTeam;
	team_error_t retval;

	require(jt != 0);
	require(jt != metalTeam);

	if( jt != curr )
	{
		int iLockRoot;
		
		lockTeam(curr);
		require(jteamInvariant(curr));
		/* Check to make sure we can leave this team */
		if( (curr->state == JTS_ALIVE) ||
		    (curr->state == JTS_REVOKING) ||
		    (th == reaperThread) )
		{
			/* Lock down the home team */
			if( curr == jthread_get_data(th)->ownerTeam )
				curr->absentThreads++;
			retval = TEAM_ERROR_NONE;
		}
		else
		{
			retval = TEAM_ERROR_NO_SOURCE;
		}
		ensure(jteamInvariant(curr));
		unlockTeam(curr);
		
		if( retval == TEAM_ERROR_NONE )
		{
			lockTeam(jt);
			require(jteamInvariant(jt));
			/* Make sure we can enter the other team */
			if( (jt->state == JTS_ALIVE) ||
			    (jt->state == JTS_REVOKING) ||
			    (th == reaperThread) )
			{
				jt->guestThreads++;
			}
			else
			{
				retval = TEAM_ERROR_NO_DESTINATION;
			}
			ensure(jteamInvariant(jt));
			unlockTeam(jt);
			/*
			 * Release the home team if we fail to reserve the
			 * other team.
			 */
			if( retval != TEAM_ERROR_NONE )
				releaseHomeTeam(curr);
		}
	}
	else
	{
		/* Reserving the current team, nothing to do */
		retval = TEAM_ERROR_NONE;
	}
	return( retval );
}

void unreserveTeam(jteam *jt)
{
	jthread_t th = jthread_current();
	threadData *td = jthread_get_data(th);
	jteam *curr = td->currentTeam;

	require(jt != 0);
	require(jt != metalTeam);

	if( jt != curr )
	{
		int iLockRoot, terminate = 0;

		lockTeam(jt);
		require(jteamInvariant(jt));
		require(jt->guestThreads > 0);
		jt->guestThreads--;
		/* Check to see if the team needs to be killed */
		if( (jt->state == JTS_DYING) &&
		    (jt->guestThreads == 0) &&
		    (jt->absentThreads == 0) &&
		    (jt->dependents == 0) )
		{
			jt->state = JTS_DEAD;
			terminate = 1;
		}
		ensure(jteamInvariant(jt));
		unlockTeam(jt);
		if( terminate )
		{
			scheduleCollectTeam(jt);
		}
		releaseHomeTeam(curr);

		ensure(jteamInvariant(jt));
	}
}

int switchToTeam(jteam *dest, int (*catastrophe)(void *eobj))
{
	jteamReservation *jr;
	int retval = 0;
	void *rpc, *fp;
	
	getCallerFrameAndPC(&fp, &rpc);
	if( pushTeamReservation_raw(dest, rpc, fp, false, &jr) == JRV_OK )
	{
		if( visitReservedTeam_raw(jthread_current(),
					  jthread_get_data(jthread_current()),
					  jr, rpc, fp, catastrophe) )
		{
			retval = 1;
		}
		else
		{
			popTeamReservation(jr);
		}
	}
	return( retval );
}

void returnFromTeam(void)
{
	leaveReservedTeam();
	popTeamReservation(0);
}

static
int javaCatastropheRecovery(void *eobj)
{
	errorInfo einfo;

	/* XXX Need to clear this out just in case. */
	getCurrentThread()->holder->obj = 0;
	returnFromTeam();
	postExceptionMessage(&einfo, TEAM_EJECTION_EXCEPTION_S,
			     "Ejection from team during visit");
	throwError(&einfo);
	return( 0 );
}

int nullCatastropheRecovery(void *eobj)
{
	returnFromTeam();
	return( 0 );
}

jbool edu_utah_janosvm_sys_TeamBackEnd_switchTo(
	struct Hedu_utah_janosvm_sys_TeamBackEnd *rt,
	int wrapCount)
{
	jteam *jt = OBJECT2NATIVE(&jteamNativeClass, &rt->base);
	jteamReservation *jr;
	jbool retval = 0;
	void *rpc, *fp;
	int lpc;
	
	getCallerFrameAndPC(&fp, &rpc);
	for( lpc = 0; lpc < (wrapCount + 1); lpc++ )
	{
	    rpc = getFrameReturnPC(fp);
	    fp = getPreviousFramePointer(fp);
	}
	if( pushTeamReservation_raw(jt, rpc, fp, true, &jr) == JRV_OK )
	{
		if( visitReservedTeam_raw(jthread_current(),
					  jthread_get_data(jthread_current()),
					  jr,
					  rpc,
					  fp,
					  javaCatastropheRecovery) )
		{
			retval = 1;
		}
		else
		{
			popTeamReservation(jr);
		}
	}
	return( retval );
}

void edu_utah_janosvm_sys_TeamBackEnd_returnFrom(void)
{
	returnFromTeam();
}

void edu_utah_janosvm_sys_VMInstance_crash(void)
{
	crashIntoTeamBoundary(0);
}

struct Hedu_utah_janosvm_sys_TeamBackEnd *
edu_utah_janosvm_sys_TeamBackEnd_getObjectOwner(struct Hjava_lang_Object *obj)
{
	struct Hedu_utah_janosvm_sys_TeamBackEnd *retval = 0;
	memResourceClient *mrc;

	if( (mrc = gc_get_resource(obj)) )
	{
		retval = (struct Hedu_utah_janosvm_sys_TeamBackEnd *)
			NATIVE2OBJECT(&jteamNativeClass,
				      mrc->client.owner);
	}
	return( retval );
}

struct Hedu_utah_janosvm_sys_TeamBackEnd *
edu_utah_janosvm_sys_TeamBackEnd_getCurrentTeam(void)
{
	jthread_t th = jthread_current();

	return( (Hedu_utah_janosvm_sys_TeamBackEnd *)
		NATIVE2OBJECT(&jteamNativeClass,
			      jthread_get_data(th)->currentTeam) );
}

void
edu_utah_janosvm_sys_TeamBackEnd_setClassLocal(
	struct Hedu_utah_janosvm_sys_TeamBackEnd *rt,
	struct Hjava_lang_Class *clazz,
	struct Hjava_lang_Object *obj)
{
	jteam *jt = OBJECT2NATIVE(&jteamNativeClass, &rt->base);
	
	setClassLocalData(jt, clazz, obj);
}

Hjava_lang_Object *
edu_utah_janosvm_sys_TeamBackEnd_getClassLocal(
	struct Hedu_utah_janosvm_sys_TeamBackEnd *rt,
	struct Hjava_lang_Class *clazz)
{
	jteam *jt = OBJECT2NATIVE(&jteamNativeClass, &rt->base);
	
	return( getClassLocalData(jt, clazz) );
}

void
edu_utah_janosvm_sys_TeamBackEnd_setCurrentClassLocal(
	struct Hjava_lang_Class *clazz,
	struct Hjava_lang_Object *obj)
{
	setClassLocalData(0, clazz, obj);
}

Hjava_lang_Object *
edu_utah_janosvm_sys_TeamBackEnd_getCurrentClassLocal(
	struct Hjava_lang_Class *clazz)
{
	return( getClassLocalData(0, clazz) );
}

struct Hedu_utah_janosvm_sys_TeamReservation *
edu_utah_janosvm_sys_TeamBackEnd_pushReservation(
	struct Hedu_utah_janosvm_sys_TeamBackEnd *rt,
	jint wrapCount)
{
	struct Hedu_utah_janosvm_sys_TeamReservation *retval = 0;
	
	if( rt )
	{
		jteam *jt = OBJECT2NATIVE(&jteamNativeClass, &rt->base);
		jteamReservation *jr;
		errorInfo einfo;
		void *rpc, *fp;
		int lpc;
		
		getCallerFrameAndPC(&fp, &rpc);
		for( lpc = 0; lpc < (wrapCount + 1); lpc++ )
		{
			rpc = getFrameReturnPC(fp);
			fp = getPreviousFramePointer(fp);
		}
		switch( pushTeamReservation_raw(jt, rpc, fp, true, &jr) )
		{
		case JRV_OK:
			retval = (struct Hedu_utah_janosvm_sys_TeamReservation *)
				&jr->entry.base;
			break;
		case JRV_RESERVE_FAILED:
			postException(&einfo, DEAD_TEAM_EXCEPTION_S);
			throwError(&einfo);
			break;
		case JRV_OVERFLOW:
			postException(&einfo, RESERVATION_OVERFLOW_ERROR_S);
			throwError(&einfo);
			break;
		}
	}
	return( retval );
}

void edu_utah_janosvm_sys_TeamBackEnd_popReservation(struct Hedu_utah_janosvm_sys_TeamReservation* tr)
{
	popTeamReservation((jteamReservation *)tr);
}

jbool edu_utah_janosvm_sys_TeamBackEnd_addDependency(
	struct Hedu_utah_janosvm_sys_TeamBackEnd *rt,
	struct Hedu_utah_janosvm_sys_TeamBackEnd *javaDep)
{
	jbool retval = 0;
	
	if( javaDep )
	{
		jteam *jt = OBJECT2NATIVE(&jteamNativeClass, &rt->base);
		jteam *dep = OBJECT2NATIVE(&jteamNativeClass, &javaDep->base);
		
		if( dep && switchToTeam(kernelTeam, 0) )
		{
			retval = addTeamDependency(jt, dep);
			returnFromTeam();
		}
	}
	return( retval );
}

void edu_utah_janosvm_sys_TeamBackEnd_notifyOnFinish(
	struct Hedu_utah_janosvm_sys_TeamBackEnd *rt)
{
	jteam *jt = OBJECT2NATIVE(&jteamNativeClass, &rt->base);

	jt->flags |= JTF_NOTIFY_ON_FINISH;
}

void edu_utah_janosvm_sys_TeamBackEnd_doTheReap(
	struct Hedu_utah_janosvm_sys_TeamBackEnd *rt)
{
	jteam *jt = OBJECT2NATIVE(&jteamNativeClass, &rt->base);

	require(jthread_current());
	
	reaperThread = jthread_current();
	
	require(jt->state == JTS_DEAD);
	require(jteamInvariant(jt));
	
#if 0
	rt.reaperCollect();
	TeamBackEnd.teamList.event(null, id, TeamEvent.BURIED);
#else
	collectNativeTeam(jt);
	
	/*
	 * Remove any references to the team object, we do this here
	 * since there are no other pointers to the java object, just
	 * the internal jteam pointer.
	 */
	gc_rm_ref(NATIVE2OBJECT(&jteamNativeClass, jt));
#endif
}

static void installTeamResource(jteam *jt, resourceClient *rc)
{
	require(jt != 0);
	require(rc != 0);
	require(rc->handler != 0);
	
	rc->owner = jt;
	if( rc->handler->internalIndex == -1 )
	{
		rc->next = jt->externalResources;
		jt->externalResources = rc;
	}
	else
	{
		jt->internalResources[rc->handler->internalIndex] = rc;
	}
}

int attachTeamResource(jteam *jt, resourceSpec *rs)
{
	resourceClient *rc;
	int retval = 0;

	require(jt != 0);
	require(rs != 0);
	require(currentTeam() == jt);
	
	if( (rc = subscribeResource(jt, rs)) )
	{
		installTeamResource(jt, rc);
		retval = 1;
	}
	return( retval );
}

void detachTeamResource(jteam *jt, resource *res, int id)
{
	require(jt != 0);
	require(res != 0);
	
	if( res->internalIndex == -1 )
	{
		resourceClient *curr, **prev;
		int done = 0;

		prev = &jt->externalResources;
		curr = jt->externalResources;
		while( curr && !done )
		{
			if( (curr->handler == res) && (curr->id == id) )
			{
				*prev = curr->next;
				curr->next = 0;
				unsubscribeResource(curr);
				done = 1;
			}
			prev = &curr->next;
			curr = curr->next;
		}
	}
	else
	{
		unsubscribeResource(jt->internalResources[res->internalIndex]);
		jt->internalResources[res->internalIndex] = 0;
	}
}

void grabTeam(jteam *jt)
{
	require(jt);
	// require(currentTeam() == kernelTeam);
	
	GC_add_ref(main_collector, NATIVE2OBJECT(&jteamNativeClass, jt));
}

jteam *grabTeamWithID(sparseID_t tid)
{
	sparseElement *se;
	jteam *retval = 0;
	int iLockRoot;

	// require(currentTeam() == kernelTeam);
	
	lockStaticMutex(&teamArrayLock);
	if( (se = findSparseElement(teamArray, tid)) )
	{
		retval = (jteam *)se->data.ptr;
		if( retval )
		{
			grabTeam(retval);
		}
	}
	unlockStaticMutex(&teamArrayLock);
	return( retval );
}

void dropTeam(jteam *jt)
{
	int rc;

	// require(currentTeam() == kernelTeam);
	
	rc = GC_rm_ref(main_collector, NATIVE2OBJECT(&jteamNativeClass, jt));

	ensure(rc);
}

jteam *currentTeam(void)
{
	jteam *retval;

	require(jthread_current() != 0);
	
	retval = jthread_get_data(jthread_current())->currentTeam;

	ensure(retval != 0);
	return( retval );
}
