/*
 * logResource.c
 * Resource to managed standard I/O for each team.
 *
 * Copyright (c) 2001, 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

#include "assert_pp.h"

#include "logResource.h"
#include "memResource.h"
#include "jteam.h"
#include "soft.h"
#include "nativeClass.h"
#include "nativeClass-inline.h"
#include "mungWallProxy.h"
#include "scripting.h"
#include "module.h"
#include "moduleList.h"

#include <misc/stringParsing.h>

#include "edu_utah_janosvm_resources_LogSpec.h"
#include "edu_utah_janosvm_resources_LogResource.h"

#define JAVA_LOG_SPEC "edu/utah/janosvm/resources/LogSpec"

/** Module declarations **/
static int initLogResourceModule(int stage);

module_t logResourceModule = {
	"logResource",
	0,
	initLogResourceModule,
	0,
	0
};

/** LogResource declarations **/
static int logSubscribe(struct _resource *res,
			struct _resourceClient *rc,
			struct _resourceSpec *rs);
static void logUnsubscribe(struct _resource *res,
			   struct _resourceClient *rc);
static resourceSpec *logCreateJSISpec(struct _resource *res, jsiMessage *jm);
static void logDeleteJSISpec(struct _resource *res,
			     struct _resourceSpec *rs);

static resourceFunctions logResourceFunctions = {
	0,		/* install */
	0,		/* uninstall */
	0,		/* configure */
	0,		/* fillSpec */
	logCreateJSISpec,
	logDeleteJSISpec,
	0,		/* defaultSpec */
	logSubscribe,
	0,		/* enable */
	logUnsubscribe
};

logResource metalLogResource;

static logSpec defaultSpec = {
	{ &metalLogResource.link },
	0
};

/** Module Definitions **/
static
void logAtExit(void)
{
	logClient *lc;
	jteam *jt;

	if( kernelTeam )
	{
		jt = kernelTeam;
	}
	else
	{
		jt = metalTeam;
	}
	if( (lc = (logClient *)
	     getTeamInternalResource(jt, &metalLogResource.link)) )
	{
		flushLogger(&lc->log);
	}
}

int initLogResourceModule(int stage)
{
	int retval = 0;

	switch( stage )
	{
	case SID_LOGRESOURCEMODULE_INITMETAL:
		installResource(&metalLogResource.link,
				RA_Name, "log",
				RA_Flags, RF_INTERNAL,
				RA_Functions, &logResourceFunctions,
				RA_ClientSize, sizeof(logClient),
				RA_SpecSize, sizeof(logSpec),
				RA_TAG_DONE);
		if( metalLogResource.defaultLogFile == NULL )
		{
			metalLogResource.defaultLogFile = stderr;
		}
		retval = attachTeamResource(metalTeam, &defaultSpec.spec);
		atexit(logAtExit);
		break;
	case SID_LOGRESOURCEMODULE:
		installResource(&metalLogResource.link,
				RA_SpecNativeClass, JAVA_LOG_SPEC,
				RA_TAG_DONE);
		retval = 1;
		break;
	}
	return( retval );
}

static
void defaultTeamFlushFunction(void *arg,
			      struct _logger *l,
			      struct _logEntry *le,
			      struct _logEntryAttribute **lea)
{
	logClient *lc = (logClient *)arg;
	
	if( le->classification && lc->output )
	{
		int lpc;
		
		fprintf(lc->output,
			"%s:%s:%d:%s[%ld.%ld] -",
			l->name,
			le->sourceFile,
			le->sourceLine,
			le->classification,
			(long)(le->recordTime.tv_sec),
			(long)(le->recordTime.tv_usec));
		for( lpc = 0; lpc < le->attributeCount; lpc++ )
		{
			if( lpc > 0 )
				fprintf(lc->output, ",");
			fprintf(lc->output, " %s:%s=",
				lea[lpc]->id,
				loggerAttributeNames[lea[lpc]->type]);
			switch( lea[lpc]->type )
			{
			case LEA_Boolean:
				fprintf(lc->output, "%s",
					lea[lpc]->value.b ?
					"true" : "false");
				break;
			case LEA_Character:
				fprintf(lc->output, "%c", lea[lpc]->value.c);
				break;
			case LEA_Integer:
				fprintf(lc->output, "%d", lea[lpc]->value.i);
				break;
			case LEA_LongLong:
				fprintf(lc->output, "%qd", lea[lpc]->value.ll);
				break;
			case LEA_Float:
				fprintf(lc->output, "%f", lea[lpc]->value.f);
				break;
			case LEA_String:
				fprintf(lc->output, "%s", lea[lpc]->value.str);
				break;
			case LEA_Pointer:
				fprintf(lc->output, "%p", lea[lpc]->value.ptr);
				break;
			default:
				assert(0);
				break;
			}
		}
		fprintf(lc->output, "\n");
	}
}

int logSubscribe(struct _resource *res,
		 struct _resourceClient *rc,
		 struct _resourceSpec *rs)
{
	logClient *lc = (logClient *)rc;
	logSpec *ls = (logSpec *)rs;
	char *fileName = 0;
	int retval = 0;
	FILE *file = 0;

	if( ls && ls->fileName )
	{
		if( ls->fileName[0] &&
		    (fileName = gc_malloc(strlen(ls->fileName) + 1,
					  GC_ALLOC_TEMP_STRING)) )
		{
			strcpy(fileName, ls->fileName);
			file = fopen(ls->fileName, "a");
		}
	}
	else
	{
		file = metalLogResource.defaultLogFile;
	}
	if( (file || (ls->fileName[0] == 0)) &&
	    initLogger(&lc->log,
		       LA_Name, rc->owner->name,
		       LA_FlushFunction, defaultTeamFlushFunction,
		       LA_FlushFunctionArg, lc,
		       LA_TAG_DONE) )
	{
		lc->fileName = fileName;
		lc->output = file;
		retval = 1;
	}
	else
	{
		if( file )
			fclose(file);
		if( fileName )
			gc_free(fileName);
	}
	return( retval );
}

void logUnsubscribe(struct _resource *res,
		    struct _resourceClient *rc)
{
	logClient *lc = (logClient *)rc;

	killLogger(&lc->log);
	if( lc->output && (lc->output != metalLogResource.defaultLogFile) )
	{
		fclose(lc->output);
	}
}

resourceSpec *logCreateJSISpec(struct _resource *res, jsiMessage *jm)
{
	resourceSpec *retval = 0;

	if( (retval = createResourceSpec(res)) )
	{
		logSpec *ls = (logSpec *)retval;

		/*
		 * Copy the default values in case they aren't specified in
		 * the message.
		 */
		*ls = defaultSpec;
		/*
		 * The message lives longer than the spec so we don't have to
		 * copy them out.
		 */
		jsiMessageGet(jm,
			      JMA_String, "log", 0, &ls->fileName,
			      JMA_TAG_DONE);
	}
	return( retval );
}

void logDeleteJSISpec(struct _resource *res, struct _resourceSpec *rs)
{
	deleteResourceSpec(rs);
}

void teamLog_real(char *srcFile, int srcLine, char *cl, char *msg)
{
	logClient *lc = (logClient *)
		getTeamInternalResource(0, &metalLogResource.link);
	logger *l = &lc->log;
	int iLockRoot;
	
	if( lc )
	{
		jthread_disable_stop();
		lockMutex(lc);
		addLogMessage_real(l, srcFile, srcLine, cl, msg);
		unlockMutex(lc);
		jthread_enable_stop();
	}
}

void teamLogEntry_real(char *srcFile, int srcLine, char *cl, ...)
{
	logClient *lc = (logClient *)
		getTeamInternalResource(0, &metalLogResource.link);
	logger *l = &lc->log;
	va_list args;
	int iLockRoot;

	if( lc )
	{
		jthread_disable_stop();
		lockMutex(lc);
		va_start(args, cl);
		addLogEntryV(l, srcFile, srcLine, cl, args);
		va_end(args);
		unlockMutex(lc);
		jthread_enable_stop();
	}
}

void addTeamLogAnalyzer(logAnalyzerMask *lam)
{
	logClient *lc = (logClient *)
		getTeamInternalResource(0, &metalLogResource.link);
	logger *l = &lc->log;
	int iLockRoot;

	jthread_disable_stop();
	lockMutex(lc);
	addLogAnalyzer(l, lam);
	unlockMutex(lc);
	jthread_enable_stop();
}

void remTeamLogAnalyzer(logAnalyzerMask *lam)
{
	logClient *lc = (logClient *)
		getTeamInternalResource(0, &metalLogResource.link);
	logger *l = &lc->log;
	int iLockRoot;

	jthread_disable_stop();
	lockMutex(lc);
	remLogAnalyzer(l, lam);
	unlockMutex(lc);
	jthread_enable_stop();
}

void edu_utah_janosvm_resources_LogSpec_nativeConstructor(
	struct Hedu_utah_janosvm_resources_LogSpec *jls,
	struct Hjava_lang_String *jFileName)
{
	logSpec *ls = OBJECT2NATIVE(&metalLogResource.link.specNativeClass,
				    (Hjava_lang_Object *)jls);
	errorInfo einfo;
	int error = 0;

	require(jFileName != 0);

	einfo.type = 0;
	jls->found = 0;
	ls->spec.handler = &metalLogResource.link;
	if( !(ls->fileName = stringJava2C(jFileName)) )
		error = 1;
	if( error )
	{
		/* The finalizer should free them... right? */
		if( einfo.type == 0 )
			postOutOfMemory(&einfo);
		throwError(&einfo);
	}
}

void edu_utah_janosvm_resources_LogSpec_nativeFinalize(
	struct Hedu_utah_janosvm_resources_LogSpec *jls)
{
	logSpec *ls = OBJECT2NATIVE(&metalLogResource.link.specNativeClass,
				    (Hjava_lang_Object *)jls);

	if( !jls->found )
	{
		gc_free(ls->fileName);
		ls->fileName = 0;
	}
}

struct Hjava_lang_String *
edu_utah_janosvm_resources_LogSpec_getFileName(
	struct Hedu_utah_janosvm_resources_LogSpec *jls)
{
	logSpec *ls = OBJECT2NATIVE(&metalLogResource.link.specNativeClass,
				    (Hjava_lang_Object *)jls);
	struct Hjava_lang_String *retval;

	if( !(retval = stringC2Java(ls->fileName)) )
	{
		errorInfo einfo;
		
		postOutOfMemory(&einfo);
		throwError(&einfo);
	}
	return( retval );
}

struct Hedu_utah_janosvm_resources_LogSpec *
edu_utah_janosvm_resources_LogSpec_current(void)
{
	struct Hedu_utah_janosvm_resources_LogSpec *retval = 0;
	logClient *lc = (logClient *)
		getTeamInternalResource(0, &metalLogResource.link);
	logSpec *ls;
	int error = 0;

	retval = (struct Hedu_utah_janosvm_resources_LogSpec *)
		newObject(metalLogResource.link.specNativeClass.peer);
	ls = OBJECT2NATIVE(&metalLogResource.link.specNativeClass,
			   (Hjava_lang_Object *)retval);
	ls->spec.handler = &metalLogResource.link;
	if( lc->output )
	{
		if( (ls->fileName = gc_malloc(strlen(lc->fileName) + 1,
					      GC_ALLOC_TEMP_STRING)) )
		{
			strcpy(ls->fileName, lc->fileName);
		}
		else
		{
			error = 1;
		}
	}
	else if( lc->output == 0 )
	{
		if( !(ls->fileName = gc_malloc(1, GC_ALLOC_TEMP_STRING)) )
			error = 1;
	}
	if( error )
	{
		errorInfo einfo;
		
		postOutOfMemory(&einfo);
		throwError(&einfo);
	}
	return( retval );
}
