/*
 * nameSpace.h
 * Manage the root namespace for a team
 *
 * Copyright (c) 2000, 2001, 2002, 2003 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

#include "gtypes.h"
#include "support.h"
#include "locks.h"
#include "stringSupport.h"
#include "jteam.h"
#include "jteamReservation.h"
#include "nameSpace.h"
#include "thread.h"
#include "nativeClass.h"
#include "classPath.h"
#include "readClass.h"
#include "itypes.h"
#include "stdioResource.h"
#include "assert_pp.h"
#include "soft.h"
#include "nativeClass-inline.h"
#include "threadData.h"
#include "threadData-inline.h"
#include "mungWallProxy.h"
#include "machine.h"
#include "logResource.h"
#include "module.h"
#include "moduleList.h"
#include "access.h"

#include <misc/stringParsing.h>
#include <jsi/jsi.h>

#include "edu_utah_janosvm_sys_Importable.h"
#include "edu_utah_janosvm_sys_PortAuthority.h"

#include "edu_utah_janosvm_resources_NameSpaceGroup.h"
#include "edu_utah_janosvm_resources_NameSpaceResource.h"
#include "edu_utah_janosvm_resources_NameSpaceSpec.h"

#define JAVA_NAME_SPACE_SPEC "edu/utah/janosvm/resources/NameSpaceSpec"

struct Hedu_utah_janosvm_sys_Importable;
nativeClass importNativeClass;
Hjava_lang_Class *ImportableClass;
Hjava_lang_Class *ExportableClass;
Hjava_lang_Class *ListClass;
Hjava_lang_Class *ListNodeClass;
Hjava_lang_Class *ObjectHolderClass;
Hjava_lang_Class *CrossTeamCloneableClass;
Hjava_lang_Class *NameSpaceGroupClass;

extern Method *FindTeamMethod;

extern Hjava_lang_Class *TeamClass;

nameSpaceResource metalNameSpaceResource;

Utf8Const *utfKernel;
Utf8Const *utfIdentity;

/** classGroup sectionType **/
typedef struct _classGroupSectionData {
	tagSectionData tsd;
	classGroupType *cgt;
} classGroupSectionData;

static sectionTypeFunctions nsClassGroupSectionFunctions;

static tagSectionType nsClassGroupSection;

int nsClassGroupSectionWalker(struct _sectionType *st,
			      struct _sectionFileData *sfd,
			      struct _sectionFile *sf)
{
	Utf8Const *utfName;
	int retval = 0;

	if( (utfName = utf8ConstNew(childSectionName(sfd), -1)) )
	{
		classGroupSectionData *cgsd =
			LOCAL_SECTION_DATA(classGroupSectionData *, st, sfd);
		
		/* Allocate the group type and add it to the resource */
		if( (cgsd->cgt = createClassGroupType(utfName,
						      cgsd->tsd.taglist)) )
		{
			cgsd->cgt->next = metalNameSpaceResource.types;
			metalNameSpaceResource.types = cgsd->cgt;
			retval = 1;
		}
		else
		{
			utf8ConstRelease(utfName);
		}
	}
	return( retval );
}

/** viewGroup sectionType **/
static sectionTypeFunctions nsViewGroupSectionFunctions;

static tagSectionType nsViewGroupSection;

int nsViewGroupSectionWalker(struct _sectionType *st,
			     struct _sectionFileData *sfd,
			     struct _sectionFile *sf)
{
	Utf8Const *utfName;
	int retval = 0;

	if( (utfName = utf8ConstNew(childSectionName(sfd), -1)) )
	{
		tagSectionData *tsd = LOCAL_SECTION_DATA(tagSectionData *,
							 st,
							 sfd);
		classGroupSectionData *cgsd;
		sectionFileData *par_sfd;

		/* Find the parent section, should be a classGroup */
		par_sfd = getSectionParent(sf, sfd);
		if( par_sfd && (par_sfd->type == &nsClassGroupSection.type) )
		{
			viewGroupType *vgt;
			
			cgsd = LOCAL_SECTION_DATA(classGroupSectionData *,
						  &nsClassGroupSection,
						  par_sfd);

			/* Add the view group to this class group */
			if( (vgt = createViewGroupType(utfName,
						       tsd->taglist)) )
			{
				vgt->next = cgsd->cgt->views;
				cgsd->cgt->views = vgt;
				retval = 1;
			}
			else
			{
				retval = 0;
			}
		}
		else
		{
			retval = 0;
		}
		if( !retval )
			utf8ConstRelease(utfName);
	}
	return( retval );
}

/** nameSpace helper functions **/

/* Add a classGroup/viewGroup pair to a nameSpaceSpec */
static
int subscribeView(nameSpaceSpec *nss,
		  sparseID_t id, Utf8Const *cgName, Utf8Const *vgName)
{
	groupNames *new_data;
	int retval = 0;

	if( (new_data = gc_realloc(nss->viewSubscriptions.data,
				   (nss->viewSubscriptions.size + 1) *
				   sizeof(groupNames),
				   GC_ALLOC_NAME_SPACE)) )
	{
		int slot;
		
		nss->viewSubscriptions.data = new_data;
		slot = nss->viewSubscriptions.size++;
		nss->viewSubscriptions.data[slot].teamID = id;
		nss->viewSubscriptions.data[slot].className = cgName;
		nss->viewSubscriptions.data[slot].viewName = vgName;
		retval = 1;

		ensure_valid_memory(new_data);
	}
	return( retval );
}

/* Add a classGroup/viewGroup pair to a nameSpaceClient */
static
int subscribeViewClient(nameSpaceClient *nsc, viewGroup *vg)
{
	viewSubscriberPool **last_pool, *pool;
	int retval = 0;
	int iLockRoot;

	lockMutex(nsc);
	last_pool = &nsc->views;
	pool = nsc->views;
	while( pool )
	{
		int lpc;

		for( lpc = 0; lpc < pool->used; lpc++ )
		{
			if( pool->data[lpc].vGroup == vg )
			{
				/* Already subscribed. */
				unlockMutex(nsc);
				return( 1 );
			}
		}
		if( pool->used < VIEW_SUBSCRIBER_POOL_SIZE )
		{
			break;
		}
		last_pool = &pool->next;
		pool = pool->next;
	}
	if( !pool )
	{
		pool = gc_malloc(sizeof(viewSubscriberPool),
				 GC_ALLOC_NAME_SPACE);
		*last_pool = pool;
	}
	if( pool )
	{
		viewSubscriber *vs;

		vs = &pool->data[pool->used];
		pool->used++;
		vs->vGroup = vg;
		retval = 1;
	}
	unlockMutex(nsc);
	return( retval );
}

/** Resource functions **/
static
int nsResourceFillSpec(struct _resource *res,
		       struct _resourceSpec *rs,
		       sectionFile *sf,
		       tagSectionData *tsd)
{
	nameSpaceSpec *nss = (nameSpaceSpec *)rs;
	int lpc, retval = 1;
	tagList *tl = tsd->taglist;

	/*
	 * Walk the tag list interpreting it as each tag is the classGroup
	 * name and the value is the viewGroup name.
	 */
	for( lpc = 0; (lpc < tl->tags.count) && retval; lpc++ )
	{
		retval = subscribeView(nss,
				       0,
				       tl->tags.data[lpc].tag,
				       tl->tags.data[lpc].data.str);
	}
	return( retval );
}

static
resourceSpec *nsCreateJSISpec(struct _resource *res, jsiMessage *jm)
{
	resourceSpec *retval = 0;
	int count;
	
	count = jsiMessageIdentifierCount(jm, "namespace");
	if( count && (retval = createResourceSpec(res)) )
	{
		nameSpaceSpec *nss = (nameSpaceSpec *)retval;
		int error = 0, lpc;

		if( (nss->viewSubscriptions.data =
		     gc_malloc(sizeof(groupNames) * count,
			       GC_ALLOC_NAME_SPACE)) )
		{
			nss->viewSubscriptions.size = count;
			for( lpc = 0; lpc < count; lpc++ )
			{
				parsedString cName, vName;
				groupNames *gn;
				char *val;

				gn = &nss->viewSubscriptions.data[lpc];
				jsiMessageGet(
					jm,
					JMA_String, "namespace", lpc, &val,
					JMA_TAG_DONE);
				if( parseString(val,
						SPO_Integer, &gn->teamID,
						SPO_Expect, ":",
						SPO_String, &cName,
						SPO_Expect, ":",
						SPO_String, &vName,
						SPO_End) )
				{
					gn->className = utf8ConstNew(
						cName.data, cName.len);
					gn->viewName = utf8ConstNew(
						vName.data, vName.len);
					if( !gn->className || !gn->viewName )
					{
						error = 1;
					}
				}
				else
				{
					error = 1;
				}
			}
		}
		else
		{
			error = 1;
		}
		if( error )
		{
			deleteResourceSpecFromJSI(retval);
			retval = 0;
		}
	}
	return( retval );
}

static
void nsDeleteJSISpec(struct _resource *res, struct _resourceSpec *rs)
{
	nameSpaceSpec *nss = (nameSpaceSpec *)rs;
	int lpc;

	for( lpc = 0; lpc < nss->viewSubscriptions.size; lpc++ )
	{
		utf8ConstRelease(nss->viewSubscriptions.data[lpc].className);
		utf8ConstRelease(nss->viewSubscriptions.data[lpc].viewName);
	}
	gc_free(nss->viewSubscriptions.data);
	deleteResourceSpec(rs);
}

static
char *nsHandleJSIQuery(struct _resourceClient *rc, jsiMessage *jm)
{
	nameSpaceClient *nsc = (nameSpaceClient *)rc;
	viewSubscriberPool *vsp;
	char *retval = 0;
	int iLockRoot;

	switch( jm->what )
	{
	case JSI_REPLY:
		lockMutex(nsc);
		vsp = nsc->views;
		while( vsp )
		{
			int lpc;

			for( lpc = 0; lpc < vsp->used; lpc++ )
			{
				jsiMessageAdd(jm,
					      JMA_Integer,
					      "namespace.team",
					      gc_get_resource(vsp->data[lpc].
							      vGroup->cGroup)->
					      client.owner->elemID,
					      JMA_String,
					      "namespace.classgroup",
					      vsp->data[lpc].vGroup->cGroup->
					      type->name->data,
					      JMA_String,
					      "namespace.viewgroup",
					      vsp->data[lpc].vGroup->type->
					      name->data,
					      JMA_TAG_DONE);
			}
			vsp = vsp->next;
		}
		unlockMutex(nsc);
		break;
	default:
		retval = "Unsupported operation";
		break;
	}
	return( retval );
}

static
Hjava_lang_Class *getSubscribedClassView(nameSpaceClient *nsc,
					 char *name,
					 viewSubscriber **vs_out)
{
	Hjava_lang_Class *retval = 0;
	viewSubscriberPool *pool;
	int lpc;
	
	pool = nsc->views;
	while( pool && !retval )
	{
		/* Check for an imported class first */
		for( lpc = 0; !retval && (lpc < pool->used); lpc++ )
		{
			if( (retval = findViewClassPart(pool->data[lpc].vGroup,
							name)) )
			{
				*vs_out = &pool->data[lpc];
			}
		}
		pool = pool->next;
	}
	return( retval );
}

static
int verifyNameSpace(nameSpaceClient *nsc)
{
	int lpc, lpc2, lpc3, lpc4, retval = 1;
	viewSubscriberPool *vsp;

	vsp = nsc->views;
	while( /* retval && */ vsp )
	{
		for( lpc = 0;
		     // retval &&
		     lpc < vsp->used;
		     lpc++ )
		{
			viewSubscriber *vs = &vsp->data[lpc];

			for( lpc2 = 0;
			     //retval &&
			     lpc2 < vs->vGroup->cViews.size;
			     lpc2++ )
			{
				classView *cv = &vs->vGroup->cViews.data[lpc2];

				if( CLASS_IS_ARRAY(cv->cl) ||
				    CLASS_IS_PRIMITIVE(cv->cl) )
				{
					continue;
				}
				for( lpc3 = 0;
				     //retval &&
				     lpc3 < cv->cl->nmethods;
				     lpc3++ )
				{
					methodView *mv = &cv->methods[lpc3];
					viewSubscriber *vs_src;
					int is_int = 1;
					
					if( mv->flags & MVF_INTERNAL )
					{
						continue;
					}
					if( (cv->cl->methods[lpc3].accflags &
					     ACC_SYNCHRONIZED) &&
					    (cv->cl->methods[lpc3].accflags &
					     ACC_STATIC) )
					{
						teamLogEntry("namespace",
							     LEA_String, "Warning", "Imported class has static synchronized method", -1,
							     LEA_String, "imported-class", cv->name->data, -1,
							     LEA_String, "imported-class-method", cv->cl->methods[lpc3].name->data, -1,
							     LEA_TAG_DONE);
					}
					if( (cv->cl->methods[lpc3].accflags &
					     ACC_PRIVATE) ||
					    ((cv->cl->methods[lpc3].accflags &
					      (ACC_PUBLIC|
					       ACC_PRIVATE|
					       ACC_PROTECTED)) == 0) )
					{
						continue;
					}
					for( lpc4 = 0;
					     //retval &&
					     lpc4 < mv->parsed_sig->nargs + 1;
					     lpc4++ )
					{
						if( getSubscribedClassView(nsc, (char *)&mv->parsed_sig->signature->data[mv->parsed_sig->ret_and_args[lpc4]], &vs_src) )
						{
							is_int = is_int &&
								(vs_src == vs);
						}
						else
						{
							teamLogEntry("namespace",
								     LEA_String, "Error", "Imported class has public reference to non-imported class", -1,
								     LEA_String, "imported-class", cv->name->data, -1,
								     LEA_String, "imported-class-method", cv->cl->methods[lpc3].name->data, -1,
								     LEA_String, "referenced-class", &mv->parsed_sig->signature->data[mv->parsed_sig->ret_and_args[lpc4]], -1, // XXX partLen()
								     LEA_TAG_DONE);
							retval = 0;
						}
					}
					if( is_int )
					{
						mv->flags |= MVF_INTERNAL;
					}
				}
			}
		}
		vsp = vsp->next;
	}
	return( retval );
}

static int subscribeViewFromSpec(nameSpaceClient *nsc,
				 groupNames *gn,
				 errorInfo *einfo)
{
	int retval = 0;
	jteam *jt;

	if( (jt = grabTeamWithID(gn->teamID)) )
	{
		jteamReservation *jr;
		
		if( (jr = pushTeamReservation(jt)) )
		{
			nameSpaceClient *jt_nsc;
			classGroup *cg;
			viewGroup *vg;

			jt_nsc = (nameSpaceClient *)getTeamInternalResource(
				jt, &metalNameSpaceResource.link);
			cg = findClassGroup(&jt_nsc->exportedClasses,
					    gn->className);
			if( cg )
			{
				vg = findViewGroup(cg, gn->viewName);
				if( vg )
				{
					if( switchToTeam(kernelTeam, 0) )
					{
						if( (nsc->client.owner == jt) ||
						    (retval = addTeamDependency(nsc->client.owner, jt)) )
						{
							subscribeViewClient(nsc, vg);
						}
						else
						{
							postOutOfMemory(einfo);
						}
						returnFromTeam();
					}
					else
					{
						if( einfo )
						{
							postExceptionMessage(
								einfo,
								DEAD_TEAM_EXCEPTION_S,
								"kernel");
						}
					}
				}
				else
				{
					if( einfo )
					{
						postExceptionMessage(
							einfo,
							"edu.utah.janosvm.resources."
							"ViewGroupNotFoundException",
							"%s",
							gn->viewName->data);
					}
					teamLogEntry("namespace",
						     LEA_String, "Error", "Couldn't find view group", -1,
						     LEA_String, "viewgroup", gn->viewName->data, -1,
						     LEA_String, "source-team", jt->name, -1,
						     LEA_Integer, "team-id", gn->teamID,
						     LEA_TAG_DONE);
				}
			}
			else
			{
				if( einfo )
				{
					postExceptionMessage(
						einfo,
						"edu.utah.janosvm.resources."
						"ClassGroupNotFoundException",
						"%s",
						gn->className->data);
				}
				teamLogEntry("namespace",
					     LEA_String, "Error", "Couldn't find class group", -1,
					     LEA_String, "classgroup", gn->className->data, -1,
					     LEA_String, "source-team", jt->name, -1,
					     LEA_Integer, "team-id", gn->teamID,
					     LEA_TAG_DONE);
			}
			popTeamReservation(jr);
		}
		else
		{
			if( einfo )
			{
				postExceptionMessage(einfo,
						     DEAD_TEAM_EXCEPTION_S,
						     "%d",
						     gn->teamID);
			}
			teamLogEntry("namespace",
				     LEA_String, "Error", "Couldn't visit team", -1,
				     LEA_Integer, "id", gn->teamID,
				     LEA_TAG_DONE);
		}
		dropTeam(jt);
	}
	else
	{
		if( einfo )
		{
			postExceptionMessage(einfo,
					     DEAD_TEAM_EXCEPTION_S,
					     "%d",
					     gn->teamID);
		}
		teamLogEntry("namespace",
			     LEA_String, "Error", "Couldn't find team", -1,
			     LEA_Integer, "team-id", gn->teamID,
			     LEA_TAG_DONE);
	}
	return( retval );
}

#define DEFAULT_TABLE_SIZE 31

static
int nsResourceSubscribe(struct _resource *res,
			struct _resourceClient *rc,
			struct _resourceSpec *rs)
{
	nameSpaceClient *nsc = (nameSpaceClient *)rc;
	nameSpaceSpec tmp_nss, *nss = (nameSpaceSpec *)rs;
	int lpc, retval = 1;
	groupNames tmp_gn;

	/* Create the root group for this client */
	nsc->cGroup = createClassGroup(0, 0);
	nsc->mapping.table = gc_malloc(sizeof(nameMapping *) *
				       (DEFAULT_TABLE_SIZE + 1),
				       GC_ALLOC_NAME_SPACE);
	assert(nsc->mapping.table != 0); // XXX
	nsc->mapping.size = DEFAULT_TABLE_SIZE;
	nsc->classLocals.table = gc_malloc(sizeof(classLocal *) *
					   (DEFAULT_TABLE_SIZE + 1),
					   GC_ALLOC_NAME_SPACE);
	assert(nsc->classLocals.table != 0); // XXX
	nsc->classLocals.size = DEFAULT_TABLE_SIZE;
	if( nss )
	{
		int foundKernel = 0;
		
		for( lpc = 0;
		     (lpc < nss->viewSubscriptions.size) && !foundKernel;
		     lpc++ )
		{
			if( (nss->viewSubscriptions.data[lpc].teamID ==
			     kernelTeam->elemID) &&
			    !strcmp(nss->viewSubscriptions.data[lpc].
				    className->data,
				    "kernel") )
			{
				foundKernel = 1;
			}
		}
		if( kernelTeam && !foundKernel )
		{
			teamLogEntry("namespace",
				     LEA_String, "Error", "Kernel class group not specified in name space specification", -1,
				     LEA_TAG_DONE);
			retval = 0;
		}
	}
	else
	{
		/* Fill in the default spec arguments. */
		nss = &tmp_nss;
		tmp_gn.teamID = 1;
		tmp_gn.className = utfKernel;
		tmp_gn.viewName = utfIdentity;
		nss->spec.handler = &metalNameSpaceResource.link;
		nss->viewSubscriptions.data = &tmp_gn;
		nss->viewSubscriptions.size = 1;
	}
	/* XXX Preallocate pools since we'll be working in the kernel */
	{
		viewSubscriberPool *pool;
		int count =
			(nss->viewSubscriptions.size +
			 VIEW_SUBSCRIBER_POOL_SIZE -
			 1) /
			VIEW_SUBSCRIBER_POOL_SIZE;

		for( lpc = 0; lpc < count; lpc++ )
		{
			pool = gc_malloc(sizeof(viewSubscriberPool),
					 GC_ALLOC_NAME_SPACE);
			assert(pool != 0); // XXX
			pool->next = nsc->views;
		        nsc->views = pool;
		}
	}
	/* Visit the kernel so we can import the groups from other teams. */
	if( switchToTeam(kernelTeam, 0) )
	{
		for( lpc = 0; lpc < nss->viewSubscriptions.size; lpc++ )
		{
			groupNames *gn = &nss->viewSubscriptions.data[lpc];
			errorInfo einfo;

			if( !subscribeViewFromSpec(nsc, gn, &einfo) )
			{
				discardErrorInfo(&einfo);
				retval = 0;
			}
		}
		returnFromTeam();
	}
	else
	{
		retval = 0;
	}
	if( retval )
	{
		if( !(retval = verifyNameSpace(nsc)) )
		{
		}
	}
	return( retval );
}

static
void nsResourceUnsubscribe(struct _resource *res, struct _resourceClient *rc)
{
	nameSpaceClient *nsc = (nameSpaceClient *)rc;
	struct Hjava_lang_Class *cl;
	classGroup *cg;
	classView *cv;
	int lpc;

	require(nsc->flags & NSCF_DEAD);
	require(nsc->imports == 0);

	/*
	 * XXX We need to manually run through and free the interface
	 * implementation indexes for each class.  We should really only
	 * do this for shared interfaces.
	 */
	for( lpc = 0; lpc < nsc->cGroup->vGroups.data[0].cViews.size; lpc++ )
	{
		cv = &nsc->cGroup->vGroups.data[0].cViews.data[lpc];
		cl = cv->cl;
		if( cl && cl->itable2dtable )
		{
			int i;

			for( i = 0; i < cl->total_interface_len; i++ )
			{
				cl->interfaces[i]->
					implementors[cl->impl_index] = -1;
			}
		}
	}
	cg = nsc->exportedClasses.groups;
	while( cg )
	{
		for( lpc = 0; lpc < cg->vGroups.data[0].cViews.size; lpc++ )
		{
			cv = &cg->vGroups.data[0].cViews.data[lpc];
			cl = cv->cl;
			if( cl && cl->itable2dtable )
			{
				int i;
		    
				for( i = 0; i < cl->total_interface_len; i++ )
				{
					cl->interfaces[i]->
						implementors[cl->impl_index] =
						-1;
				}
			}
		}
		cg = cg->next;
	}
}

static resourceFunctions nsResourceFunctions = {
	0,
	0,
	0,
	nsResourceFillSpec,
	nsCreateJSISpec,
	nsDeleteJSISpec,
	0,
	nsResourceSubscribe,
	0,
	nsResourceUnsubscribe,
	nsHandleJSIQuery,
};

/** edu.utah.janosvm.sys.PortAuthority native functions **/
void edu_utah_janosvm_sys_PortAuthority_revokeAllImports(void)
{
	nameSpaceClient *nsc = (nameSpaceClient *)
		getTeamInternalResource(0, &metalNameSpaceResource.link);
	struct Hedu_utah_janosvm_sys_Importable *imp;
	memImportLink *mil;
	int iLockRoot;

	require(currentTeam() == nsc->client.owner);
	require(!(nsc->flags & NSCF_DEAD));

	/*
	 * Everything is allocated inside the team, so just disconnect any
	 * imported objects.  XXX This should be moved elsewhere.
	 */
	lockMutex(nsc);
	nsc->flags |= NSCF_DEAD;
	mil = nsc->imports;
	unlockMutex(nsc);
	while( mil )
	{
		imp = (struct Hedu_utah_janosvm_sys_Importable *)
			NATIVE2OBJECT(&importNativeClass, mil);
		GC_finalizeObject((Collector *)
				  teamCollector(nsc->client.owner), imp);
		/* XXX ick */
		lockMutex(nsc);
		{
			require(mil != nsc->imports);
			require(mil->link == (memImportLink *)~0);
			
			mil = nsc->imports;
		}
		unlockMutex(nsc);
	}
}

jbool edu_utah_janosvm_sys_PortAuthority_register(struct Hedu_utah_janosvm_sys_Importable *imp)
{
	memImportLink *mil = OBJECT2NATIVE(&importNativeClass,
					   (Hjava_lang_Object *)imp);
	nameSpaceClient *nsc;
	int iLockRoot;
	jbool retval = 0;

	nsc = (nameSpaceClient *)
		getTeamInternalResource(0, &metalNameSpaceResource.link);
	lockMutex(nsc);
	if( !(nsc->flags & NSCF_DEAD) )
	{
		mil->link = nsc->imports;
		nsc->imports = mil;
		retval = 1;
	}
	unlockMutex(nsc);
	return( retval );
}

void edu_utah_janosvm_sys_PortAuthority_unregister(struct Hedu_utah_janosvm_sys_Importable *imp)
{
	memImportLink *mil = OBJECT2NATIVE(&importNativeClass,
					   (Hjava_lang_Object *)imp);
	memImportLink *curr, **prev;
	nameSpaceClient *nsc;
	int iLockRoot;

	nsc = (nameSpaceClient *)
		getTeamInternalResource(0, &metalNameSpaceResource.link);
	lockMutex(nsc);
	prev = &nsc->imports;
	curr = nsc->imports;
	while( curr && (curr != mil) )
	{
		prev = &curr->link;
		curr = curr->link;
	}
	if( curr )
	{
		*prev = curr->link;
		curr->link = (memImportLink *)~0;
	}
	unlockMutex(nsc);
}

jbool edu_utah_janosvm_sys_PortAuthority_isValid(struct Hedu_utah_janosvm_sys_Importable *imp)
{
	memImportLink *mil = OBJECT2NATIVE(&importNativeClass,
					   (Hjava_lang_Object *)imp);
	
	return (mil->link != (memImportLink *)~0);
}

/** Module declarations **/
static int initNameSpaceResourceModule(int stage);

module_t nameSpaceModule = {
	"namespace",
	0,
	initNameSpaceResourceModule,
	0,
	0
};

static int initNameSpaceResourceModule(int stage)
{
	switch( stage )
	{
	case SID_NAMESPACEMODULE_INITMETAL:
		{
			nameSpaceSpec nss;

			utfKernel = utf8ConstNew("kernel", -1);
			utfIdentity = utf8ConstNew("identity", -1);
			
			installResource(
				&metalNameSpaceResource.link,
				RA_Name, "namespace",
				RA_Flags, RF_INTERNAL,
				RA_Functions, &nsResourceFunctions,
				RA_ClientSize, sizeof(nameSpaceClient),
				RA_SpecSize, sizeof(nameSpaceSpec),
				RA_TAG_DONE);
			
			nss.spec.handler = &metalNameSpaceResource.link;
			nss.viewSubscriptions.data = 0;
			nss.viewSubscriptions.size = 0;
			attachTeamResource(metalTeam, &nss.spec);
			break;
		}
	case SID_NAMESPACEMODULE:
		{
			errorInfo einfo;

			installResource(
				&metalNameSpaceResource.link,
				RA_ResourceSectionName, "namespace-resource",
				RA_SpecSectionName, "namespace",
				RA_SpecNativeClass, JAVA_NAME_SPACE_SPEC,
				RA_TAG_DONE);
			installSectionType(
				&nsClassGroupSection.type,
				STA_Name, "classgroup",
				STA_Template, &tagsFileSectionTemplate,
				STA_Functions, &nsClassGroupSectionFunctions,
				STA_Size, (sizeof(classGroupSectionData) -
					   sizeof(tagSectionData)),
				STA_Walk, nsClassGroupSectionWalker,
				STA_TAG_DONE);
			installSectionType(
				&nsViewGroupSection.type,
				STA_Name, "viewgroup",
				STA_Template, &tagsFileSectionTemplate,
				STA_Functions, &nsViewGroupSectionFunctions,
				STA_Walk, nsViewGroupSectionWalker,
				STA_TAG_DONE);
			importNativeClass.size = sizeof(memImportLink);
			/* XXX Load any exempt classes */
			ListClass = lookupClass(
				"edu/utah/janosvm/sys/List",
				NULL,
				&einfo);
			ListNodeClass = lookupClass(
				"edu/utah/janosvm/sys/ListNode",
				NULL,
				&einfo);
			ObjectHolderClass = lookupClass(
				"java/lang/ThreadObjectHolder",
				NULL,
				&einfo);
			CrossTeamCloneableClass = lookupClass(
				"edu/utah/janosvm/sys/CrossTeamCloneable",
				NULL,
				&einfo);
			NameSpaceGroupClass = lookupClass(
				"edu/utah/janosvm/resources/NameSpaceGroup",
				NULL,
				&einfo);
			if( (ImportableClass =
			     lookupClass("edu/utah/janosvm/sys/Importable",
					 NULL,
					 &einfo)) )
			{
				bindNativeClass(&importNativeClass,
						ImportableClass,
						&einfo);
				ExportableClass = lookupClass(
					"edu/utah/janosvm/sys/Exportable",
					NULL,
					&einfo);
			}
		}
		break;
	}
	return( 1 );
}

Hjava_lang_Class *bindClassMapping(nameMapping *nm, errorInfo *einfo)
{
	Hjava_lang_Class *retval = 0;
	viewSubscriberPool *pool;
	nameSpaceClient *nsc;
	char *cname;
	int lpc;
	
	require(nm != 0);
	require(nm->loader == 0);
	require(einfo != 0);

	cname = (char *)nm->name->data;
	nsc = (nameSpaceClient *)
		getTeamInternalResource(0, &metalNameSpaceResource.link);
	pool = nsc->views;
	while( pool && !retval )
	{
		/* Check for an imported class first */
		for( lpc = 0; !retval && (lpc < pool->used); lpc++ )
		{
			struct Hjava_lang_Class *cl;
			
			if( (cl = findViewClass(pool->data[lpc].vGroup,
						nm->name)) )
			{
				if( !CLASS_IS_PRIMITIVE(cl) )
				{
					retval = cl;
					retval = classMappingLoaded(nm,
								    retval);
				}
			}
		}
		pool = pool->next;
	}
	if( !retval )
	{
		char *buf;

		/* No matching import, try the class path */
		if( (buf = gc_malloc(strlen(cname) +
				     1 + /* `.' */
				     5 + /* `class' */
				     1,
				     GC_ALLOC_NAME_SPACE)) )
		{
			classPathClient *cpc;
			classFile cf;

			{
				sprintf(buf, "%s.class", cname);

				ensure_valid_memory(buf);
			}
			cpc = (classPathClient *)
				getTeamInternalResource(0,
					&metalClassPathResource.link);
			switch( loadClassPathFile(cpc, buf, &cf, einfo) )
			{
			case FTH_LOAD_ERROR:
			case FTH_LOAD_NOT_FOUND:
				/*
				 * loadClassPathFile will post an IOException,
				 * but class loaders can only throw
				 * ClassNotFoundExceptions, or errors.
				 */
				discardErrorInfo(einfo);
				postExceptionMessage(
					einfo,
					JAVA_LANG(ClassNotFoundException),
					"%s",
					nm->name->data);
				if( kernelTeam )
				{
#if 0
					teamLogEntry("namespace",
						     LEA_String, "Error", "Class not found in path", -1,
						     LEA_String, "class", nm->name->data, -1,
						     LEA_TAG_DONE);
#endif
				}
				break;
			case FTH_LOAD_FOUND:
				/* Found the class try to instantiate it */
				if( (retval = newClass()) )
				{
					utf8ConstAssign(retval->primaryView.
							name,
							nm->name);
					retval = readClass(retval,
							   &cf,
							   NULL,
							   einfo);
					if( retval )
					{
					}
					else
					{
#if 0
						teamLogEntry("namespace",
							     LEA_String, "Error", "Cannot read in class", -1,
							     LEA_String, "class", nm->name->data, -1,
							     LEA_TAG_DONE);
#endif
					}
				}
				else
				{
					postOutOfMemory(einfo);
				}
				gc_free((void *)cf.base);
				break;
			}
			gc_free(buf);
		}
		else
		{
			postOutOfMemory(einfo);
		}
	}
	/*
	 * XXX This is kinda stupid, we should be able to fail more gracefully
	 * than checking each time we try to bind a class
	 */
	if( !retval &&
	    ((strcmp(cname, "java/lang/ClassNotFoundException") == 0) ||
	     (strcmp(cname, "java/lang/Object") == 0)) )
	{
		fprintf(stderr,
			"Cannot find essential class '%s' in class library ..."
			" aborting.\n",
			cname);
		ABORT();
	}
	return( retval );
}

/* Lookup a nameMapping in the given client */
static
nameMapping *internalFindClassMapping(nameSpaceClient *nsc,
				      Utf8Const *name,
				      int table_index,
				      Hjava_lang_ClassLoader *loader)
{
	nameMapping *curr, *retval = 0;

	curr = nsc->mapping.table[table_index];
	while( curr && !retval )
	{
		if( utf8ConstEqual(curr->name, name) &&
		    (curr->loader == loader) )
		{
			retval = curr;
		}
		curr = curr->next;
	}
	return( retval );
}

KAFFE_INLINE
int hashMappingName(nameSpaceClient *nsc, Utf8Const *name)
{
	int retval;

	require(nsc != 0);
	require_valid_memory(nsc);
	require_valid_memory(nsc->mapping.table);
	require(name != 0);
	
	retval = utf8ConstHashValue(name) & nsc->mapping.size;

	ensure(retval >= 0);
	ensure(retval < (nsc->mapping.size + 1));

	return( retval );
}

nameMapping *findClassMapping(Utf8Const *name, Hjava_lang_ClassLoader *loader)
{
	nameMapping *retval = 0;

	nameSpaceClient *nsc;
	int idx;
	int iLockRoot;

	nsc = (nameSpaceClient *)
		getTeamInternalResource(0, &metalNameSpaceResource.link);
	lockMutex(&nsc->mapping);
	idx = hashMappingName(nsc, name);
	retval = internalFindClassMapping(nsc, name, idx, loader);
	unlockMutex(&nsc->mapping);
	return( retval );
}

Hjava_lang_Class *findValidClassMapping(Utf8Const *name,
					Hjava_lang_ClassLoader *loader)
{
	Hjava_lang_Class *retval = 0;
	nameMapping *nm;
	
	if( (nm = findClassMapping(name, loader)) )
	{
		int iLockRoot;
		int done = 0;
		
		lockMutex(nm);
		while( !done )
		{
			switch( nm->state )
			{
			case NMS_EMPTY:
				done = 1;
				break;
			case NMS_DONE:
				retval = nm->data.cl;
				done = 1;
				break;
			case NMS_SEARCHING:
				if( nm->data.thread == jthread_current() )
				{
					done = 1;
					break;
				}
				/* FALLTHROUGH */
			default:
				waitCond(nm, 0);
				break;
			}
		}
		unlockMutex(nm);
	}
	return( retval );
}

nameMapping *getClassMapping(Utf8Const *name,
			     Hjava_lang_ClassLoader *loader,
			     errorInfo *einfo)
{
	nameMapping *retval = 0;
	nameSpaceClient *nsc;
	int idx;
	int iLockRoot;

	nsc = (nameSpaceClient *)
		getTeamInternalResource(0, &metalNameSpaceResource.link);
	lockMutex(&nsc->mapping);
	idx = hashMappingName(nsc, name);
	if( (retval = internalFindClassMapping(nsc, name, idx, loader)) )
	{
		/* Found a match, return it */
	}
	else if( (retval = gc_malloc(sizeof(nameMapping),
				     GC_ALLOC_NAME_SPACE)) )
	{
		/* Need to make a new one */
		retval->next = nsc->mapping.table[idx];
		nsc->mapping.table[idx] = retval;
		utf8ConstAddRef(name);
		retval->name = name;
		retval->loader = loader;

		ensure_valid_memory(nsc);
		ensure_valid_memory(nsc->mapping.table);
	}
	else
	{
		postOutOfMemory(einfo);
	}
	
	ensure(!retval ||
	       internalFindClassMapping(nsc, name, idx, loader));
	
	unlockMutex(&nsc->mapping);
	return( retval );
}

static
nameDependency *findNameDependency(nameSpaceClient *nsc, jthread_t jt)
{
	nameDependency *curr, *retval = 0;

	for( curr = nsc->dependencies; curr && !retval; curr = curr->next )
	{
		if( curr->thread == jt )
		{
			retval = curr;
		}
	}
	return( retval );
}

static
int addNameDependency(nameDependency *nd)
{
	nameSpaceClient *nsc;
	int retval = 1;
	int iLockRoot;

	require(nd != 0);

	nsc = (nameSpaceClient *)
		getTeamInternalResource(0, &metalNameSpaceResource.link);
	lockMutex(&nsc->mapping);
	{
		nameDependency *curr;
		
		nd->next = nsc->dependencies;
		nsc->dependencies = nd;

		for( curr = findNameDependency(nsc, nd->mapping->data.thread);
		     curr && retval;
		     curr = findNameDependency(nsc,
					       curr->mapping->data.thread) )
		{
			if( curr->mapping->data.thread == nd->thread )
			{
				retval = 0;
			}
		}
	}
	unlockMutex(&nsc->mapping);
	return( retval );
}

static
void remNameDependency(nameMapping *nm)
{
	nameSpaceClient *nsc;
	int iLockRoot;

	require(nm != 0);
	
	nsc = (nameSpaceClient *)
		getTeamInternalResource(0, &metalNameSpaceResource.link);
	lockMutex(&nsc->mapping);
	{
		nameDependency **last, *curr;

		last = &nsc->dependencies;
		curr = nsc->dependencies;
		while( curr && (curr->mapping != nm) )
		{
			last = &curr->next;
			curr = curr->next;
		}
		if( curr )
		{
			*last = curr->next;
		}
	}
	unlockMutex(&nsc->mapping);
}

int classMappingSearch(nameMapping *nm,
		       Hjava_lang_Class **out_cl,
		       errorInfo *einfo)
{
	int done = 0, retval = 1;
	nameDependency nd;
	jthread_t jt;
	int iLockRoot;

	jt = jthread_current();
	while( !done )
	{
		lockMutex(nm);
		switch( nm->state )
		{
		case NMS_EMPTY:
			/* This thread's responsibility. */
			nm->state = NMS_SEARCHING;
			nm->data.thread = jt;
			done = 1;
			break;
		case NMS_SEARCHING:
			if( nm->data.thread == jt )
			{
				done = 1;
				break;
			}
			waitCond(nm, 0);
			break;
		case NMS_LOADING:
			/*
			 * Another thread is loading it, make sure there is not
			 * a cycle.
			 */
			nd.thread = jt;
			nd.mapping = nm;
			if( (nm->data.thread == jt) ||
			    !addNameDependency(&nd) )
			{
				/* Circularity. */
				done = 1;
				retval = 0;
				postExceptionMessage(
					einfo,
					JAVA_LANG(ClassCircularityError),
					"%s",
					nm->name->data);
			}
			else
			{
				waitCond(nm, 0);
			}
			remNameDependency(nm);
			break;
		case NMS_LOADED:
			/*
			 * Another thread loaded it, however, its not finished
			 * linking yet.
			 */
			waitCond(nm, 0);
			break;
		case NMS_DONE:
			/* Its already been loaded and linked. */
			*out_cl = nm->data.cl;
			done = 1;
			break;
		}
		unlockMutex(nm);
	}
	return( retval );
}

int classMappingLoad(nameMapping *nm,
		     Hjava_lang_Class **out_cl,
		     errorInfo *einfo)
{
	int done = 0, retval = 1;
	nameDependency nd;
	jthread_t jt;
	int iLockRoot;

	*out_cl = 0;
	jt = jthread_current();
	while( !done )
	{
		lockMutex(nm);
		switch( nm->state )
		{
		case NMS_EMPTY:
		case NMS_SEARCHING:
			/* This thread's responsibility. */
			nm->state = NMS_LOADING;
			nm->data.thread = jt;
			done = 1;
			break;
		case NMS_LOADING:
			/*
			 * Another thread is loading it, make sure there is not
			 * a cycle.
			 */
			nd.thread = jt;
			nd.mapping = nm;
			if( (nm->data.thread == jt) ||
			    !addNameDependency(&nd) )
			{
				/* Circularity. */
				done = 1;
				retval = 0;
				postExceptionMessage(
					einfo,
					JAVA_LANG(ClassCircularityError),
					"%s",
					nm->name->data);
			}
			else
			{
				waitCond(nm, 0);
			}
			remNameDependency(nm);
			break;
		case NMS_LOADED:
			/*
			 * Another thread loaded it, however, its not finished
			 * linking yet.
			 */
			waitCond(nm, 0);
			break;
		case NMS_DONE:
			/* Its already been loaded and linked. */
			*out_cl = nm->data.cl;
			done = 1;
			break;
		}
		unlockMutex(nm);
	}
	return( retval );
}

Hjava_lang_Class *classMappingLoaded(nameMapping *nm, Hjava_lang_Class *cl)
{
	Hjava_lang_Class *retval = 0;
	int iLockRoot;

	require(nm != 0);
	require(cl != 0);
	
	lockMutex(nm);
	{
		switch( nm->state )
		{
		case NMS_LOADING:
			/* FALLTHROUGH */
		case NMS_SEARCHING:
			if( cl->state < CSTATE_PREPARED )
			{
				nm->state = NMS_LOADED;
			}
			else
			{
				nm->state = NMS_DONE;
			}
			nm->data.cl = cl;
			retval = cl;
			break;
		default:
			/* Ignore. */
			retval = nm->data.cl;
			break;
		}
		broadcastCond(nm);
	}
	unlockMutex(nm);

	return( retval );
}

void setClassMappingState(nameMapping *nm, name_mapping_state_t nms)
{
	int iLockRoot;
	
	require(nm != 0);
	
	lockMutex(nm);
	{
		switch( nm->state )
		{
		case NMS_SEARCHING:
			nm->state = nms;
			break;
		case NMS_LOADING:
			/* FALLTHROUGH */
		case NMS_LOADED:
			nm->state = nms;
			break;
		case NMS_DONE:
			break;
		case NMS_EMPTY:
			break;
		default:
			assert(0);
			break;
		}
		broadcastCond(nm);
	}
	unlockMutex(nm);
}

void clearClassMapping(Utf8Const *name, Hjava_lang_ClassLoader *loader)
{
	nameSpaceClient *nsc;
	nameMapping *nm = 0;
	int idx;
	int iLockRoot;
	
	nsc = (nameSpaceClient *)
		getTeamInternalResource(0, &metalNameSpaceResource.link);
	if( name )
	{
		lockMutex(&nsc->mapping);
		idx = hashMappingName(nsc, name);
		if( (nm = internalFindClassMapping(nsc, name, idx, loader)) )
		{
			nm->data.cl = 0;
		}
		unlockMutex(&nsc->mapping);
	}
}

void removeLoaderMappings(Hjava_lang_ClassLoader *loader)
{
	nameMapping *curr, *next, **prev;
	nameSpaceClient *nsc;
	int lpc;
	int iLockRoot;
	
	nsc = (nameSpaceClient *)
		getTeamInternalResource(0, &metalNameSpaceResource.link);
	lockMutex(&nsc->mapping);
	for( lpc = 0; lpc < nsc->mapping.size; lpc++ )
	{
		prev = &nsc->mapping.table[lpc];
		curr = nsc->mapping.table[lpc];
		while( curr )
		{
			next = curr->next;
			if( curr->loader == loader )
			{
				/*
				 * Since we're killing the current node, theres
				 * no need to advance the prev pointer.
				 */
				*prev = next;
				gc_free(curr);
			}
			else
			{
				prev = &curr->next;
			}
			curr = next;
		}
	}
	unlockMutex(&nsc->mapping);
}

#if defined(KAFFE_STATS) || defined(KAFFE_PROFILER) || defined(DEBUG)
void walkNameSpace(nameSpaceClient *nsc,
		   int (*walker)(Hjava_lang_Class *clazz, void *),
		   void *param)
{
	nameMapping *nm;
	int lpc;
	
	if( !nsc )
	{
		nsc = (nameSpaceClient *)
			getTeamInternalResource(0,
				&metalNameSpaceResource.link);
	}
	for( lpc = 0; lpc < nsc->mapping.size; lpc++ )
	{
		nm = nsc->mapping.table[lpc];
		while( nm )
		{
			if( nm->state == NMS_DONE && nm->data.cl )
			{
				walker(nm->data.cl, param);
			}
			nm = nm->next;
		}
	}
}
#endif

const char* getClassTraceName(Hjava_lang_Class* cl)
{
	Utf8Const *name = 0;
	nameSpaceClient *nsc;

	if (cl == 0)
		return "<nullclass>";

	nsc = (nameSpaceClient *)
		getTeamInternalResource(0, &metalNameSpaceResource.link);

	if( (cl->cGroup == 0) || !nsc )
	{
		/* XXX This really shouldn't happen... */
		if ( cl->primaryView.name )
			return cl->primaryView.name->data;
		return "<unresolvedname>";
	}

	/*
	 * Default to the primary view name.  Current team may not
	 * have a view on this class, and thus technically shouldn't be
	 * able to see its name, but that makes debugging too difficult.
	 */
	name = cl->primaryView.name;

	if( cl->cGroup != nsc->cGroup )
	{
		bool found = false;
		viewSubscriberPool *pool;

		pool = nsc->views;
		while( pool && !found )
		{
			int lpc;
			for( lpc = 0;
			     !found && (lpc < pool->used);
			     lpc++ )
			{
				if( pool->data[lpc].vGroup->cGroup ==
				    cl->cGroup )
				{
					found = true;
					name = pool->data[lpc].vGroup->
						cViews.data[cl->viewSlot].name;
				}
			}
			pool = pool->next;
		}
	}

	if (name != 0)
		return name->data;
	return "<unresolvedname>";
}

classView *getClassView(Hjava_lang_Class *cl, errorInfo *einfo)
{
	classView *retval = 0;
	nameSpaceClient *nsc;
	int lpc;

	require(cl != 0);
	require(einfo != 0);

	nsc = (nameSpaceClient *)
		getTeamInternalResource(0, &metalNameSpaceResource.link);
	if( (cl->cGroup == 0) || !nsc )
	{
		/* XXX This really shouldn't happen... */
		return( &cl->primaryView );
	}
	if( cl->cGroup == nsc->cGroup )
	{
		retval = &cl->primaryView;
	}
	else
	{
		viewSubscriberPool *pool;

		pool = nsc->views;
		while( pool && !retval )
		{
			for( lpc = 0;
			     !retval && (lpc < pool->used);
			     lpc++ )
			{
				if( pool->data[lpc].vGroup->cGroup ==
				    cl->cGroup )
				{
					retval = &pool->data[lpc].vGroup->
						cViews.data[cl->viewSlot];
				}
			}
			pool = pool->next;
		}
	}
	if( !retval )
	{
		memResourceClient *mrc;

		mrc = gc_get_resource(cl);
		postExceptionMessage(
			einfo,
			"edu.utah.janosvm.sys.SegmentationFaultError",
			"Current team (`%s', %u) has no view on class "
			"\"%s\" from team (`%s', %u).",
			currentTeam()->name,
			currentTeam()->elemID,
			getClassTraceName(cl),
			mrc->client.owner->name,
			mrc->client.owner->elemID);

		retval = 0;
	}

	ensure((retval == 0) || (retval->cl != 0));
	ensure((retval == 0) || (retval->name != 0));
	/*
	 * XXX cannot check that methods are non-null, or whatnot
	 * because arrays and primitives store unrelated crap
	 * the in the methods field of the class.
	 *
	 * ensure((retval == 0) ||
	 *     (retval->methods == 0) == (cl->methods == 0));
	 */
	return( retval );
}

const char* getMethodTraceSig(const Method* meth)
{
	Hjava_lang_Class *cl;
	methodView* mv;
	int idx;

	cl = meth->class;
	if (cl == 0)
		return "<nullclass>";

	idx = meth - cl->methods;
	mv = &cl->primaryView.methods[idx];
	if (mv == 0)
		return "<nullmethview>";

	return METHOD_VIEW_SIGD(mv);
}

methodView *getMethodView(const classView* cv, const Method* meth)
{
	methodView *retval = 0;
	int idx;

	require(cv != 0);
	require(meth != 0);
	require(meth->class == cv->cl);

	idx = meth - cv->cl->methods;
	retval = &cv->methods[idx];

	ensure(retval != 0);
	return( retval );
}

methodView *getPrimaryMethodView(const Method *meth)
{
	Hjava_lang_Class *cl = meth->class;
	methodView *retval = 0;
	classView *cv;
	int idx;

	idx = meth - cl->methods;
	cv = &cl->primaryView;
	retval = &cv->methods[idx];
	return( retval );
}

KAFFE_INLINE
int hashClassLocal(nameSpaceClient *nsc, Utf8Const *name)
{
	int retval;

	require(nsc != 0);
	require(name != 0);
	
	retval = utf8ConstHashValue(name) & nsc->classLocals.size;

	ensure(retval >= 0);
	ensure(retval < (nsc->classLocals.size + 1));

	return( retval );
}

int setClassLocalData(jteam *jt,
		      Hjava_lang_Class *cl,
		      struct Hjava_lang_Object *obj)
{
	int hash, retval = 0;
	nameSpaceClient *nsc;
	classLocal *local;
	int iLockRoot;

	nsc = (nameSpaceClient *)
		getTeamInternalResource(jt, &metalNameSpaceResource.link);
	hash = hashClassLocal(nsc, cl->primaryView.name);
	lockMutex(&nsc->mapping);
	local = nsc->classLocals.table[hash];
	while( local && (local->cl != cl) )
		local = local->next;
	if( !local &&
	    (local = gc_malloc(sizeof(classLocal), GC_ALLOC_NAME_SPACE)) )
	{
		local->next = nsc->classLocals.table[hash];
		nsc->classLocals.table[hash] = local;
		local->cl = cl;
		local->data = 0;

		ensure_valid_memory(nsc);
		ensure_valid_memory(nsc->classLocals.table);
	}
	unlockMutex(&nsc->mapping);
	if( local )
	{
		if( local->data )
			GC_rm_ref((Collector *)teamCollector(jt), local->data);
		local->data = obj;
		if( obj )
			GC_add_ref((Collector *)teamCollector(jt), obj);
		retval = 1;
	}
	return( retval );
}

struct Hjava_lang_Object *getClassLocalData(jteam *jt, Hjava_lang_Class *cl)
{
	struct Hjava_lang_Object *retval = 0;
	nameSpaceClient *nsc;
	classLocal *local;
	int hash;
	
	nsc = (nameSpaceClient *)
		getTeamInternalResource(jt, &metalNameSpaceResource.link);
	hash = hashClassLocal(nsc, cl->primaryView.name);
	local = nsc->classLocals.table[hash];
	while( local && (local->cl != cl) )
		local = local->next;
	if( local )
	{
		retval = local->data;
	}
	return( retval );
}

void checkClass(Hjava_lang_Class *cl, Hjava_lang_ClassLoader *loader)
{
	/* XXX Is this just KaffeOS legacy */
}

void finalizeClassLoader(Collector *collector, void *obj)
{
	Hjava_lang_ClassLoader *loader = (Hjava_lang_ClassLoader *)obj;

	removeLoaderMappings(loader);
	deleteClassGroup((classGroup *)loader->classGroup);
}

static
struct Hjava_lang_Class *matchPrimitive(Utf8Const *utfname)
{
	struct Hjava_lang_Class *retval = 0;
	const char *name = utfname->data;

	if( !strcmp(name, "boolean") )
		retval = booleanClass;
	else if( !strcmp(name, "byte") )
		retval = byteClass;
	else if( !strcmp(name, "char") )
		retval = charClass;
	else if( !strcmp(name, "double") )
		retval = doubleClass;
	else if( !strcmp(name, "float") )
		retval = floatClass;
	else if( !strcmp(name, "int") )
		retval = intClass;
	else if( !strcmp(name, "long") )
		retval = longClass;
	else if( !strcmp(name, "short") )
		retval = shortClass;
	else if( !strcmp(name, "void") )
		retval = voidClass;
	return( retval );
}

static
int exportClasses(nameSpaceClient *nsc, classGroup *cg, errorInfo *einfo)
{
	int retval = 1;
	tagList *tl;
	
	/* Walk the exported classes specified in the config file */
	if( (tl = cg->type->defs) && (tl->tags.count != 0) )
	{
		Hjava_lang_Class *cl;
		int lpc;
		
		for( lpc = 0; retval && (lpc < tl->tags.count); lpc++ )
		{
			Utf8Const *className;
			
			className = tl->tags.data[lpc].tag;
			/* Load each class and add it to the group */
			if( !(cl = matchPrimitive(className)) )
			{
				if( (className->data[0] == '[') &&
				    (cl = loadArray(className, 0, einfo)) )
				{
				}
				else
				{
					cl = loadClass(className, 0, einfo);
				}
			}
			if( cl )
			{
				if( !cl->cGroup->cgl )
				{
					int iLockRoot;
					
					lockMutex(nsc);
					remGroupClass(cl);
					addGroupClass(cg, cl);
					unlockMutex(nsc);
				}
				else
				{
					retval = 0;
					postExceptionMessage(
						einfo,
						"java.lang."
						"ClassNotFoundException",
						"Class `%s' already exported",
						className->data);
				}
			}
			else
			{
				teamLogEntry("namespace",
					     LEA_String, "Error", "Class not found", -1,
					     LEA_String, "class", className->data, -1,
					     LEA_TAG_DONE);
				retval = 0;
			}
		}
	}
	else
	{
		postExceptionMessage(einfo,
				     "edu.utah.janosvm.resources."
				     "InvalidClassGroupException",
				     "Class group `%s' is empty.",
				     cg->type->name->data);
		retval = 0;
	}
	return( retval );
}

int exportClassGroup(Utf8Const *cgName, errorInfo *einfo)
{
	classGroupType *type;
	nameSpaceClient *nsc;
	int retval = 0;
	classGroup *cg;
	
	nsc = (nameSpaceClient *)
		getTeamInternalResource(0, &metalNameSpaceResource.link);
	/* Try to find the class group type to import */
	if( (type = findClassGroupType(metalNameSpaceResource.types, cgName)) )
	{
		if( (cg = createClassGroup(type, nsc->cGroup)) )
		{
			/* Subscribe ourselves to the new group */
			subscribeViewClient(nsc, &cg->vGroups.data[0]);
			/* Export the individual classes */
			if( exportClasses(nsc, cg, einfo) )
			{
				/* Add the group to our exports */
				addClassGroup(&nsc->exportedClasses, cg);
				retval = 1;
			}
			else
			{
				/* XXX unsubscribeViewClient */
				teamLogEntry("namespace",
					     LEA_String, "Error", "Unable to export class group", -1,
					     LEA_String, "classGroup",
					     cgName->data, -1,
					     LEA_TAG_DONE);
			}
		}
		else
		{
			postOutOfMemory(einfo);
		}
	}
	else
	{
		postExceptionMessage(
			einfo,
			"edu.utah.janosvm.resources."
			"ClassGroupNotFoundException",
			"%s",
			cgName->data);
		teamLogEntry("namespace",
			     LEA_String, "Error", "Could not find class group for export.", -1,
			     LEA_String, "classGroup", cgName->data, -1,
			     LEA_TAG_DONE);
	}
	return( retval );
}

Method *findMethodFromPC(uintp pc)
{
	memResourceClient *mrc = gc_get_resource((void *)pc);
	Method *retval = 0;

	/*
	 * Get a reference to the team that owns this code, we shouldn't have
	 * to worry about it going away since we're using the code...  So its
	 * either exported or we're visiting the team.
	 */
	if( mrc )
	{
		void *pc_key;

		if( (pc_key = GC_getObjectBase((Collector *)mrc->collector,
					       (void *)pc)) )
		{
			jitCodeHeader *jch;

			jch = (jitCodeHeader *)pc_key;
			retval = jch->method;
		}
	}
	return( retval );
}

struct Hjava_lang_Class *getNativeCallerClass(void)
{
	struct Hjava_lang_Class *retval = 0;
	void *fp, *pc;
	Method *meth;

	getCallerFrameAndPC(&fp, &pc);
	pc = getFrameReturnPC(getPreviousFramePointer(fp));
	if( (meth = findMethodFromPC((uintp)pc)) )
	{
		retval = meth->class;
	}
	return( retval );
}

void edu_utah_janosvm_resources_NameSpaceSpec_nativeConstructor(
	struct Hedu_utah_janosvm_resources_NameSpaceSpec *jnss,
	HArrayOfObject *nsg)
{
	nameSpaceSpec *nss = OBJECT2NATIVE(&metalNameSpaceResource.link.
					   specNativeClass,
					   (Hjava_lang_Object *)jnss);
	struct Hedu_utah_janosvm_resources_NameSpaceGroup *curr;
	Utf8Const *cg, *vg;
	int error = 0, lpc;
	errorInfo einfo;
	sparseID_t id;

	require(nsg != 0);

	nss->spec.handler = &metalNameSpaceResource.link;
	for( lpc = 0; !error && (lpc < ARRAY_SIZE(nsg)); lpc++ )
	{
		extern Method *GetTeamIDMethod;
		
		curr = ((struct Hedu_utah_janosvm_resources_NameSpaceGroup **)
			ARRAY_DATA(nsg))[lpc];
		if( curr == NULL )
			continue;

		/* XXX This seems gross. */
		id = do_execute_java_method(unhand(curr)->team,
					    0,
					    0,
					    GetTeamIDMethod, 0).i;

		require(curr->classGroup);
		require(curr->viewGroup);
		
		cg = stringJava2Utf8ConstReplace(unhand(curr)->classGroup,
						 0, 0);
		vg = stringJava2Utf8ConstReplace(unhand(curr)->viewGroup,
						 0, 0);
		if( cg && vg )
		{
			error = !subscribeView(nss, id, cg, vg);
		}
		else
		{
			error = 1;
		}
		if( error )
		{
			postOutOfMemory(&einfo);
			if( cg )
				utf8ConstRelease(cg);
			if( vg )
				utf8ConstRelease(vg);
		}
	}
	if( error )
	{
		throwError(&einfo);
	}
}

void edu_utah_janosvm_resources_NameSpaceSpec_nativeFinalizer(
	struct Hedu_utah_janosvm_resources_NameSpaceSpec *jnss)
{
	struct Hedu_utah_janosvm_resources_ResourceSpec *jrs =
		(struct Hedu_utah_janosvm_resources_ResourceSpec *)jnss;
	nameSpaceSpec *nss = OBJECT2NATIVE(&metalNameSpaceResource.link.
					   specNativeClass,
					   (Hjava_lang_Object *)jnss);

	if( !jrs->found )
	{
		int lpc;
		
		for( lpc = 0; lpc < nss->viewSubscriptions.size; lpc++ )
		{
			utf8ConstRelease(nss->viewSubscriptions.data[lpc].
					 className);
			utf8ConstRelease(nss->viewSubscriptions.data[lpc].
					 viewName);
		}
		gc_free(nss->viewSubscriptions.data);
	}
}

jint
edu_utah_janosvm_resources_NameSpaceSpec_getGroupCount(
	struct Hedu_utah_janosvm_resources_NameSpaceSpec *jnss)
{
	nameSpaceSpec *nss = OBJECT2NATIVE(&metalNameSpaceResource.link.
					   specNativeClass,
					   (Hjava_lang_Object *)jnss);

	return( nss->viewSubscriptions.size );
}

struct Hedu_utah_janosvm_resources_NameSpaceGroup *
edu_utah_janosvm_resources_NameSpaceSpec_getGroup(
	struct Hedu_utah_janosvm_resources_NameSpaceSpec *jnss,
	jint index)
{
	nameSpaceSpec *nss = OBJECT2NATIVE(&metalNameSpaceResource.link.
					   specNativeClass,
					   (Hjava_lang_Object *)jnss);
	struct Hedu_utah_janosvm_resources_NameSpaceGroup *retval;
	errorInfo einfo;
	groupNames *gn;
	jteam *jt;
	
	if( (index < 0) || (index >= nss->viewSubscriptions.size) )
	{
	    postExceptionMessage(&einfo,
				 JAVA_LANG(IllegalArgumentException),
				 "Invalid index: %d",
				 index);
	    throwError(&einfo);
	}
	gn = &nss->viewSubscriptions.data[index];
	retval = (struct Hedu_utah_janosvm_resources_NameSpaceGroup *)
		newObject(NameSpaceGroupClass);
	jt = grabTeamWithID(gn->teamID);
	retval->team = do_execute_java_method(0, 0, 0,
					      FindTeamMethod,
					      1,
					      0,
					      NATIVE2OBJECT(&jteamNativeClass,
							    jt)).l;
	retval->classGroup = checkPtr(stringC2Java(gn->className->data));
	retval->viewGroup = checkPtr(stringC2Java(gn->viewName->data));
	return( retval );
}

struct Hedu_utah_janosvm_resources_NameSpaceSpec *
edu_utah_janosvm_resources_NameSpaceSpec_current(void)
{
	struct Hedu_utah_janosvm_resources_NameSpaceSpec *retval;
	nameSpaceClient *nsc = (nameSpaceClient *)
		getTeamInternalResource(0, &metalNameSpaceResource.link);
	viewSubscriberPool *vsp;
	nameSpaceSpec *nss;
	int total = 0;
	int iLockRoot;
	
	retval = (struct Hedu_utah_janosvm_resources_NameSpaceSpec *)
		newObject(metalNameSpaceResource.link.specNativeClass.peer);
	retval->found = 1;
	nss = OBJECT2NATIVE(&metalNameSpaceResource.link.specNativeClass,
			    (Hjava_lang_Object *)retval);
	nss->spec.handler = &metalNameSpaceResource.link;
	
	lockMutex(nsc);
	vsp = nsc->views;
	while( vsp )
	{
		total += vsp->used;
		vsp = vsp->next;
	}
	if( (nss->viewSubscriptions.data =
	     gc_malloc(sizeof(groupNames) * total, GC_ALLOC_NAME_SPACE)) )
	{
		groupNames *gn = nss->viewSubscriptions.data;
		int lpc;
		
		nss->viewSubscriptions.size = total;
		vsp = nsc->views;
		while( vsp )
		{
			for( lpc = 0; lpc < vsp->used; lpc++ )
			{
				gn->teamID = gc_get_resource(vsp->data[lpc].
							     vGroup->cGroup)->
					client.owner->elemID;
				gn->className = vsp->data[lpc].vGroup->cGroup->
					type->name;
				gn->viewName = vsp->data[lpc].vGroup->type->
					name;
				gn++;
			}
			vsp = vsp->next;
		}
	}
	else
	{
		retval = 0;
	}
	unlockMutex(nsc);
	
	return( retval );
}

Hjava_lang_Object *
edu_utah_janosvm_resources_NameSpaceSpec_clone(
	struct Hedu_utah_janosvm_resources_NameSpaceSpec *jnss)
{
	struct Hedu_utah_janosvm_resources_NameSpaceSpec *retval;
	nameSpaceSpec *nss, *retval_nss;
	int lpc;
	
	retval = (struct Hedu_utah_janosvm_resources_NameSpaceSpec *)
		newObject(metalNameSpaceResource.link.specNativeClass.peer);
	nss = OBJECT2NATIVE(&metalNameSpaceResource.link.specNativeClass,
			    (Hjava_lang_Object *)jnss);
	retval_nss = OBJECT2NATIVE(&metalNameSpaceResource.link.
				   specNativeClass,
				   (Hjava_lang_Object *)retval);
	retval_nss->spec.handler = &metalNameSpaceResource.link;

	retval_nss->viewSubscriptions.data =
		checkPtr(gc_malloc(sizeof(groupNames) *
				   nss->viewSubscriptions.size,
				   GC_ALLOC_NAME_SPACE));
	retval_nss->viewSubscriptions.size = nss->viewSubscriptions.size;
	for( lpc = 0; lpc < nss->viewSubscriptions.size; lpc++ )
	{
		utf8ConstAddRef(nss->viewSubscriptions.data[lpc].className);
		utf8ConstAddRef(nss->viewSubscriptions.data[lpc].viewName);
		retval_nss->viewSubscriptions.data[lpc] =
			nss->viewSubscriptions.data[lpc];
	}
	return( &retval->base );
}

void edu_utah_janosvm_resources_NameSpaceResource_exportClassGroup(
	struct Hjava_lang_String *jType)
{
	errorInfo einfo;
	int ok = 0;
	
	if( jType != 0 )
	{
		Utf8Const *type;
		
		if( (type = stringJava2Utf8ConstReplace(jType, 0, 0)) )
		{
			ok = exportClassGroup(type, &einfo);
			utf8ConstRelease(type);
		}
		else
		{
			postOutOfMemory(&einfo);
		}
	}
	else
	{
		postException(&einfo, JAVA_LANG(NullPointerException));
	}
	if( !ok )
	{
		throwError(&einfo);
	}
}

void edu_utah_janosvm_resources_NameSpaceResource_importClassGroup(
	struct Hedu_utah_janosvm_resources_NameSpaceGroup *jnsg)
{
	extern Method *GetTeamIDMethod;
	
	nameSpaceClient *nsc = (nameSpaceClient *)
		getTeamInternalResource(0, &metalNameSpaceResource.link);
	Utf8Const *cg, *vg;
	errorInfo einfo;
	groupNames gn;
	sparseID_t id;
	int error = 0;
	
	/* XXX This seems gross. */
	id = do_execute_java_method(unhand(jnsg)->team,
				    0,
				    0,
				    GetTeamIDMethod, 0).i;
	cg = stringJava2Utf8ConstReplace(unhand(jnsg)->classGroup, 0, 0);
	vg = stringJava2Utf8ConstReplace(unhand(jnsg)->viewGroup, 0, 0);

	gn.teamID = id;
	gn.className = cg;
	gn.viewName = vg;

	if( cg && vg )
	{
		jteamReservation *jr;
		
		if( (jr = pushTeamReservation(kernelTeam)) )
		{
			error = !subscribeViewFromSpec(nsc, &gn, &einfo);
			popTeamReservation(jr);
		}
		else
		{
			postExceptionMessage(&einfo,
					     DEAD_TEAM_EXCEPTION_S,
					     "kernel");
		}
	}
	else
	{
		postOutOfMemory(&einfo);
		error = 1;
	}
	if( cg )
	{
		utf8ConstRelease(cg);
	}
	if( vg )
	{
		utf8ConstRelease(vg);
	}
	if( error )
	{
		throwError(&einfo);
	}
}

// XXX this is gross
static void stuffStringIn(const char* cstr, HArrayOfChar *sbuf, int* offset)
{
	const char* p;
	for (p = cstr; *p != '\0'; p++)
	{
		require((*offset) < obj_length(sbuf)); // XXX handle better
		unhand_array(sbuf)->body[(*offset)++] = (jchar)(0x00FF & *p);
	}
}

int edu_utah_janosvm_resources_NameSpaceSpec_nativeToString(
	struct Hedu_utah_janosvm_resources_NameSpaceSpec *jnss,
	HArrayOfChar *sbuf)
{
	int i;
	int sbufoff;
	const nameSpaceSpec *nss = OBJECT2NATIVE(&metalNameSpaceResource.link.
						 specNativeClass,
						 (Hjava_lang_Object *)jnss);
	require(jnss);
	require(nss);
	require(sbuf);
	require(obj_length(sbuf) > 0);

	
	sbufoff = 0;
	stuffStringIn("NameSpaceSpec[", sbuf, &sbufoff);

	for (i = 0; i < nss->viewSubscriptions.size; i++)
	{
		Utf8Const *className = nss->viewSubscriptions.data[i].className;
		Utf8Const *viewName = nss->viewSubscriptions.data[i].viewName;

		if (i != 0)
			stuffStringIn(", ", sbuf, &sbufoff);

		// XXX ignores limits, ignores non-ascii utf8 chars
		if (utf8ConstEqual(className, viewName))
			stuffStringIn(className->data, sbuf, &sbufoff);
		else
		{
			stuffStringIn(className->data, sbuf, &sbufoff);
			stuffStringIn(" as ", sbuf, &sbufoff);
			stuffStringIn(viewName->data, sbuf, &sbufoff);
		}
	}

	stuffStringIn("]", sbuf, &sbufoff);

	return sbufoff;
}

