/*
 * compileQueue.h
 * Manage the (re)compile queue.
 * 
 * Copyright (c) 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

#include "config.h"

#include "assert_pp.h"
#include "native.h"
#include "errors.h"
#include "compileQueue.h"

#include "gc.h"
#include <janosvm/jteam.h>

int cqInitCompileQueue(compileQueue *cq)
{
	int lpc, retval = 1;

	require(cq != 0);

	cq->lock = 0;
	for( lpc = 0; lpc < CQ_MAX; lpc++ )
	{
		lnNewList((struct lnMinList *)&cq->queues[lpc]);
	}
	return( retval );
}

void cqKillCompileQueue(compileQueue *cq)
{
}

void cqAddCode(compileQueue *cq, compile_type_t type, jitCodeHeader *jch)
{
	int iLockRoot;

	require(cq != 0);
	require(jch != 0);
	require(!kernelTeam ||
		(currentTeam() == gc_get_resource(jch)->client.owner));
	
	lockMutex(cq);
	if( jch->flags & JCF_DYING )
	{
		/*
		 * The method was just removed from the queue and is probably
		 * being recompiled.
		 */
	}
	else
	{
		struct lnMinList *ml = 0;
		
		if( jch->flags & JCF_ON_LIST )
		{
			lnRemove((struct lnMinNode *)&jch->link);
		}
		ml = (struct lnMinList *)&cq->queues[type];
		switch( type )
		{
		case CQ_RECOMPILE:
			/*
			 * The method contains lazy lookups that have been
			 * used, therefore it is sub-optimal.
			 */
			jch->flags |= JCF_SUB_OPTIMAL;
			break;
		case CQ_SPECULATE:
			break;
		case CQ_FINISHED:
			/*
			 * The method may still be sub-optimal, but we only
			 * care if the sub-optimal parts are actually used.
			 * Therefore, we assume it is optimal until a lazy
			 * lookup is performed.
			 */
			jch->flags &= ~JCF_SUB_OPTIMAL;
			break;
		case CQ_MAX:
			assert(0);
			break;
		}
		lnAddHead(ml, (struct lnMinNode *)&jch->link);
		jch->flags |= JCF_ON_LIST;
	}
	unlockMutex(cq);
}

void cqRemCode(compileQueue *cq, jitCodeHeader *jch)
{
	int iLockRoot;
	
	lockMutex(cq);
	if( jch->flags & JCF_ON_LIST )
	{
		lnRemove((struct lnMinNode *)&jch->link);
		jch->flags &= ~JCF_ON_LIST;
		jch->flags |= JCF_DYING;
	}
	unlockMutex(cq);
}

jitCodeHeader *cqNextCompile(compileQueue *cq)
{
	jitCodeHeader *retval;
	int iLockRoot;

	require(cq != 0);
	
	lockMutex(cq);
	if( !(retval = (jitCodeHeader *)
	      lnRemHead((struct lnMinList *)&cq->queues[CQ_RECOMPILE])) )
	{
		retval = (jitCodeHeader *)
			lnRemHead((struct lnMinList *)
				  &cq->queues[CQ_SPECULATE]);
	}
	if( retval )
	{
		retval->flags &= ~JCF_ON_LIST;
		retval->flags |= JCF_DYING;
	}
	unlockMutex(cq);
	return( retval );
}

void cqIterate(compileQueue *cq, cqCodeIterator_t iterator, void *arg)
{
	int iLockRoot;
	int lpc;
	
	require(cq != 0);
	require(iterator != 0);

	lockMutex(cq);
	for( lpc = 0; lpc < CQ_MAX; lpc++ )
	{
		jitCodeHeader *jch;
		
		jch = (jitCodeHeader *)cq->queues[lpc].lh_Head;
		while( jch->link.ln_Succ )
		{
			iterator(cq, jch, arg);
			jch = (jitCodeHeader *)jch->link.ln_Succ;
		}
	}
	unlockMutex(cq);
}
