/*
 * compileQueue.h
 * Manage the (re)compile queue.
 *
 * Copyright (c) 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

#ifndef _compileQueue_h
#define _compileQueue_h

/*
 * The "compile queue" is used to manage any method compilations that are not
 * directly triggered by user interaction.  For example, if the VM were idle
 * for a period of time it could begin to speculatively (re)compile code to
 * improve performance. Currently, this facility is only used to recompile code
 * that contains lazy field/method lookups.
 */

#include "locks.h"
#include "jit3/machine.h"
#include <janosvm/listNode.h>

/*
 * Types of compiling to be done:
 *   CQ_RECOMPILE - Recompile a sub-optimal method.
 *   CQ_SPECULATE - Do a speculative compile of a method.
 *   CQ_FINISHED - The method has finished compiling.
 */
typedef enum {
	CQ_RECOMPILE,
	CQ_SPECULATE,
	CQ_FINISHED,
	CQ_MAX,
} compile_type_t;

/*
 * Root structure for the compile queue.
 */
typedef struct _compileQueue {
	iLock *lock;
	struct lnList queues[CQ_MAX];
} compileQueue;

/*
 * Initialize a compileQueue object.
 *
 * cq - The compileQueue object to initialize.
 * returns - True if the initialization was successful.
 */
int cqInitCompileQueue(compileQueue *cq);
/*
 * Finalize a compileQueue object.
 *
 * cq - The compileQueue object to finalize.
 */
void cqKillCompileQueue(compileQueue *cq);

/*
 * Add code to a compileQueue.
 *
 * cq - The queue to add the jit'ed code to.
 * type - The type of processing to perform on the code (e.g. recompile,
 *   speculative compile).
 * jch - The code to add to the queue.
 */
void cqAddCode(compileQueue *cq, compile_type_t type, jitCodeHeader *jch);
/*
 * Remove code from a compileQueue.
 *
 * cq - The queue to add the jit'ed code to.
 * jch - The code to add to the queue.
 */
void cqRemCode(compileQueue *cq, jitCodeHeader *jch);
/*
 * Get the next method to be (re)compiled.
 *
 * cq - The queue to containing methods to be (re)compiled.
 * returns - NULL or a jitCodeHeader object describing a method to be compiled.
 */
jitCodeHeader *cqNextCompile(compileQueue *cq);

/*
 * Iterator function type.
 *
 * cq - The compileQueue being iterated over.
 * jch - The next jitCodeHeader object to be processed.
 * arg - User specified argument.
 */
typedef void (*cqCodeIterator_t)(compileQueue *cq,
				 jitCodeHeader *jch,
				 void *arg);

/*
 * Iterate over a compileQueue.
 *
 * cq - The queue to iterate over.
 * iterator - The iterator function to call for each jitCodeHeader in the
 * queue.
 * arg - User specified argument to be passed to the iterator.
 */
void cqIterate(compileQueue *cq, cqCodeIterator_t iterator, void *arg);

/*
 * Call back used to retrieve the current compileQueue.
 *
 * returns - The current compile queue.
 */
extern compileQueue *currentCompileQueue(void);

/*
 * Call back used to iterate over the global list of compile queues.
 *
 * ptr - An iteration pointer.  When initially iterating over the list the
 *   pointer should be initialized to zero.
 * returns - A compileQueue or NULL if the end of the list has been
 *   encountered.
 */
extern compileQueue *nextCompileQueue(void **ptr);

#endif
