/*
 * lookup.c
 * Various lookup calls for resolving objects, methods, exceptions, etc.
 *
 * Copyright (c) 1996, 1997
 *	Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */

#include "debug.h"
#include "config.h"
#include "config-std.h"
#include "config-mem.h"
#include "jtypes.h"
#include "errors.h"
#include "gtypes.h"
#include "access.h"
#include "object.h"
#include "constants.h"
#include "classMethod.h"
#include "lookup.h"
#include "exception.h"
#include "thread.h"
#include "baseClasses.h"
#include "stringSupport.h"
#include "machine.h"
#include "locks.h"
#include "soft.h"
#include <janosvm/nameSpace.h>

/*
 * Lookup a method reference and get the various interesting bits.
 *
 * Return false if unsuccessful because of a malformed class file.
 * Returns true otherwise. 
 *
 * Note that even if it returns true, the method may not be found.
 * call->method is set to NULL in this case.
 */
bool
getMethodSignatureClass(constIndex idx, Hjava_lang_Class* this, bool loadclass, int isSpecial, callInfo* call, errorInfo *einfo)
{
	constants* pool;
	constIndex ci;
	constIndex ni;
	Hjava_lang_Class* class;
	classView* cView;
	Utf8Const* name;
	Utf8Const* sig;
	Method* mptr;
	classInfo cinfo;
	int i;

	call->class = 0;
	call->method = 0;
	call->signature = 0;
	call->name = 0;
	call->cname = 0;

	pool = CLASS_CONSTANTS(this);
	if (pool->tags[idx] != CONSTANT_Methodref &&
	    pool->tags[idx] != CONSTANT_InterfaceMethodref) {
DBG(RESERROR,	dprintf("No Methodref found for idx=%d\n", idx);	)
		postExceptionMessage(einfo, JAVA_LANG(ClassFormatError),
			"method name unknown, tag = %d", pool->tags[idx]);
                return (false);
	}

	ni = METHODREF_NAMEANDTYPE(idx, pool);
	name = WORD2UTF(pool->data[NAMEANDTYPE_NAME(ni, pool)]);
	sig = WORD2UTF(pool->data[NAMEANDTYPE_SIGNATURE(ni, pool)]);
	ci = METHODREF_CLASS(idx, pool);
	
	call->name = name;
	call->signature = sig;

	if (getClass(ci, this, loadclass, &cinfo, einfo) == false) {
		call->cname = cinfo.cname;
		countInsAndOuts(sig->data,
				&call->in,
				&call->out,
				&call->rettype);
		if (loadclass)
			return (false);
		else
			return (true);
	}
	else {
		class = cinfo.clazz;
	}
	cView = getClassView(class, einfo);
	if (cView == NULL) {
		call->cname = WORD2UTF(pool->data[ci]);
		countInsAndOuts(sig->data, &call->in, &call->out, &call->rettype);
		/*
		 * XXX Have to use postNoClassDefFoundError() so that
		 * verification can proceed despite this error.  Loses
		 * a bunch of info about the error...
		 */
		discardErrorInfo(einfo);
		postNoClassDefFoundError(einfo, call->cname->data);
		return false;
	}
	call->cname = cView->name;
	if (class->state == CSTATE_FAILED) {
		if (loadclass)
		{
			postExceptionMessage(einfo,
					     JAVA_LANG(NoClassDefFoundError),
					     "%s: static initializer failed",
					     call->cname->data);
			return (false);
		}
		else
		{
			return (true);
		}
	}
	assert(class->state >= CSTATE_LINKED);
	
	if (isSpecial == 1) {
		if (!utf8ConstEqual(name, constructor_name) && class !=
		    this && instanceof(class, this)) {
			class = this->superclass;
		}
	}
	
	call->class = class;
	call->cname = cView->name;
	call->method = 0;
	/* Find method - we don't use findMethod(...) yet since this
	 * will initialise our class (and we don't want to do that).
	 */
	mptr = 0;
	for (; class != 0; class = class->superclass) {
		cView = getClassView(class, einfo);
		if (cView == NULL) {
			countInsAndOuts(sig->data, &call->in, &call->out, &call->rettype);
			/*
			 * XXX Have to use postNoClassDefFoundError() so that
			 * verification can proceed despite this error.  Loses
			 * a bunch of info about the error...
			 */
			discardErrorInfo(einfo);
			postNoClassDefFoundError(einfo, call->cname->data);
			return false;
		}

		mptr = findMethodLocal(cView, name, sig);
		if (mptr != NULL) {
			if (GC_CHECK(this, mptr->class, einfo)) {
				return (false);
			}
			call->method = mptr;
			break;
		}
	}
	
	/* If we've not found anything and we're searching interfaces,
	 * search them too.
	 */
	if (mptr == 0 && isSpecial == 2) {
		class = call->class;
		for (i = class->total_interface_len - 1; i >= 0; i--) {
			classView* icView = getClassView(class->interfaces[i], einfo);
			if (icView == NULL)
			{
				countInsAndOuts(sig->data, &call->in, &call->out, &call->rettype);
				/*
				 * XXX Have to use postNoClassDefFoundError() so that
				 * verification can proceed despite this error.  Loses
				 * a bunch of info about the error...
				 */
				discardErrorInfo(einfo);
				postNoClassDefFoundError(einfo, call->cname->data);
				return false;
			}
			
			mptr = findMethodLocal(icView, name, sig);
			if (mptr != 0) {
				call->method = mptr;
				break;
			}
		}
	}

	/* Calculate in's and out's */
	countInsAndOuts(sig->data, &call->in, &call->out, &call->rettype);

DBG(MLOOKUP,	
	if (loadClass)
		dprintf("getMethodSignatureClass(%s,%s,%s) -> %s\n",
			getClassTraceName(call->class),
			name->data, sig->data, 
			(call->method ? "success" : "failure"));	)
	return (true);
}

/*
 * Get and link the class to which constant pool index idx in class this
 * refers.  The returned class object is at least LINKED.
 */
bool
getClass(constIndex idx, Hjava_lang_Class* this, bool loadclass,
	 classInfo *cinfo,
	 errorInfo *einfo)
{
	constants* pool;
	Utf8Const *name;
	Hjava_lang_Class* class;
	int tag;
	int iLockRoot;

	pool = CLASS_CONSTANTS(this);

	tag = pool->tags[idx];

	cinfo->clazz = 0;
	cinfo->cname = 0;
	
	switch (tag) {
	case CONSTANT_ResolvedClass:
		cinfo->clazz = CLASS_CLASS(idx, pool);
		cinfo->cname = 0;
		return (true);

	case CONSTANT_Class:
		/* The class may be resolved by another thread so we better
		 * lock and get the tag & name again just in case.  If we
		 * have been resolved then we just return the answer.
		 */
		lockClass(this);
		tag = pool->tags[idx];
		cinfo->cname = name = WORD2UTF(pool->data[idx]);
		unlockClass(this);

		if (tag == CONSTANT_ResolvedClass) {
			cinfo->clazz = CLASS_CLASS(idx, pool);
			cinfo->cname = 0;
			return (true);
		}
		break;

	default:
		postException(einfo, JAVA_LANG(ClassFormatError));
		return (false);
	}

	/* Find the specified class.
	 * NB: the name is *not* a signature.
	 */
	if (name->data[0] == '[') {
		class = loadArray(name, this->loader, einfo);
		if (!class) {
			if( (einfo->type & KERR_EXCEPTION) &&
			    !strcmp(einfo->classname,
				    JAVA_LANG(ClassNotFoundException)) )
			{
				errorInfo einfo_copy = *einfo;
				
				postNoClassDefFoundError(einfo, einfo->mess);
				discardErrorInfo(&einfo_copy);
			}
			return (false);
		}
	}
	else if (loadclass) {
		class = loadClass(name, this->loader, einfo);
		if (!class) {
			if( (einfo->type & KERR_EXCEPTION) &&
			    !strcmp(einfo->classname,
				    JAVA_LANG(ClassNotFoundException)) )
			{
				errorInfo einfo_copy = *einfo;
				
				postNoClassDefFoundError(einfo, einfo->mess);
				discardErrorInfo(&einfo_copy);
			}
			return (false);
		}
	}
	else if (!(class = findValidClassMapping(name, this->loader)) ||
		 (processClass(class, CSTATE_LINKED, einfo) == false)) {
		return (false);
	}

	assert((class->state == CSTATE_FAILED) ||
	       (class->state >= CSTATE_LINKED));

	/* Lock the class while we update the constant pool.  Someone
	 * may have done this already but we don't care.
	 */
	lockClass(this);
	if( !GC_SET_SAFE(pool, pool->data[idx], (ConstSlot)class, einfo) )
	{
		pool->tags[idx] = CONSTANT_ResolvedClass;
	}
	else
	{
		class = 0;
	}
	unlockClass(this);

	if( class )
	{
		cinfo->clazz = class;
		return (true);
	}
	else
	{
		cinfo->clazz = 0;
		return (false);
	}
}

bool
getField(constIndex idx, Hjava_lang_Class* this, bool isStatic, bool loadclass, fieldInfo* ret, errorInfo *einfo)
{
	constants* pool;
	constIndex ci;
	constIndex ni;
	classInfo cinfo;
	Field* field;
	Hjava_lang_Class* class;

	ret->field = 0;
	ret->class = 0;
	
	pool = CLASS_CONSTANTS(this);
	if (pool->tags[idx] != CONSTANT_Fieldref) {
DBG(RESERROR,	dprintf("No Fieldref found\n");				)
		postExceptionMessage(einfo, JAVA_LANG(ClassFormatError),
			"tag was %d", pool->tags[idx]);
		return (false);
	}

	ci = FIELDREF_CLASS(idx, pool);

	ni = FIELDREF_NAMEANDTYPE(idx, pool);

	ret->cname = WORD2UTF(pool->data[ci]);
	ret->name = WORD2UTF(pool->data[NAMEANDTYPE_NAME(ni, pool)]);
	ret->signature = WORD2UTF(pool->data[NAMEANDTYPE_SIGNATURE(ni, pool)]);

	if (getClass(ci, this, loadclass, &cinfo, einfo) == false) {
		if (loadclass)
			return (false);
		else
			return (true);
	}
	class = cinfo.clazz;

DBG(FLOOKUP,	dprintf("*** getField(%s,%s,%s)\n",
			getClassTraceName(class),
			WORD2UTF(pool->data[NAMEANDTYPE_NAME(ni, pool)])->data, 
			WORD2UTF(pool->data[NAMEANDTYPE_SIGNATURE(ni, pool)])->data);
    )

	ret->class = class;
	field = lookupClassField(class, WORD2UTF(pool->data[NAMEANDTYPE_NAME(ni, pool)]), isStatic, einfo);
	if (field == 0) {
		return (false);
	}
	if (!utf8ConstEqual(field->signature, ret->signature)) {
		postExceptionMessage(einfo,
				     JAVA_LANG(NoSuchFieldError),
				     "%s.%s %s",
				     ret->cname->data,
				     ret->name->data,
				     ret->signature->data);
		return (false);
	}

	/* XXX Should we verify that ret->signature matches field? */

	/* XXX Should we verify that ret->signature matches field? */

	ret->field = field;
	return (true);
}

/*
 * Lookup a method (and translate) in the specified class.  If no method matches
 * the name, NULL is returned.
 */
Method*
findMethodLocal(classView* cv, Utf8Const* name, Utf8Const* signature)
{
	Hjava_lang_Class* class;
	Method* mptr;
	int n;

	require(cv != 0);
	require(name != 0);
	require(signature != 0);

	class = cv->cl;

	/*
	 * Lookup method - this could be alot more efficient but never mind.
	 * Also there is no attempt to honour PUBLIC, PRIVATE, etc.
	 */
	n = CLASS_NMETHODS(class);
	for (mptr = CLASS_METHODS(class); --n >= 0; ++mptr) {
		methodView* mView = getMethodView(cv, mptr);
		if (utf8ConstEqual (name, mptr->name) && utf8ConstEqual (signature, METHOD_VIEW_SIG(mView))) {
DBG(MLOOKUP,
			dprintf("findMethodLocal(%s,%s,%s) -> %p\n",
				getClassTraceName(class),
				name->data, signature->data, mptr); )

			return (mptr);
		}
	}

DBG(MLOOKUP,
	dprintf("findMethodLocal(%s,%s,%s) -> NOT FOUND\n",
		getClassTraceName(class),
		name->data, signature->data); )

	return NULL;
}

/*
 * Find a method that might otherwise be masked because the current team does
 * not have a view on it, but it does have a view on the parent class.
 */
static Method*
findMethodIndex(Hjava_lang_Class *class, int idx)
{
	Method *retval = 0;
	int lpc;
	
	for( lpc = 0; (lpc < class->nmethods) && (retval == 0); lpc++ )
	{
		if( class->methods[lpc].idx == idx )
		{
			retval = &class->methods[lpc];
		}
	}
	return( retval );
}

/*
 * Lookup a method (and translate) in the specified class.
 */
Method*
findMethod(Hjava_lang_Class* class, Utf8Const* name, Utf8Const* signature, errorInfo *einfo)
{
	Hjava_lang_Class* class_orig = class;
	bool skipped = true;
	bool success;

	/*
	 * Waz CSTATE_LINKED - Must resolve constants before we do any
	 * translation.  Might not be right though ... XXX
	 */
	if (class->state < CSTATE_USABLE) {
		success = processClass(class, CSTATE_COMPLETE, einfo);
		if (!success)
			return (0);
	}

	/*
	 * Lookup method - this could be alot more efficient but never mind.
	 * Also there is no attempt to honour PUBLIC, PRIVATE, etc.
	 */
	for (; class != 0; class = class->superclass) {
		Method* mptr;
		classView* cView;

		cView = getClassView(class, einfo);
		/*
		 * The parent's may be visible, so try them.  If they all fail
		 * the einfo should still be set as a segfault, although the
		 * class name might be a bit screwy.
		 */
		if (cView == 0)
		{
			skipped = true;
			continue;
		}

		mptr = findMethodLocal(cView, name, signature);
		if (mptr != NULL) {
			if( skipped && (mptr->idx >= 0) )
			{
				int idx;

				idx = mptr->idx;
				mptr = 0;
				for( class = class_orig;
				     (class != 0) && (mptr == 0);
				     class = class->superclass )
				{
					mptr = findMethodIndex(class, idx);
				}
			}
			return mptr;
		}
	}
	postExceptionMessage(einfo,
			     JAVA_LANG(NoSuchMethodError),
			     "%s.%s",
			     CLASS_CNAME(class_orig),
			     name->data);
	return (0);
}
