/*
 * soft.c
 * Soft instruction support.
 *
 * Copyright (c) 1996, 1997
 *	Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 */

#include "debug.h"
#define	TDBG(s)

#include "config.h"
#include "config-std.h"
#include "config-math.h"
#include "config-mem.h"
#include <stdarg.h>
#include "gtypes.h"
#include "bytecode.h"
#include "slots.h"
#include "access.h"
#include "object.h"
#include "constants.h"
#include "errors.h"
#include "classMethod.h"
#include "baseClasses.h"
#include "lookup.h"
#include "exception.h"
#include "locks.h"
#include "soft.h"
#include "external.h"
#include "thread.h"
#include "baseClasses.h"
#include "itypes.h"
#include "machine.h"
#include "support.h"
#include "utf8const.h"
#include "compileQueue.h"
#include "threadData-inline.h"
#include <janosvm/nameSpace.h>
#include "fp.h"

/*
 * soft_new
 */
void*
soft_new(Hjava_lang_Class* c)
{
	Hjava_lang_Object* obj;
	errorInfo info;

	jthread_guardcheck();

	if (c->state != CSTATE_COMPLETE && processClass(c, CSTATE_COMPLETE, &info) == false) {
		goto bad;
	}
	obj = newObjectChecked(c, &info);
	if (obj == 0) {
		goto bad;
	}

DBG(NEWINSTR,	
	dprintf("New object of type %s (%d,%p)\n", 
		getClassTraceName(c), c->bfsize, obj); )

	return (obj);
bad:
	throwError(&info);
	return (0);
}

void*
soft_lazynew(Hjava_lang_Class* c, int idx)
{
	Hjava_lang_Object *retval = 0;
	memResourceClient *mrc;
	jitCodeHeader *jch;
	jteam *owner = 0;
	errorInfo einfo;
	classInfo cinfo;
	void *fp, *pc;

	jthread_guardcheck();

	cinfo.clazz = 0;
	getCallerFrameAndPC(&fp, &pc);
	jch = findJITCodeHeader((uintp)pc);
	if( kernelTeam && (mrc = gc_get_resource(jch)) )
	{
		owner = mrc->client.owner;
	}
	if( owner && (switchToTeam(owner, 0) == 0) )
	{
		postException(&einfo, DEAD_TEAM_EXCEPTION_S);
		throwError(&einfo);
	}
	else if( getClass(idx, c, true, &cinfo, &einfo) &&
		 processClass(cinfo.clazz, CSTATE_COMPLETE, &einfo) )
	{
		cqAddCode(currentCompileQueue(), CQ_RECOMPILE, jch);
	}
	else
	{
		cinfo.clazz = 0;
	}
	if( owner )
	{
		returnFromTeam();
	}
	if( cinfo.clazz )
	{
		retval = newObjectChecked(cinfo.clazz, &einfo);
	}
	if( !retval )
	{
		throwError(&einfo);
	}
	return( retval );
}

/*
 * soft_newarray
 */
void*
soft_newarray(jint type, jint size)
{
	Hjava_lang_Object* obj;
	errorInfo info;

	jthread_guardcheck();

	if (size < 0) {
		char buf[16];

		sprintf(buf, "%d", size);
		throwException(NegativeArraySizeException(buf));
	}

	obj = newArrayChecked(TYPE_CLASS(type), size, &info);
	if (obj == 0) {
		throwError(&info);
	}

DBG(NEWINSTR,	
	dprintf("New array of %s [%d] (%p)\n", 
		getClassTraceName(TYPE_CLASS(type)), size, obj); )

	return (obj);
}

/*
 * soft_anewarray
 */
void*
soft_anewarray(Hjava_lang_Class* elclass, jint size)
{
	Hjava_lang_Object* obj;
	errorInfo info;

	jthread_guardcheck();

	if (size < 0) {
		char buf[16];

		sprintf(buf, "%d", size);
		throwException(NegativeArraySizeException(buf));
	}

	obj = newArrayChecked(elclass, size, &info);
	if (obj == 0) {
		throwError(&info);
	}

DBG(NEWINSTR,	
	dprintf("New array object [%d] of %s (%p)\n", size,
		getClassTraceName(elclass), obj); )
	return (obj);
}

/*
 * soft_multianewarray.
 */
#define	MAXDIMS	16

#if defined(INTERPRETER)
void*
soft_multianewarray(Hjava_lang_Class* class, jint dims, slots* args)
{
	errorInfo einfo;
        int array[MAXDIMS];
        Hjava_lang_Object* obj;
        jint arg;
	int i;
	int* arraydims;

	jthread_guardcheck();

        if (dims < MAXDIMS) {
		arraydims = array;
	}
	else {
		arraydims = checkPtr(KCALLOC(dims+1, sizeof(int)));
	}

	/* stack grows up, so move to the first dimension */
	args -= dims-1;

        /* Extract the dimensions into an array */
        for (i = 0; i < dims; i++) {
		arg = args[i].v.tint;
                if (arg < 0) {
			char buf[16];
			
			sprintf(buf, "%d", arg);
                        throwException(NegativeArraySizeException(buf));
		}
                arraydims[i] = arg;
        }
        arraydims[i] = -1;

        /* Mmm, okay now build the array using the wonders of recursion */
        obj = newMultiArrayChecked(class, arraydims, &einfo);

	if (arraydims != array) {
		KFREE(arraydims);
	}

	if (!obj) {
		throwError(&einfo);
	}

        /* Return the base object */
	return (obj);
}
#endif

#if defined(TRANSLATOR)
void*
soft_vmultianewarray(Hjava_lang_Class* class, jint dims, va_list ap)
{
	errorInfo einfo;
	int array[MAXDIMS];
	int i;
	jint arg;
	Hjava_lang_Object* obj;
	int* arraydims;

	jthread_guardcheck();

        if (dims < MAXDIMS-1) {
		arraydims = array;
	}
	else {
		arraydims = checkPtr(KCALLOC(dims+1, sizeof(int)));
	}

	/* Extract the dimensions into an array */
	for (i = 0; i < dims; i++) {
		arg = va_arg(ap, jint);
		if (arg < 0) {
			char buf[16];
			
			if (arraydims != array) {
				KFREE(arraydims);
			}
			sprintf(buf, "%d", arg);
                        throwException(NegativeArraySizeException(buf));
		}
		arraydims[i] = arg;
	}
	arraydims[i] = -1;

	/* Mmm, okay now build the array using the wonders of recursion */
        obj = newMultiArrayChecked(class, arraydims, &einfo);

	if (arraydims != array) {
		KFREE(arraydims);
	}

	if (!obj) {
		throwError(&einfo);
	}

	/* Return the base object */
	return (obj);
}

void*
soft_multianewarray(Hjava_lang_Class* class, jint dims, ...)
{
	void* obj;
	va_list ap;
	
	jthread_guardcheck();

	va_start(ap, dims);
	obj = soft_vmultianewarray(class, dims, ap);
	va_end(ap);
	return (obj);
}
#endif

/*
 * soft_lookupinterfacemethod.
 */
void*
soft_lookupinterfacemethod(Hjava_lang_Object* obj, Hjava_lang_Class* ifclass, int idx)
{
	Hjava_lang_Class* cls;
	void*	ncode;
	register int i;
	register short* implementors;

	jthread_guardcheck();

	cls = OBJECT_CLASS(obj);
	implementors = ifclass->implementors;
	i = cls->impl_index;

#if 1	/* it should never be necessary to initialize the class here
	 * cause we're invoking a method on an existing, created object
	 */
        assert (cls->state >= CSTATE_USABLE);
#else
	/* initialize class if necessary */
        if (cls->state < CSTATE_USABLE) {
		errorInfo info;
		if (processClass(cls, CSTATE_COMPLETE, &info) == false) {
			throwError(&info);
		}
	}
#endif

	/* handle invocations on java.lang.Object methods via INVOKEINTERFACE */
	if (implementors == 0 || i > implementors[0]) {
		goto notfound;
	}

	/* skip word at the beginning of itable2dtable */
	ncode = cls->itable2dtable[implementors[i] + idx + 1];

	/* This means this class does not implement this interface method
	 * at all.  This is something we detect at the time the interface
	 * dispatch table is built.  To avoid this test, we could instead
	 * point at a nosuchmethod routine there.  However, we would have
	 * to find a way to pass along the information which methods it
	 * is that's missing (or create multiple nosuch_method routines,
	 * given that they should be rare---minus possible DoS.)
	 */
	if (ncode == (void *)-1) {
		goto notfound;
	}
	assert(ncode != NULL);
	return (ncode);

notfound:
	/*
	 * Compilers following the latest version of the JLS emit a
	 * INVOKEINTERFACE instruction for methods that aren't defined in
	 * an interface, but inherited from Object.
	 *
	 * In this case, the JVM must
	 * a) check that the object really implements the interface given
	 * b) find and invoke the method on object.
	 *
	 * The best way to jit that would be a checkcast <interface_type>
	 * followed by an INVOKEVIRTUAL.
	 *
	 * For now, we simply detect the case where an object method is called
	 * and find it by hand using its (name, signature).
	 */
	{
		errorInfo info;
		Utf8Const* methSig;
		
		/*
		 * Finding a method sig is a bit more complicated than
		 * it should be:
		 */
		Method* meth = CLASS_METHODS(ifclass) + idx;
		classView* methClassView = getClassView(meth->class, &info);
		methodView* methView = getMethodView(methClassView, meth);
		methSig = METHOD_VIEW_SIG(methView);

		assert(methClassView != 0);

		if (ifclass == ObjectClass) {
			Method* indirectMeth;
			indirectMeth = findMethod(cls, meth->name, methSig, &info);
			if (indirectMeth == 0) {
				throwError(&info);
			}
			return (METHOD_INDIRECTMETHOD(indirectMeth));
		}
	
		soft_nosuchmethod(cls, meth->name, methSig);
	}
	return (0);
}

inline
jint
instanceof_class(Hjava_lang_Class* c, Hjava_lang_Class* oc)
{
	Hjava_lang_Class* tc;

	/* Check for superclass matches */
	for (tc = oc->superclass; tc != 0; tc = tc->superclass) {
		if (c == tc) {
			return (1);
		}
	}
	return (0);
}

inline
jint
instanceof_interface(Hjava_lang_Class* c, Hjava_lang_Class* oc)
{
	int i;

	/* Check 'total' interface list */
	for (i = oc->total_interface_len - 1; i >= 0; i--) {
		if (c == oc->interfaces[i]) {
			return (1);
		}
	}
	return (0);
}

inline
jint
instanceof_array(Hjava_lang_Class* c, Hjava_lang_Class* oc)
{
	/* Skip as many arrays of arrays as we can.  We stop when we find
	 * a base class in either. */
	while (CLASS_IS_ARRAY(c) && CLASS_IS_ARRAY(oc)) {
		c = CLASS_ELEMENT_TYPE(c);
		oc = CLASS_ELEMENT_TYPE(oc);
	}

	/* If we are still casting to an array then we have failed already */
	if (CLASS_IS_ARRAY(c))
		return (0);

	/* If a base type, they must match exact. */
	if (CLASS_IS_PRIMITIVE(c)) {
		return (c == oc);
	}

	/* Casting to an object of some description. */
	if (CLASS_IS_ARRAY(oc)) {
		/* The only thing we can cast an array to is java/lang/Object.
		 * Checking this here willl save time.
		 */
		return (c == ObjectClass);
	}

	/* Cannot cast to a primitive class. */
	if (CLASS_IS_PRIMITIVE(oc)) {
		return (0);
	}

	/* Casting one object to another.  */
	return (instanceof(c, oc));
}

jint
instanceof(Hjava_lang_Class* c, Hjava_lang_Class* oc)
{
	/* Handle the simplest case first - they are the same */
	if (c == oc) {
		return (1);
	}
	/* Else if an array check that */
	else if (CLASS_IS_ARRAY(c)) {
		return (instanceof_array(c, oc));
	}
	/* Else if an interface check that */
	else if (CLASS_IS_INTERFACE(c)) {
		return (instanceof_interface(c, oc));
	}
	/* Else is must to a class */
	else {
		return (instanceof_class(c, oc));
	}
}

/*
 * soft_instanceof.
 */
jint
soft_instanceof(Hjava_lang_Class* c, Hjava_lang_Object* o)
{
	jthread_guardcheck();

	/* Null object are never instances of anything */
	if (o == 0) {
		return (0);
	}

	return (instanceof(c, OBJECT_CLASS(o)));
}

/*
 * soft_checkcast.
 */
void*
soft_checkcast(Hjava_lang_Class* c, Hjava_lang_Object* o)
{
	jthread_guardcheck();

	if (o != 0 && !instanceof(c, OBJECT_CLASS(o))) {
		/*
		 * Let's be a bit more informative as to why the class
		 * cast exception happened.
		 */
		Hjava_lang_Throwable* ccexc;
		const char *fromtype = CLASS_CNAME(OBJECT_CLASS(o));
		const char *totype = CLASS_CNAME(c);
		char *format = "can't cast `%s' to `%s'";
		char *buf = checkPtr(gc_malloc(strlen(fromtype) 
			+ strlen(totype) + strlen(format), GC_ALLOC_SOFT));
		sprintf(buf, format, fromtype, totype);
		ccexc = ClassCastException(buf);
		gc_free(buf);
		throwException(ccexc);
	}
	return (o);
}


/*
 * soft_athrow.
 */
void
soft_athrow(Hjava_lang_Object* o)
{
	jthread_guardcheck();

	if (o == 0) {
		soft_nullpointer();
	}
	else {
		/* NB: This will keep the stacktrace that was built
		 * when the exception was constructed.
		 * If you wanted a new stacktrace corresponding to the
		 * site where the exception is thrown, you'd use
		 * `throwException' instead.  However, this would slow down
		 * exceptions because two stacktrace have to be constructed.
		 *
		 * XXX Uh, I think thats how its supposed to work, creating
		 * a new back trace doesn't seem like just an optimization.
		 */
		throwExternalException((Hjava_lang_Throwable*)o);
	}
}

/*
 * soft_badarrayindex.
 */
void
soft_badarrayindex(int idx)
{
	char buf[16];
	
	jthread_guardcheck();

	sprintf(buf, "%d", idx);
	throwException(ArrayIndexOutOfBoundsException(buf));
}

/*
 * soft_nullpointer.
 */
void
soft_nullpointer(void)
{
	jthread_guardcheck();

	throwException(NullPointerException);
}

/*
 * soft_divzero.
 */
void
soft_divzero(void)
{
	jthread_guardcheck();

	throwException(ArithmeticException);
}

/*
 * soft_stackoverflow.
 */
void
soft_stackoverflow(void)
{
	Hjava_lang_Throwable *th;
	errorInfo einfo;

	for(;;);
	
	jthread_guardcheck();

	/* XXX Dear lord this sucks! */
	jthread_relaxstack(1);
	th = (Hjava_lang_Throwable *)
		newObjectChecked(javaLangStackOverflowError, &einfo);
	jthread_relaxstack(0);
	throwException(th);
}

/*
 * soft_nosuchclass.
 */
void
soft_nosuchclass(Utf8Const* c)
{
	char buf[256];

	jthread_guardcheck();

	sprintf(buf, "%.80s", c->data);
	throwException(NoClassDefFoundError(buf));
}

/*
 * soft_nosuchmethod.
 */
void
soft_nosuchmethod(Hjava_lang_Class* c, Utf8Const* n, Utf8Const* s)
{
	char buf[256];

	jthread_guardcheck();

	sprintf(buf, "%.80s.%.80s%.80s", CLASS_CNAME(c), n->data, s->data);
	throwException(NoSuchMethodError(buf));
}

/*
 * soft_nosuchfield.
 */
void
soft_nosuchfield(Utf8Const* c, Utf8Const* n)
{
	char buf[256];

	jthread_guardcheck();

	sprintf(buf, "%.100s.%.100s", c->data, n->data);
	throwException(NoSuchFieldError(buf));
}

void
soft_linkage(Utf8Const *c, Utf8Const* n)
{
	char buf[256];

	jthread_guardcheck();

	sprintf(buf, "%.100s.%.100s", c->data, n->data);
	throwException(LinkageError(buf));
}

void
soft_illegalaccess(Utf8Const *c, Utf8Const* n)
{
	char buf[256];

	jthread_guardcheck();

	sprintf(buf, "%.100s.%.100s", c->data, n->data);
	throwException(IllegalAccessError(buf));
}

void
soft_incompatibleclasschange(Utf8Const *c, Utf8Const* n)
{
	char buf[256];

	jthread_guardcheck();

	sprintf(buf, "%.100s.%.100s", c->data, n->data);
	throwException(IncompatibleClassChangeError(buf));
}

void*
soft_lookupstaticmethod(struct Hjava_lang_Class *cl, int const_idx)
{
	memResourceClient *mrc;
	jitCodeHeader *jch;
	jteam *owner = 0;
	void *retval = 0;
	errorInfo einfo;
	void *fp, *pc;
	callInfo ci;
	
	jthread_guardcheck();
	
	getCallerFrameAndPC(&fp, &pc);
	jch = findJITCodeHeader((uintp)pc);
	if( kernelTeam && (mrc = gc_get_resource(jch)) )
	{
		owner = mrc->client.owner;
	}
	if( owner && (switchToTeam(owner, 0) == 0) )
	{
		postException(&einfo, DEAD_TEAM_EXCEPTION_S);
		throwError(&einfo);
	}
	else if( getMethodSignatureClass(const_idx,
					 cl,
					 true,
					 false,
					 &ci,
					 &einfo) &&
		 processClass(ci.method->class, CSTATE_COMPLETE, &einfo) )
	{
		if( !(ci.method->accflags & ACC_STATIC) )
		{
			if( owner )
			{
				returnFromTeam();
			}
			soft_incompatibleclasschange(ci.cname, ci.name);
		}
		else if( !checkMethodAccess(cl,
					    ci.class,
					    ci.method) )
		{
			if( owner )
			{
				returnFromTeam();
			}
			soft_illegalaccess(ci.cname, ci.name);
		}
		else
		{
			retval = METHOD_NATIVECODE(ci.method);
		}
		cqAddCode(currentCompileQueue(),
			  CQ_RECOMPILE,
			  jch);
		
		ensure(retval != 0);
	}
	if( owner )
	{
		returnFromTeam();
	}
	if( !retval )
	{
		throwError(&einfo);
	}
	return( retval );
}

void*
soft_lookupstaticfield(struct Hjava_lang_Class *cl,
		       int const_idx,
		       int getter)
{
	memResourceClient *mrc;
	jitCodeHeader *jch;
	jteam *owner = 0;
	void *retval = 0;
	errorInfo einfo;
	void *fp, *pc;
	fieldInfo fi;
	
	jthread_guardcheck();
	
	getCallerFrameAndPC(&fp, &pc);
	while( (jch = findJITCodeHeader((uintp)pc)) == 0 )
	{
		pc = getFrameReturnPC(fp);
		fp = getPreviousFramePointer(fp);
	}
	if( kernelTeam && (mrc = gc_get_resource(jch)) )
	{
		owner = mrc->client.owner;
	}
	if( owner && (switchToTeam(owner, 0) == 0) )
	{
		postException(&einfo, DEAD_TEAM_EXCEPTION_S);
		throwError(&einfo);
	}
	else if( getField(const_idx, cl, true, true, &fi, &einfo) )
	{
		soft_initialise_class(fi.class);
		
		if( !(fi.field->accflags & ACC_STATIC) )
		{
			if( owner )
			{
				returnFromTeam();
			}
			soft_incompatibleclasschange(fi.cname, fi.name);
		}
		else if( !checkFieldAccess(cl,
					   fi.class,
					   fi.field) )
		{
			if( owner )
			{
				returnFromTeam();
			}
			soft_illegalaccess(fi.cname, fi.name);
		}
		else
		{
			jvalue buf;
			
			cqAddCode(currentCompileQueue(),
				  CQ_RECOMPILE,
				  jch);

			if( getter && fi.field->getTeamLocalMethod )
			{
				/*
				 * It has a team local getter, call it instead
				 * of returning the field address.
				 */
				buf = do_execute_java_method(
					0,
					0,
					0,
					fi.field->getTeamLocalMethod,
					1);
				/* XXX kinda nasty... */
				retval = &buf;
			}
			else
			{
				retval = FIELD_ADDRESS(fi.field);
			}

			ensure(retval != 0);
		}
	}
	if( owner )
	{
		returnFromTeam();
	}
	if( !retval )
	{
		throwError(&einfo);
	}
	return( retval );
}

void *soft_getstatic(struct Hjava_lang_Class *cl, int const_idx)
{
	void *retval;

	retval = soft_lookupstaticfield(cl, const_idx, 1);
	return( retval );
}

int soft_getstatic1(struct Hjava_lang_Class *cl, int const_idx)
{
	int *retval;

	retval = soft_lookupstaticfield(cl, const_idx, 1);
	return( *retval );
}

jlong soft_getstaticlong(struct Hjava_lang_Class *cl, int const_idx)
{
	jlong *retval;

	retval = soft_lookupstaticfield(cl, const_idx, 1);
	return( *retval );
}

jdouble soft_getstaticdouble(struct Hjava_lang_Class *cl, int const_idx)
{
	jdouble *retval;

	retval = soft_lookupstaticfield(cl, const_idx, 1);
	return( *retval );
}

void *soft_putstatic(struct Hjava_lang_Class *cl, int const_idx)
{
	void *retval;

	retval = soft_lookupstaticfield(cl, const_idx, 0);
	return( retval );
}

void*
soft_lookupmethod(struct Hjava_lang_Object *obj,
		  Utf8Const *className,
		  Utf8Const *methodName,
		  Utf8Const *methodSig,
		  struct Hjava_lang_Class *cl)
{
	memResourceClient *mrc;
	jitCodeHeader *jch;
	jteam *owner = 0;
	void *retval = 0;
	errorInfo einfo;
	void *fp, *pc;
	Method *meth;
	
	jthread_guardcheck();

#if defined(CREATE_NULLPOINTER_CHECKS)
	if (obj == 0) {
		soft_nullpointer();
	}
#endif

	getCallerFrameAndPC(&fp, &pc);
	jch = findJITCodeHeader((uintp)pc);
	if( kernelTeam && (mrc = gc_get_resource(jch)) )
	{
		owner = mrc->client.owner;
	}
	if( owner && (switchToTeam(owner, 0) == 0) )
	{
		postException(&einfo, DEAD_TEAM_EXCEPTION_S);
		throwError(&einfo);
	}
	else if( (meth = findMethod(OBJECT_CLASS(obj),
				    methodName,
				    methodSig,
				    &einfo)) )
	{
		if( meth->accflags & ACC_STATIC )
		{
			if( owner )
			{
				returnFromTeam();
			}
			soft_incompatibleclasschange(className, methodName);
		}
		else if( !checkMethodAccess(cl, OBJECT_CLASS(obj), meth) )
		{
			if( owner )
			{
				returnFromTeam();
			}
			soft_illegalaccess(className, methodName);
		}
		else
		{
			cqAddCode(currentCompileQueue(),
				  CQ_RECOMPILE,
				  jch);

			retval = METHOD_NATIVECODE(meth);

			ensure(retval != 0);
		}
	}
	if( owner )
	{
		returnFromTeam();
	}
	if( !retval )
	{
		throwError(&einfo);
	}
	return( retval );
}

void*
soft_lookupsupermethod(struct Hjava_lang_Object *obj,
		       Utf8Const *className,
		       Utf8Const *methodName,
		       Utf8Const *methodSig,
		       struct Hjava_lang_Class *cl)
{
	memResourceClient *mrc;
	Hjava_lang_Class *scl;
	jitCodeHeader *jch;
	jteam *owner = 0;
	void *retval = 0;
	errorInfo einfo;
	void *fp, *pc;
	Method *meth;
	
	jthread_guardcheck();

#if defined(CREATE_NULLPOINTER_CHECKS)
	if (obj == 0) {
		soft_nullpointer();
	}
#endif

	getCallerFrameAndPC(&fp, &pc);
	jch = findJITCodeHeader((uintp)pc);
	if( kernelTeam && (mrc = gc_get_resource(jch)) )
	{
		owner = mrc->client.owner;
	}
	if( owner && (switchToTeam(owner, 0) == 0) )
	{
		postException(&einfo, DEAD_TEAM_EXCEPTION_S);
		throwError(&einfo);
	}
	else if( (scl = loadClass(className,
				  OBJECT_CLASS(obj)->loader,
				  &einfo)) &&
		 (meth = findMethod(scl,
				    methodName,
				    methodSig,
				    &einfo)) )
	{
		classView* mcView = getClassView(meth->class, &einfo);

		if (mcView == 0)
		{
		}
		else if( meth->accflags & ACC_STATIC )
		{
			if( owner )
			{
				returnFromTeam();
			}
			soft_incompatibleclasschange(mcView->name, meth->name);
		}
		else if( !checkMethodAccess(cl, OBJECT_CLASS(obj), meth) )
		{
			if( owner )
			{
				returnFromTeam();
			}
			soft_illegalaccess(mcView->name, meth->name);
		}
		else
		{
			cqAddCode(currentCompileQueue(),
				  CQ_RECOMPILE,
				  jch);

			retval = METHOD_NATIVECODE(meth);

			ensure(retval != 0);
		}
	}
	if( owner )
	{
		returnFromTeam();
	}
	if( !retval )
	{
		throwError(&einfo);
	}
	return( retval );
}

void*
soft_lookupfield(struct Hjava_lang_Object *obj,
		 Utf8Const *fieldName,
		 Utf8Const *fieldSig,
		 struct Hjava_lang_Class *cl)
{
	memResourceClient *mrc;
	jitCodeHeader *jch;
	jteam *owner = 0;
	void *retval = 0;
	errorInfo einfo;
	void *fp, *pc;
	Field *ft;
	
	jthread_guardcheck();
	
	getCallerFrameAndPC(&fp, &pc);
	jch = findJITCodeHeader((uintp)pc);
	if( kernelTeam && (mrc = gc_get_resource(jch)) )
	{
		owner = mrc->client.owner;
	}
	if( owner && (switchToTeam(owner, 0) == 0) )
	{
		postException(&einfo, DEAD_TEAM_EXCEPTION_S);
		throwError(&einfo);
	}
	else if( (ft = lookupClassField(OBJECT_CLASS(obj),
					fieldName,
					false,
					&einfo)) )
	{
		classView* fcView = getClassView(OBJECT_CLASS(obj), &einfo);

		if (fcView == 0)
		{
			if( owner )
			{
				returnFromTeam();
			}
			throwError(&einfo);
		}
		else if( !utf8ConstEqual(ft->signature, fieldSig) )
		{
			if( owner )
			{
				returnFromTeam();
			}
			soft_nosuchfield(fcView->name, fieldName);
		}
		else if( ft->accflags & ACC_STATIC )
		{
			if( owner )
			{
				returnFromTeam();
			}
			soft_incompatibleclasschange(fcView->name, ft->name);
		}
		else if( !checkFieldAccess(cl, OBJECT_CLASS(obj), ft) )
		{
			if( owner )
			{
				returnFromTeam();
			}
			soft_illegalaccess(fcView->name, ft->name);
		}
		else
		{
			cqAddCode(currentCompileQueue(),
				  CQ_RECOMPILE,
				  jch);

			retval = ((char *)obj) + FIELD_BOFFSET(ft);
			
			ensure(retval != 0);
		}
	}
	if( owner )
	{
		returnFromTeam();
	}
	if( !retval )
	{
		throwError(&einfo);
	}
	return( retval );
}

/*
 * soft_initialise_class.
 */
void
soft_initialise_class(Hjava_lang_Class* c)
{
	jthread_guardcheck();

	/* We check this outside the processClass to save a subroutine call */
	if (c->state != CSTATE_COMPLETE) {
		errorInfo info;
		if (processClass(c, CSTATE_COMPLETE, &info) == false) {
			throwError(&info);
		}
	}
}

#if defined(TRANSLATOR)
/*
 * Trampolines come in here - do the translation and replace the trampoline.
 */
nativecode*
soft_fixup_trampoline(FIXUP_TRAMPOLINE_DECL)
{
	Method* meth;
	void **where;
	void *tramp;
	errorInfo info;

	/* FIXUP_TRAMPOLINE_INIT sets tramp and where */
	FIXUP_TRAMPOLINE_INIT;
	tramp = *where;

	jthread_guardcheck();

	if( !jthread_stackcheck(REDZONE + (REDZONE / 4)) )
	{
		soft_stackoverflow();
	}
	
DBG(MOREJIT,
	dprintf("soft_fixup_trampoline(): meth %p, where %p\n",
		meth, where);
    );

	/* If this class needs initializing, do it now.  */
	if (meth->class->state < CSTATE_USABLE &&
	    processClass(meth->class, CSTATE_COMPLETE, &info) == false) {
		require((info.throwable == 0) ||
			(OBJECT_CLASS(&info.throwable->base) !=
			 javaLangClassNotFoundException));

		throwError(&info);
	}

	/* Generate code on demand.  */
	if (!METHOD_TRANSLATED(meth)) {
		int success;

		success = translate(meth, &info);
		if (success == false) {
			require((info.throwable == 0) ||
				(OBJECT_CLASS(&info.throwable->base) !=
				 javaLangClassNotFoundException));
			
			throwError(&info);
		}
	}

	/* Update the origin of this trampoline --- this could be a
	 * dispatch table, interface dispatch table, or a gcj trampoline
	 *
	 * This is rather tricky.  We may jump through the same trampoline
	 * twice.  For instance, class A refers to Double.longBitsToDouble,
	 * but Double.<clinit> also jumps through the same trampoline
	 * before it is patched.
	 */
	*where = METHOD_NATIVECODE(meth);

#if 0
	if (METHOD_PRE_COMPILED(meth)) {
		nativecode* ncode = METHOD_TRUE_NCODE(meth);
		nativecode* ocode = METHOD_NATIVECODE(meth);
		METHOD_NATIVECODE(meth) = ncode;
		/* Update the dtable entries for all classes if this isn't a
	   	   static method.  */
		if (meth->idx >= 0 && ocode != ncode) {
			meth->class->dtable->method[meth->idx] = ncode;
		}
		SET_METHOD_PRE_COMPILED(meth, 0);
	}
#endif

TDBG(	dprintf("Calling %s:%s%s @ %p\n", meth->class->name->data, meth->name->data, METHOD_SIGD(meth), METHOD_NATIVECODE(meth));	)

DBG(MOREJIT,
	dprintf("soft_fixup_trampoline(): return %p\n",
		METHOD_NATIVECODE(meth));
    );

	return (METHOD_NATIVECODE(meth));
}
#endif

/*
 * Check we can store 'obj' into the 'array'.
 */
void
soft_checkarraystore(Hjava_lang_Object* array, Hjava_lang_Object* obj)
{
	jthread_guardcheck();

	if( obj != 0 )
	{
		Hjava_lang_Class *element_class;

		element_class = CLASS_ELEMENT_TYPE(OBJECT_CLASS(array));
		if( (element_class != OBJECT_CLASS(obj)) &&
		    (soft_instanceof(element_class, obj) == 0) )
		{
			throwException(ArrayStoreException);
		}
	}
}

/*
 * soft_dcmpg
 */
jint
soft_dcmpg(jdouble v1, jdouble v2)
{
	jint ret;
	
	if ((!isinf(v1) && isnan(v1)) || (!isinf(v2) && isnan(v2))) {
		ret = 1;
	}
	else if (v1 > v2) {
		ret = 1;
	}
	else if (v1 == v2) {
		ret = 0;
	}
	else {
		ret = -1;
	}

	return (ret);
}

/*
 * soft_dcmpl
 */
jint
soft_dcmpl(jdouble v1, jdouble v2)
{
        jint ret;
	
	if ((!isinf(v1) && isnan(v1)) || (!isinf(v2) && isnan(v2))) {
		ret = -1;
	}
        else if (v1 > v2) {
                ret = 1;
        }
        else if (v1 == v2) {
                ret = 0;
        }
        else {
                ret = -1;
        }
	return (ret);
}

/*
 * soft_fcmpg
 */
jint
soft_fcmpg(jfloat v1, jfloat v2)
{
        jint ret;
	jint v1bits;
	jint v2bits;

	v1bits = floatToInt(v1);
	v2bits = floatToInt(v2);

        if (FISNAN(v1bits) || FISNAN(v2bits)) {
		ret = 1;
	}
        else if (v1 > v2) {
                ret = 1;
        }
        else if (v1 == v2) {
                ret = 0;
        }
        else {
                ret = -1;
        }
	return (ret);
}

/*
 * soft_fcmpg
 */
jint
soft_fcmpl(jfloat v1, jfloat v2)
{
        jint ret;
	jint v1bits;
	jint v2bits;

	v1bits = floatToInt(v1);
	v2bits = floatToInt(v2);

        if (FISNAN(v1bits) || FISNAN(v2bits)) {
		ret = -1;
	}
        else if (v1 > v2) {
                ret = 1;
        }
        else if (v1 == v2) {
                ret = 0;
        }
        else {
                ret = -1;
        }
	return (ret);
}

jlong
soft_lmul(jlong v1, jlong v2)
{
	return (v1 * v2);
}

jlong
soft_ldiv(jlong v1, jlong v2)
{
	return (v1 / v2);
}

jlong
soft_lrem(jlong v1, jlong v2)
{
	return (v1 % v2);
}

jfloat
soft_fadd(jfloat v1, jfloat v2)
{
	return floatAdd(v1, v2);
}

jdouble
soft_faddl(jdouble v1, jdouble v2)
{
	return doubleAdd(v1, v2);
}

jfloat
soft_fsub(jfloat v1, jfloat v2)
{
	return floatSubtract(v1, v2);
}

jdouble
soft_fsubl(jdouble v1, jdouble v2)
{
	return doubleSubtract(v1, v2);
}


jfloat
soft_fmul(jfloat v1, jfloat v2)
{
	return floatMultiply(v1, v2);
}

jdouble
soft_fmull(jdouble v1, jdouble v2)
{
	return doubleMultiply(v1, v2);
}

jfloat
soft_fdiv(jfloat v1, jfloat v2)
{
	return floatDivide(v1, v2);
}

jdouble
soft_fdivl(jdouble v1, jdouble v2)
{
	return doubleDivide(v1, v2);
}

jfloat
soft_frem(jfloat v1, jfloat v2)
{
	return (javaRemainderf(v1, v2));
}

jdouble
soft_freml(jdouble v1, jdouble v2)
{
	return (javaRemainder(v1, v2));
}

jlong
soft_lshll(jlong v1, jint v2)
{
	return (v1 << (v2 & 63));
}

jlong
soft_ashrl(jlong v1, jint v2)
{
	return (v1 >> (v2 & 63));
}

jlong
soft_lshrl(jlong v1, jint v2)
{
	return (((uint64)v1) >> (v2 & 63));
}

jint
soft_lcmp(jlong v1, jlong v2)
{
#if 0
	jlong lcc = v2 - v1;
	if (lcc < 0) {
		return (-1);
	}
	else if (lcc > 0) {
		return (1);
	}
	else {
		return (0);
	}
#endif
	if (v2 < v1) {
		return (-1);
	}
	else if (v2 > v1) {
		return (1);
	}
	else {
		return (0);
	}
}

jint
soft_mul(jint v1, jint v2)
{
	return (v1*v2);
}

jint
soft_div(jint v1, jint v2)
{
	return (v1/v2);
}

jint
soft_rem(jint v1, jint v2)
{
	return (v1%v2);
}

jfloat
soft_cvtlf(jlong v)
{
	return ((jfloat)v);
}

jfloat
soft_cvtif(jint v)
{
	return ((jfloat)v);
}

jdouble
soft_cvtid(jint v)
{
	return ((jdouble)v);
}

jdouble
soft_cvtld(jlong v)
{
	return ((jdouble)v);
}

jdouble
soft_cvtfd(jfloat v)
{
	jint vbits;

	vbits = floatToInt(v);
        if (FISNAN(vbits)) {
		return (longToDouble(DNANBITS));
	}
	else {
		return ((jdouble)v);
	}
}

jfloat
soft_cvtdf(jdouble v)
{
	jlong vbits;

	vbits = doubleToLong(v);
        if (DISNAN(vbits)) {
		return (intToFloat(FNANBITS));
	}
	else {
		return ((jfloat)v);
	}
}

/*
 * The following functions round the float/double to an int/long.
 * They round the value toward zero.
 */

jlong
soft_cvtfl(jfloat v)
{
	jint vbits;

	vbits = floatToInt(v);
        if (FISNAN(vbits)) {
		return ((jlong)0);
	}

	if (v < 0.0) {
		v = ceil(v);
	}
	else {
		v = floor(v);
	}
	/* If too small return smallest long */
	if (v <= -9223372036854775808.0) {
	    return ((jlong)1) << 63;
	}
	/* If too big return biggest long */
	else if (v >= 9223372036854775807.0) {
	    return ~(((jlong)1) << 63);
	}
	else {
	    return ((jlong)v);
	}
}

jlong
soft_cvtdl(jdouble v)
{
	jlong vbits;

	vbits = doubleToLong(v);
        if (DISNAN(vbits)) {
		return ((jlong)0);
	}

	if (v < 0.0) {
		v = ceil(v);
	}
	else {
		v = floor(v);
	}
	/* If too small return smallest long */
	if (v <= -9223372036854775808.0) {
	    return ((jlong)1) << 63;
	}
	/* If too big return biggest long */
	else if (v >= 9223372036854775807.0) {
	    return ~(((jlong)1) << 63);
	}
	else {
	    return ((jlong)v);
	}
}

jint
soft_cvtfi(jfloat v)
{
        jint vbits;

	vbits = floatToInt(v);
        if (FISNAN(vbits)) {
		return (0);
	}

	if (v < 0.0) {
		v = ceil(v);
	}
	else {
		v = floor(v);
	}
	/* If too small return smallest int */
	if (v <= -2147483648.0) {
		return (-2147483647-1);
	}
	/* If too big return biggest int */
	else if (v >= 2147483647.0) {
		return (2147483647);
	}
	else {
		return ((jint)v);
	}
}

jint
soft_cvtdi(jdouble v)
{
        jlong vbits;

	vbits = doubleToLong(v);
        if (DISNAN(vbits)) {
		return (0);
	}

	if (v < 0.0) {
		v = ceil(v);
	}
	else {
		v = floor(v);
	}
	/* If too small return smallest int */
	if (v <= -2147483648.0) {
		return (-2147483647-1);
	}
	/* If too big return biggest int */
	else if (v >= 2147483647.0) {
		return (2147483647);
	}
	else {
		return ((jint)v);
	}
}

void
soft_debug1(void* a0, void* a1, void* a2)
{
}

void
soft_debug2(void* a0, void* a1, void* a2)
{
}

void
soft_trace(Method* meth, void* args)
{
	const char* msigd;
	const char* classd;
	errorInfo einfo;
	classView* cView;
	
	cView = getClassView(meth->class, &einfo);
	if (cView == 0)
	{
		classd = "<unviewable>";
		msigd = "(<unviewable>)";
	}
	else
	{
		methodView* methView = getMethodView(cView, meth);

		classd = cView->name->data;
		msigd = METHOD_VIEW_SIGD(methView);
	}
	dprintf("soft_trace: %s.%s%s\n", classd, meth->name->data, msigd);
}
