/*
 * Copyright (c) 1999, 2000, 2001, 2002 The University of Utah.  All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file.
 *
 * Written by the Flux Research Group at the University of Utah.
 * Authors: Godmar Back, Leigh Stoller, Patrick Tullmann
 */

/*
 * jthread creation functions
 */

#include "anjthread_types.h"
#include "anjthread_types_priv.h"
#include "gc.h"
#include "anjthread_globals.h"
#include "anjthread_intr.h"
#include "anjthread_create.h"
#include "debug.h"

#include <janosvm/jteamReservation.h>

#if defined(THREAD_ANALYZER)
#include <janosvm/leak.h>
#include <janosvm/logResource.h>

void threadAnalyzerFunction(struct _logAnalyzerMask *lam,
			    struct _logger *l,
			    struct _logEntry *le,
			    struct _logEntryAttribute **attrs)
{
	jthread_t th = (jthread_t)lam;

	if( !le->classification )
		return;
	if( !strcmp(le->classification, "leak") )
	{
		if( (attrs[1]->value.ptr == th->jlThread) )
		{
			if( !strcmp(attrs[0]->id, "unregister-leak") )
			{
				assert(th->status == JTHREAD_STATUS_INNODEOS);
			}
			le->classification = 0;
		}
		else
		{
			Hjava_lang_Thread *jlt = attrs[1]->value.ptr;
			
			th = (jthread_t)jlt->PrivateInfo;
			if( th->threadData.ownerTeam != currentTeam() )
				le->classification = 0;
		}
	}
}

static
logAnalyzer threadAnalyzer = {
	threadAnalyzerFunction
};
#endif

static void
anjthread_startthread(void *arg) __NORETURN__;

/*
 * Start function for explicitly created threads.  Most threads
 * are "partial" threads, and they come in through the function
 * anjthread_enterkaffe()
 *
 * This function install the cleanup handler, sets jthread-specific 
 * data and calls the actual work function.
 */
static void
anjthread_startthread(void *arg)
{
	jthread_t jtid = (jthread_t)arg;

DBG(JTHREAD, 
	dprintf("anjthread_startthread(jth=%p)\n", jtid); 
    )
	assert(jtid);
	assert(jtid->stackptr);

	an_mutex_lock(&jthread_creationLock);
	an_thread_sethandler(_jthread_intr_handler);
	jtid->flags |= JTHREAD_FLAG_INSTALLED_HANDLER;
	DBGIF(an_thread_setdata(jtid));
	an_mutex_unlock(&jthread_creationLock);

	/*
	 * Change thread status.  Check to make sure thread
	 * wasn't killed/interrupted before starting.
	 */
	assert(jtid->status == JTHREAD_STATUS_NEWBORN);
	jtid->status = JTHREAD_STATUS_RUNNING;
	if( jtid->threadData.flags & TDF_ZOMBIE )
	{
		jthread_slaughter(jtid);
	}

DBG(JTHREAD, 
	dprintf("anjthread_startthread(jth=%p) calling func (%p)\n",
		jtid, jtid->func); 
    )

	/* Invoke the thread's actual start function */
	jtid->func(jtid->jlThread);

	panic("Thread start function has returned");
}

/*
 * Create a new jthread and explicitly start it running the given
 * callback.
 */
jthread_t
jthread_create(unsigned int pri, void (*func)(void *), int daemon,
	       void *jlThread, size_t threadStackSize)
{
	jthread_t tid;
	void* th;
	int rc;

	/* XXX Check against OSKit minimum stack limit */
	if (threadStackSize < (5 * 0x1000))
		threadStackSize = (5 * 0x1000);

	/* Allocate the tid */
        tid = gc_malloc(sizeof (*tid), GC_ALLOC_THREADCTX);
	if (tid == NULL)
		return NULL; /* XXX good enough? */

	/* Allocate the thread's stack */
	tid->stackptr = gc_malloc(threadStackSize, GC_ALLOC_THREADSTACK);
	if (tid->stackptr == NULL)
	{
		gc_free(tid);
		return NULL;
	}

	/* Basic tid initialization */
        tid->jlThread = jlThread;
        tid->func = func; /* XXX duplicated at too many levels. */
        tid->status = JTHREAD_STATUS_NEWBORN;
	tid->flags |= JTHREAD_FLAG_DONT_STOP;
	tid->stopCounter = 1;

	/* Setup the new thread's Team data */
	rc = inheritThreadData(&tid->threadData,
				&(jthread_current()->threadData),
				tid);
	if (rc == 0)
	{
		gc_free(tid->stackptr);
		gc_free(tid);
		return NULL;
	}

	/* Create the NodeOS thread. */
	th = an_thread_create(&(tid->native_thread),
			      tid->stackptr,
			      threadStackSize);
	if (th == NULL)
	{
		gc_free(tid->stackptr);
		gc_free(tid);
		return NULL;
	}

#if defined(THREAD_ANALYZER)
	tid->lam.mask = 0;
	tid->lam.analyzer = &threadAnalyzer;
	addTeamLogAnalyzer(&tid->lam);
	registerLeak(jlThread, "jt");
#endif
	
	an_mutex_lock(&jthread_creationLock);
	
	/* Start the NodeOS thread. */
	rc = an_thread_start(&(tid->native_thread), an_domain_current(),
			     anjthread_startthread, tid);
	if (rc != 0)
	{
		an_error_t err;
		
		an_mutex_unlock(&jthread_creationLock);
		err = an_thread_destroy(&(tid->native_thread));
		assert(err == 0);
		gc_free(tid->stackptr);
		gc_free(tid);
		return NULL; /* will be returned after unlocking */
	}

	assert(tid->nextlive == 0);
        tid->nextlive = jthread_liveThreads;
        jthread_liveThreads = tid;

        jthread_talive++;  
        if ((tid->daemon = daemon) != 0) {
                jthread_tdaemon++;
        }

	an_mutex_unlock(&jthread_creationLock);

DBG(JTHREAD,
	dprintf("jthread_create(pri=%d, daemon=%d) -> jth=%p nth=%p\n",
		pri, daemon, tid, &(tid->native_thread));
    )
	
	assert(tid);
        return (tid);
}

static void errorInPartialThread(void *arg)
{
	panic("errorInPartialThread()");
}

/*
 * Create a new jthread, but do not start it.  Partial threads
 * are queued up in the NodeOS, on a threadpool, and will be
 * injected back into the VM through a callback function
 * (usu. on an inChannel).  All callbacks should come
 * anjthread_enterkaffe().
 *
 * XXX should return int error code!
 */
int
anjthread_createpartialthread(jthread_t tid,
			      void *jlThread,
			      size_t threadStackSize)
{
	void* th;
	int err;
	
        assert(tid);
        assert(jlThread);
	/* assert(threadStackSize); */

	tid->flags |= JTHREAD_FLAG_PARTIAL;
        tid->jlThread = jlThread;
        tid->func = errorInPartialThread;

	/* Create the NodeOS thread object. */

	/* XXX Check against OSKit minimum stack limit */
	if (threadStackSize < (5 * 0x1000))
		threadStackSize = (5 * 0x1000);

	tid->stackptr = gc_malloc(threadStackSize, GC_ALLOC_THREADSTACK);
	if (!tid->stackptr)
	{
		return 0;
	}

        tid->status = JTHREAD_STATUS_INNODEOS;
	th = an_thread_create(&(tid->native_thread),
			      tid->stackptr,
			      threadStackSize);
	if (th == NULL)
	{
		gc_free(tid->stackptr);
		return 0;
	}

	/*
	 * Setup the thread's Team state.
	 */
	err = inheritThreadData(&tid->threadData,
				&(jthread_current()->threadData),
				tid);
	if (err == 0)
	{
		gc_free(tid->stackptr);
		an_thread_destroy(&(tid->native_thread));
		return 0;
	}
	
	an_mutex_lock(&jthread_creationLock);
	jthread_talive++;  /* XXX move this */
	assert(tid->nextlive == 0);
        tid->nextlive = jthread_liveThreads;
        jthread_liveThreads = tid;
	an_mutex_unlock(&jthread_creationLock);
	
DBG(JTHREAD,
    dprintf(__FUNCTION__ "(tid=%p, jlThread=%p, threadStackSize=%d) done.\n",
	    tid, jlThread, threadStackSize);
    )
	return 1;
}


/**
 * This function never returns.  All partial threads should enter
 * the VM through this function, or the specialized entercallback
 * function.
 */
void
anjthread_enterkaffe(void *arg, jmethodID runmethod, ...)
{
	struct Hjava_lang_Object* startArg = arg;
	jthread_t tid = jthread_current();
	va_list argptr;

DBG(JTHREAD,
    	dprintf(__FUNCTION__ "(arg=%p, meth=%p, ...) jtid=%p\n",
		arg, runmethod, tid);
    )
	assert(tid->stackptr);
        tid->daemon = 0;

	jthread_disable_stop();
	
	an_thread_sethandler(_jthread_intr_handler);
	
#if 0
	an_mutex_lock(&jthread_creationLock);


	/* From jthread_create: */
	assert(tid->nextlive == 0);
        tid->nextlive = jthread_liveThreads;
        jthread_liveThreads = tid;

	/*
	 * XXX do *not* increment talive here.  Our caller does that.
	 * (For implicit threads, the immediate calling function does
	 * it; for explicit threads, the function that kicks this
	 * thread does it.
	 */

	an_mutex_unlock(&jthread_creationLock);
#endif
	
	/*
	 * Change thread status.  Check to make sure thread
	 * wasn't killed/interrupted before starting.
	 */
	jthread_spinon(tid);
	assert(tid->status == JTHREAD_STATUS_INNODEOS);
        tid->status = JTHREAD_STATUS_RUNNING;
	if ((tid->flags & JTHREAD_FLAG_INTR_MASK) != 0)
		jthread_handleInterrupt(tid);
	jthread_spinoff(tid);


	/* XXX assert that thread's team state is in reasonable shape. */

	/* XXX check that team isn't dying? */

	/* XXX function defined in ../../thread.c: */
	va_start(argptr, runmethod);
	threadEnterVM_valist(jthread_getcookie(jthread_current()),
			     startArg, runmethod, argptr);
	va_end(argptr);

	panic("threadEnterVM_valist should not return");
}

#ifdef TRANSLATOR
typedef struct {
	void *receiver;
	Method *receive;
	void *packet;
	int offset;
} callback_data;

static void
callback_fn(void *_data)
{
	callback_data *data = (callback_data *) _data;
	void (*receive)(void *, void *, int, void *)
		= (void*) METHOD_INDIRECTMETHOD(data->receive);
	printf("jreceive %p %p %p %p %d %p\n",
	       data->receive,
	       receive,
	       data->receiver,
	       data->packet,
	       data->offset,
	       jthread_stacklimit());
	receive(data->receiver, data->packet, data->offset,
		jthread_stacklimit());
}
#endif

/**
 * This function never returns.  
 */
void
anjthread_entercallback(void *arg, jmethodID runmethod,
			void *buf, int offset)
{
#ifndef TRANSLATOR
	anjthread_enterkaffe(arg, runmethod, buf, offset);
#else
	callback_data data = { arg, (Method*) runmethod, buf, offset };
	jthread_t tid = jthread_current();

DBG(JTHREAD,
    	dprintf(__FUNCTION__ "(arg=%p, meth=%p, buf=%p, offset=%d) jtid=%p\n",
		arg, runmethod, tid, buf, offset);
    )
	assert(tid->stackptr);
        tid->daemon = 0;

	an_mutex_lock(&jthread_creationLock);

	an_thread_sethandler(_jthread_intr_handler);

	/* From jthread_create: */
	assert(tid->nextlive == 0);
        tid->nextlive = jthread_liveThreads;
        jthread_liveThreads = tid;

	/*
	 * XXX do *not* increment talive here.  Our caller does that.
	 * (For implicit threads, the immediate calling function does
	 * it; for explicit threads, the function that kicks this
	 * thread does it.
	 */

	an_mutex_unlock(&jthread_creationLock);
	
	/*
	 * Change thread status.  Check to make sure thread
	 * wasn't killed/interrupted before starting.
	 */
	jthread_spinon(tid);
	assert(tid->status == JTHREAD_STATUS_INNODEOS);
        tid->status = JTHREAD_STATUS_RUNNING;
	if ((tid->flags & JTHREAD_FLAG_INTR_MASK) != 0)
		jthread_handleInterrupt(tid);
	jthread_spinoff(tid);


	/* XXX assert that thread's team state is in reasonable shape. */

	/* XXX check that team isn't dying? */

	/* XXX function defined in ../../thread.c: */
	allowTeamThread(currentTeam());
	threadEnterVM_fn(jthread_getcookie(jthread_current()),
			 callback_fn, &data);

	panic("threadEnterVM_fn should not return");
#endif
}

void
anjthread_enterkaffe_c(void)
{
	jthread_t tid = jthread_current();
	
	assert(tid->stackptr);
        tid->daemon = 0;

        tid->status = JTHREAD_STATUS_RUNNING;
	
	an_thread_sethandler(_jthread_intr_handler);
	tid->flags |= JTHREAD_FLAG_INSTALLED_HANDLER;

	if( tid->flags & JTHREAD_FLAG_CAGED )
		an_thread_exit();
	pushTeamReservation(currentTeam());
}

void
anjthread_exitkaffe_c(void)
{
	jthread_t tid = jthread_current();
	
	rollbackReservationStack((void *)~0, (void *)0, (void *)~0);
        tid->status = JTHREAD_STATUS_INNODEOS;
}
