/*
 * jactivation.h
 *
 * Copyright (c) 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

#ifndef _jactivation_h
#define _jactivation_h

#include <stdio.h>

#include "assert_pp.h"
#include "jcpu.h"
#include "jthreadState.h"

/*
 * Possible states a jactivation can be in:
 *   unused - Free for use.
 *   running - The currently running thread.
 *   stalled - Contains a stalled thread.
 */
typedef enum {
	JAS_MIN,
	JAS_UNUSED,
	JAS_RUNNING,
	JAS_STALLED,
	JAS_MAX
} jactivation_state_t;

/*
 * A jactivation is used contain a running or suspended jthread.
 *
 * ja_Next - The next jactivation in the free list.
 * ja_State - The jactivation_state_t value for this activation.
 * ja_Thread - The thread contained by this activation.
 */
typedef struct _jactivation {
	struct _jactivation *ja_Next;
	jactivation_state_t ja_State;
	jthreadState *ja_Thread;
} jactivation;

/*
 * A jactivationList contains a list of unused jactivations.
 *
 * jal_Unused - A singly linked list of unused jactivations.
 */
typedef struct _jactivationList {
	jactivation *jal_Unused;
} jactivationList;

/*
 * A jactivationState holds the global state for jactivations.
 *
 * jas_Current - The currently running jactivation or NULL if the process
 *   is executing startup/interrupt code.  NOTE: This value can change
 *   during the course of execution of a thread, so don't dereference it
 *   outside of a jcpuDisable()/jcpuEnable() pair.
 * jas_CurrentThread - The currently running jthread.  NOTE: This is here
 *   because the thread could be switched after jas_Current has been read and
 *   before it could be dereferenced.
 */
typedef struct _jactivationState {
	jactivation *jas_Current;
	jthreadState *jas_CurrentThread;
} jactivationState;

/*
 * The global state for jactivations.
 */
extern jactivationState jactivation_state;

/*
 * Allocate a jactivation from the given free list.
 *
 * jal - The list to allocate the activation from.
 * js - The jthread that the activation should contain.
 * returns - A jactivation obtained from `jal'.
 */
jactivation *jactivationAlloc(jactivationList *jal, jthreadState *js);

/*
 * Print the value of the jactivation.
 *
 * file - The file to print to.
 * ja - The jactivation to describe.
 */
void jactivationPrint(FILE *file, jactivation *ja);

/*
 * Print the value of the jactivationList.
 *
 * file - The file to print to.
 * jal - The jactivationList to describe.
 */
void jactivationListPrint(FILE *file, jactivationList *jal);

/*
 * Print the value of the jactivationState.
 *
 * file - The file to print to.
 */
void jactivationStatePrint(FILE *file);

/*
 * Return a jactivation to the free list.
 *
 * jal - The jactivationList to add `ja' to.
 * ja - The jactivation to free.
 */
static inline
void jactivationFree(jactivationList *jal, jactivation *ja)
{
	require(jal != 0);

	require(ja != 0);
	require(ja->ja_State == JAS_STALLED);
	
	require(jactivation_state.jas_Current != ja);

	ja->ja_State = JAS_UNUSED;
	ja->ja_Thread = 0;
	ja->ja_Next = jal->jal_Unused;
	jal->jal_Unused = ja;
}

/*
 * Stall a currently running jactivation.
 *
 * Note: This needs to be inlined because otherwise we'd snapshot a stack
 * frame that wouldn't be there on return from a function.
 */
#define jactivationStall() \
({ \
	int jas_retval; \
\
	require(jcpuDisabled()); \
	require(jactivation_state.jas_Current != 0); \
\
	if( jcpuContextCapture(&jactivation_state.jas_Current-> \
			       ja_Thread->jts_Context) ) \
	{ \
		require(jactivation_state.jas_Current-> \
			ja_State == JAS_STALLED); \
\
		jactivation_state.jas_Current->ja_State = JAS_RUNNING; \
		jas_retval = 1; \
	} \
	else \
	{ \
		require(jactivation_state.jas_Current->ja_State == \
			JAS_RUNNING); \
\
		jcpuAccountingEnd(&jactivation_state.jas_Current-> \
				  ja_Thread->jts_CPUAccount); \
		jactivation_state.jas_Current->ja_State = JAS_STALLED; \
		jactivation_state.jas_Current = 0; \
		jactivation_state.jas_CurrentThread = 0; \
		jas_retval = 0; \
	} \
\
	ensure(jcpuDisabled()); \
	jas_retval; \
})

/*
 * Resume a stalled jactivation.
 *
 * ja - The jactivation to resume.
 */
static inline
void jactivationResume(jactivation *ja)
{
	require(jcpuDisabled());
	
	require(ja != 0);
	require(ja->ja_State == JAS_STALLED);
	require(ja->ja_Thread != 0);

	require(jactivation_state.jas_Current != ja);

	jactivation_state.jas_Current = ja;
	jactivation_state.jas_CurrentThread = ja->ja_Thread;
	jcpuAccountingStart(&ja->ja_Thread->jts_CPUAccount);
	jcpuContextRestore(&ja->ja_Thread->jts_Context);

	ensure(0);
}

/*
 * returns - The current jactivation or NULL if we're executing in
 * kernel/interrupt mode.
 */
static inline
jactivation *jactivationCurrent(void)
{
	require(jcpuDisabled());
	
	return( jactivation_state.jas_Current );
}

/*
 * returns - The current jthreadState or NULL if we're executing in
 * kernel/interrupt mode.
 */
static inline
jthreadState *jactivationCurrentThread(void)
{
	return( jactivation_state.jas_CurrentThread );
}

#endif
