/*
 * jcpuAccounting.h
 *
 * Copyright (c) 2002, 2003 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

#ifndef _jcpu_accounting_h
#define _jcpu_accounting_h

#include "assert_pp.h"

#include "jcpu.h"

/**
 * The type used for the CPU accumulators.
 */
typedef unsigned long long jcpu_accum_t;

/*
 * A jcpuAccount tracks the amount of CPU used by a sked/thread.
 *
 * jca_Parent - A parent account that should be updated with the total amount
 *   of used CPU time.  Can be NULL.
 * jca_StartUsed - The start time for tracking used CPU.
 * jca_TotalUsed - The total used CPU time.
 * jca_Birth - The time when the accounting structure was initialized.
 * jca_Death - The time when the accounting structure was killed.
 */
typedef struct _jcpuAccount {
	struct _jcpuAccount *jca_Parent;
	
	jcpu_accum_t jca_StartUsed;
	jcpu_accum_t jca_TotalUsed;

	jcpu_accum_t jca_Birth;
	jcpu_accum_t jca_Death;
	unsigned long jca_ResetCount;
} jcpuAccount;

/*
 * Print the value of the jcpuAccount.
 *
 * file - The file to print to.
 * jca - The account to print.
 */
void jcpuAccountingPrint(FILE *file, jcpuAccount *jca);

/*
 * Initialize a jcpuAccount object.
 *
 * jca - The jcpuAcccount to initialize.
 */
static inline
void jcpuAccountingInit(jcpuAccount *jca)
{
	require(jca != 0);

	jca->jca_StartUsed = 0;
	jca->jca_TotalUsed = 0;
	jca->jca_Birth = jcpuTime();
	jca->jca_Death = 0;
	jca->jca_ResetCount = 0;
}

/*
 * Reset the used CPU time in the given account.
 *
 * jca - An initialized jcpuAccount.
 */
static inline
void jcpuAccountingReset(jcpuAccount *jca)
{
	require(jca != 0);

	jca->jca_StartUsed = 0;
	jca->jca_TotalUsed = 0;
	jca->jca_ResetCount += 1;
}

/*
 * Finalize a jcpuAccount object.
 *
 * jca - The jcpuAccuont to finalize.
 */
static inline
void jcpuAccountingKill(jcpuAccount *jca)
{
	jcpu_accum_t curr;
	
	require(jca != 0);

	curr = jcpuTime();
	jca->jca_Death = curr;
}

/*
 * Start tracking used CPU time and stop tracking unused time.
 *
 * jca - The jcpuAccount to add the used time to.
 */
static inline
void jcpuAccountingStart(jcpuAccount *jca)
{
	require(jca != 0);

	jca->jca_StartUsed = jcpuUsedTime();
}

/*
 * Stop tracking used CPU time and start tracking unused time.
 *
 * jca - The jcpuAccount to add the used time to.
 */
static inline
void jcpuAccountingEnd(jcpuAccount *jca)
{
	jcpu_accum_t curr, diff;

	require(jca != 0);

	curr = jcpuUsedTime();
	{
		require(curr >= jca->jca_StartUsed);
		
		diff = curr - jca->jca_StartUsed;
		jca->jca_TotalUsed += diff;
		if( jca->jca_Parent )
		{
			jca->jca_Parent->jca_TotalUsed += diff;
		}
	}
}

/*
 * Compute the idle time for an account.
 *
 * jca - The jcpuAccount to compute idle time for.
 * returns - The idle time for the account.
 */
static inline
jcpu_accum_t jcpuAccountingIdle(jcpuAccount *jca)
{
	jcpu_accum_t retval;

	require(jca != 0);

	retval = (jcpuTime() - jca->jca_Birth) - jca->jca_TotalUsed;
	return( retval );
}

/*
 * Symbol used to indicate an infinite limit for jcpuAccountingCheck.
 */
#define JCPU_ACCOUNT_INFINITE_LIMIT ((jcpu_accum_t)0)

/*
 * Check the total used CPU time against a limit.
 *
 * jca - The jcpuAccount to check.
 * limit - The limit on used CPU time.
 * returns - True if the total CPU time used by the account is less than the
 * limit.
 */
static inline
int jcpuAccountingCheck(jcpuAccount *jca, jcpu_accum_t limit)
{
	int retval = 0;

	require(jca != 0);

	if( (limit == JCPU_ACCOUNT_INFINITE_LIMIT) ||
	    (jca->jca_TotalUsed < limit) )
	{
		retval = 1;
	}
	return( retval );
}

#endif
