/*
 * jcpuContext.h
 *
 * Copyright (c) 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

#ifndef _jcpu_context_h
#define _jcpu_context_h

#include <errno.h>

/* XXX Push the kaffe krap out of here. */
#include "config.h"
#include "config-setjmp.h"
#include "config-std.h"
#include "config-mem.h"
#include "config-io.h"
#include "config-signal.h"
#include "gtypes.h"		/* for jlong */
#include "md.h"

#include "assert_pp.h"

#include "jsignalSet.h"

#if !defined(SP_OFFSET)
#error "Need to specify a stack pointer offset"
#endif

#if !defined(GET_SP)
#define GET_SP(env) (((void **)(env))[SP_OFFSET])
#endif
#if !defined(SET_SP)
#define SET_SP(env, val) ((void **)(env))[SP_OFFSET] = (val)
#endif

/* Set the base pointer in a jmp_buf if we can (only a convenience) */
#if defined(BP_OFFSET)
#define SET_BP(E, V)    ((void**)(E))[BP_OFFSET] = (V)
#define GET_BP(E)    (((void**)(E))[BP_OFFSET])
#endif

/* The number of bytes to copy from an old stack to a new stack. */
#if !defined(JTHREAD_STACK_COPY)
#define JTHREAD_STACK_COPY (32 * 4)
#endif

#if !defined(LOAD_FP)
#define LOAD_FP(fp)
#define SAVE_FP(fp)
#endif

#if !defined(JTHREAD_JMPBUF)
#include <setjmp.h>
#define JCPU_JMPBUF jmp_buf
#define JCPU_SETJMP setjmp
#define JCPU_LONGJMP longjmp
#else
#define JCPU_JMPBUF JTHREAD_JMPBUF
#define JCPU_SETJMP JTHREAD_SETJMP
#define JCPU_LONGJMP JTHREAD_LONGJMP
#endif

/*
 * Global state for jcpuContexts.
 *
 * jccs_AsyncSignals - The current asynchronous signal set.
 */
typedef struct _jcpuContextState {
	jsignalSet *jccs_AsyncSignals;
} jcpuContextState;

extern jcpuContextState jcpu_context_state;

/*
 * The value passed to longjmp when switching contexts.
 */
#define JCPU_RESUME_CODE 1

/*
 * jcpuContext flags.
 *   JCF_ROOT_FRAME - Indicates this context captures the top frame on a stack.
 *   JCF_ACTIVE - Indicates that this context is currently running.
 */
enum {
	JCB_ROOT_FRAME,
	JCB_ACTIVE,
};

enum {
	JCF_ROOT_FRAME = (1L << JCB_ROOT_FRAME),
	JCF_ACTIVE = (1L << JCB_ACTIVE),
};

/*
 * A jcpuContext captures and stores the state of a running thread.
 *
 * jcc_Flags - Holds the set of JCF_* flags.
 * jcc_Env - A buffer for use with setjmp/longjmp.
 * jcc_StackTop - The last recorded top of the stack.
 * jcc_FPState - Optional floating point state.
 * jcc_ErrorNumber - Holds the unix "errno" value.
 * jcc_AsyncSignals - Manages asynchronous signals for this context.  The
 *   contents of this are not specified here.
 */
typedef struct _jcpuContext {
	unsigned long jcc_Flags;
	JCPU_JMPBUF jcc_Env;
	void *jcc_StackTop;
#if defined(SAVED_FP_SIZE)
	char jcc_FPState[SAVED_FP_SIZE];
#endif
	int jcc_ErrorNumber;
	jsignalSet jcc_AsyncSignals;
} jcpuContext;

/*
 * Initialize a jcpuContext object with the given stack.
 *
 * jcc - The jcpuContext object to initialize.
 * newStack - The stack to install in the new context.
 */
void jcpuContextInit(jcpuContext *jcc, void *newStack);

/*
 * Restore a previously captured jcpuContext.
 *
 * jcc - The jcpuContext to restore.
 */
void jcpuContextRestore(jcpuContext *jcc);

/*
 * Print the value of the jcpuContext to the file.
 *
 * file - The file to print to.
 * jcc - The jcpuContext to print.
 */
void jcpuContextPrint(FILE *file, jcpuContext *jcc);

/*
 * A call back function which executes in the root frame of a new context.
 * Typically, this function would update any thread specific state and execute
 * the thread's start function.
 */
extern void jcpuContextRootFrame(void);

/*
 * A call back function used to handle asynchronous signals sent to the current
 * thread.
 *
 * recvmask - The mask of signals sent to the thread.
 */
extern void jcpuContextAsyncSignals(jsignal_set_t recvmask);

/*
 * Capture the state of the current thread and store it in the given context.
 *
 * jcc - The jcpuContext used to store the state.
 */
#define jcpuContextCapture(jcc) \
({ \
	int jcc_retval; \
\
	require(jcpuDisabled()); \
\
	require((jcc) != 0); \
	require(!((jcc)->jcc_Flags & JCF_ROOT_FRAME)); \
	require((jcc)->jcc_Flags & JCF_ACTIVE); \
\
	SAVE_FP((jcc)->jcc_FPState); \
	if( JCPU_SETJMP((jcc)->jcc_Env) == JCPU_RESUME_CODE ) \
	{ \
		require((jcc)->jcc_Flags & JCF_ACTIVE); \
\
		LOAD_FP((jcc)->jcc_FPState); \
		jcc_retval = 1; \
	} \
	else \
	{ \
		(jcc)->jcc_Flags &= ~JCF_ACTIVE; \
		(jcc)->jcc_StackTop = GET_SP((jcc)->jcc_Env); \
		(jcc)->jcc_ErrorNumber = errno; \
		jcc_retval = 0; \
	} \
\
	ensure(jcpuDisabled()); \
\
	jcc_retval; \
})

#define UNIX_JTHREADS2_CONTEXT_CONFIG_BASE \
	" (JCPU_SETJMP=" __XSTRING(JCPU_SETJMP) ")" \
	" (JCPU_LONGJMP=" __XSTRING(JCPU_LONGJMP) ")" \
	" (SP_OFFSET=" __XSTRING(SP_OFFSET) ")" \
	" (SAVED_FP_SIZE=" __XSTRING(SAVED_FP_SIZE) ")" \
	" (JTHREAD_STACK_COPY=" __XSTRING(JTHREAD_STACK_COPY) ")"

#if defined(BP_OFFSET)
#define UNIX_JTHREADS2_CONTEXT_CONFIG_BP_OFFSET \
	" (BP_OFFSET=" __XSTRING(BP_OFFSET) ")"
#else
#define UNIX_JTHREADS2_CONTEXT_CONFIG_BP_OFFSET
#endif

#if defined(FP_OFFSET)
#define UNIX_JTHREADS2_CONTEXT_CONFIG_FP_OFFSET \
	" (FP_OFFSET=" __XSTRING(FP_OFFSET) ")"
#else
#define UNIX_JTHREADS2_CONTEXT_CONFIG_FP_OFFSET
#endif

#if defined(SAVE_FP)
#define UNIX_JTHREADS2_CONTEXT_CONFIG_SAVE_FP " (SAVE_FP)"
#else
#define UNIX_JTHREADS2_CONTEXT_CONFIG_SAVE_FP
#endif

#if defined(NEED_STACK_ALIGN)
#define UNIX_JTHREADS2_CONTEXT_CONFIG_NEED_STACK_ALIGN " (NEED_STACK_ALIGN)"
#else
#define UNIX_JTHREADS2_CONTEXT_CONFIG_NEED_STACK_ALIGN
#endif

#define UNIX_JTHREADS2_CONTEXT_CONFIG \
	UNIX_JTHREADS2_CONTEXT_CONFIG_BASE \
	UNIX_JTHREADS2_CONTEXT_CONFIG_BP_OFFSET \
	UNIX_JTHREADS2_CONTEXT_CONFIG_FP_OFFSET \
	UNIX_JTHREADS2_CONTEXT_CONFIG_SAVE_FP \
	UNIX_JTHREADS2_CONTEXT_CONFIG_NEED_STACK_ALIGN

#endif
