/*
 * jcpuSignal.h
 *
 * Copyright (c) 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

#ifndef _jcpu_signal_h
#define _jcpu_signal_h

#include <stdio.h>
#include <signal.h>

/*
 * Signal set types:
 *   all - All of the signals that are handled by jthreads.
 *   sync - Synchronous signals (e.g. SIGFP, SIGSEGV).
 *   alt_stack - Signals to be taken on an alternate stack.
 *   async - Asynchronous signals (e.g. SIGALRM).
 */
typedef enum {
	JCPU_ALL_SIGNALS,
	JCPU_SYNC_SIGNALS,
	JCPU_ALT_STACK_SIGNALS,
	JCPU_ASYNC_SIGNALS,
	JCPU_SIGNAL_SET_TYPE_COUNT
} jcpu_signal_set_type_t;

/*
 * Global signal state.
 *
 * jss_Signals - Preinitialized sigset_t's for the different types.
 * jss_QueuedSignalCount - The number of queued signals.
 * jss_QueuedSignals - The queued signals.
 */
typedef struct _jcpuSignalState {
	sigset_t jss_SignalSets[JCPU_SIGNAL_SET_TYPE_COUNT];
	int jss_QueuedSignalCount;
	sigset_t jss_QueuedSignals;
} jcpuSignalState;

#if defined(SIGVTALRM)
#define JCPU_SIGNAL_ASYNC_COUNT 6
#else
#define JCPU_SIGNAL_ASYNC_COUNT 5
#endif
#define JCPU_SIGNAL_SYNC_COUNT 4
#define JCPU_SIGNAL_ALT_STACK_COUNT 2
#define JCPU_SIGNAL_IGNORED_COUNT 1

/*
 * Array containing the values of all the asynchronous signals that will be
 * handled.
 */
extern int jcpuSignalAsync[JCPU_SIGNAL_ASYNC_COUNT];

/*
 * Array containing the values of all the synchronous signals that will be
 * handled.
 */
extern int jcpuSignalSync[JCPU_SIGNAL_SYNC_COUNT];

/*
 * Array containing the values of all the synchronous signals that will be
 * handled on an alternate stack.
 */
extern int jcpuSignalAltStack[JCPU_SIGNAL_ALT_STACK_COUNT];

/*
 * Array containing the values of all the ignored signals.
 */
extern int jcpuSignalIgnored[JCPU_SIGNAL_IGNORED_COUNT];

extern jcpuSignalState jcpu_signal_state;

/*
 * Initialize the global signal state.
 */
int jcpuSignalInit(void);

/*
 * Register a handler for a signal.
 *
 * sig - The signal type.
 * handler - The signal handler to install.
 */
void jcpuSignalRegister(int sig, void *handler);

/*
 * type - The signal set type to block.
 */
static inline
void jcpuSignalBlock(jcpu_signal_set_type_t type)
{
	require(type >= 0);
	require(type < JCPU_SIGNAL_SET_TYPE_COUNT);
	
	sigprocmask(SIG_BLOCK, &jcpu_signal_state.jss_SignalSets[type], 0);
}

/*
 * type - The signal set type to unblock.
 */
static inline
void jcpuSignalUnblock(jcpu_signal_set_type_t type)
{
	require(type >= 0);
	require(type < JCPU_SIGNAL_SET_TYPE_COUNT);
	
	sigprocmask(SIG_UNBLOCK, &jcpu_signal_state.jss_SignalSets[type], 0);
}

/*
 * Queue a signal for future processing.
 *
 * sig - The signal number to queue.
 */
static inline
void jcpuSignalEnqueue(int sig)
{
	require(jcpu_signal_state.jss_QueuedSignalCount >= 0);
	require(jcpu_signal_state.jss_QueuedSignalCount <=
		JCPU_SIGNAL_ASYNC_COUNT);
	
	require(sigismember(&jcpu_signal_state.
			    jss_SignalSets[JCPU_ASYNC_SIGNALS],
			    sig));

	if( !sigismember(&jcpu_signal_state.jss_QueuedSignals, sig) )
	{
		sigaddset(&jcpu_signal_state.jss_QueuedSignals, sig);
		jcpu_signal_state.jss_QueuedSignalCount++;
	}

	ensure(jcpu_signal_state.jss_QueuedSignalCount > 0);
	ensure(jcpu_signal_state.jss_QueuedSignalCount <=
	       JCPU_SIGNAL_ASYNC_COUNT);
}

/*
 * Dequeue a previously queued signal.
 *
 * sig - The signal number to dequeue.
 */
static inline
void jcpuSignalDequeue(int sig)
{
	require(jcpu_signal_state.jss_QueuedSignalCount > 0 );
	require(jcpu_signal_state.jss_QueuedSignalCount <=
		JCPU_SIGNAL_ASYNC_COUNT);
	
	require(sigismember(&jcpu_signal_state.
			    jss_SignalSets[JCPU_ASYNC_SIGNALS],
			    sig));
	
	sigdelset(&jcpu_signal_state.jss_QueuedSignals, sig);
	jcpu_signal_state.jss_QueuedSignalCount--;

	ensure(jcpu_signal_state.jss_QueuedSignalCount >= 0);
	ensure(jcpu_signal_state.jss_QueuedSignalCount <
	       JCPU_SIGNAL_ASYNC_COUNT);
}

/*
 * sig - The signal to check.
 * return - True if the signal was previously queued.
 */
static inline
int jcpuSignalIsQueued(int sig)
{
	require(sigismember(&jcpu_signal_state.
			    jss_SignalSets[JCPU_ASYNC_SIGNALS],
			    sig));
	
	return( sigismember(&jcpu_signal_state.jss_QueuedSignals, sig) );
}

#if defined(SIGVTALRM)
#define UNIX_JTHREADS2_SIGNAL_CONFIG_SIGVTALRM " (SIGVTALRM)"
#else
#define UNIX_JTHREADS2_SIGNAL_CONFIG_SIGVTALRM
#endif

#if defined(SA_SIGINFO)
#define UNIX_JTHREADS2_SIGNAL_CONFIG_SA_SIGINFO " (SA_SIGINFO)"
#else
#define UNIX_JTHREADS2_SIGNAL_CONFIG_SA_SIGINFO
#endif

#if defined(SA_RESTART)
#define UNIX_JTHREADS2_SIGNAL_CONFIG_SA_RESTART " (SA_RESTART)"
#else
#define UNIX_JTHREADS2_SIGNAL_CONFIG_SA_RESTART
#endif

#define UNIX_JTHREADS2_SIGNAL_CONFIG \
	" (MINSIGSTKSZ=" __XSTRING(MINSIGSTKSZ) ")" \
	UNIX_JTHREADS2_SIGNAL_CONFIG_SIGVTALRM \
	UNIX_JTHREADS2_SIGNAL_CONFIG_SA_RESTART \
	UNIX_JTHREADS2_SIGNAL_CONFIG_SA_SIGINFO

#endif
