/*
 * jcpu_unix.c
 *
 * Copyright (c) 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

#include <assert.h>
#include <stdarg.h>
#include <signal.h>
#include <sys/time.h>

#include "jtimer_unix.h"
#include "jio_unix.h"

#include "jcpu_unix.h"

/* XXX Find a better place for this. */
#include <xprof/xprofiler.h>

jcpuState jcpu_state;

static
void jcpuAsyncHandler(SIGNAL_ARGS(sig, sc))
{
#if defined(KAFFE_XPROFILER)
	if( sig == SIGVTALRM )
	{
		SIGNAL_CONTEXT_POINTER(scp) =
			GET_SIGNAL_CONTEXT_POINTER(sc);
		
		profileHit((char *)SIGNAL_PC(scp));
	}
#endif
	jcpuSignalEnqueue(sig);
	if( jcpu_state.jcs_DisableCount == 0 )
	{
		jcpuDisable();
		
		if( !(jsked_state.jss_Flags & JSSF_IN_SWAP) )
		{
			jcpuUpdateUsedTime();

			jcpuSignalUnblock(JCPU_ASYNC_SIGNALS);
		}
		
		jcpuHandleQueuedSignals();
		
		jcpuEnable();
	}
	/*
	 * XXX Why the hell don't we queue jcpuEvents here?
	 * We'll even be able to capture the data in the sigcontext...
	 */
}

/* Thread quantum */
#define JCPU_QUANTUM 10000 /* 10 ms */

static
void jcpuActivateTimeSlicing(void)
{
	struct itimerval tv;

	tv.it_interval.tv_sec = 0;
	tv.it_value.tv_sec = 0;
	tv.it_interval.tv_usec = JCPU_QUANTUM;
	tv.it_value.tv_usec = JCPU_QUANTUM;
	setitimer(ITIMER_VIRTUAL, &tv, 0);
}

static
void jcpuDeactivateTimeSlicing(void)
{
	struct itimerval tv;

	memset(&tv, 0, sizeof(tv));
	setitimer(ITIMER_VIRTUAL, &tv, 0);
}

int jcpuInit(int tag, ...)
{
	static int initialized = 0;

	int retval = 0;
	va_list args;

	va_start(args, tag);
	if( !initialized )
	{
		if( jcpuSignalInit() )
		{
			int lpc;
			
			for( lpc = 0;
			     lpc < (sizeof(jcpuSignalAsync) / sizeof(int));
			     lpc++ )
			{
				jcpuSignalRegister(jcpuSignalAsync[lpc],
						   jcpuAsyncHandler);
			}
			
			jtimerInit();
			
			while( tag != JCPUA_TAG_DONE )
			{
				switch( tag )
				{
				default:
					assert(0);
					break;
				}
				tag = va_arg(args, int);
			}
			
			jcpuActivateTimeSlicing();

			initialized = 1;
			
			retval = 1;
		}
	}
	else
	{
		retval = 1;
	}
	va_end(args);
	return( retval );
}

static
jcpuEvent *jcpuHandleAsyncSignal(int sig, jcpuEvent *jce)
{
	jcpuEvent *retval = 0;

	require(jce != 0);
	
	jcpuSignalDequeue(sig);
	switch( sig )
	{
	case SIGINT:
	case SIGTERM:
		jce->jce_Type = JCPU_EVENT_RESET;
		retval = jce;
		break;
	case SIGIO:
		jioQueryFDs(0); /* Poll all the file descriptors */
		jce->jce_Type = JCPU_EVENT_IO;
		retval = jce;
		break;
	case SIGALRM:
		jtimerReset(&jtimer_state);
		jce->jce_Type = JCPU_EVENT_TIMER;
		retval = jce;
		break;
	case SIGCHLD:
		jce->jce_Type = JCPU_EVENT_CHILD;
		retval = jce;
		break;
#if defined(SIGVTALRM)
	case SIGVTALRM:
		if( jcpu_state.jcs_ForbidCount == 0 )
		{
			jce->jce_Type = JCPU_EVENT_PREEMPT;
			retval = jce;
		}
		break;
#endif
	default:
		fprintf(stderr,
			"jcpu_unix.c:Internal Error:"
			"Unhandled CPU signal %d\n",
			sig);
		ensure(0);
		break;
	}
	return( retval );
}

void jcpuHandleQueuedSignals(void)
{
	if( jcpu_signal_state.jss_QueuedSignalCount )
	{
		jcpuEvent *jce, *head = 0;
		int lpc, eventCount = 0;
		
		for( lpc = 0;
		     (lpc < (sizeof(jcpuSignalAsync) / sizeof(int))) &&
			     jcpu_signal_state.jss_QueuedSignalCount;
		     lpc++ )
		{
			if( jcpuSignalIsQueued(jcpuSignalAsync[lpc]) )
			{
				if( (jce = jcpuHandleAsyncSignal(
					jcpuSignalAsync[lpc],
					&jcpu_state.
					jcs_EventList[eventCount])) )
				{
					eventCount += 1;
					jce->jce_Next = head;
					head = jce;
				}
			}
		}
		jcpuAsyncEventHandler(head);
	}
}
