/*
 * jcpu_unix.h
 *
 * Copyright (c) 2002, 2003 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

#ifndef _jcpu_unix_h
#define _jcpu_unix_h

#include <signal.h>
#include <sys/time.h>

#include <config.h>
#if defined(HAVE_GETRUSAGE)
#include <sys/resource.h>
#endif

#include "assert_pp.h"

#include "jcpuEvent.h"
#include "jcpuSignal.h"
#include "jcpuContext.h"

#define JCPU_SPIN_LOCK_DECL

#if !defined(SIGNAL_ARGS)
#define SIGNAL_ARGS(sig, sc) int sig, void *sc
#endif

/*
 * The global jcpu state.
 *
 * jcs_DisableCount - The number of times interrupts have been disabled.
 * jcs_EventList - Preallocated jcpuEvents.
 * jcs_CurrentUsedTime - Cache for the currently used amount of CPU time.
 */
typedef struct _jcpuState {
	unsigned int jcs_DisableCount;
	unsigned int jcs_ForbidCount;
	jcpuEvent jcs_EventList[JCPU_EVENT_MAX];
	unsigned long long jcs_StartTime;
	unsigned long long jcs_CurrentUsedTime;
} jcpuState;

extern jcpuState jcpu_state;

enum {
	JCPUA_TAG_DONE,
};

/*
 * Initialize the global jcpu state.
 */
int jcpuInit(int tag, ...);

/*
 * Function used to handle queued signals.
 */
void jcpuHandleQueuedSignals(void);

/*
 * Disabled interrupts.
 */
static inline
void jcpuDisable(void)
{
	require(jcpu_state.jcs_DisableCount == 0);
	
	jcpu_state.jcs_DisableCount++;
}

/*
 * Enable interrupts.
 */
static inline
void jcpuEnable(void)
{
	require(jcpu_state.jcs_DisableCount == 1);

	if( jcpu_state.jcs_DisableCount == 1 )
	{
		jsignal_set_t recvmask = 0;
		
		/*
		 * Block the signals so we can avoid the race between
		 * the handling of queued signals and setting
		 * DisableCount to zero.
		 */
		// jcpuSignalBlock(JCPU_ASYNC_SIGNALS);
		
		/* Handle queued interrupts */
		if( jcpu_signal_state.jss_QueuedSignalCount )
		{
			jcpuHandleQueuedSignals();
		}
		else
		{
			jcpu_state.jcs_DisableCount = 0;
		}

		if( jcpu_context_state.jccs_AsyncSignals )
		{
			recvmask = jsignalRecv(jcpu_context_state.
					       jccs_AsyncSignals);
		}

		jcpu_state.jcs_DisableCount = 0;

		if( recvmask )
		{
			jcpuContextAsyncSignals(recvmask);
		}
		// jcpuSignalUnblock(JCPU_ASYNC_SIGNALS);
	}
	else
	{
		jcpu_state.jcs_DisableCount--;
	}
}

/*
 * return - The number of times interrupts have been disabled.
 */
static inline
unsigned int jcpuDisableCount(void)
{
	require(jcpu_state.jcs_DisableCount >= 0);
	require(jcpu_state.jcs_DisableCount <= 1);
	
	return( jcpu_state.jcs_DisableCount );
}

/*
 * return - The true if interrupts have been disabled.
 */
static inline
int jcpuDisabled(void)
{
	require(jcpu_state.jcs_DisableCount >= 0);
	require(jcpu_state.jcs_DisableCount <= 1);
	
	return( jcpu_state.jcs_DisableCount );
}

static inline
void jcpuSpinLock(void *obj)
{
	require(jcpu_state.jcs_ForbidCount >= 0);
	require(jcpu_state.jcs_ForbidCount < 10);
	
	jcpu_state.jcs_ForbidCount++;
}

static inline
void jcpuSpinUnlock(void *obj)
{
	require(jcpu_state.jcs_ForbidCount >= 1);
	require(jcpu_state.jcs_ForbidCount < 10);
	
	jcpu_state.jcs_ForbidCount--;
}

/*
 * return - The current time.  The value returned has no specific units until
 * it is run through the jcpuConvertTime function.
 */
static inline
unsigned long long jcpuTime(void)
{
	unsigned long long retval;
	
	{
		struct timeval tv;

		gettimeofday(&tv, 0);
		retval = (unsigned long long)tv.tv_usec;
		retval += ((unsigned long long)tv.tv_sec *
			   (unsigned long long)(1000 * 1000));

		retval -= jcpu_state.jcs_StartTime;
	}
	return( retval );
}

/*
 * return - The current time used by the process.  The value returned has no
 * specific units until it is run through the jcpuConvertTime function.
 */
static inline
unsigned long long jcpuUsedTime(void)
{
	return jcpu_state.jcs_CurrentUsedTime;
}

/*
 * Update the cached used CPU time value.
 */
static inline
void jcpuUpdateUsedTime(void)
{
	unsigned long long new_time;
	
#if defined(i386) && defined(NOT_SMP)
#define readitmr()	\
({ \
	unsigned long low, high; \
	asm volatile( \
	".byte 0x0f; .byte 0x31" \
	: "=d" (high), "=a" (low)); \
	low; \
})
	new_time = readitmr();
#undef readitmr
#elif defined(HAVE_GETRUSAGE)
	{
		struct rusage ru;

		switch( getrusage(RUSAGE_SELF, &ru) )
		{
		case -1:
			ensure(0);
			break;
		}
		new_time = (unsigned long long)ru.ru_utime.tv_usec;
		new_time += (unsigned long long)ru.ru_stime.tv_usec;
		new_time += ((unsigned long long)ru.ru_utime.tv_sec *
			   (unsigned long long)(1000 * 1000));
		new_time += ((unsigned long long)ru.ru_stime.tv_sec *
			   (unsigned long long)(1000 * 1000));
	}
#else
#if 1
	{
		new_time = clock();
	}
#else
	{
		struct timeval tv;

		gettimeofday(&tv, 0);
		new_time = (unsigned long long)tv.tv_usec;
		new_time += ((unsigned long long)tv.tv_sec *
			   (unsigned long long)(1000 * 1000));
	}
#endif
#endif
	jcpu_state.jcs_CurrentUsedTime = new_time;
}

/*
 * Types of time units supported:
 *   JCPU_TIME_MS - milliseconds.
 */
typedef enum {
	JCPU_TIME_MIN,
	JCPU_TIME_MS,
	JCPU_TIME_MAX,
} jcpu_time_type_t;

/*
 * Convert between the value returned by jcpuTime and something more
 * conventional.
 *
 * type - The type of time value to convert "value" to.
 * value - The value to convert.
 * return - The converted value.
 */
static inline
unsigned long long jcpuConvertTime(jcpu_time_type_t type,
				   unsigned long long value)
{
	unsigned long long retval;

	require(type > JCPU_TIME_MIN);
	require(type < JCPU_TIME_MAX);
	
#if defined(i386) && defined(NOT_SMP)
#error "Implement me"
#else
	switch( type )
	{
	case JCPU_TIME_MS:
#if defined(HAVE_GETRUSAGE)
		retval = value / 1000;
#else
		retval = (value / CLOCKS_PER_SEC) * 1000;
#endif
		break;
	default:
		require(0);
		break;
	}
#endif
	return( retval );
}

static inline
void jcpuTimeInit(void)
{
	if( jcpu_state.jcs_StartTime == 0 )
	{
		jcpu_state.jcs_StartTime = jcpuTime();
		jcpuUpdateUsedTime();
	}
}

/*
 * Callback used to allocate memory.
 *
 * size - The number of bytes to allocate.
 * type - The type of memory being allocated.
 */
extern void *jcpuMalloc(size_t size, int type);

extern void *jcpuRealloc(void *old, size_t size, int type);

/*
 * Callback used to free memory.
 *
 * mem - Memory previously allocated with jcpuMalloc.
 * type - The type of memory being allocated.
 */
extern void jcpuFree(void *mem);
/*
 * Callback used to free memory that does not automatically record its size
 * and type.
 *
 * mem - Memory previously allocated with jcpuMalloc.
 * size - The size of memory allocated with jcpuMalloc.
 * type - The type of memory allocated with jcpuMalloc.
 */
extern void jcpuFreeExplicit(void *mem, size_t size, int type);

/*
 * Callback used to handle a singly linked list of jcpuEvents.
 */
extern void jcpuAsyncEventHandler(jcpuEvent *jce);

#if defined(HAVE_GETRUSAGE)
#define UNIX_JTHREADS2_JCPU_CONFIG_GETRUSAGE " (HAVE_GETRUSAGE)"
#else
#define UNIX_JTHREADS2_JCPU_CONFIG_GETRUSAGE
#endif

#define UNIX_JTHREADS2_JCPU_CONFIG \
	" (SIGNAL_ARGS=" __XSTRING((SIGNAL_ARGS(sig, sc))) ")" \
	UNIX_JTHREADS2_JCPU_CONFIG_GETRUSAGE

#endif
