/*
 * jioAccounting.h
 *
 * Copyright (c) 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

#ifndef _jio_accounting_h
#define _jio_accounting_h

#include "assert_pp.h"

/*
 * The types of accounts tracked.
 *
 *   JIO_ACCOUNT_IN - Input.
 *   JIO_ACCOUNT_OUT - Output.
 */
typedef enum {
	JIO_ACCOUNT_IN,
	JIO_ACCOUNT_OUT,
	JIO_ACCOUNT_MAX,
} jio_accounting_type_t;

/*
 * The type used for the I/O accumulators.
 */
typedef unsigned long long jio_account_t;

/*
 * A jioAccount tracks the amount of I/O done for an object.
 *
 * jia_Accumulator - The input and output accumulators.
 */
typedef struct _jioAccount {
	jio_account_t jia_Accumulator[JIO_ACCOUNT_MAX];
} jioAccount;

/*
 * Macro used to initialize a statically allocated jioAccount.
 */
#define METAL_JIOACCOUNT_DATA \
{ \
	{ 0, 0 } \
}

/*
 * Initialize a jioAccount.
 *
 * jia - The account to initialized.
 */
void jioAccountingInit(jioAccount *jia);
/*
 * Finalize a jioAccount.
 *
 * jia - A previously initialized account.
 */
void jioAccountingKill(jioAccount *jia);
/*
 * Print the contents of a jioAccount.
 *
 * file - The file to print the jioAccount to.
 * jia - The account to print.
 */
void jioAccountingPrint(FILE *file, jioAccount *jia);

/*
 * Add an amount to a jioAccount's accumulator.
 *
 * jia - An initialized jioAccount.
 * type - The accumulator type.
 * amount - The amount to add to the accumulator.
 */
static inline
void jioAccountingAdd(jioAccount *jia,
		      jio_accounting_type_t type,
		      jio_account_t amount)
{
	require(jia != 0);
	require(type >= JIO_ACCOUNT_IN);
	require(type < JIO_ACCOUNT_MAX);

	jia->jia_Accumulator[type] += amount;
}

/*
 * Add one account's accumulators to another one.
 *
 * dst - An initialized jioAccount that will receive the other's accumulator
 *   values.
 * src - An initialized jioAccount that is to be added to 'dst'.  After this
 *   operation has completed, src's accumulators will be zeroed out.
 */
static inline
void jioAccountingAbsorb(jioAccount *dst, jioAccount *src)
{
	int lpc;
	
	require(dst != 0);
	require(src != 0);

	for( lpc = 0; lpc < JIO_ACCOUNT_MAX; lpc++ )
	{
		dst->jia_Accumulator[lpc] += src->jia_Accumulator[lpc];
		src->jia_Accumulator[lpc] = 0;
	}
}

#endif
