/*
 * jio_file.c
 *
 * Copyright (c) 2002, 2003 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

#include <fcntl.h>
#include <sys/param.h>
#include <sys/types.h>
#include <sys/socket.h>

#include "config.h"
#include "config-std.h"

#include "assert_pp.h"

#include "listNode-inline.h"

#include "jsked.h"
#include "jthreadState.h"
#include "jthread.h"
#include "jio_file.h"

/*
 * Type of jioTimedIO request:
 *   TYPE_STD_IN - Standard read operation.
 *   TYPE_STD_OUT - Standard write operation.
 *   TYPE_RECV_DATAGRAM - recvfrom.
 *   TYPE_SEND_DATAGRAM - sendto.
 *   TYPE_ACCEPT - accept.
 */
enum {
	TYPE_STD_IN,
	TYPE_STD_OUT,
	TYPE_ACCEPT,
	TYPE_DATAGRAM,
	TYPE_RECV_DATAGRAM,
	TYPE_SEND_DATAGRAM,
};

/*
 * State values for the jioFileState object:
 *   JIO_FILE_STATE_IDLE - Operation was successful.
 *   JIO_FILE_STATE_ERROR - Operation failed.
 *   JIO_FILE_STATE_DO - Do the requested operation.
 *   JIO_FILE_STATE_DO_ASYNC - Redo the operation and send a jioRequest.
 */
typedef enum {
	JIO_FILE_STATE_MIN,
	JIO_FILE_STATE_IDLE,
	JIO_FILE_STATE_ERROR,
	JIO_FILE_STATE_DO,
	JIO_FILE_STATE_DO_ASYNC,
	JIO_FILE_STATE_MAX
} jio_file_state_t;

/*
 * A jioFileState object is used to store the current state of a file
 * operation.
 */
typedef struct _jioFileState {
	jio_file_state_t jfs_State;
} jioFileState;

/* XXX Refactor me. */
static
int jioTimedIO(int type, int fd, void *buf, size_t len, int timeout,
	       ssize_t *out,
	       int (*func)(int, void *, size_t),
	       int flags, struct sockaddr *saddr, socklen_t *slen)
{
	jio_accounting_type_t acc_type = JIO_ACCOUNT_IN;
	jthread_t curr = jthread_current();
	int bytes_io = 0, retval = 0;
	jsked *js = jskedContext();
	jioRequestPort *jrp;
	jioFileState jfs;
	jioRequest *jr;

	require(fd >= 0);
	require(fd < FD_SETSIZE);

	jrp = &js->js_IORequestPort;
	jfs.jfs_State = JIO_FILE_STATE_DO;
	while( (jfs.jfs_State != JIO_FILE_STATE_IDLE) &&
	       (jfs.jfs_State != JIO_FILE_STATE_ERROR) )
	{
		int rc = 0;
		
		switch( jfs.jfs_State )
		{
		case JIO_FILE_STATE_DO:
			/* Do the operation directly. */
			switch( type )
			{
			case TYPE_STD_IN:
				rc = (bytes_io = func(fd, buf, len)) >= 0;
				acc_type = JIO_ACCOUNT_IN;
				break;
			case TYPE_STD_OUT:
				rc = (bytes_io = func(fd, buf, len)) >= 0;
				acc_type = JIO_ACCOUNT_OUT;
				break;
			case TYPE_RECV_DATAGRAM:
				rc = (bytes_io = recvfrom(fd,
							  buf,
							  len,
							  flags,
							  saddr,
							  slen)) >= 0;
				acc_type = JIO_ACCOUNT_IN;
				break;
			case TYPE_SEND_DATAGRAM:
				rc = (bytes_io = sendto(fd,
							buf,
							len,
							flags,
							saddr,
							*slen)) >= 0;
				acc_type = JIO_ACCOUNT_OUT;
				break;
			case TYPE_ACCEPT:
#if 1
				rc = 0;
				errno = EAGAIN;
#else
				jcpuDisable();
				{
					int newFD;

					rc = (newFD = accept(fd, saddr, slen))
					    >= 0;
					if( rc )
					{
						if( (jrp->jrp_FDCount + 1) >
						    jrp->jrp_FDMax )
						{
							close(newFD);
							retval = EMFILE;
						}
						else
						{
							jioAllocFD(
								jrp,
								JIO_FD_TYPE_SOCKET,
								newFD);
							*((int *)buf) = newFD;
							bytes_io = 1;
						}
					}
				}
				jcpuEnable();
#endif
				acc_type = JIO_ACCOUNT_IN;
				break;
			default:
				assert(0);
				break;
			}
			if( rc )
			{
				/* Success */
				jfs.jfs_State = JIO_FILE_STATE_IDLE;
			}
			else
			{
				switch( errno )
				{
				case EINTR:
					/* Try again */
					break;
				case EAGAIN:
				case EINPROGRESS:
					jfs.jfs_State =
						JIO_FILE_STATE_DO_ASYNC;
					break;
				case EALREADY:
					jfs.jfs_State = JIO_FILE_STATE_ERROR;
					retval = EISCONN;
					break;
				default:
					jfs.jfs_State = JIO_FILE_STATE_ERROR;
					retval = errno;
					break;
				}
			}
			break;
		case JIO_FILE_STATE_DO_ASYNC:
			/* Initialize the thread's jioRequest */
			jr = &curr->jt_IORequest;
			jr->jr_Message.jm_Flags = 0;
			jr->jr_FD = fd;
			jcpuDisable();
			/*
			 * Send the request immediately and rerun the
			 * operation.
			 *
			 * NOTE: We do this because of the race between
			 * setting asynchronous signalling on the file and
			 * doing the operation.
			 */
			jioSend(jrp, jr);
			switch( type )
			{
			case TYPE_STD_IN:
			case TYPE_STD_OUT:
				rc = (bytes_io = func(fd, buf, len)) >= 0;
				break;
			case TYPE_RECV_DATAGRAM:
				rc = (bytes_io = recvfrom(fd,
							  buf,
							  len,
							  flags,
							  saddr,
							  slen)) >= 0;
				break;
			case TYPE_SEND_DATAGRAM:
				rc = (bytes_io = sendto(fd,
							buf,
							len,
							flags,
							saddr,
							*slen)) >= 0;
				break;
			case TYPE_ACCEPT:
				{
					int newFD;

					rc = (newFD = accept(fd, saddr, slen))
					    >= 0;
					if( rc )
					{
						jioFD *jfd;
						
						if( (jfd = jioAllocFD(jrp)) )
						{
							jioSharedFD *jsfd;
							
							jsfd = jioAllocSharedFD(newFD, JIO_FD_TYPE_SOCKET);
							jioBindFD(jrp, jfd, jsfd);
							*((int *)buf) = newFD;
							bytes_io = 1;
						}
						else
						{
							close(newFD);
							retval = EMFILE;
						}
					}
				}
				break;
			default:
				assert(0);
				break;
			}
			if( rc )
			{
				jfs.jfs_State = JIO_FILE_STATE_IDLE;
			}
			else
			{
				jtimer *jt = &curr->jt_TimerRequest;
				jsignal_set_t recvmask, waitmask;

				/*
				 * Wait for the I/O reply port signal or an
				 * asynchronous interrupt from the program.
				 */
				waitmask = (JTSIG_IO |
					    JTSIG_INTERRUPT |
					    JTSIG_KILL);
				switch( errno )
				{
				case EINTR:
					jfs.jfs_State = JIO_FILE_STATE_DO;
					break;
				case EAGAIN:
					if( timeout )
					{
						unsigned long long ct;

						ct = jtimerCurrent();
						/* Also wait for a timer. */
						jt->jt_Message.jm_Flags = 0;
						jt->jt_Flags = 0;
						jt->jt_AbsoluteTime =
							ct +
							timeout;
						if( jt->jt_AbsoluteTime > ct )
						{
							jtimerAdd(&js->
								  js_SubTimers,
								  jt);
							waitmask |=
								JTSIG_TIMER;
						}
					}
					recvmask = jthreadStateWait(
						&curr->jt_State, waitmask);
					if( (recvmask & JTSIG_INTERRUPT) ||
					    (recvmask & JTSIG_KILL) )
					{
						jfs.jfs_State =
							JIO_FILE_STATE_ERROR;
						retval = EINTR;
					}
					if( recvmask & JTSIG_TIMER )
					{
						jfs.jfs_State =
							JIO_FILE_STATE_ERROR;
						retval = ETIMEDOUT;
					}
					if( recvmask & JTSIG_IO )
					{
						/* Success. */
						jfs.jfs_State =
							JIO_FILE_STATE_DO;
						retval = 0;
					}
					if( timeout )
					{
						jtimerRem(&js->js_SubTimers,
							  jt);
					}
					break;
				default:
					jfs.jfs_State = JIO_FILE_STATE_ERROR;
					retval = errno;
					break;
				}
			}
			/* Cancel the request if it is still queued. */
			jioCancel(jr);
			lnRemove(&jr->jr_Message.jm_Link);
			if( jr->jr_Flags & JIOF_CANCELLED )
			{
				/*
				 * The jioCancel will call jmessageReply,
				 * which sets our JTSIG_IO signal, so we
				 * need to clear it.
				 */
				jthreadStateWait(&curr->jt_State, JTSIG_IO);
			}
			jcpuEnable();
			break;
		default:
			require(0);
			break;
		}
	}
	if( bytes_io )
	{
		jioSharedFD *jsfd;

		jsfd = &jio_state.jis_FDs[fd];
		switch( jsfd->jsfd_Type )
		{
		case JIO_FD_TYPE_FILE:
		case JIO_FD_TYPE_STDIO:
			jioAccountingAdd(&jsfd->jsfd_Account,
					 acc_type,
					 bytes_io);
			jioAccountingAdd(&js->
					 js_IOAccount[JSKED_IOACCOUNT_DISK],
					 acc_type,
					 bytes_io);
			jioAccountingAdd(&jsked_state.
					 jss_IOAccount[JSKED_IOACCOUNT_DISK],
					 acc_type,
					 bytes_io);
			break;
		case JIO_FD_TYPE_PIPE:
			jioAccountingAdd(&jsfd->jsfd_Account,
					 acc_type,
					 bytes_io);
			jioAccountingAdd(&js->
					 js_IOAccount[JSKED_IOACCOUNT_IPC],
					 acc_type,
					 bytes_io);
			jioAccountingAdd(&jsked_state.
					 jss_IOAccount[JSKED_IOACCOUNT_IPC],
					 acc_type,
					 bytes_io);
			break;
		case JIO_FD_TYPE_SOCKET:
			if( type > TYPE_DATAGRAM )
			{
				jioAccountingAdd(&jsfd->jsfd_Account,
						 acc_type,
						 1);
				jioAccountingAdd(
					&js->js_IOAccount
					[JSKED_IOACCOUNT_NETWORK_PACKET],
					acc_type,
					1);
				jioAccountingAdd(
					&jsked_state.jss_IOAccount
					[JSKED_IOACCOUNT_NETWORK_PACKET],
					acc_type,
					1);
			}
			jioAccountingAdd(&jsfd->jsfd_Account,
					 acc_type,
					 bytes_io);
			jioAccountingAdd(&js->js_IOAccount
					 [JSKED_IOACCOUNT_NETWORK],
					 acc_type,
					 bytes_io);
			jioAccountingAdd(&jsked_state.jss_IOAccount
					 [JSKED_IOACCOUNT_NETWORK],
					 acc_type,
					 bytes_io);
			break;
		default:
			break;
		}
	}
	if( out != 0 )
	{
		*out = bytes_io;
	}
	return( retval );
}

int jioRead(int fd, void *buf, size_t len, ssize_t *out)
{
	int retval;

	if( fd >= 0 )
	{
		jthread_current()->jt_IORequest.jr_Flags = JIOF_READ;
		retval = jioTimedIO(TYPE_STD_IN, fd, buf, len, 0, out, read,
				    0, 0, 0);
	}
	else
	{
		retval = EBADF;
	}
	return( retval );
}

int jioReadAndWait(int fd, void *buf, size_t len, int sample_time,
		   ssize_t *out)
{
	int retval = 0;

	if( fd >= 0 )
	{
		jthread_current()->jt_IORequest.jr_Flags = JIOF_READ;
		while( (retval = jioTimedIO(TYPE_STD_IN,
					    fd, buf, len, sample_time, out,
					    read,
					    0, 0, 0)) == -1 );
	}
	else
	{
		retval = EBADF;
	}
	return( retval );
}

int jioTimedRead(int fd, void *buf, size_t len, int timeout, ssize_t *out)
{
	int retval;

	if( fd >= 0 )
	{
		jthread_current()->jt_IORequest.jr_Flags = JIOF_READ;
		retval = jioTimedIO(TYPE_STD_IN,
				    fd, buf, len, timeout, out, read,
				    0, 0, 0);
	}
	else
	{
		retval = EBADF;
	}
	return( retval );
}

int jioRecvfrom(int fd, void *buf, size_t len, int flags,
		struct sockaddr *saddr, int *slen, int timeout, ssize_t *out)
{
	int retval;

	if( fd >= 0 )
	{
		jthread_current()->jt_IORequest.jr_Flags = JIOF_READ;
		retval = jioTimedIO(TYPE_RECV_DATAGRAM,
				    fd, buf, len, timeout, out, 0,
				    flags, saddr, slen);
	}
	else
	{
		retval = EBADF;
	}
	return( retval );
}

int jioWrite(int fd, const void *buf, size_t len, ssize_t *out)
{
	int retval;

	if( fd >= 0 )
	{
		jthread_current()->jt_IORequest.jr_Flags = JIOF_WRITE;
		retval = jioTimedIO(TYPE_STD_OUT, fd, (void *)buf, len, 0, out,
				    (int (*)(int, void *, size_t))write,
				    0, 0, 0);
	}
	else
	{
		retval = EBADF;
	}
	return( retval );
}

int jioTimedWrite(int fd, const void *buf, size_t len, int timeout,
		  ssize_t *out)
{
	int retval;

	if( fd >= 0 )
	{
		jthread_current()->jt_IORequest.jr_Flags = JIOF_WRITE;
		retval = jioTimedIO(TYPE_STD_OUT,
				    fd, (void *)buf, len, timeout, out,
				    (int (*)(int, void *, size_t))write,
				    0, 0, 0);
	}
	else
	{
		retval = EBADF;
	}
	return( retval );
}

int jioSendto(int fd, const void* buf, size_t len, int flags,
	      const struct sockaddr *addr, int addrlen, ssize_t *out)
{
	int retval = 0;

	if( fd >= 0 )
	{
		jthread_current()->jt_IORequest.jr_Flags = JIOF_WRITE;
		retval = jioTimedIO(TYPE_SEND_DATAGRAM,
				    fd, (void *)buf, len, 0, out,
				    0,
				    flags, (struct sockaddr *)addr, &addrlen);
	}
	else
	{
		retval = EBADF;
	}
	return( retval );
}

int jioTimedConnect(int fd,
		    struct sockaddr *name,
		    int namelen,
		    int timeout)
{
	int retval;

	if( fd >= 0 )
	{
		jthread_current()->jt_IORequest.jr_Flags = JIOF_WRITE;
		if( (retval = jioTimedIO(TYPE_STD_OUT,
					 fd, name, namelen, timeout, 0,
					 (int (*)(int, void *, size_t))connect,
					 0, 0, 0)) )
		{
			switch( retval )
			{
			case ETIMEDOUT:
				shutdown(fd, SHUT_RDWR);
				break;
			default:
				break;
			}
		}
	}
	else
	{
		retval = EBADF;
	}
	return( retval );
	
}

int jioAccept(int fd,
	      struct sockaddr *name,
	      int *len,
	      int timeout,
	      int *out)
{
	int retval;

	if( fd >= 0 )
	{
		jthread_current()->jt_IORequest.jr_Flags = JIOF_READ;
		retval = jioTimedIO(TYPE_ACCEPT,
				    fd, out, 0, timeout, 0,
				    0,
				    0, name, len);
	}
	else
	{
		retval = EBADF;
	}
	return( retval );
}

int jioSocket(int af, int type, int proto, int *out)
{
	jsked *js = jskedContext();
	jioRequestPort *jrp;
	int fd, retval = 0;
	jioFD *jfd;
	
	jrp = &js->js_IORequestPort;
	jcpuDisable();
	if( (jfd = jioAllocFD(jrp)) )
	{
		if( (fd = socket(af, type, proto)) >= 0 )
		{
			jioSharedFD *jsfd;
			
			jsfd = jioAllocSharedFD(fd, JIO_FD_TYPE_SOCKET);
			jioBindFD(jrp, jfd, jsfd);
			*out = fd;
		}
		else
		{
			retval = errno;
		}
	}
	else
	{
		retval = EMFILE;
	}
	jcpuEnable();
	return( retval );
}

int jioOpen(const char *path, int flags, int mode, int *out)
{
	jsked *js = jskedContext();
	jioRequestPort *jrp;
	int fd, retval = 0;
	jioFD *jfd;
	
	require(path != 0);
	require(out != 0);

	*out = -1;
	jrp = &js->js_IORequestPort;
	jcpuDisable();
	if( (jfd = jioAllocFD(jrp)) )
	{
		if( (path = jioResolvePath(jrp,
					   path,
					   jio_state.jis_ResolvedPath1,
					   MAXPATHLEN)) &&
		    (fd = open(path,
			       flags
#if defined(O_BINARY)
			       | O_BINARY
#endif
			       , mode)) >= 0 )
		{
			jioSharedFD *jsfd;
			
			jsfd = jioAllocSharedFD(fd, JIO_FD_TYPE_FILE);
			jioBindFD(jrp, jfd, jsfd);
			*out = fd;
		}
		else
		{
			retval = errno;
		}
	}
	else
	{
		retval = EMFILE;
	}
	jcpuEnable();
	
	ensure((*out == -1) ||
	       (*out >= 0) ||
	       (*out < FD_SETSIZE));
	return( retval );
}

int jioPipe(int *files)
{
	jsked *js = jskedContext();
	jioFD *jfd[2] = { 0, 0 };
	jioRequestPort *jrp;
	int rc, retval = 0;

	require(files != 0);

	jrp = &js->js_IORequestPort;
	files[0] = -1;
	files[1] = -1;
	jcpuDisable();
	if( (jfd[0] = jioAllocFD(jrp)) &&
	    (jfd[1] = jioAllocFD(jrp)) )
	{
		if( (rc = pipe(files)) >= 0 )
		{
			jioSharedFD *jsfd[2];
			
			jsfd[0] = jioAllocSharedFD(files[0], JIO_FD_TYPE_FILE);
			jsfd[1] = jioAllocSharedFD(files[1], JIO_FD_TYPE_FILE);

			jioBindFD(jrp, jfd[0], jsfd[0]);
			jioBindFD(jrp, jfd[1], jsfd[1]);
		}
		else
		{
			retval = errno;
		}
	}
	else
	{
		retval = EMFILE;
	}
	jcpuEnable();
	return( retval );
}

int jioDup(int fd, int *out_fd)
{
	jsked *js = jskedContext();
	jioRequestPort *jrp;
	int retval = 0;
	jioFD *jfd;

	jrp = &js->js_IORequestPort;
	*out_fd = -1;
	jcpuDisable();
	if( (fd < 0) ||
	    (fd >= FD_SETSIZE) ||
	    (jio_state.jis_FDs[fd].jsfd_Type == JIO_FD_TYPE_FREE) )
	{
		retval = EBADF;
	}
	else if( (jfd = jioAllocFD(jrp)) )
	{
		jioSharedFD *jsfd;

		jsfd = &jio_state.jis_FDs[fd];
		jioBindFD(jrp, jfd, jsfd);
		*out_fd = fd;
	}
	else
	{
		retval = EMFILE;
	}
	jcpuEnable();

	ensure((*out_fd == -1) ||
	       (*out_fd >= 0) ||
	       (*out_fd < FD_SETSIZE));
	return( retval );
}

int jioClose(int fd)
{
	jsked *js = jskedContext();
	int retval = 0;
	
	if( (fd >= 0) && (fd < FD_SETSIZE) )
	{
		jcpuDisable();
		jioFreeFD(&js->js_IORequestPort, fd);
		jcpuEnable();
	}
	else
	{
		retval = EBADF;
	}
	return( retval );
}
