/*
 * jio_unix.h
 *
 * Copyright (c) 2002, 2003 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

#include <assert.h>
#include <stdarg.h>
#include <fcntl.h>
#include <errno.h>
#include <string.h>
#include <limits.h>

#include <sys/poll.h>

#include "assert_pp.h"

#include "config.h"
#include "config-std.h"

#include "listNode.h"
#include "listNode-inline.h"

#include "jcpu.h"
#include "jio_unix.h"

#include "jthread_config.h"

#if !defined(ALLOC_TYPE_FD)
#define ALLOC_TYPE_FD -1
#endif

jioState jio_state;

int jioInit(int tag, ...)
{
	int retval = 1;
	va_list args;

	va_start(args, tag);
	if( jio_state.jis_RequestPorts.lh_Head == NULL )
	{
		lnNewList(&jio_state.jis_RequestPorts);
		jio_state.jis_FDMax = -1;
		while( tag != JIOA_TAG_DONE )
		{
			switch( tag )
			{
			default:
				assert(0);
				break;
			}
			tag = va_arg(args, int);
		}
	}
	va_end(args);
	return( retval );
}

void jioInitRequestPort(jioRequestPort *jrp)
{
	require(jrp != 0);
	
	jmessagePortInit(&jrp->jrp_MessagePort, 0, -1);
	jrp->jrp_Flags |= JRPF_INITED;
}

void jioAddRequestPort(jioRequestPort *jrp)
{
	require(jcpuDisabled());
	
	require(jrp != 0);
	require(!(jrp->jrp_Flags & JRPF_ACTIVE));
	
	jrp->jrp_Flags |= JRPF_ACTIVE;
	lnAddTail(&jio_state.jis_RequestPorts, &jrp->jrp_Link);
}

void jioRemRequestPort(jioRequestPort *jrp)
{
	require(jcpuDisabled());
	
	require(jrp != 0);
	require(jrp->jrp_Flags & JRPF_ACTIVE);

	jrp->jrp_Flags &= ~JRPF_ACTIVE;
	lnRemove(&jrp->jrp_Link);
}

void jioKillRequestPort(jioRequestPort *jrp)
{
	jioFD *jfd;
	int lpc;
	
	require(jrp != 0);
	require(!jcpuDisabled());

	jcpuDisable();
	jfd = jrp->jrp_FDs.Data;
	for( lpc = 0; lpc < jrp->jrp_FDs.Max; lpc++ )
	{
		jioSharedFD *jsfd;

		if( (jsfd = jfd[lpc].jfd_FD) )
		{
			jfd[lpc].jfd_FD->jsfd_Users -= 1;
			jfd[lpc].jfd_FD = 0;
			if( jsfd->jsfd_Users == 0 )
			{
				jioFreeSharedFD(jsfd - jio_state.jis_FDs);
			}
		}
	}
	jrp->jrp_Flags &= ~JRPF_INITED;
	jcpuEnable();
	jcpuFree(jrp->jrp_FDs.Data);
	jrp->jrp_FDs.Data = 0;
}

static
void jioCopyFDs(jioFD *new_data, jioFD *old_data, unsigned int count)
{
	if( old_data )
	{
		int new_index = 0, lpc;
		
		for( lpc = 0; lpc < count; lpc++ )
		{
			if( old_data[lpc].jfd_FD )
			{
				new_data[new_index] = old_data[lpc];
				new_index += 1;
			}
		}
	}
}

int jioSetFDMax(jioRequestPort *jrp, unsigned int max)
{
	int retval = 0;

	if( max < FD_SETSIZE )
	{
		jioFD *new_data;

		if( (new_data = jcpuMalloc(max * sizeof(jioFD),
					   ALLOC_TYPE_FD)) )
		{
			jioFD *old_data;

			jcpuDisable();
			if( jrp->jrp_FDs.Count > max )
			{
				old_data = new_data;
				retval = EINVAL;
			}
			else
			{
				old_data = jrp->jrp_FDs.Data;
				jioCopyFDs(new_data, old_data, max);
				jrp->jrp_FDs.Data = new_data;
				jrp->jrp_FDs.Max = max;
			}
			jcpuEnable();
			jcpuFree(old_data);
		}
		else
		{
			retval = ENOMEM;
		}
	}
	else
	{
		retval = EINVAL;
	}
	return( retval );
}

jioSharedFD *jioAllocSharedFD(int fd, jio_fd_type_t type)
{
	jioSharedFD *retval = 0;

	require(fd >= 0);
	require(fd < FD_SETSIZE);
	require(type > JIO_FD_TYPE_MIN);
	require(type < JIO_FD_TYPE_MAX);
	require(jcpuDisabled());

	retval = &jio_state.jis_FDs[fd];
	{
		require(retval->jsfd_Type == JIO_FD_TYPE_FREE);

		retval->jsfd_Type = type;
		retval->jsfd_Users = 0;
		retval->jsfd_Waiters = 0;
		jioAccountingInit(&retval->jsfd_Account);

		jioRegisterFD(fd, type);
	}
	
	return( retval );
}

void jioFreeSharedFD(int fd)
{
	jioSharedFD *jsfd;

	require(fd >= 0);
	require(fd < FD_SETSIZE);
	require(jcpuDisabled());

	jsfd = &jio_state.jis_FDs[fd];
	{
		require(jsfd->jsfd_Users == 0);
		require(jsfd->jsfd_Type != JIO_FD_TYPE_FREE);

		jioUnregisterFD((jsfd - jio_state.jis_FDs), jsfd->jsfd_Type);
		jioAccountingKill(&jsfd->jsfd_Account);
		jsfd->jsfd_Type = JIO_FD_TYPE_FREE;
	}
}

jioFD *jioAllocFD(jioRequestPort *jrp)
{
	jioFD *retval = 0;

	require(jrp != 0);
	require(jrp->jrp_Flags & JRPF_INITED);
	require(jcpuDisabled());
	
	if( jrp->jrp_FDs.Count < jrp->jrp_FDs.Max )
	{
		int lpc;

		for( lpc = 0; (lpc < jrp->jrp_FDs.Max) && !retval; lpc++ )
		{
			if( jrp->jrp_FDs.Data[lpc].jfd_FD == 0 )
			{
				retval = &jrp->jrp_FDs.Data[lpc];
			}
		}
	}
	return( retval );
}

void jioFreeFD(jioRequestPort *jrp, int fd)
{
	jioFD *jfd;
	
	require(jrp != 0);
	require(fd >= 0);
	require(fd < FD_SETSIZE);
	require(jcpuDisabled());

	if( (jfd = jioFindFD(jrp, fd)) )
	{
		jioSharedFD *jsfd;

		require(jfd->jfd_FD != 0);

		jsfd = jfd->jfd_FD;
		jfd->jfd_FD = 0;
		jsfd->jsfd_Users -= 1;
		if( jsfd->jsfd_Users == 0 )
		{
			jioFreeSharedFD(fd);
		}
		jrp->jrp_FDs.Count -= 1;
	}
	else
	{
		require(0);
	}
}

jioFD *jioFindFD(jioRequestPort *jrp, int fd)
{
	jioSharedFD *jsfd;
	jioFD *retval = 0;
	int lpc;
	
	require(fd >= 0);
	require(fd < FD_SETSIZE);
	require(jcpuDisabled());
	
	jsfd = &jio_state.jis_FDs[fd];
	for( lpc = 0; (lpc < jrp->jrp_FDs.Max) && !retval; lpc++ )
	{
		if( jrp->jrp_FDs.Data[lpc].jfd_FD == jsfd )
		{
			retval = &jrp->jrp_FDs.Data[lpc];
		}
	}
	return( retval );
}

void jioBindFD(jioRequestPort *jrp, jioFD *jfd, jioSharedFD *jsfd)
{
	require(jfd != 0);
	require(jsfd != 0);
	require(jcpuDisabled());

	jrp->jrp_FDs.Count += 1;
	jfd->jfd_FD = jsfd;
	jsfd->jsfd_Users += 1;
}

int jioSetRootDirectory(jioRequestPort *jrp, char *path)
{
	int retval = 0;

	require(jrp != 0);
	require(path != 0);
	
	jcpuDisable();
	if( path[0] == '\0' )
	{
		jrp->jrp_RootDirectory[0] = '\0';
	}
	else if( realpath(path, jrp->jrp_RootDirectory) == 0 )
	{
		retval = errno;
	}
	else
	{
		jrp->jrp_RootDirectoryLen = strlen(jrp->jrp_RootDirectory);
		if( jrp->jrp_RootDirectory[jrp->jrp_RootDirectoryLen - 1] !=
		    '/' )
		{
			strcat(jrp->jrp_RootDirectory, "/");
			jrp->jrp_RootDirectoryLen += 1;
		}
	}
	jcpuEnable();

	return( retval );
}

typedef enum {
	JIO_PATH_END,
	JIO_PATH_SAME,
	JIO_PATH_DOWN,
	JIO_PATH_UP,
} path_depth_t;

static
path_depth_t jioPathComponent(const char **path)
{
	path_depth_t retval = JIO_PATH_DOWN;
	const char *component_end;

	if( *path[0] == '\0' )
	{
		retval = JIO_PATH_END;
	}
	else
	{
		if( (component_end = strchr(*path, '/')) == 0 )
		{
			component_end = *path + strlen(*path);
		}
		if( (*path) + 1 == component_end )
		{
			retval = JIO_PATH_SAME;
		}
		else if( strncmp(*path, "..", component_end - (*path)) == 0 )
		{
			retval = JIO_PATH_UP;
		}
		else if( strncmp(*path, ".", component_end - (*path)) == 0 )
		{
			retval = JIO_PATH_SAME;
		}
		else
		{
			retval = JIO_PATH_DOWN;
		}
		*path = component_end + 1;
	}
	return( retval );
}

char *jioResolvePath(jioRequestPort *jrp,
		     const char *path,
		     char *resolved_path,
		     size_t resolved_size)
{
	static char rooted_path[MAXPATHLEN];
	
	char *retval = 0;

	require(jrp != 0);
	require(path != 0);
	require(resolved_path != 0);
	require(resolved_size >= MAXPATHLEN);
	require(jcpuDisabled());

	if( jrp->jrp_RootDirectoryLen == 0 )
	{
		errno = EACCES;
	}
	else if( path[0] == '\0' )
	{
		errno = EINVAL;
	}
	else if( !strcmp(jrp->jrp_RootDirectory, "/") )
	{
		strncpy(resolved_path, path, resolved_size);
		retval = path;
	}
	else
	{
		const char *start_path = path, *next_path = path;
		int depth = 0, done = 0;
		char *last_comp = 0;

		do {
			switch( jioPathComponent(&next_path) )
			{
			case JIO_PATH_END:
				done = 1;
				break;
			case JIO_PATH_SAME:
				break;
			case JIO_PATH_DOWN:
				depth += 1;
				break;
			case JIO_PATH_UP:
				if( depth > 0 )
				{
					depth -= 1;
				}
				else
				{
					start_path = next_path;
				}
				break;
			}
		} while( !done );
		strcpy(rooted_path, jrp->jrp_RootDirectory);
		if( depth == 0 )
		{
			/* They're at depth zero, or above... */
			retval = resolved_path;
			strcpy(retval, rooted_path);
		}
		else
		{
			strncpy(&rooted_path[jrp->jrp_RootDirectoryLen],
				start_path,
				MAXPATHLEN - jrp->jrp_RootDirectoryLen);
			if( (last_comp = strrchr(rooted_path, '/')) )
			{
				last_comp[0] = '\0';
			}
			if( (retval = realpath(rooted_path, resolved_path)) )
			{
				if( last_comp )
				{
					last_comp[0] = '/';
					strcat(retval, last_comp);
				}
			}
			else
			{
				/* realpath() sets the errno. */
			}
		}
	}

	return( retval );
}

inline
void jioWatchFD(int fd, jio_watch_type_t type)
{
	require(FD_ISSET(fd, &jio_state.jis_AllocatedFDs));
	require(type > JIO_WATCH_TYPE_MIN);
	require(type < JIO_WATCH_TYPE_MAX);

	switch( type )
	{
	case JIO_WATCH_TYPE_CANCEL_READ:
		FD_CLR(fd, &jio_state.jis_PendingFDs.read);
		break;
	case JIO_WATCH_TYPE_CANCEL_WRITE:
		FD_CLR(fd, &jio_state.jis_PendingFDs.write);
		break;
	case JIO_WATCH_TYPE_READ:
		FD_SET(fd, &jio_state.jis_PendingFDs.read);
		break;
	case JIO_WATCH_TYPE_WRITE:
		FD_SET(fd, &jio_state.jis_PendingFDs.write);
		break;
	default:
		assert(0);
		break;
	}
}

void jioSend(jioRequestPort *jrp, jioRequest *jr)
{
	require(jrp != 0);
	require(jr != 0);
	require(jcpuDisabled());
	require(jrp->jrp_Flags & JRPF_ACTIVE);

	jr->jr_Message.jm_Flags = 0;
	jr->jr_Flags &= ~JIOF_CANCELLED;
	if( jr->jr_Flags & JIOF_READ )
		jioWatchFD(jr->jr_FD, JIO_WATCH_TYPE_READ);
	if( jr->jr_Flags & JIOF_WRITE )
		jioWatchFD(jr->jr_FD, JIO_WATCH_TYPE_WRITE);
	jioSetFDAttr(jr->jr_FD, JIO_FD_ATTR_ASYNC_ON);
	jmessagePut(&jrp->jrp_MessagePort, &jr->jr_Message);
}

void jioCancel(jioRequest *jr)
{
	require(jcpuDisabled());

	require(jr != 0);
	
	if( !(jr->jr_Message.jm_Flags & JMF_REPLY) )
	{
		/* Still on the system port. */
		lnRemove(&jr->jr_Message.jm_Link);
		jr->jr_Flags |= JIOF_CANCELLED;
		jmessageReply(&jr->jr_Message);
		jioSetFDAttr(jr->jr_FD, JIO_FD_ATTR_ASYNC_OFF);
	}
	/*
	 * XXX What to do about canceling the watch here?
	 *
	 * Nothing, it might be in use by someone else.
	 */
}

void jioHandleRequests(void)
{
	jioRequestPort *jrp;

	jrp = (jioRequestPort *)jio_state.jis_RequestPorts.lh_Head;
	while( jrp->jrp_Link.ln_Succ )
	{
		jioRequest *jr, *jr_next;

		require(jrp->jrp_Flags & JRPF_ACTIVE);
		
		jr = (jioRequest *)jrp->jrp_MessagePort.jmp_Messages.lh_Head;
		jr_next = (jioRequest *)jr->jr_Message.jm_Link.ln_Succ;
		while( jr_next )
		{
			int recv = 0;
			
			if( (jr->jr_Flags & JIOF_READ) &&
			    FD_ISSET(jr->jr_FD,
				     &jio_state.jis_ReadyFDs.read) )
			{
				recv = 1;
			}
			if( (jr->jr_Flags & JIOF_WRITE) &&
			    FD_ISSET(jr->jr_FD,
				     &jio_state.jis_ReadyFDs.write) )
			{
				recv = 1;
			}

			if( recv )
			{
				lnRemove(&jr->jr_Message.jm_Link);
				jmessageReply(&jr->jr_Message);
				jioSetFDAttr(jr->jr_FD, JIO_FD_ATTR_ASYNC_OFF);
			}
			jr = jr_next;
			jr_next = (jioRequest *)jr->jr_Message.jm_Link.ln_Succ;
		}
		jrp = (jioRequestPort *)jrp->jrp_Link.ln_Succ;
	}
	FD_ZERO(&jio_state.jis_ReadyFDs.read);
	FD_ZERO(&jio_state.jis_ReadyFDs.write);
}

int jioRegisterFD(int fd, jio_fd_type_t type)
{
	int retval = -1;

#if (defined(FIOSSAIOSTAT) && !(defined(hpux) && defined(FIOASYNC)))
	int on = 1;
#endif
	int rc;
	
#if (defined(FIOSSAIOOWN) && !(defined(hpux) && defined(F_SETOWN))) || \
    defined(F_SETOWN)
	static int pid = -1;
	
	if( pid == -1 )
		pid = getpid();
#endif
	
	require(jcpuDisabled());
	require((fd == -1) || (fd >= 0));
	require(!FD_ISSET(fd, &jio_state.jis_AllocatedFDs));
	require(type > JIO_FD_TYPE_MIN);
	require(type < JIO_FD_TYPE_MAX);

	if( fd == -1 )
	{
	}
	else
	{
#if defined(F_SETFD)
		/* set close-on-exec flag for this file descriptor */
		if ((rc = fcntl(fd, F_SETFD, 1)) < 0) {
			perror("F_SETFD");
			return( rc );
		}
#endif
		/* Make non-blocking */
		if ((rc = fcntl(fd, F_GETFL, 0)) < 0) {
			perror("F_GETFL");
			return( rc );
		}
		
		/*
		 * Apparently, this can fail, for instance when we stdout is 
		 * redirected to /dev/null. (On FreeBSD)
		 */
		fcntl(fd, F_SETFL, rc | O_NONBLOCK);
		
#if defined(FIOSSAIOSTAT) && !(defined(hpux) && defined(FIOASYNC))
		rc = ioctl(fd, FIOSSAIOSTAT, &on);
		if (rc < 0 && errno != ENOTTY) {
			/* Defines ENOTTY to be an acceptable error */
			perror("FIOSSAIOSTAT");
			return( rc );
		}
#endif
		
#if !(defined(O_ASYNC) || defined(FIOASYNC) ||  \
      defined(FASYNC) || defined(FIOSSAIOSTAT))
#error	Could not put socket in async mode
#endif
		/*
		 * Allow socket to signal this process when new data is
		 * available
		 */
#if defined(FIOSSAIOOWN) && !(defined(hpux) && defined(F_SETOWN))
		if( (type == JIO_FD_TYPE_STDIO) ||
		    (type == JIO_FD_TYPE_PIPE) ||
		    (type == JIO_FD_TYPE_SOCKET) )
		{
			rc = ioctl(fd, FIOSSAIOOWN, &pid);
#if 0
			if( (rc == -1) && (errno != ENOTTY) )
			{
				perror("Error doing FIOSSAIOWN");
			}
#endif
		}
		
#elif defined(F_SETOWN)
		if( (type == JIO_FD_TYPE_STDIO) ||
		    (type == JIO_FD_TYPE_PIPE) ||
		    (type == JIO_FD_TYPE_SOCKET) )
		{
			rc = fcntl(fd, F_SETOWN, pid);
			/*
			 * On some systems, this will flag an error if fd is
			 * not a socket.
			 */
#if 0
			if( rc < 0 )
			{
				perror("F_SETOWN");
			}
#endif
		}
#endif
		FD_SET(fd, &jio_state.jis_AllocatedFDs);
		if( fd > jio_state.jis_FDMax )
			jio_state.jis_FDMax = fd;

		retval = fd;
	}
	return( retval );
}

void jioUnregisterFD(int fd, jio_fd_type_t type)
{
	require(jcpuDisabled());
	
	if( fd != -1 )
	{
		int lpc;
		
		require(fd >= 0);
		require(FD_ISSET(fd, &jio_state.jis_AllocatedFDs));
		
		fcntl(fd, F_SETFL, fcntl(fd, F_GETFL, 0) & ~O_NONBLOCK);
		
		switch( type )
		{
		case JIO_FD_TYPE_STDIO:
			break;
		default:
			/* XXX do shutdown() here? */
			while( (close(fd) == -1) && (errno == EINTR) );
			break;
		}
		FD_CLR(fd, &jio_state.jis_AllocatedFDs);
		FD_CLR(fd, &jio_state.jis_PendingFDs.read);
		FD_CLR(fd, &jio_state.jis_PendingFDs.write);
		FD_CLR(fd, &jio_state.jis_ReadyFDs.read);
		FD_CLR(fd, &jio_state.jis_ReadyFDs.write);
		if( fd == jio_state.jis_FDMax )
		{
			for( lpc = fd - 1;
			     (lpc >= 0) &&
			     !FD_ISSET(lpc, &jio_state.jis_AllocatedFDs);
			     lpc-- )
			{
			}
			jio_state.jis_FDMax = lpc;
		}
		
		ensure(jio_state.jis_FDMax >= -1);
	}
}

void jioSetFDAttr(int fd, jio_fd_attr_t attr)
{
	jioSharedFD *jsfd = &jio_state.jis_FDs[fd];
	int val;

	require(fd >= 0);
	require(fd < FD_SETSIZE);
	require(FD_ISSET(fd, &jio_state.jis_AllocatedFDs));
	require(attr > JIO_FD_ATTR_MIN);
	require(attr < JIO_FD_ATTR_MAX);
	require(jcpuDisabled());
	
	switch( attr )
	{
	case JIO_FD_ATTR_ASYNC_ON:
		if( jsfd->jsfd_Waiters == 0 )
		{
			val = fcntl(fd, F_GETFL, 0);
			val |=
#if defined(O_ASYNC)
				O_ASYNC
#elif defined(FASYNC)
				FASYNC
#endif
				;
			fcntl(fd, F_SETFL, val);
		}
		jsfd->jsfd_Waiters += 1;
		break;
	case JIO_FD_ATTR_ASYNC_OFF:
		jsfd->jsfd_Waiters -= 1;
		if( jsfd->jsfd_Waiters == 0 )
		{
			jioWatchFD(fd, JIO_WATCH_TYPE_CANCEL_READ);
			jioWatchFD(fd, JIO_WATCH_TYPE_CANCEL_WRITE);
			val = fcntl(fd, F_GETFL, 0);
			val &= ~(
#if defined(O_ASYNC)
				 O_ASYNC
#elif defined(FASYNC)
				 FASYNC
#else
				 0
#endif
				 );
			fcntl(fd, F_SETFL, val);
		}
		break;
	default:
		require(0);
		break;
	}
}

void jioValidFDs(void)
{
	int lpc;

	for( lpc = 0; lpc <= jio_state.jis_FDMax; lpc++ )
	{
		if( FD_ISSET(lpc, &jio_state.jis_PendingFDs.read) ||
		    FD_ISSET(lpc, &jio_state.jis_PendingFDs.write) )
		{
			int rc;

			rc = fcntl(lpc, F_GETFL, 0);

			ensure(rc != -1);
#if defined(O_ASYNC)
			ensure(rc & O_ASYNC);
#elif defined(FASYNC)
			ensure(rc & FASYNC);
#endif
		}
	}
}

void jioQueryFDs(int block)
{
	static struct pollfd pollarray[FD_SETSIZE]; /* XXX Static? */
	static int again = 0;
	int rc, lpc, used;

	require(!again);
	require(block == 0);
	require(jio_state.jis_FDMax >= -1);
	
	again = 1;
	for( lpc = 0, used = 0; lpc <= jio_state.jis_FDMax; lpc++ )
	{
		short events = 0;

		if( FD_ISSET(lpc, &jio_state.jis_PendingFDs.read) )
			events |= POLLIN;
		if( FD_ISSET(lpc, &jio_state.jis_PendingFDs.write) )
			events |= POLLOUT;
		if( events != 0 )
		{
			pollarray[used].fd = lpc;
			pollarray[used].events = events;
			used++;
		}
	}
	do {
		rc = poll(pollarray, used, block ? -1 : 0);
	} while( (rc == -1) && (errno == EINTR) );
	if( rc > 0 )
	{
		for( lpc = 0; (lpc < used) && (rc > 0); lpc++ )
		{
			if( pollarray[lpc].revents )
			{
				rc -= 1;
			}
			if( pollarray[lpc].revents & POLLIN )
			{
				FD_CLR(pollarray[lpc].fd,
				       &jio_state.jis_PendingFDs.read);
				FD_SET(pollarray[lpc].fd,
				       &jio_state.jis_ReadyFDs.read);
			}
			if( pollarray[lpc].revents & POLLOUT )
			{
				FD_CLR(pollarray[lpc].fd,
				       &jio_state.jis_PendingFDs.write);
				FD_SET(pollarray[lpc].fd,
				       &jio_state.jis_ReadyFDs.write);
			}
		}
	}
	else if( rc < 0 )
	{
		switch( errno )
		{
		case EFAULT:
			fprintf(stderr,
				"jio_unix.c:Internal error:"
				"Bad file descriptor array pointer\n");
			ensure(0);
			break;
		case EINVAL:
			fprintf(stderr,
				"jio_unix.c:Internal error:"
				"Bad timeout value\n");
			ensure(0);
			break;
		default:
			fprintf(stderr,
				"jio_unix.c:Internal error:"
				"Unknown errno value: %d\n", errno);
			ensure(0);
			break;
		}
	}
	again = 0;
}

static
char *jioFDTypeStrings[] = {
	"",
	"file",
	"socket",
};

void jioFDPrint(FILE *file, jioFD *jfd)
{
	require(file != 0);
	require(jfd != 0);

	fprintf(file,
		"jioFD(%p:%d)[jfd_Type=%s]",
		jfd,
		jfd->jfd_FD - jio_state.jis_FDs,
		jioFDTypeStrings[jfd->jfd_FD->jsfd_Type]);
}

void jioRequestPrint(FILE *file, jioRequest *jr)
{
	fprintf(file,
		"jioRequest(%p)[jr_Message={ ",
		jr);
	jmessagePrint(file, &jr->jr_Message);
	fprintf(file,
		" }; "
		"jr_Flags=%lx; "
		"jr_FD=%d]",
		jr->jr_Flags,
		jr->jr_FD);
}

void jioRequestPortPrint(FILE *file, jioRequestPort *jrp)
{
	int lpc;

	require(file != 0);
	require(jrp != 0);
	
	fprintf(file,
		"jioRequestPort(%p)[jrp_Flags=%lx; "
		"jrp_AllocatedFDs={ ",
		jrp,
		jrp->jrp_Flags);
	for( lpc = 0; lpc < jrp->jrp_FDs.Max; lpc++ )
	{
		if( jrp->jrp_FDs.Data[lpc].jfd_FD )
		{
			jioFDPrint(file, &jrp->jrp_FDs.Data[lpc]);
		}
	}
	fprintf(file,
		"}; "
		"jrp_MessagePort={ ");
	jmessagePortPrint(file, &jrp->jrp_MessagePort);
	fprintf(file, " }]");
}

void jioStatePrint(FILE *file)
{
	jioRequestPort *jrp;
	int lpc;

	require(file != 0);
	
	fprintf(file, "jio_state[jis_RequestPorts={ ");
	jrp = (jioRequestPort *)jio_state.jis_RequestPorts.lh_Head;
	while( jrp->jrp_Link.ln_Succ )
	{
		jioRequestPortPrint(file, jrp);
		fprintf(file, " ");
		jrp = (jioRequestPort *)jrp->jrp_Link.ln_Succ;
	}
	fprintf(file,
		"}; jis_FDMax=%d; jis_AllocatedFDs={ ",
		jio_state.jis_FDMax);
	for( lpc = 0; lpc < jio_state.jis_FDMax; lpc++ )
	{
		if( FD_ISSET(lpc, &jio_state.jis_AllocatedFDs) )
		{
			fprintf(file, "%d ", lpc);
		}
	}
	fprintf(file, "}; jis_PendingFDs.read={ ");
	for( lpc = 0; lpc < jio_state.jis_FDMax; lpc++ )
	{
		if( FD_ISSET(lpc, &jio_state.jis_PendingFDs.read) )
		{
			fprintf(file, "%d ", lpc);
		}
	}
	fprintf(file, "}; jis_PendingFDs.write={ ");
	for( lpc = 0; lpc < jio_state.jis_FDMax; lpc++ )
	{
		if( FD_ISSET(lpc, &jio_state.jis_PendingFDs.write) )
		{
			fprintf(file, "%d ", lpc);
		}
	}
	fprintf(file, "}; jis_ReadyFDs.read={ ");
	for( lpc = 0; lpc < jio_state.jis_FDMax; lpc++ )
	{
		if( FD_ISSET(lpc, &jio_state.jis_ReadyFDs.read) )
		{
			fprintf(file, "%d ", lpc);
		}
	}
	fprintf(file, "}; jis_ReadyFDs.write={ ");
	for( lpc = 0; lpc < jio_state.jis_FDMax; lpc++ )
	{
		if( FD_ISSET(lpc, &jio_state.jis_ReadyFDs.write) )
		{
			fprintf(file, "%d ", lpc);
		}
	}
	fprintf(file, "}]");
}
