/*
 * jio_unix.h
 *
 * Copyright (c) 2002, 2003 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

#ifndef _jio_unix_h
#define _jio_unix_h

#include <stdio.h>
#include <limits.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>

#include "defs.h"

#include "jmessage.h"
#include "jioAccounting.h"

struct _jioRequestPort;

/*
 * Flags for jioRequest messages.
 *   JIOF_READ - A readability test.
 *   JIOF_WRITE - A writability test.
 *   JIOF_CANCELLED - The request was cancelled.
 */
enum {
	JIOB_READ,
	JIOB_WRITE,
	JIOB_CANCELLED,
};

enum {
	JIOF_READ = (1L << JIOB_READ),
	JIOF_WRITE = (1L << JIOB_WRITE),
	JIOF_CANCELLED = (1L << JIOB_CANCELLED),
};

/*
 * A jioRequest is used to make I/O requests on the system.
 *
 * jr_Message - jmessage header.
 * jr_Flags - Holds the JIOF_* flag values.
 * jr_FD - The file descriptor to use when processing the request.
 */
typedef struct _jioRequest {
	jmessage jr_Message;
	unsigned long jr_Flags;
	int jr_FD;
} jioRequest;

/*
 * Types of file descriptors:
 *   JIO_FD_TYPE_FILE - A regular file.
 *   JIO_FD_TYPE_SOCKET - A socket.
 */
typedef enum {
	JIO_FD_TYPE_MIN = -1,
	JIO_FD_TYPE_FREE,
	JIO_FD_TYPE_STDIO,
	JIO_FD_TYPE_FILE,
	JIO_FD_TYPE_PIPE,
	JIO_FD_TYPE_SOCKET,
	JIO_FD_TYPE_MAX
} jio_fd_type_t;

/*
 * A jioFD object is used to track file descriptors allocated by a particular
 * jsked.
 *
 * jfd_Next - Link to the next jioFD in the list.
 * jfd_Type - The type of file descriptor.
 */
typedef struct _jioSharedFD {
	jio_fd_type_t jsfd_Type;
	jioAccount jsfd_Account;
	unsigned int jsfd_Waiters;
	unsigned int jsfd_Users;
} jioSharedFD;

typedef struct _jioFD {
	jioSharedFD *jfd_FD;
} jioFD;

/*
 * Flags for jioRequestPorts:
 *   JRPF_ACTIVE - The port is active.
 */
enum {
	JRPB_INITED,
	JRPB_ACTIVE,
};

enum {
	JRPF_INITED = (1L << JRPB_INITED),
	JRPF_ACTIVE = (1L << JRPB_ACTIVE),
};

/*
 * Default number of active descriptors allowed for a port.
 */
#define JIO_DEFAULT_FD_MAX 128

/*
 * A jioRequestPort object is used to multiplex requests for an individual
 * jsked.
 *
 * jrp_Link - Link to the other ports in the global list.
 * jrp_Flags - Holds the JRPF_* flags.
 * jrp_AllocatedFDs - The list of jioFD's managed by the port.
 * jrp_FDCount - The number of open file descriptors.
 * jrp_FDMax - The maximum number of open file descriptors allowed.
 * jrp_FDPeak - The peak number of open file descriptors.
 * jrp_MessagePort - The port for receiving jioRequest messages.
 * jrp_RootDirectory - The root directory for this port.  Any files opened are
 *   relative to this point.  An empty string means no access.
 * jrp_RootDirectoryLen - The length of the string in jrp_RootDirectory.
 */
typedef struct _jioRequestPort {
	struct lnMinNode jrp_Link;
	unsigned long jrp_Flags;
	struct {
		jioFD *Data;
		unsigned int Count;
		unsigned int Peak;
		unsigned int Max;
	} jrp_FDs;
	jmessagePort jrp_MessagePort;
	char jrp_RootDirectory[MAXPATHLEN];
	unsigned int jrp_RootDirectoryLen;
} jioRequestPort;

/*
 * The global state.
 *
 * jis_RequestPorts - The list of active jioRequestPorts in the system.
 * jis_FDMax - The maximum file descriptor value in the system.
 * jis_FDs - Statically allocated jioFD objects.
 * jis_AllocatedFDs - Bitmask tracking the set of allocated FD's.
 * jis_PendingFDs.read - Bitmask tracking the set of FD's that are waiting
 *   for data.
 * jis_PendingFDs.write - Bitmask tracking the set of FD's that are waiting
 *   for output buffers to clear.
 * jis_ReadyFDs.read - Bitmask tracking the set of FD's that received the
 *   data they were waiting for.
 * jis_ReadyFDs.write - Bitmask tracking the set of FD's that are again able
 *   accept output.
 * jis_ResolvedPath[12] - Scratch memory that can be used with jioResolvePath.
 */
typedef struct _jioState {
	struct lnMinList jis_RequestPorts;
	int jis_FDMax;
	jioSharedFD jis_FDs[FD_SETSIZE];
	fd_set jis_AllocatedFDs;
	struct {
		fd_set read;
		fd_set write;
	} jis_PendingFDs;
	struct {
		fd_set read;
		fd_set write;
	} jis_ReadyFDs;
	char jis_ResolvedPath1[MAXPATHLEN];
	char jis_ResolvedPath2[MAXPATHLEN];
} jioState;

extern jioState jio_state;

enum {
	JIOA_TAG_DONE,
};

/*
 * Initialize the global jio state.
 */
int jioInit(int tag, ...);

/*
 * Initialize a jioRequestPort.
 */
void jioInitRequestPort(jioRequestPort *jrp);

/*
 * Add a jioRequestPort to the system.
 */
void jioAddRequestPort(jioRequestPort *jrp);

/*
 * Remove a jioRequestPort from the system.
 */
void jioRemRequestPort(jioRequestPort *jrp);

/*
 * Deallocate any memory attached to the given jioRequestPort.
 */
void jioKillRequestPort(jioRequestPort *jrp);

int jioSetFDMax(jioRequestPort *jrp, unsigned int max);

/*
 * Allocate a jioFD object and add it to the jioRequestPort.
 *
 * jrp - The jioRequestPort that will track the jioFD.
 * type - The type of file descriptor.
 * return - An initialized jioFD object.
 */
jioSharedFD *jioAllocSharedFD(int fd, jio_fd_type_t type);

/*
 * Free the jioFD object that is tracking the given file descriptor.
 *
 * jrp - The jioRequestPort that is tracking jioFD object.
 * fd - The file descriptor to free.
 */
void jioFreeSharedFD(int fd);

jioFD *jioAllocFD(jioRequestPort *jrp);

void jioFreeFD(jioRequestPort *jrp, int fd);

jioFD *jioFindFD(jioRequestPort *jrp, int fd);

void jioBindFD(jioRequestPort *jrp, jioFD *jfd, jioSharedFD *jsfd);

/*
 * Set the root directory for the given request port to the given path.
 *
 * jrp - The jioRequestPort whose root directory should be changed.
 * path - The path to the root directory.
 * returns - The zero for success, an errno value otherwise.
 */
int jioSetRootDirectory(jioRequestPort *jrp, char *path);

/*
 * Resolve a path relative to a port's root directory.
 *
 * jrp - The jioRequestPort to resolve the path against.
 * path - The path to resolve.
 * resolved_path - A preallocated buffer that will receive the resolved path.
 * resolved_size - The size of the 'resolved_path' buffer.  Must be atleast
 *   MAXPATHLEN bytes.
 */
char *jioResolvePath(jioRequestPort *jrp,
		     const char *path,
		     char *resolved_path,
		     size_t resolved_size);

/*
 * Send a jioRequest message to a jioRequestPort.
 *
 * jrp - The jioRequestPort that is handling requests for the file descriptor.
 * jr - The jioRequest to send.
 */
void jioSend(jioRequestPort *jrp, jioRequest *jr);

/*
 * Cancel a previously sent jioRequest.
 *
 * jr - The request to cancel.
 */
void jioCancel(jioRequest *jr);

/*
 * Process the queued jioRequest's using recently polled file descriptor
 * status.
 */
void jioHandleRequests(void);

/*
 * Register a file descriptor with the system.  Registration involves:
 *   Setting the file descriptor to non-blocking mode.
 *   Setting the close-on-exec flag.
 *   Updating the jio_state.jis_AllocatedFDs set.
 *   Updating the jio_state.jis_FDMax value.
 *
 * fd - The file descriptor to register.
 * type - The type of file descriptor.
 */
int jioRegisterFD(int fd, jio_fd_type_t type);

/*
 * Unregister a file descriptor.
 *
 * fd - The file descriptor to unregister.
 */
void jioUnregisterFD(int fd, jio_fd_type_t type);

/*
 * Set of file descriptor attributes that can be manipulated:
 *   JIO_FD_ATTR_ASYNC_ON - Turn on asynchronous signalling for descriptor.
 *   JIO_FD_ATTR_ASYNC_OFF - Turn off asynchronous signalling for descriptor.
 */
typedef enum {
	JIO_FD_ATTR_MIN,
	JIO_FD_ATTR_ASYNC_ON,
	JIO_FD_ATTR_ASYNC_OFF,
	JIO_FD_ATTR_MAX
} jio_fd_attr_t;

/*
 * Set a file descriptor attribute.
 *
 * fd - The file descriptor to manipulate.
 * attr - The attribute to change.
 */
void jioSetFDAttr(int fd, jio_fd_attr_t attr);

/*
 * Type of events to watch for on a file descriptor:
 *   JIO_WATCH_TYPE_READ - Watch for readability.
 *   JIO_WATCH_TYPE_WRITE - Watch for writability.
 *   JIO_WATCH_TYPE_CANCEL_READ - Cancel watching for readability.
 *   JIO_WATCH_TYPE_CANCEL_WRITE - Cancel watching for writability.
 */
typedef enum {
	JIO_WATCH_TYPE_MIN,
	JIO_WATCH_TYPE_CANCEL_READ,
	JIO_WATCH_TYPE_CANCEL_WRITE,
	JIO_WATCH_TYPE_READ,
	JIO_WATCH_TYPE_WRITE,
	JIO_WATCH_TYPE_MAX
} jio_watch_type_t;

/*
 * fd - The file descriptor to change the watch status for.
 * type - The type of events to watch/not watch for.
 */
void jioWatchFD(int fd, jio_watch_type_t type);

/*
 * Poll the current state of the watched file descriptors.
 *
 * block - Block waiting for a change in the watched file descriptor state.
 */
void jioQueryFDs(int block);

/*
 * Print the value of a jioFD object.
 *
 * file - The file to print to.
 * jfd - The jioFD object to print.
 */
void jioFDPrint(FILE *file, jioFD *jfd);

/*
 * Print the value of a jioRequest object.
 *
 * file - The file to print to.
 * jr - The jioRequest object to print.
 */
void jioRequestPrint(FILE *file, jioRequest *jr);

/*
 * Print the value of a jioRequestPort object.
 *
 * file - The file to print to.
 * jrp - The jioRequestPort object to print.
 */
void jioRequestPortPrint(FILE *file, jioRequestPort *jrp);

/*
 * Print the global jio state.
 *
 * file - The file to print to.
 */
void jioStatePrint(FILE *file);

#endif
