/*
 * jmesssage.h
 *
 * Copyright (c) 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

#ifndef _jmessage_h
#define _jmessage_h

#include <stdio.h>

#include "assert_pp.h"

#include "listNode.h"
#include "jsignalSet.h"
#include "jthreadState.h"

/*
 * A jmessagePort is used to receive messages sent by threads in the system.
 *
 * jmp_Messages - The list of messages sent to this port.
 * jmp_Thread - The thread to be signalled upon receipt of messages.
 * jmp_SignalNumber - The signal number to use when notifying the thread.
 */
typedef struct _jmessagePort {
	struct lnMinList jmp_Messages;
	jthreadState *jmp_Thread;
	int jmp_SignalNumber;
} jmessagePort;

/*
 * jmessage flags:
 *   JMF_REPLY - The message is a reply to an earlier request.
 */
enum {
	JMB_REPLY,
};

enum {
	JMF_REPLY = (1L << JMB_REPLY),
};

/*
 * A jmessage is a header for messages sent to a jmessagePort.
 *
 * jm_Link - Link to other nodes in the list.
 * jm_Flags - Holds one or more of the JMF_* flags.
 * jm_ReplyPort - The port where any replies are to be sent.
 */
typedef struct _jmessage {
	struct lnMinNode jm_Link;
	unsigned long jm_Flags;
	jmessagePort *jm_ReplyPort;
} jmessage;

/*
 * Print the value of the message to the file.
 *
 * file - The file to print to.
 * jm - The jmessage to print.
 */
void jmessagePrint(FILE *file, jmessage *jm);

/*
 * Print the value of the message port.
 *
 * file - The file to print to.
 * jmp - The jmessagePort to print.
 */
void jmessagePortPrint(FILE *file, jmessagePort *jmp);

/*
 * Initialize a jmessagePort with the given values.
 *
 * jmp - The jmessagePort object to initialize.
 * jts - The thread to notify upon receipt of a message.  Can be NULL.
 * signalNumber - The signal number to use when notifying a thread.
 */
static inline
void jmessagePortInit(jmessagePort *jmp, jthreadState *jts, int signalNumber)
{
	require(jmp != 0);
	require((jts == 0) ||
		((signalNumber >= MIN_JSIGNAL_NUMBER) &&
		 (signalNumber <= MAX_JSIGNAL_NUMBER)));

	lnNewList(&jmp->jmp_Messages);
	jmp->jmp_Thread = jts;
	jmp->jmp_SignalNumber = signalNumber;
}

/*
 * Put a message on a jmessagePort and notify the jmessagePort's thread, if
 * there is one.
 *
 * jmp - The destination port.
 * jm - The message to add to the port.
 */
static inline
void jmessagePut(jmessagePort *jmp, jmessage *jm)
{
	require(jmp != 0);
	require(jm != 0);
	
	lnAddTail(&jmp->jmp_Messages, &jm->jm_Link);
	if( jmp->jmp_Thread )
		jthreadStateSignal(jmp->jmp_Thread, jmp->jmp_SignalNumber);
}

/*
 * Reply to a previously sent message.  The message will have its JMF_REPLY
 * flag set and added to jm->jm_ReplyPort.
 *
 * jm - The request message with the reply data.
 */
static inline
void jmessageReply(jmessage *jm)
{
	require(jm != 0);
	require(jm->jm_ReplyPort != 0);
	
	jm->jm_Flags |= JMF_REPLY;
	jmessagePut(jm->jm_ReplyPort, jm);
}

/*
 * Remove and return a message from the given jmessagePort.  If there are no
 * more messages on the port a NULL will be returned.
 *
 * jmp - The jmessagePort to remove messages from.
 * returns - A jmessage or NULL if there are no more messages.
 */
static inline
jmessage *jmessageGet(jmessagePort *jmp)
{
	jmessage *retval;

	require(jmp != 0);
	
	if( (retval = (jmessage *)lnRemHead(&jmp->jmp_Messages)) )
	{
		retval->jm_Flags = 0;
	}
	return( retval );
}

/*
 * Wait for a jmessage to be received by a jmessagePort.  If messages are
 * already queued on the port, one will be immediately removed and returned.
 *
 * jmp - The jmessagePort to wait on.
 * returns - A jmessage received by the port.
 */
static inline
jmessage *jmessageWaitFor(jmessagePort *jmp)
{
	jmessage *retval;

	require(jmp != 0);
	require(jmp->jmp_Thread != 0);
	require(jmp->jmp_SignalNumber >= 0);

	while( !(retval = (jmessage *)lnRemHead(&jmp->jmp_Messages)) )
	{
		jthreadStateWait(jmp->jmp_Thread,
				 (1L << jmp->jmp_SignalNumber));

		/*
		 * Receiving the signal doesn't mean there is a message since
		 * multiple messages can be processed at once, but the signals
		 * for each message aren't.
		 *
		 *   recv msg -> set signal
		 *   waitFor msg -> clr signal
		 *   recv msg -> set signal
		 *   get msg
		 *   waitFor msg -> clr signal, but no message!
		 */
	}
	return( retval );
}

#endif
