/*
 * jsignalSet.h
 *
 * Copyright (c) 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

#ifndef _jsignal_set_h
#define _jsignal_set_h

#include "assert_pp.h"

/*
 * The type used to store signal bits.
 */
typedef unsigned long jsignal_set_t;

/*
 * The minimum and maximum possible signal numbers.
 */
#define MIN_JSIGNAL_NUMBER (0)
#define MAX_JSIGNAL_NUMBER ((sizeof(jsignal_set_t) * 8) - 1)

/*
 * A jsignalSet maintains the set of signals that a thread is interested.
 *
 * jss_AllocMask - The set of signals that have been allocated.
 * jss_WaitMask - The set of signals that a thread is waiting for.
 * jss_RecvMask - The set of signals that have been received by a thread.
 */
typedef struct _jsignalSet {
	jsignal_set_t jss_AllocMask;
	jsignal_set_t jss_WaitMask;
	jsignal_set_t jss_RecvMask;
} jsignalSet;

/*
 * Allocate a signal number.
 *
 * jss - The jsignalSet object to allocate a signal from.
 * signalNumber - The signalNumber to allocate or -1 if a specific signal is
 *   required.
 * returns - The allocated signal number or -1 if the signal couldn't be
 *   allocated.
 */
int jsignalAlloc(jsignalSet *jss, int signalNumber);

/*
 * Free a previously allocated signal number.
 *
 * jss - The jsignalSet object that the signal was allocated from.
 * signalNumber - The signal number to free.  A value of -1 will be ignored.
 */
void jsignalFree(jsignalSet *jss, int signalNumber);

/*
 * Print the value of the jsignalSet to the file.
 *
 * file - The file to print to.
 * jss - The jsignalSet object to print.
 */
void jsignalPrint(FILE *file, jsignalSet *jss);

/*
 * Initialize a jsignalSet object.
 *
 * jss - The jsignalSet object to initialize.
 */
static inline
void jsignalInit(jsignalSet *jss)
{
	require(jss != 0);
	
	jss->jss_AllocMask = 0;
	jss->jss_WaitMask = 0;
	jss->jss_RecvMask = 0;
}

/*
 * Send a signal to a jsignalSet object.
 *
 * jss - The destination jsignalSet object.
 * signalNumber - The signal number to set in the jsignalSet object.  NOTE:
 *   This is the bit number to set, _not_ the bitmask to set.
 * returns - True if another thread is waiting on the jsignalSet and needs to
 *   be resumed.
 */
static inline
int jsignalSend(jsignalSet *jss, int signalNumber)
{
	jsignal_set_t bitmask = (1L << signalNumber);
	int retval = 0;

	require(jss != 0);
	require(signalNumber >= MIN_JSIGNAL_NUMBER);
	require(signalNumber <= MAX_JSIGNAL_NUMBER);
	require(jss->jss_AllocMask & bitmask);

	jss->jss_RecvMask |= bitmask;
	if( jss->jss_WaitMask & bitmask )
	{
		retval = 1;
	}
	return( retval );
}

/*
 * Update the jsignalSet object with the set of signals that the current thread
 * should be notified of when they are sent.  In turn, it can also be used to
 * update the set of signals that the thread is not interested in.
 *
 * jss - The jsignalSet object to update.
 * careMask - The set of signals that the thread cares to be notified of.
 * dontCareMask - The set of signals that the thread is not interested in
 *   being notified of.
 */
static inline
void jsignalWatch(jsignalSet *jss,
		  jsignal_set_t careMask,
		  jsignal_set_t dontCareMask)
{
	require(jss != 0);
	require((jss->jss_AllocMask & careMask) == careMask);
	require((jss->jss_AllocMask & dontCareMask) == dontCareMask);
	
	jss->jss_WaitMask |= careMask;
	jss->jss_WaitMask &= ~dontCareMask;
}

/*
 * Consume any interesting signals received by a jsignalSet.  If any signals
 * were received that were not set in the jss_WaitMask they will remain set
 * in the jss_RecvMask.
 *
 * jss - The jsignalSet to consume signals from.
 * returns - The set of signals received.
 */
static inline
jsignal_set_t jsignalRecv(jsignalSet *jss)
{
	jsignal_set_t retval = 0;

	require(jss != 0);

	retval = jss->jss_RecvMask & jss->jss_WaitMask;
	jss->jss_RecvMask &= ~jss->jss_WaitMask;

	ensure((jss->jss_AllocMask & retval) == retval);
	
	return( retval );
}

#endif
