/*
 * jsked.c
 *
 * Copyright (c) 2002, 2003 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

#include <assert.h>

#include "jsked.h"

char *jskedIOAccountNames[JSKED_IOACCOUNT_MAX] = {
	"ipc",
	"disk",
	"network",
	"network_packet",
};

jskedState jsked_state;

int jskedInitState(void)
{
	int retval = 1;

	if( !(jsked_state.jss_Flags & JSSF_INITIALIZED) )
	{
		int lpc;

		jcpuTimeInit();
		jcpuAccountingInit(&jsked_state.jss_CPUAccount);
		jcpuAccountingStart(&jsked_state.jss_CPUAccount);
		for( lpc = 0; lpc < JSKED_IOACCOUNT_MAX; lpc++ )
		{
			jioAccountingInit(&jsked_state.jss_IOAccount[lpc]);
		}
		lnNewList(&jsked_state.jss_DisabledEntities);
		lnNewList(&jsked_state.jss_EnabledEntities);
		retval = jskedInitCBState();

		jsked_state.jss_Flags |= JSSF_INITIALIZED;
	}
	return( retval );
}

int jskedInit(jsked *js,
	      jthreadState *schedulerThread,
	      jthreadState *(*create)(struct _jsked *js,
				      char *spec,
				      size_t size,
				      int tag,
				      ...),
	      void (*delete)(struct _jsked *js, jthreadState *jts))
{
	int lpc, retval = 1;

	require(js != 0);
	require(schedulerThread != 0);
	require(create != 0);
	require(delete != 0);
	
	js->js_State = JSKED_STATE_INACTIVE;

	js->js_NoStopCount = 0;
	
	jsignalInit(&js->js_Signals);
	for( lpc = 0; lpc < JSSIGB_MAX; lpc++ )
	{
		int rc;

		rc = jsignalAlloc(&js->js_Signals, lpc);

		ensure(rc != -1);
	}

	lnNewList(&js->js_Threads);
	
	jmessagePortInit(&js->js_ExternalPort, 0, -1);

	jioInitRequestPort(&js->js_IORequestPort);

	jcpuAccountingInit(&js->js_CPUAccount);
	js->js_CPULimit = JCPU_ACCOUNT_INFINITE_LIMIT;
	for( lpc = 0; lpc < JSKED_IOACCOUNT_MAX; lpc++ )
	{
		jioAccountingInit(&js->js_IOAccount[lpc]);
	}
	
	schedulerThread->jts_Flags |= JTF_EXCLUSIVE;
	schedulerThread->jts_OwnerSked = js;
	schedulerThread->jts_ContextSked = js;
	schedulerThread->jts_CPUAccount.jca_Parent = &js->js_CPUAccount;

	js->js_Activations.jal_Unused = 0;
	
	jtimerListInit(&js->js_SubTimers);
	js->js_SubTimers.jtl_Parent = &jtimer_state;
	js->js_SubTimers.jtl_CurrentTimer.jt_UserData = js;
	
	js->js_ReceivedBlockEvents = 0;
	js->js_ReceivedResumeEvents = 0;

	js->js_PreemptEvent.jae_Type = JACTIVATION_EVENT_PREEMPT;
	
	js->js_SchedulerThread = schedulerThread;
	js->js_CreateThread = create;
	js->js_DeleteThread = delete;
	return( retval );
}

void jskedKill(jsked *js)
{
	require(js != 0);

	jioKillRequestPort(&js->js_IORequestPort);
}

/*
 * Initialize the first jsked in the system.
 *
 * js - The jsked object to initialize.
 * return - True if the initialization was successful.
 */
static int jskedFirst(jsked *js)
{
	jthreadState *mainThread;
	int retval = 0;
	
	jsked_state.jss_ProcessEntity = js;
	/* Create a thread object for the process context. */
	if( (mainThread = jskedMainThread(js)) )
	{
		jactivationEvent *jae;

		/* Tell the scheduler about the new thread. */
		jae = &mainThread->jts_ResumeEvent;
		jae->jae_Type = JACTIVATION_EVENT_SPAWN;
		jae->jae_Next = js->js_ReceivedResumeEvents;
		js->js_ReceivedResumeEvents = jae;
		jskedSignal(js, JSSIGB_RESUME);
		if( jcpuContextCapture(&mainThread->jts_Context) )
		{
			js->js_CurrentActivation->ja_State = JAS_RUNNING;
		}
		else
		{
			jskedSwap();
		}
		retval = 1;
	}
	return( retval );
}

int jskedAdd(jsked *js)
{
	int retval = 0;
	
	require(js != 0);
	require(js->js_State == JSKED_STATE_INACTIVE);
	
	jskedLockState();
	
	js->js_State = JSKED_STATE_READY;
	if( (retval = jskedReserve(js)) )
	{
		static int firstOne = 1;
		
		lnAddTail(&jsked_state.jss_EnabledEntities, &js->js_Link);
		jioAddRequestPort(&js->js_IORequestPort);
		if( firstOne )
		{
			/*
			 * This is the first jsked, need to start up a thread
			 * representing the processes "main" thread.
			 */
			firstOne = 0;
			if( !jskedFirst(js) )
			{
				retval = 0;
			}
		}
	}

	jskedUnlockState();
	
	return( retval );
}

void jskedRem(jsked *js)
{
	require(js != 0);
	require(js->js_State != JSKED_STATE_INACTIVE);

	jskedLockState();
	{
		jtimerRem(&jtimer_state, &js->js_SubTimers.jtl_CurrentTimer);
		jioRemRequestPort(&js->js_IORequestPort);
		lnRemove(&js->js_Link);
		if( jskedCurrent() == js )
		{
			js->js_State = JSKED_STATE_INACTIVE;
			jskedSwap();
		}
		else
		{
			jskedUnreserve(js);
		}
	}
	jskedUnlockState();
}

void jskedResumeSchedulerThread(jsked *js)
{
	jthreadState *jts = js->js_SchedulerThread;

	require(jcpuDisabled());
	require((js->js_State == JSKED_STATE_READY) ||
		(js->js_State == JSKED_STATE_RUNNING));
	require(js->js_CurrentActivation == 0);
	
	jsked_state.jss_CurrentEntity = js;
	js->js_State = JSKED_STATE_RUNNING;
	if( (js->js_Signals.jss_WaitMask == 0) &&
	    (js->js_Signals.jss_RecvMask & js->js_ShortSignalMask) &&
	    !(js->js_Signals.jss_RecvMask & ~js->js_ShortSignalMask) )
	{
		jsignal_set_t recvmask;
		
		jsignalWatch(&js->js_Signals,
			     js->js_ShortSignalMask,
			     0);
		{
			recvmask = jsignalRecv(&js->js_Signals);

			ensure(recvmask != 0);
		}
		jsignalWatch(&js->js_Signals,
			     0,
			     js->js_ShortSignalMask);
		jts = js->js_ShortFunction(js, js->js_ShortArgument, recvmask);
		if( jts == js->js_SchedulerThread )
		{
			js->js_Signals.jss_RecvMask |= recvmask;
		}
	}
	js->js_CurrentActivation = jactivationAlloc(&js->js_Activations, jts);
	jactivationResume(js->js_CurrentActivation);
}

void jskedSwap(void)
{
	jsked *js;
	
	require(jcpuDisabled());

	while( (js = jskedPickNext()) == 0 )
	{
		/* No jsked to run, block waiting for file(s)/timer(s). */
		jsked_state.jss_CurrentEntity = JSKED_KERNEL_ENTITY;
#if 0
		/*
		 * XXX Excessive and doesn't solve the race between
		 * jcpuEnable() and doing the poll/select.
		 */
		jioQueryFDs(1);
#elif 0
		/*
		 * XXX Same jcpuEnable() and doing the poll/select.
		 */
		jcpuEnable();
		sleep(0x0fffffff); /* XXX */
		jcpuDisable();
#elif 1
		/*
		 * XXX This seems to work, there isn't an interrupt race so
		 * we don't have to do the jthread v1 wouldlosewakeup thing.
		 * Also, we're not doing extra work with the poll.
		 */
		{
			static sigset_t emptySet;
			
			/*
			 * Block async's now, they'll be unblocked by
			 * sigsuspend
			 */
			jcpuSignalBlock(JCPU_ASYNC_SIGNALS);
			jsked_state.jss_Flags |= JSSF_IN_SWAP;
			jsked_state.jss_Flags &= ~JSSF_RETRY_SWAP;
			jcpuEnable();
			if( !(jsked_state.jss_Flags & JSSF_RETRY_SWAP) )
			{
				extern void jioValidFDs(void);
				
				jcpuAccountingEnd(&jsked_state.jss_CPUAccount);
				if( lnEmptyList(&jsked_state.jss_EnabledEntities) )
				{
					jskedExit(0);
				}
				jioValidFDs();
				sigsuspend(&emptySet);
				jcpuAccountingStart(&jsked_state.jss_CPUAccount);
			}
			else
			{
				/*
				 * The jcpuEnable() caused some signals to be
				 * processed so we need to immediately rerun
				 * pick next.
				 */
			}
			jcpuDisable();
			jsked_state.jss_Flags &= ~JSSF_IN_SWAP;
			jcpuSignalUnblock(JCPU_ASYNC_SIGNALS);
		}
#endif
	}

	if( js->js_CurrentActivation )
	{
		require(js->js_CurrentActivation->ja_Thread->jts_Flags &
			JTF_EXCLUSIVE);
		
		jsked_state.jss_CurrentEntity = js;
		/* The scheduler thread must already be running. */
		jactivationResume(js->js_CurrentActivation);
	}
	else
	{
		/* Restart the scheduler thread. */
		jskedResumeSchedulerThread(js);
	}
}

jsignal_set_t jskedWait(jsked *js, jsignal_set_t mask)
{
	unsigned long retval;

	require(js != 0);
	require(mask != 0);
	
	jcpuDisable();
	jsignalWatch(&js->js_Signals, mask, 0);
	if( !(retval = jsignalRecv(&js->js_Signals)) )
	{
		/* Haven't already received any of the signals. */
		js->js_State = JSKED_STATE_WAITING;
		if( jactivationStall() )
		{
			retval = jsignalRecv(&js->js_Signals);

			ensure(retval != 0);
		}
		else
		{
			require(js->js_Activations.jal_Unused != 0);
			
			jactivationFree(&js->js_Activations,
					js->js_CurrentActivation);
			js->js_CurrentActivation = 0;
			jskedSwap();

			ensure(0);
		}
	}
	jsignalWatch(&js->js_Signals, 0, mask);
	jcpuEnable();
	return( retval );
}

void jskedSignal(jsked *js, jsked_signal_t sigNumber)
{
	require(js != 0);
	require(sigNumber >= 0);
	require(sigNumber < JSSIGB_MAX);
	
	if( jsignalSend(&js->js_Signals, sigNumber) )
	{
		/* They were waiting on the signal, resume the jsked. */
		switch( js->js_State )
		{
		case JSKED_STATE_RUNNING:
			ensure(0);
			break;
		case JSKED_STATE_WAITING:
			js->js_State = JSKED_STATE_READY;
			jskedResume(js);
			break;
		case JSKED_STATE_READY:
			break;
		default:
			assert(0);
			break;
		}
	}
	switch( sigNumber )
	{
	case JSSIGB_PREEMPT:
		{
			jactivationEvent *jae;

			require(js->js_PreemptEvent.jae_Subject.activation ==
				0);
			require(!(js->js_CurrentActivation->ja_Thread->
				  jts_Flags & JTF_EXCLUSIVE));
			require(js->js_CurrentActivation->ja_State ==
				JAS_STALLED);

			jae = &js->js_PreemptEvent;
			jae->jae_Subject.activation = js->js_CurrentActivation;
			js->js_CurrentActivation = 0;
		}
		break;
	case JSSIGB_TIMER:
		jtimerReset(&js->js_SubTimers);
		break;
	default:
		break;
	}
}

static
void jskedBroadcastIO(void)
{
	jioHandleRequests();
}

static
void jskedBroadcastTimer(void)
{
	jtimer *jt;
	jsked *js;

	/*
	 * Remove all the received timers from the list and signal the
	 * jskeds.
	 */
	while( (jt = (jtimer *)lnRemHead(&jtimer_state.jtl_ReceivedTimers)) )
	{
		jt->jt_Flags = 0;
		js = jt->jt_UserData;
		jskedSignal(js, JSSIGB_TIMER);
	}
}

void jskedBroadcast(int sigNumber)
{
	switch( sigNumber )
	{
	case JSSIGB_TIMER:
		jskedBroadcastTimer();
		break;
	case JSSIGB_IO:
		jskedBroadcastIO();
		break;
	default:
		{
			jsked *js;
			
			js = (jsked *)jsked_state.jss_EnabledEntities.lh_Head;
			while( js->js_Link.ln_Succ )
			{
				jskedSignal(js, sigNumber);
				js = (jsked *)js->js_Link.ln_Succ;
			}
		}
		break;
	}
}

static
void jskedDoExternal(jsked *js, jsked_external_t type)
{
	jskedExternalMessage jem;
	jmessagePort jmp;
	
	require(jskedCurrent() != js);

	jmessagePortInit(&jmp, jactivationCurrentThread(), JTSIGB_ONCE);
	jem.jem_Message.jm_Link.ln_Pred = 0;
	jem.jem_Message.jm_Link.ln_Succ = 0;
	jem.jem_Message.jm_Flags = 0;
	jem.jem_Message.jm_ReplyPort = &jmp;
	jem.jem_Type = type;
	jskedPutExternal(js, &jem);
	jcpuDisable();
	{
		jmessageWaitFor(&jmp);
	}
	jcpuEnable();
}

void jskedDisable(jsked *js)
{
	jskedDoExternal(js, JSKED_EXTERNAL_DISABLE);
}

void jskedEnable(jsked *js)
{
	jskedDoExternal(js, JSKED_EXTERNAL_ENABLE);
}

static
char *jskedStateStrings[] = {
	"",
	"inactive",
	"disabled",
	"waiting",
	"ready",
	"running"
};

void jskedPrint(FILE *file, jsked *js)
{
	jactivationEvent *jae;
	int lpc;
	
	fprintf(file,
		"jsked(%p)[js_State=%s; "
		"js_Signals={",
		js,
		jskedStateStrings[js->js_State]);
	jsignalPrint(file, &js->js_Signals);
	fprintf(file, "}; js_CPUAccount={");
	jcpuAccountingPrint(file, &js->js_CPUAccount);
	fprintf(file, "}; js_IOAccount={ ");
	for( lpc = 0; lpc < JSKED_IOACCOUNT_MAX; lpc++ )
	{
		jioAccountingPrint(file, &js->js_IOAccount[lpc]);
		fprintf(file, " ");
	}
	fprintf(file, "}; js_Activations={");
	jactivationListPrint(file, &js->js_Activations);
	fprintf(file, "}; js_CurrentActivation={");
	jactivationPrint(file, js->js_CurrentActivation);
	if( js->js_State != JSKED_STATE_INACTIVE )
	{
		fprintf(file, "}; js_IORequestPort={");
		jioRequestPortPrint(file, &js->js_IORequestPort);
	}
	fprintf(file, "}; js_SubTimers={");
	jtimerListPrint(file, &js->js_SubTimers);
	fprintf(file, "}; js_ReceivedBlockEvents={ ");
	jae = js->js_ReceivedBlockEvents;
	while( jae )
	{
		jactivationEventPrint(file, jae);
		fprintf(file, " ");
		jae = jae->jae_Next;
	}
	fprintf(file, "}; js_ReceivedResumeEvents={ ");
	jae = js->js_ReceivedResumeEvents;
	while( jae )
	{
		jactivationEventPrint(file, jae);
		fprintf(file, " ");
		jae = jae->jae_Next;
	}
	fprintf(file, "}; js_PreemptEvent={");
	jactivationEventPrint(file, &js->js_PreemptEvent);
	fprintf(file, "}; js_SchedulerThread={");
	jthreadStatePrint(file, js->js_SchedulerThread);
	fprintf(file, "}]");
}

/** jcpu hooks **/

void jcpuAsyncEventHandler(jcpuEvent *jce)
{
	jsked *js = jskedCurrent();

	jcpuAccountingEnd(&jsked_state.jss_CPUAccount);
	jcpuAccountingStart(&jsked_state.jss_CPUAccount);
	if( (jsked_state.jss_CurrentEntity == JSKED_KERNEL_ENTITY) &&
	    lnEmptyList(&jsked_state.jss_EnabledEntities) )
	{
		/* Too early. */
	}
	else if( js &&
		 js->js_CurrentActivation &&
		 jactivationStall() )
	{
		/* Returned from stall, return to the jsked's thread. */
	}
	else
	{
		/* Process events. */
		while( jce )
		{
			switch( jce->jce_Type )
			{
			case JCPU_EVENT_PREEMPT:
				break;
			case JCPU_EVENT_RESET:
				jskedBroadcast(JSSIGB_RESET);
				break;
			case JCPU_EVENT_TIMER:
				jskedBroadcast(JSSIGB_TIMER);
				break;
			case JCPU_EVENT_IO:
				jskedBroadcast(JSSIGB_IO);
				break;
			case JCPU_EVENT_CHILD:
				jskedBroadcast(JSSIGB_CHILD);
				break;
			default:
				break;
			}
			jce = jce->jce_Next;
		}
		if( js &&
		    js->js_CurrentActivation &&
		    !(js->js_CurrentActivation->ja_Thread->jts_Flags &
		      JTF_EXCLUSIVE) )
		{
			/* Preempt the current thread. */
			jskedSignal(js, JSSIGB_PREEMPT);
		}
		if( jsked_state.jss_Flags & JSSF_IN_SWAP )
		{
			/*
			 * XXX We're in a jskedSwap, tell it to retry the
			 * jskedPickNext().
			 */
			jsked_state.jss_Flags |= JSSF_RETRY_SWAP;
		}
		else
		{
			jskedSwap();
		    
			ensure(0);
		}
	}
}

/** jthreadState hooks **/

unsigned long jthreadStateWait(jthreadState *jts, unsigned long waitmask)
{
	int disabled = jcpuDisabled();
	unsigned long retval;

	require(jts != 0);
	
	if( !disabled )
		jcpuDisable();
	{
		jsignalWatch(&jts->jts_Signals, waitmask, 0);
		if( !(retval = jsignalRecv(&jts->jts_Signals)) )
		{
			jcpuUpdateUsedTime();
			
			if( jactivationStall() )
			{
				retval = jsignalRecv(&jts->jts_Signals);
			}
			else
			{
				jactivationEvent jae;
				jactivation *ja;
				jsked *js;

				/* Tell the scheduler thread we've blocked. */
				js = jts->jts_OwnerSked;
				ja = js->js_CurrentActivation;
				js->js_CurrentActivation = 0;
				jts->jts_ResumeEvent.jae_Type =
					JACTIVATION_EVENT_UNUSED;
				jae.jae_Subject.activation = ja;
				jae.jae_Next = js->js_ReceivedBlockEvents;
				jae.jae_Type = JACTIVATION_EVENT_BLOCK;
				js->js_ReceivedBlockEvents = &jae;
				jskedSignal(js, JSSIGB_BLOCK);
				jskedResumeSchedulerThread(js);

				ensure(0);
			}
		}
		jsignalWatch(&jts->jts_Signals, 0, waitmask);
	}
	if( !disabled )
		jcpuEnable();
	return( retval );
}

void jthreadStateSignal(jthreadState *jts, int signalNumber)
{
	int disabled = jcpuDisabled();

	require(jts != 0);
	require(signalNumber >= 0);
	require(signalNumber < (sizeof(int) * 8));
	
	if( !disabled )
		jcpuDisable();
	{
		if( jsignalSend(&jts->jts_Signals, signalNumber) )
		{
			jactivationEvent *jae = &jts->jts_ResumeEvent;
			jsked *js;
			
			js = jts->jts_OwnerSked;
			if( jae->jae_Type == JACTIVATION_EVENT_UNUSED )
			{
				/*
				 * Tell the scheduler thread to resume the
				 * target thread.
				 */
				jae->jae_Subject.thread = jts;
				jae->jae_Next = js->js_ReceivedResumeEvents;
				jae->jae_Type = JACTIVATION_EVENT_RESUME;
				js->js_ReceivedResumeEvents = jae;
				jskedSignal(js, JSSIGB_RESUME);
			}
			else
			{
				/*
				 * The thread has already received a resume
				 * event, no need to send another.
				 *
				 * XXX This is ok right?
				 */
			}
		}
	}
	if( !disabled )
		jcpuEnable();
}
