/*
 * jsked.h
 *
 * Copyright (c) 2002, 2003 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

#ifndef _jsked_h
#define _jsked_h

#include <stdio.h>
#include <stdarg.h>

#include "listNode.h"

#include "jcpu.h"
#include "jio.h"
#include "jtimer_unix.h"
#include "jcpuAccounting.h"
#include "jioAccounting.h"
#include "jsignalSet.h"
#include "jthreadState.h"
#include "jactivation.h"
#include "jactivationEvent.h"

struct _jskedReservation;

/*
 * Preallocated jsignalSet signals:
 *   JSSIGB_EXTERNAL - An external message has been received.
 *   JSSIGB_IO - An I/O event. (Internal, it currently doesn't reach the
 *               scheduler thread.)
 *   JSSIGB_RESET - A termination/reset/interrupt event.
 *   JSSIGB_TIMER - A timer event. (Internal)
 *   JSSIGB_BLOCK - A block event from a thread.
 *   JSSIGB_RESUME - A resume event for a thread.
 *   JSSIGB_SKED_TRAP - A scheduler trap event.
 */
typedef enum {
	JSSIGB_EXTERNAL,
	JSSIGB_KERNEL_TRAP,
	JSSIGB_DEBUG_TRAP,
	JSSIGB_RESET,
	JSSIGB_IO,
	JSSIGB_CHILD,
	JSSIGB_TIMER,
	JSSIGB_BLOCK,
	JSSIGB_RESUME,
	JSSIGB_PREEMPT,
	JSSIGB_SKED_TRAP,
	JSSIGB_MAX,
} jsked_signal_t;

enum {
	JSSIG_EXTERNAL = (1L << JSSIGB_EXTERNAL),
	JSSIG_KERNEL_TRAP = (1L << JSSIGB_KERNEL_TRAP),
	JSSIG_DEBUG_TRAP = (1L << JSSIGB_DEBUG_TRAP),
	JSSIG_RESET = (1L << JSSIGB_RESET),
	JSSIG_IO = (1L << JSSIGB_IO),
	JSSIG_CHILD = (1L << JSSIGB_CHILD),
	JSSIG_TIMER = (1L << JSSIGB_TIMER),
	JSSIG_BLOCK = (1L << JSSIGB_BLOCK),
	JSSIG_RESUME = (1L << JSSIGB_RESUME),
	JSSIG_PREEMPT = (1L << JSSIGB_PREEMPT),
	JSSIG_SKED_TRAP = (1L << JSSIGB_SKED_TRAP),
};

typedef enum {
	JSKED_STATE_MIN,
	JSKED_STATE_INACTIVE,
	JSKED_STATE_DISABLED,
	JSKED_STATE_WAITING,
	JSKED_STATE_READY,
	JSKED_STATE_RUNNING,
	JSKED_STATE_MAX
} jsked_state_t;

/*
 * jskedExternalMessage types:
 *   JSKED_EXTERNAL_DISABLE - Disable the jsked.
 *   JSKED_EXTERNAL_ENABLE - Enable the jsked.
 *   JSKED_EXTERNAL_CHANGE_SPEC - Change a thread's scheduling specification.
 */
typedef enum {
	JSKED_EXTERNAL_MIN,
	JSKED_EXTERNAL_DISABLE,
	JSKED_EXTERNAL_ENABLE,
	JSKED_EXTERNAL_CHANGE_SPEC,
	JSKED_EXTERNAL_MAX,
} jsked_external_t;

/*
 * A jskedExternalMessage is used to control one jsked from another.
 *
 * jem_Message - The jmessage base.
 * jem_Type - The type of external message.
 */
typedef struct _jskedExternalMessage {
	jmessage jem_Message;
	jsked_external_t jem_Type;
} jskedExternalMessage;

/*
 * A jskedThreadMessage is used to manipulate a thread in another jsked.
 *
 * jtm_Message - The jskedExternalMessage header.
 * jtm_Target - The thread to be manipulated.
 * jtm_Payload.spec - A new scheduling specification for the thread.
 */
typedef struct _jskedThreadMessage {
	jskedExternalMessage jtm_Message;
	jthreadState *jtm_Target;
	union {
		char *spec;
	} jtm_Payload;
} jskedThreadMessage;

/*
 * I/O accounts tracked by a jsked.
 *
 * JSKED_IOACCOUNT_IPC - IPC stats (pipes).
 * JSKED_IOACCOUNT_DISK - File access stats.
 * JSKED_IOACCOUNT_NETWORK - Network access stats.
 * JSKED_IOACCOUNT_NETWORK_PACKET - Network datagram access stats.
 */
typedef enum {
	JSKED_IOACCOUNT_IPC,
	JSKED_IOACCOUNT_DISK,
	JSKED_IOACCOUNT_NETWORK,
	JSKED_IOACCOUNT_NETWORK_PACKET,
	JSKED_IOACCOUNT_MAX,
} jsked_ioaccount_t;

extern char *jskedIOAccountNames[JSKED_IOACCOUNT_MAX];

/*
 * A jsked object represents a simple virtual processor and provides a
 * separation between groups of threads.
 *
 * js_Link - The link into the global jsked list.
 * js_State - The current state of the jsked.
 * js_Signals - The jsked's signal set.
 * js_Account - The jsked's CPU account.
 * js_Threads - The list of threads controlled by this jsked.
 * js_Activations - The list of preallocated jactivation's.
 * js_CurrentActivation - The current activation executing within the jsked.
 * js_ExternalPort - A jmessagePort used to receive external messages.
 * js_IORequestPort - A jmessagePort used to receive I/O request messages.
 * js_SubTimers - A jtimerList used to track timer events for threads in the
 *   jsked.
 * js_ReceivedBlockEvents - A singly linked list of block jactivationEvents.
 * js_ReceivedResumeEvents - A singly linked list of resume jactivationEvents.
 * js_ReceivedTrapEvents - A singly linked list of trap jactivationEvents.
 * js_PreemptEvent - A preallocated preempt jactivationEvent.
 * js_SchedulerThread - The jsked's scheduler thread.
 * js_ShortSignalMask - Short-circuit signal mask, if the jsked receives one
 *   or more of these signals and no others, a scheduler function will be
 *   invoked instead of doing a context switch to the scheduler thread.
 * js_ShortArgument - The argument to the short scheduler function.
 * js_ShortFunction - The function to call when short circuiting a context
 *   switch.
 * js_CreateThread - The function used to allocate jthread's for this jsked.
 * js_DeleteThread - The function used to free jthread's for this jsked.
 */
typedef struct _jsked {
	struct lnMinNode js_Link;

	jsked_state_t js_State;
	
	unsigned long js_NoStopCount;
	
	jsignalSet js_Signals;
	jcpuAccount js_CPUAccount;
	jcpu_accum_t js_CPULimit;
	jioAccount js_IOAccount[JSKED_IOACCOUNT_MAX];
	
	struct lnMinList js_Threads;
	
	jactivationList js_Activations;
	jactivation *js_CurrentActivation;

	jmessagePort js_ExternalPort;

	jioRequestPort js_IORequestPort;

	jtimerList js_SubTimers;
	
	jactivationEvent *js_ReceivedBlockEvents;
	jactivationEvent *js_ReceivedResumeEvents;
	jactivationTrapEvent *js_ReceivedTrapEvents;

	jactivationEvent js_PreemptEvent;

	jthreadState *js_SchedulerThread;
	jsignal_set_t js_ShortSignalMask;
	void *js_ShortArgument;
	jthreadState *(*js_ShortFunction)(struct _jsked *js,
					  void *arg,
					  jsignal_set_t recvmask);

	jthreadState *(*js_CreateThread)(struct _jsked *js,
					 char *spec,
					 size_t size,
					 int tag,
					 ...);
	void (*js_DeleteThread)(struct _jsked *js, jthreadState *jts);

	void *js_UserData;
} jsked;

/*
 * The "kernel" jsked.
 */
#define JSKED_KERNEL_ENTITY ((jsked *)NULL)

/*
 * Flags values for jsked_state.jss_Flags.
 *   IN_SWAP - Indicates the process is in jskedSwap and any interrupt handlers
 *   should not call into jskedSwap.
 *   RETRY_SWAP - Indicates that the process, currently in a jskedSwap,
 *   received a signal and should try again to find a runnable jsked.
 */
enum {
	JSSB_INITIALIZED,
	JSSB_IN_SWAP,
	JSSB_RETRY_SWAP,
};

enum {
	JSSF_INITIALIZED = (1L << JSSB_INITIALIZED),
	JSSF_IN_SWAP = (1L << JSSB_IN_SWAP),
	JSSF_RETRY_SWAP = (1L << JSSB_RETRY_SWAP),
};

/*
 * The global jsked state.
 *
 * jss_CPUAccount - Global CPU account.
 * jss_IOAccount - Global IO account.
 * jss_ProcessEntity - The first jsked and the one primarily associated with
 *   the whole process.
 * jss_EnabledEntities - The list of jsked's in the system.
 * jss_CurrentEntity - The currently executing jsked.
 */
typedef struct _jskedState {
	unsigned long jss_Flags;
	jcpuAccount jss_CPUAccount;
	jioAccount jss_IOAccount[JSKED_IOACCOUNT_MAX];
	unsigned long jss_ReadyThreadCount;
	jsked *jss_ProcessEntity;
	struct lnMinList jss_DisabledEntities;
	struct lnMinList jss_EnabledEntities;
	jsked *jss_CurrentEntity;
} jskedState;

extern jskedState jsked_state;

/*
 * Initialize the global jsked state.
 */
int jskedInitState(void);

static inline
void jskedLockState(void)
{
	jcpuDisable();
}

static inline
void jskedUnlockState(void)
{
	jcpuEnable();
}

/*
 * Initialize a jsked object.
 *
 * js - The jsked object to initialize.
 * jts - The scheduler thread.
 * create - The function that should be used when creating threads for this
 *   jsked.
 * delete - The function that should be used to delete threads created for
 *   this jsked.
 */
int jskedInit(jsked *js,
	      jthreadState *jts,
	      jthreadState *(*create)(struct _jsked *js,
				      char *spec,
				      size_t size,
				      int tag,
				      ...),
	      void (*delete)(struct _jsked *js, jthreadState *jts));

/*
 * Deinitialize a jsked object.
 *
 * js - An initialized jsked object.
 */
void jskedKill(jsked *js);

/*
 * Add a jsked to the system.
 *
 * js - An initialized jsked object.
 */
int jskedAdd(jsked *js);

/*
 * Remvoe a jsked from the system.
 *
 * js - A previously added jsked object.
 */
void jskedRem(jsked *js);

/*
 * Resume the scheduler thread in the currently running jsked.
 *
 * js - A live jsked object.
 */
void jskedResumeSchedulerThread(jsked *js);

/*
 * Block the current jsked and pick a new one to run in its place or block
 * the process waiting for an external event.
 */
void jskedSwap(void);

/*
 * Block the current jsked waiting for the given set of signals.  This function
 * is used by the scheduler thread when it has no threads to schedule in its
 * place and needs to wait for outside input.
 *
 * js - The currently executing jsked.
 * mask - The bitmask of signals to wait for.
 * return - The bitmask of received signals.
 */
jsignal_set_t jskedWait(jsked *js, jsignal_set_t mask);

/*
 * Send a signal to the given jsked.
 *
 * js - The jsked to signal.
 * sigNumber - The signal number to send to the jsked.
 */
void jskedSignal(jsked *js, jsked_signal_t sigNumber);

/*
 * Broadcast a signal to all the jsked's in the system.
 *
 * sigNumber - The signal number to broadcast.
 */
void jskedBroadcast(int sigNumber);

/*
 * Print the value of the jsked object.
 *
 * file - The file to print to.
 * js - The jsked object to print.
 */
void jskedPrint(FILE *file, jsked *js);

/*
 * Put a jskedExternalMessage on a jsked's js_ExternalPort.
 *
 * js - The target jsked.
 * jem - The message to add to the port.
 */
static inline
void jskedPutExternal(jsked *js, jskedExternalMessage *jem)
{
	jcpuDisable();
	{
		jmessagePut(&js->js_ExternalPort, &jem->jem_Message);
		jskedSignal(js, JSSIGB_EXTERNAL);
	}
	jcpuEnable();
}

/*
 * Get a jskedExternalMessage from a jsked's js_ExternalPort.
 *
 * js - The target jsked.
 * return - A jskedExternalMessage or NULL if there are no more messages.
 */
static inline
jskedExternalMessage *jskedGetExternal(jsked *js)
{
	jskedExternalMessage *retval;
	
	jcpuDisable();
	{
		retval = (jskedExternalMessage *)
			jmessageGet(&js->js_ExternalPort);
	}
	jcpuEnable();
	return( retval );
}

/*
 * Send a JSKED_EXTERNAL_DISABLE message to the given jsked and wait for the
 * reply.
 *
 * js - The jsked to disable.
 */
void jskedDisable(jsked *js);

/*
 * Send a JSKED_EXTERNAL_ENABLE message to the given jsked and wait for the
 * reply.
 *
 * js - The jsked to enable.
 */
void jskedEnable(jsked *js);

/*
 * return - The jsked that is executing the current thread.
 */
static inline
jsked *jskedCurrent(void)
{
	return( jsked_state.jss_CurrentEntity );
}

/*
 * return - The jsked that should handle requests for resources in the current
 * thread.
 */
static inline
jsked *jskedContext(void)
{
	return( jactivationCurrentThread()->jts_ContextSked );
}

/*
 * Initialize the jsked implementation state.
 */
extern int jskedInitCBState(void);

/*
 * Callback used to create a new jsked object.
 *
 * jr - The reservation spec.
 * jts - The scheduler thread.
 * create - The function that should be used when creating threads for this
 *   jsked.
 * delete - The function that should be used to delete threads created for
 *   this jsked.
 */
extern jsked *jskedCreate(struct _jskedReservation *jr,
			  jthreadState *jts,
			  jthreadState *(*create)(struct _jsked *js,
						  char *spec,
						  size_t size,
						  int tag,
						  ...),
			  void (*delete)(struct _jsked *js,
					 jthreadState *jts));

/*
 * Callback used to delete a jsked.
 *
 * js - A jsked object created with jskedCreate.
 */
extern void jskedDelete(jsked *js);

extern int jskedReserve(jsked *js);
extern void jskedUnreserve(jsked *js);
extern void jskedGetReservation(jsked *js, struct _jskedReservation *jr);
extern int jskedChangeReservation(jsked *js, struct _jskedReservation *jr);
extern void jskedSuspend(jsked *js);
extern void jskedResume(jsked *js);

/**
 * Call back used to notify the system of an overdraw in the CPU account.
 */
extern void jskedCPUOverdraw(jsked *js, jcpu_accum_t amount);
extern void jskedIOOverdraw(jsked *js, jsked_ioaccount_t type);

/*
 * Callback used to pick the next jsked to run.
 *
 * return - The next jsked to run or NULL if there are no ready jskeds.
 */
extern jsked *jskedPickNext(void);

/*
 * Callback used to create and initialize the process thread.
 *
 * js - The process jsked.
 * return - A new jthreadState with atleast JTSA_ProcessThread set to one.
 */
extern jthreadState *jskedMainThread(jsked *js);

extern void jskedExit(int status);

#endif
