/*
 * jskedPriority.c
 *
 * Copyright (c) 2002, 2003 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

#include "jsked.h"

#if defined(JSKED_PRIORITY)

#include "jskedPriority.h"

#include "jthread_config.h"

#if !defined(ALLOC_TYPE_SKED)
#define ALLOC_TYPE_SKED -1
#endif

/* XXX Lame round-robin jsked scheduler */

typedef struct _jskedPriorityHeader {
	struct lnNode jsph_Link;
} jskedPriorityHeader;

typedef struct _jskedPriority {
	jskedPriorityHeader jsp_Header;
	jsked jsp_Sked;
	
	jactivation jsp_Activations[0];
} jskedPriority;

typedef struct _jskedPriorityState {
	struct lnList jsps_ReadyEntities;
	struct lnList jsps_WaitEntities;
} jskedPriorityState;

static
jskedPriorityState jsked_priority_state;

#define JSKED_TO_PRIORITY(js) \
	((jskedPriority *)(((jskedPriorityHeader *)(js)) - 1))

int jskedInitCBState(void)
{
	lnNewList((struct lnMinList *)&jsked_priority_state.
		  jsps_ReadyEntities);
	lnNewList((struct lnMinList *)&jsked_priority_state.jsps_WaitEntities);
	return( 1 );
}

jsked *jskedCreate(jskedPriorityReservation *jpr,
		   jthreadState *schedulerThread,
		   jthreadState *(*create)(struct _jsked *js,
					   char *spec,
					   size_t size,
					   int tag,
					   ...),
		   void (*delete)(struct _jsked *js, jthreadState *jts))
{
	jskedPriority *jsp;
	jsked *retval = 0;

	require(schedulerThread != 0);
	require(create != 0);
	require(delete != 0);

	if( (jsp = jcpuMalloc(sizeof(jskedPriority) +
			      (sizeof(jactivation) * 2),
			      ALLOC_TYPE_SKED)) )
	{
		if( jskedInit(&jsp->jsp_Sked,
			      schedulerThread,
			      create,
			      delete) )
		{
			jactivation *ja;

			if( jpr )
			{
				jsp->jsp_Header.jsph_Link.ln_Pri =
					jpr->jpr_Priority;
			}
			retval = &jsp->jsp_Sked;
			ja = &jsp->jsp_Activations[0];
			ja->ja_Next = &jsp->jsp_Activations[1];
			jsp->jsp_Activations[1].ja_Next = 0;
			jsp->jsp_Sked.js_Activations.jal_Unused = ja;
		}
		else
		{
			jcpuFree(jsp);
		}
	}
	return( retval );
}

void jskedDelete(jsked *js)
{
	if( js )
	{
		jskedPriority *jsp;

		jskedKill(js);
		jsp = JSKED_TO_PRIORITY(js);
		jcpuFree(jsp);
	}
}

int jskedReserve(jsked *js)
{
	jskedPriority *jsp;
	
	require(js != 0);

	jsp = JSKED_TO_PRIORITY(js);
	lnAddTail((struct lnMinList *)&jsked_priority_state.jsps_ReadyEntities,
		  (struct lnMinNode *)&jsp->jsp_Header.jsph_Link);
	return( 1 );
}

void jskedUnreserve(jsked *js)
{
	jskedPriority *jsp;

	require(js != 0);

	switch( js->js_State )
	{
	case JSKED_STATE_RUNNING:
		break;
	case JSKED_STATE_READY:
	case JSKED_STATE_WAITING:
		jsp = JSKED_TO_PRIORITY(js);
		lnRemove((struct lnMinNode *)&jsp->jsp_Header.jsph_Link);
		break;
	default:
		require(0);
		break;
	}
	js->js_State = JSKED_STATE_INACTIVE;
}

void jskedGetReservation(jsked *js, jskedPriorityReservation *jpr)
{
	jskedPriority *jsp;
	
	require(js != 0);
	require(jpr != 0);

	jsp = JSKED_TO_PRIORITY(js);
	jpr->jpr_Priority = jsp->jsp_Header.jsph_Link.ln_Pri;
}

int jskedChangeReservation(jsked *js, jskedPriorityReservation *jpr)
{
	require(js != 0);
	require(jpr != 0);
	
	return( 1 );
}

void jskedSuspend(jsked *js)
{
	jskedPriority *jsp;

	require(js != 0);

	jsp = JSKED_TO_PRIORITY(js);
	lnRemove((struct lnMinNode *)&jsp->jsp_Header.jsph_Link);
	lnAddTail((struct lnMinList *)&jsked_priority_state.jsps_WaitEntities,
		  (struct lnMinNode *)&jsp->jsp_Header.jsph_Link);
}

void jskedResume(jsked *js)
{
	jskedPriority *jsp;

	require(js != 0);

	jsp = JSKED_TO_PRIORITY(js);
	lnRemove((struct lnMinNode *)&jsp->jsp_Header.jsph_Link);
	lnAddTail((struct lnMinList *)&jsked_priority_state.jsps_ReadyEntities,
		  (struct lnMinNode *)&jsp->jsp_Header.jsph_Link);
}

jsked *jskedPickNext(void)
{
	jsked *js = jsked_state.jss_CurrentEntity;
	jskedPriority *jsp;
	jsked *retval = 0;
	
	if( js )
	{
		struct lnList *list = 0;

		jsp = JSKED_TO_PRIORITY(js);
		switch( js->js_State )
		{
		case JSKED_STATE_RUNNING:
			js->js_State = JSKED_STATE_READY;
			/* FALLTHROUGH */
		case JSKED_STATE_READY:
			list = &jsked_priority_state.jsps_ReadyEntities;
			break;
		case JSKED_STATE_WAITING:
			list = &jsked_priority_state.jsps_WaitEntities;
			break;
		case JSKED_STATE_INACTIVE:
			break;
		default:
			require(0);
			break;
		}
		if( list )
		{
			lnAddTail((struct lnMinList *)list,
				  (struct lnMinNode *)&jsp->jsp_Header.
				  jsph_Link);
		}
		if( !jcpuAccountingCheck(&js->js_CPUAccount, js->js_CPULimit) )
		{
			jcpu_accum_t amount;

			amount = js->js_CPUAccount.jca_TotalUsed;
			jcpuAccountingReset(&js->js_CPUAccount);
			jskedCPUOverdraw(js, amount);
		}
	}
	if( (jsp = (jskedPriority *)
	     lnRemHead((struct lnMinList *)&jsked_priority_state.
		       jsps_ReadyEntities)) )
	{
		retval = &jsp->jsp_Sked;
	}
	return( retval );
}

#endif
