/*
 * jskedStride.c
 *
 * Copyright (c) 2002, 2003 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

#include "config.h"
#include "config-std.h"

#include "listNode-inline.h"

#include "jsked.h"

#if defined(JSKED_STRIDE)

#include "jskedStride.h"

#include "jthread_config.h"

#if !defined(ALLOC_TYPE_SKED)
#define ALLOC_TYPE_SKED -1
#endif

/*
 * Timer. The resolution is 10ms.
 */
#define JSKED_HZ	100
#define JSKED_TICK	(1000/JSKED_HZ)

/*
 * Stride constants.
 */
#define STRIDE1		(1 << 20)
#define SCALE		1000000
#define QUANTUM		(JSKED_TICK * SCALE)

typedef struct _jskedStrideHeader {
	struct lnMinNode jssh_Link;
	int jssh_Tickets;
	int jssh_Stride;
	int jssh_Remain;
	long long jssh_Pass;
	/* XXX Use the jsked Accounting info! */
	long long jssh_Start;
} jskedStrideHeader;

typedef struct _jskedStride {
	jskedStrideHeader jss_Header;
	jsked jss_Sked;
	
	jactivation jss_Activations[0];
} jskedStride;

typedef struct _jskedStrideState {
	struct lnMinList jsss_ReadyEntities;
	struct lnMinList jsss_WaitEntities;
	int jsss_Tickets;
	int jsss_Stride;
	long long jsss_Pass;
	long long jsss_LastUpdate;
} jskedStrideState;

static
jskedStrideState jsked_stride_state;

#define JSKED_TO_STRIDE(js) \
	((jskedStride *)(((jskedStrideHeader *)(js)) - 1))

int jskedInitCBState(void)
{
	lnNewList(&jsked_stride_state.jsss_ReadyEntities);
	lnNewList(&jsked_stride_state.jsss_WaitEntities);
	return( 1 );
}

jsked *jskedCreate(jskedStrideReservation *jsr,
		   jthreadState *schedulerThread,
		   jthreadState *(*create)(struct _jsked *js,
					   char *spec,
					   size_t size,
					   int tag,
					   ...),
		   void (*delete)(struct _jsked *js, jthreadState *jts))
{
	jskedStride *jss;
	jsked *retval = 0;

	require(schedulerThread != 0);
	require(create != 0);
	require(delete != 0);

	if( (jss = jcpuMalloc(sizeof(jskedStride) + (sizeof(jactivation) * 2),
			      ALLOC_TYPE_SKED)) )
	{
		if( jskedInit(&jss->jss_Sked,
			      schedulerThread,
			      create,
			      delete) )
		{
			int tickets = 1000;
			jactivation *ja;
			
			if( jsr )
			{
				require(jsr->jsr_Tickets > 0);
				
				tickets = jsr->jsr_Tickets;
			}
			/* Initialize stride data. */
			jss->jss_Header.jssh_Tickets = tickets;
			jss->jss_Header.jssh_Stride = STRIDE1 / tickets;
			jss->jss_Header.jssh_Remain =
				jss->jss_Header.jssh_Stride;

			/* Initialize activations for the thread scheduler. */
			ja = &jss->jss_Activations[0];
			ja->ja_Next = &jss->jss_Activations[1];
			jss->jss_Activations[1].ja_Next = 0;
			jss->jss_Sked.js_Activations.jal_Unused = ja;
			
			retval = &jss->jss_Sked;
		}
		else
		{
			jcpuFree(jss);
		}
	}
	return( retval );
}

void jskedDelete(jsked *js)
{
	if( js )
	{
		jskedStride *jss;

		jskedKill(js);
		jss = JSKED_TO_STRIDE(js);
		jcpuFree(jss);
	}
}

static
void jskedStateUpdatePass(void)
{
	long long elapsed;

	elapsed = jcpuUsedTime() - jsked_stride_state.jsss_LastUpdate;
	jsked_stride_state.jsss_LastUpdate += elapsed;
	jsked_stride_state.jsss_Pass +=
		(((long long) jsked_stride_state.jsss_Stride * SCALE) /
		 QUANTUM) * elapsed;

	ensure(jsked_stride_state.jsss_Pass >= 0);
}

static
void jskedStateUpdateTickets(int delta)
{
	jsked_stride_state.jsss_Tickets += delta;

	require(jsked_stride_state.jsss_Tickets >= 0);
	
	if( jsked_stride_state.jsss_Tickets )
	{
		jsked_stride_state.jsss_Stride = STRIDE1 /
			jsked_stride_state.jsss_Tickets;
	}
	else
	{
		jsked_stride_state.jsss_Stride = STRIDE1;
	}
	
	ensure(jsked_stride_state.jsss_Tickets >= 0);
}

static
void jskedStateInsert(jskedStride *jss)
{
	struct lnMinNode *curr;
	jskedStride *curr_jss;
	int done = 0;

	require(jss != 0);

	if( jss->jss_Sked.js_State == JSKED_STATE_READY )
	{
		curr = jsked_stride_state.jsss_ReadyEntities.lh_Head;
		while( curr->ln_Succ && !done )
		{
			curr_jss = (jskedStride *)curr;
			if( jss->jss_Header.jssh_Pass <
			    curr_jss->jss_Header.jssh_Pass )
			{
				lnInsert(curr_jss->jss_Header.
					 jssh_Link.ln_Pred,
					 &jss->jss_Header.jssh_Link);
				done = 1;
			}
			curr = curr->ln_Succ;
		}
	}
	if( !done )
	{
		switch( jss->jss_Sked.js_State )
		{
		case JSKED_STATE_READY:
			lnAddTail(&jsked_stride_state.jsss_ReadyEntities,
				  &jss->jss_Header.jssh_Link);
			break;
		case JSKED_STATE_WAITING:
			lnAddTail(&jsked_stride_state.jsss_WaitEntities,
				  &jss->jss_Header.jssh_Link);
			break;
		default:
			assert(0);
			break;
		}
	}
}

static
void jskedStrideUpdatePass(jskedStride *jss)
{
	long long elapsed;
	int stride;

	require(jss != 0);
	
	stride = jss->jss_Header.jssh_Stride;
	elapsed = jcpuUsedTime() - jss->jss_Header.jssh_Start;
	jss->jss_Header.jssh_Pass += ((stride * SCALE) / QUANTUM) * elapsed;

	assert(jss->jss_Header.jssh_Pass >= 0);
}

int jskedReserve(jsked *js)
{
	jskedStride *jss;
	
	require(js != 0);

	jss = JSKED_TO_STRIDE(js);
	jskedStateUpdatePass();
	jss->jss_Header.jssh_Pass = jsked_stride_state.jsss_Pass +
		jss->jss_Header.jssh_Remain;
	jskedStateUpdateTickets(jss->jss_Header.jssh_Tickets);
	jskedStateInsert(jss);
	return( 1 );
}

void jskedUnreserve(jsked *js)
{
	jskedStride *jss;
	
	require(js != 0);
	
	jss = JSKED_TO_STRIDE(js);
	switch( js->js_State )
	{
	case JSKED_STATE_INACTIVE:
	case JSKED_STATE_RUNNING:
	case JSKED_STATE_READY:
		jskedStateUpdatePass();
		jss->jss_Header.jssh_Remain = jss->jss_Header.jssh_Pass -
			jsked_stride_state.jsss_Pass;
		jskedStateUpdateTickets(-jss->jss_Header.jssh_Tickets);
		/* FALLTHROUGH */
	case JSKED_STATE_WAITING:
		if( (js->js_State != JSKED_STATE_RUNNING) &&
		    (js->js_State != JSKED_STATE_INACTIVE) )
		{
			lnRemove(&jss->jss_Header.jssh_Link);
		}
		break;
	default:
		require(0);
		break;
	}
	js->js_State = JSKED_STATE_INACTIVE;
}

void jskedGetReservation(jsked *js, jskedStrideReservation *jsr)
{
	jskedStride *jss;
	
	require(js != 0);
	require(jsr != 0);

	jss = JSKED_TO_STRIDE(js);
	jsr->jsr_Tickets = jss->jss_Header.jssh_Tickets;

	ensure(jsr->jsr_Tickets > 0);
}

int jskedChangeReservation(jsked *js, jskedStrideReservation *jsr)
{
	int remain, stride;
	jskedStride *jss;
	int retval = 1;
	
	require(js != 0);
	require(jsr != 0);
	require(jsr->jsr_Tickets > 0);

	jcpuDisable();
	jss = JSKED_TO_STRIDE(js);
	switch( js->js_State )
	{
	case JSKED_STATE_READY:
	case JSKED_STATE_WAITING:
		jskedStateUpdatePass();
		jss->jss_Header.jssh_Remain =
			jss->jss_Header.jssh_Pass -
			jsked_stride_state.jsss_Pass;
		lnRemove(&jss->jss_Header.jssh_Link);
		/* FALLTHROUGH */
	case JSKED_STATE_RUNNING:
		jskedStateUpdateTickets(-jss->jss_Header.jssh_Tickets);
		break;
	default:
		assert(0);
		break;
	}
	stride = STRIDE1 / jsr->jsr_Tickets;
	remain = (jss->jss_Header.jssh_Remain * stride) /
		jss->jss_Header.jssh_Stride;

	jss->jss_Header.jssh_Tickets = jsr->jsr_Tickets;
	jss->jss_Header.jssh_Stride = stride;
	jss->jss_Header.jssh_Remain = remain;

	switch( js->js_State )
	{
	case JSKED_STATE_READY:
	case JSKED_STATE_WAITING:
		jskedReserve(js);
		break;
	case JSKED_STATE_RUNNING:
		jskedStateUpdateTickets(jss->jss_Header.jssh_Tickets);
		break;
	default:
		assert(0);
		break;
	}
	jcpuEnable();
	return( retval );
}

void jskedSuspend(jsked *js)
{
	jskedStride *jss;

	require(js != 0);

	jss = JSKED_TO_STRIDE(js);
	lnRemove(&jss->jss_Header.jssh_Link);
	lnAddTail(&jsked_stride_state.jsss_WaitEntities,
		  &jss->jss_Header.jssh_Link);
}

void jskedResume(jsked *js)
{
	jskedStride *jss;

	require(js != 0);

	jss = JSKED_TO_STRIDE(js);
	lnRemove(&jss->jss_Header.jssh_Link);
	jskedReserve(&jss->jss_Sked);
}

jsked *jskedPickNext(void)
{
	jsked *js = jsked_state.jss_CurrentEntity;
	jskedStride *jss;
	jsked *retval = 0;
	
	if( js )
	{
		jss = JSKED_TO_STRIDE(js);
		jskedStrideUpdatePass(jss);
		switch( js->js_State )
		{
		case JSKED_STATE_RUNNING:
			js->js_State = JSKED_STATE_READY;
			/* FALLTHROUGH */
		case JSKED_STATE_READY:
			jskedStateInsert(jss);
			break;
		case JSKED_STATE_INACTIVE:
			jskedUnreserve(js);
			break;
		case JSKED_STATE_WAITING:
			jskedStateUpdatePass();
			jss->jss_Header.jssh_Remain =
				jss->jss_Header.jssh_Pass -
				jsked_stride_state.jsss_Pass;
			jskedStateUpdateTickets(-jss->jss_Header.jssh_Tickets);
			lnAddTail(&jsked_stride_state.jsss_WaitEntities,
				  &jss->jss_Header.jssh_Link);
			break;
		default:
			require(0);
			break;
		}
		if( !jcpuAccountingCheck(&js->js_CPUAccount, js->js_CPULimit) )
		{
			jcpu_accum_t amount;

			amount = js->js_CPUAccount.jca_TotalUsed;
			jcpuAccountingReset(&js->js_CPUAccount);
			jskedCPUOverdraw(js, amount);
		}
	}
	if( (jss = (jskedStride *)
	     lnRemHead(&jsked_stride_state.jsss_ReadyEntities)) )
	{
		jss->jss_Header.jssh_Start = jcpuUsedTime();
		retval = &jss->jss_Sked;
	}

	ensure((retval == 0) ||
	       (retval->js_State == JSKED_STATE_READY) ||
	       (retval->js_State == JSKED_STATE_RUNNING));
	return( retval );
}

#endif /* defined(JSKED_STRIDE) */
