/*
 * jthread.c
 *
 * Copyright (c) 2002, 2003 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

#include "config.h"

#if defined(HAVE_MMAP)
#include <sys/types.h>
#include <sys/mman.h>
#endif

#include "gc.h"

#include "jmutex.h"
#include "jthreadScheduler.h"

#include <janosvm/sparseArray.h>

#include "jthread.h"

jthread metalJThreadState = {
	METAL_THREAD_STATE_DATA,	/* jt_State */
	JTHREAD_STATUS_ALIVE,		/* jt_Status */
	{ { 0 } },			/* jt_IOReplyPort */
	{ { { 0 } } },			/* jt_IORequest */
	{ { 0 } },			/* jt_TimerReplyPort */
	{ { { 0 } } },			/* jt_TimerRequest */
	0,				/* jt_ExceptionStackTop */
	0,				/* jt_ExceptionFrameTop */
	METAL_THREAD_DATA,		/* jt_ThreadData */
	0,				/* jt_Next */
};

static struct {
	void (*jts_Destructor)(void *);
	void (*jts_OnStop)(void);
	void (*jts_OnDeadlock)(void);

	void (*jts_AtExit)(void);
	
	exchandler_t jts_NullHandler;
	exchandler_t jts_FloatingHandler;
	exchandler_t jts_StackOverflowHandler;
	exchandler_t jts_AbortHandler;
	
	exceptionFrame jts_ExceptionFrame;

	sparseArray *jts_Threads;
	
	jsthreadConfig jts_RootSchedulerConfig;
} jthread_state;

#if defined(INTERPRETER)
#define	DEFINEFRAME()		/* Does nothing */
#define	DEFINEFRAME_COPY(x)	/* Does nothing */
#define	EXCEPTIONPROTO		int sig
#define	EXCEPTIONFRAME(f, c)	/* Does nothing */
#define	EXCEPTIONFRAMEPTR	0
#elif defined(TRANSLATOR)
#define	DEFINEFRAME()		exceptionFrame frame
#define	DEFINEFRAME_COPY(x)	exceptionFrame frame = (x)
#define	EXCEPTIONFRAMEPTR	&frame
#endif /* TRANSLATOR */

#define IS_NULL_REFERENCE(x) \
	((((x) >= ((void *)0))) && ((x) < ((void *)getpagesize())))

static
void jthreadAsyncSignalHandler(jthreadState *jts, jsignal_set_t recvmask)
{
	if( recvmask & (1L << SIGKILL) )
	{
		/* This thread was cancelled. */
	}
}

void jcpuContextRootFrame(void)
{
	jthreadState *jts;
	jactivation *ja;

	require(jcpuDisabled());

	ja = jactivationCurrent();
	/*
	 * We need to set JAS_RUNNING here since we're coming out of the
	 * jcpuContextInit and _not_ a jactivationStall.
	 */
	ja->ja_State = JAS_RUNNING;
	jts = ja->ja_Thread;
	
	if( jthreadStackInRange(&jts->jts_ExceptionStack, &jts) )
	{
		/* Exception root frame. */
		
		/* Copy the frame from the signal handler. */
		DEFINEFRAME_COPY(jthread_state.jts_ExceptionFrame);

		jcpuSignalUnblock(JCPU_ALL_SIGNALS);

		jcpuEnable();

		/* Call the exception function (null/stackOverflow). */
		jts->jts_Function(EXCEPTIONFRAMEPTR);
	}
	else
	{
		/* Thread root frame. */
		
		require(jts->jts_Context.jcc_Flags & JCF_ROOT_FRAME);
		
		jts->jts_Context.jcc_Flags &= ~JCF_ROOT_FRAME;
		
		/* Jthread's expect to startup unstoppable. */
		jthread_disable_stop();
		
		/*
		 * Same thing here, as with above, the typical stall/resume
		 * context would be jcpuDisabled() so we enable here.
		 */
		jcpuEnable();

		/* Finally, call the thread's function... */
		jts->jts_ReturnValue = jts->jts_Function(jts->jts_Argument);
	}

	jthread_exit();

	ensure(0);
}

void *jcpuMalloc(size_t size, int type)
{
	void *retval = 0;
	
	switch( type )
	{
		/*
		 * If we have mmap(), use it to allocate stacks and their
		 * guard pages.
		 */
#if defined(HAVE_MMAP)
	case ALLOC_TYPE_STACK:
		/* Align the given size to a page boundary. */
		size = (size + (getpagesize() - 1)) & -getpagesize();
#if defined(MAP_STACK)
		/* Try FreeBSD's MAP_STACK */
		if( (retval = gc_mmap(0,
				      size + getpagesize(),
				      PROT_READ | PROT_WRITE,
				      MAP_STACK,
				      -1,
				      0,
				      ALLOC_TYPE_STACK)) != MAP_FAILED )
		{
			void *bottom, *top;

			bottom = retval;
			top = (void *)(((char *)bottom) - size);
			if( mprotect(((char *)top) - getpagesize(),
				     getpagesize(),
				     0) == -1 )
			{
				perror("Guard page mprotect");
				
				ensure(0);
			}
			retval = top;
			
			ensure(bottom > top);
		}
		else
		{
			/* No joy, must not be supported by the kernel... */
		}
#elif defined(MAP_NORESERVE)
		if( (retval = gc_mmap(0,
				      size + getpagesize(),
				      PROT_READ | PROT_WRITE,
				      MAP_ANONYMOUS | MAP_NORESERVE,
				      -1,
				      0,
				      ALLOC_TYPE_STACK)) != MAP_FAILED )
		{
			void *bottom, *top;

			retval = ((char *)retval) + getpagesize();
			top = retval;
			bottom = (void *)(((char *)top) + size);
			if( mprotect(((char *)top) - getpagesize(),
				     getpagesize(),
				     0) == -1 )
			{
				perror("Guard page mprotect");
				
				ensure(0);
			}
			
			ensure(bottom > top);
		}
		else
		{
			/* No joy, must not be supported by the kernel... */
		}
#endif
		/* If nothing's worked yet, try the basic config below. */
		if( retval != MAP_FAILED )
		{
		}
		else if( (retval = gc_mmap(0,
					   size + getpagesize(),
					   PROT_READ | PROT_WRITE,
					   MAP_ANON | MAP_PRIVATE,
					   -1,
					   0,
					   ALLOC_TYPE_STACK)) != MAP_FAILED )
		{
			if( mprotect(retval, getpagesize(), 0) == -1 )
			{
				perror("Guard page mprotect");
				
				ensure(0);
			}
			retval = ((char *)retval) + getpagesize();
		}
		else
		{
			perror("Couldn't mmap stack");
			
			retval = 0;
		}
		break;
#endif
	default:
		retval = gc_malloc(size, type);
		break;
	}
	return( retval );
}

void *jcpuRealloc(void *old, size_t size, int type)
{
	return gc_realloc(old, size, type);
}

void jcpuFree(void *mem)
{
	if( mem )
	{
		/* XXX Would like to just use gc_free() here. */
		gc_free_in_heap(mem);
	}
}

void jcpuFreeExplicit(void *mem, size_t size, int type)
{
	switch( type )
	{
#if defined(HAVE_MMAP)
	case ALLOC_TYPE_STACK:
		gc_munmap(((char *)mem) - getpagesize(),
			  size + getpagesize());
		break;
#endif
	default:
		/* XXX Would like to just use gc_free() here. */
		gc_free_in_heap(mem);
		break;
	}
}

void jthread_init(int preemptive,
		  int maxpr,
		  int minpr,
		  
		  /* XXX Remove below */
		  void *(*_allocator)(size_t),
		  void (*_deallocator)(void *),
		  /* XXX Remove above */
		  
		  void (*_destructor)(void *),
		  void (*_onstop)(void),
		  void (*_ondeadlock)(void))
{
	if( jskedInitState() &&
	    jioInit(JIOA_TAG_DONE) &&
	    jcpuInit(JCPUA_TAG_DONE) &&
	    (jthread_state.jts_Threads =
	     createSparseArray(SAA_Flags, SAF_NO_DELAY,
			       SAA_TAG_DONE)) )
	{
		jthreadState *jts;
		jsked *js;

		jthread_state.jts_Destructor = _destructor;
		jthread_state.jts_OnStop = _onstop;
		jthread_state.jts_OnDeadlock = _ondeadlock;
		jts = jsthreadSchedulerRoot(&jthread_state.
					    jts_RootSchedulerConfig);
		if( (js = jskedCreate(&jskedKernelReservation,
				      jts,
				      jsthreadSchedulerCreateThread,
				      jsthreadSchedulerDeleteThread)) )
		{
			int lpc;

			jioSetFDMax(&js->js_IORequestPort,
				    JIO_DEFAULT_FD_MAX);
			jskedAdd(js);

			jcpuDisable();
			for( lpc = 0; lpc < 3; lpc++ )
			{
				jioSharedFD *jsfd;
				jioFD *jfd;
				
				jsfd = jioAllocSharedFD(lpc,
							JIO_FD_TYPE_STDIO);
				jfd = jioAllocFD(&js->js_IORequestPort);
				jioBindFD(&js->js_IORequestPort, jfd, jsfd);
			}
			jcpuEnable();
		}
		else
		{
			require(0);
		}
	}
	else
	{
		require(0);
	}
}

int jthreadInit(jthread_t jt, jthread_type_t type)
{
	int retval = 1;
	int lpc;

	require(type > JTHREAD_TYPE_MIN);
	require(type < JTHREAD_TYPE_MAX);
	
	jt->jt_Status = JTHREAD_STATUS_LIMBO;
	for( lpc = JTSIGB_MAX; lpc < JTSIGB_MAX2; lpc++ )
	{
		int rc;
		
		rc = jsignalAlloc(&jt->jt_State.jts_Signals, lpc);
		
		ensure(rc != -1);
	}
	jmessagePortInit(&jt->jt_IOReplyPort, &jt->jt_State, JTSIGB_IO);
	jt->jt_IORequest.jr_Message.jm_ReplyPort = &jt->jt_IOReplyPort;
	jmessagePortInit(&jt->jt_TimerReplyPort,
			 &jt->jt_State,
			 JTSIGB_TIMER);
	jt->jt_TimerRequest.jt_Message.jm_ReplyPort = &jt->jt_TimerReplyPort;

	//jmutex_initialise(&jt->jt_JoinLock);
	//jcondvar_initialise(&jt->jt_JoinCond);

	switch( type )
	{
	case JTHREAD_TYPE_ROOT:
		initThreadData(&jt->jt_ThreadData, jt);
		break;
	case JTHREAD_TYPE_SCHEDULER:
		/* XXX Should we do something for scheduler threads? */
		break;
	case JTHREAD_TYPE_NORMAL:
		retval = inheritThreadData(&jt->jt_ThreadData,
					   &jthread_current()->jt_ThreadData,
					   jt);
		break;
	default:
		assert(0);
		break;
	}
	return( retval );
}

void jthreadDisable(jthread_t jt)
{
	if( jt->jt_ThreadData.freeLock )
	{
		/* XXX Big hack! */
		jt->jt_ThreadData.freeLock->link.owner =
			jt->jt_ThreadData.ownerTeam;
	}
}

static
void jthreadSyncHandler(EXCEPTIONPROTO)
{
	jthread_t jt;
	void *new_sp;
	
	DEFINEFRAME();

	if( jcpuDisabled() )
	{
		/* XXX do something more here */
		exit(-1);
	}

	jt = jthread_current();
	/* XXX ctx is defined by EXCEPTIONPROTO */
	EXCEPTIONFRAME(frame, ctx);
	switch( sig )
	{
	case SIGSEGV:
	case SIGBUS:
		/* Null/StackOverflow */

		/* Save the main thread's top for the gc. */
		jt->jt_ExceptionStackTop = (void *)EXCEPTIONSP(ctx);
		jt->jt_ExceptionFrameTop = (void *)EXCEPTIONBP(ctx);
		/* Copy the exception frame data. */
		jthread_state.jts_ExceptionFrame = *EXCEPTIONFRAMEPTR;
		/* Figure out the exception type and set the function. */
		if( IS_NULL_REFERENCE(si->si_addr) )
		{
			jt->jt_State.jts_Function =
				(void *)jthread_state.jts_NullHandler;
		}
		else
		{
			jt->jt_State.jts_Function =
				(void *)jthread_state.jts_StackOverflowHandler;
		}
		jcpuDisable();

		/* Setup the signal return context. */
		new_sp = jt->jt_State.jts_ExceptionStack.jts_Bottom
#if defined(STACK_GROWS_UP)
			+
#else
			-
#endif
			32;
		EXCEPTION_ALT_FRAME(ctx, jcpuContextRootFrame, new_sp);
		jt->jt_State.jts_Context.jcc_StackTop =
			(void *)EXCEPTIONSP(ctx);
		/* Return to the exception handling stack. */
		break;
	case SIGABRT:
		jthread_state.jts_AbortHandler(EXCEPTIONFRAMEPTR);
		break;
	case SIGFPE:
		jcpuSignalUnblock(JCPU_SYNC_SIGNALS);
		
		jthread_state.jts_FloatingHandler(EXCEPTIONFRAMEPTR);
		break;
	default:
		break;
	}
}

void jthread_initexceptions(exchandler_t nullHandler,
			    exchandler_t floatingHandler,
			    exchandler_t stackOverflowHandler,
			    exchandler_t abortHandler)
{
	int lpc;
	
	jthread_state.jts_NullHandler = nullHandler;
	jthread_state.jts_FloatingHandler = floatingHandler;
	jthread_state.jts_StackOverflowHandler = stackOverflowHandler;
	jthread_state.jts_AbortHandler = abortHandler;

	for( lpc = 0; lpc < (sizeof(jcpuSignalSync) / sizeof(int)); lpc++ )
	{
		jcpuSignalRegister(jcpuSignalSync[lpc], jthreadSyncHandler);
	}
}

void jthread_atexit(void (*f)(void))
{
	jthread_state.jts_AtExit = f;
}

void jthread_walkLiveThreads(void *arg,
			     void (*func)(void *arg,
					  jthread_t jtid,
					  void *jlThread))
{
	struct lnMinNode *curr, *succ;
	jthreadState *jts;
	jsked *js;
	
	require(jcpuDisabled());

	js = (jsked *)jsked_state.jss_EnabledEntities.lh_Head;
	while( js->js_Link.ln_Succ )
	{
		curr = js->js_Threads.lh_Head;
		while( curr->ln_Succ )
		{
			succ = curr->ln_Succ;
			jts = (jthreadState *)curr;
			func(arg, (jthread_t)jts, jts->jts_UserData);
			curr = succ;
		}
		js = (jsked *)js->js_Link.ln_Succ;
	}
}

void jthread_walk_jthreads(void *arg, void (*func)(void *arg, jthread_t jtid))
{
	struct lnMinNode *curr, *succ;
	jthreadState *jts;
	jsked *js;
	
	require(jcpuDisabled());

	js = (jsked *)jsked_state.jss_EnabledEntities.lh_Head;
	while( js->js_Link.ln_Succ )
	{
		curr = js->js_Threads.lh_Head;
		while( curr->ln_Succ )
		{
			succ = curr->ln_Succ;
			jts = (jthreadState *)curr;
			func(arg, (jthread_t)jts);
			curr = succ;
		}
		js = (jsked *)js->js_Link.ln_Succ;
	}
}

jthread_t jthread_createfirst(size_t mainThreadStackSize,
			      unsigned char prio,
			      void *jlThread)
{
	jthread_t retval = jthread_current();
	
	retval->jt_State.jts_UserData = jlThread;
	initThreadData(&retval->jt_ThreadData, retval);
	return( retval );
}

jthread_t jthread_create(unsigned char pri,
			 void (*func)(void *),
			 int daemon,
			 void *jlThread,
			 size_t threadStackSize)
{
	jsked *js = jskedContext();
	jthread_t retval;
	int ipri = pri;

	if( (retval = (jthread_t)
	     js->js_CreateThread(js,
				 (char *)&ipri,
				 sizeof(struct _jthread),
				 JTSA_StackSize, threadStackSize,
				 JTSA_UserData, jlThread,
				 JTSA_Daemon, daemon,
				 JTSA_Function, func,
				 JTSA_Argument, jlThread,
				 JTSA_AsyncSignalHandler,
					jthreadAsyncSignalHandler,
				 JTSA_TAG_DONE)) )
	{
		jcpuDisable();
		{
			jactivationEvent *jae;

			jae = &retval->jt_State.jts_ResumeEvent;
			jae->jae_Type = JACTIVATION_EVENT_SPAWN;
			jae->jae_Next = js->js_ReceivedResumeEvents;
			js->js_ReceivedResumeEvents = jae;
			jskedSignal(js, JSSIGB_RESUME);
		}
		jcpuEnable();
	}
	return( retval );
}

void jthread_exit(void)
{
	jthread_t jt;

	jt = jthread_current();
	if( jthread_state.jts_Destructor && jt->jt_State.jts_UserData )
		jthread_state.jts_Destructor(jthread_getcookie(jt));
	jthread_slaughter(jt);

	ensure(0);
}

void jthread_destroy(jthread_t jt)
{
	jsked *js = jt->jt_State.jts_OwnerSked;

	require(jt->jt_Status != JTHREAD_STATUS_BURIED);

	destroyThreadData(&jt->jt_ThreadData);
	jt->jt_Status = JTHREAD_STATUS_BURIED;
	js->js_DeleteThread(js, &jt->jt_State);
}

bool jthread_alive(jthread_t jt)
{
	bool retval = false;

	retval = (jt->jt_Status == JTHREAD_STATUS_ALIVE);
	return( retval );
}

bool jthread_killed(void)
{
	return( jthread_current()->jt_Status == JTHREAD_STATUS_DYING );
}

void jthread_yield(void)
{
	jcpuDisable();
	{
		if( jactivationStall() )
		{
		}
		else
		{
			jsked *js = jskedCurrent();
			
			jskedSignal(js, JSSIGB_PREEMPT);
			js->js_CurrentActivation = 0;
			jskedResumeSchedulerThread(js);
		}
	}
	jcpuEnable();
}

void jthread_stop(jthread_t jt)
{
	// XXX Unimplemented
}

int jthread_slaughter(jthread_t jt)
{
	int retval = 0;
	jsked *js;

	require(jt != 0);
	
	js = jskedCurrent();
	if( jthread_current() == jt )
	{
		killThreadData(&jt->jt_ThreadData, jt);
		//jmutex_lock(&jt->jt_JoinLock);
		jt->jt_Status = JTHREAD_STATUS_DYING;
		//jcondvar_broadcast(&jt->jt_JoinCond, &jt->jt_JoinLock);
		//jmutex_unlock(&jt->jt_JoinLock);
		jcpuDisable();
		{
			jactivationEvent jae;

			if( jactivationStall() )
			{
				require(0);
			}
			else
			{
				jsked *js;
				
				js = jskedCurrent();
				jae.jae_Type = JACTIVATION_EVENT_DEATH;
				jae.jae_Subject.activation =
					js->js_CurrentActivation;
				js->js_CurrentActivation = 0;
				jae.jae_Next = js->js_ReceivedBlockEvents;
				js->js_ReceivedBlockEvents = &jae;
				jskedSignal(js, JSSIGB_BLOCK);
				jskedResumeSchedulerThread(js);

				ensure(0);
			}
		}
		jcpuEnable();
	}
	else
	{
		require(0);
	}
	return( retval );
}

void jthread_interrupt(jthread_t jt)
{
	int disabled = jcpuDisabled();

	if( !disabled )
		jcpuDisable();
	{
		if( jt != jthread_current() )
			jthreadStateSignal(&jt->jt_State, JTSIGB_INTERRUPT);
	}
	if( !disabled )
		jcpuEnable();
}

#if 0
void jthread_join(jthread_t jt, void **value_ptr)
{
	require(jt != 0);

	jmutex_lock(&jt->jt_Lock);
	while( jt->jt_Status != JTHREAD_STATUS_DYING )
	{
		jcondvar_wait(&jt->jt_JoinCond, &jt->jt_JoinLock, 0);
	}
	if( value_ptr )
	{
		*value_ptr = jt->jt_State.jts_ReturnValue;
	}
	jmutex_unlock(&jt->jt_Lock);
}
#endif

/** jsked hooks **/

jthreadState *jskedMainThread(jsked *js)
{
	jthreadState *retval;
	int pri = 5;

	/* XXX */
	retval = js->js_CreateThread(js,
				     (char *)&pri,
				     sizeof(struct _jthread),
				     JTSA_ProcessThread, 1,
				     JTSA_StackSize, 1024 * 1024,
				     JTSA_AsyncSignalHandler,
				 	     jthreadAsyncSignalHandler,
				     JTSA_TAG_DONE);
	return( retval );
}

void jskedExit(int status)
{
	int disabled = jcpuDisabled();
	int lpc;
	
	jactivation_state.jas_CurrentThread = &metalJThreadState.jt_State;
	kernelTeam = 0;
	if( !disabled )
		jcpuDisable();
	for( lpc = 0; lpc < 3; lpc++ )
	{
		jioUnregisterFD(lpc, JIO_FD_TYPE_STDIO);
	}
	EXIT(status);
}

/** jmutex hooks **/

void jmutex_lock(jmutex *jm)
{
	int disabled = jcpuDisabled();
	
	require(jm != 0);

	if( !disabled )
		jcpuDisable();
	{
		jthread_t curr = jthread_current();

		require(curr->jt_Next == 0);
		
		if( jm->jm_Holder == 0 )
		{
			jm->jm_Holder = curr;
		}
		else
		{
			jsignal_set_t recvmask;
			
			curr->jt_Next = jm->jm_Waiters;
			jm->jm_Waiters = curr;
			recvmask = jthreadStateWait(&curr->jt_State,
						    JTSIG_LOCK | JTSIG_KILL);
			if( recvmask & JTSIG_LOCK )
			{
				/* We got the lock. */
			}
			else if( recvmask & JTSIG_KILL )
			{
				jthread_t jt, *jt_prev;
				
				/* We did not get the lock, we were killed. */
				jt_prev = &jm->jm_Waiters;
				jt = jm->jm_Waiters;
				while( jt != curr )
				{
					jt_prev = &jt->jt_Next;
					jt = jt->jt_Next;
				}
				if( jt == curr )
				{
					*jt_prev = jt->jt_Next;
				}
				else
				{
					assert(0);
				}
			}
		}
		
		ensure(jm->jm_Holder == curr);
		ensure(curr->jt_Next == 0);
	}
	if( !disabled )
		jcpuEnable();
}

void jmutex_unlock(jmutex *jm)
{
	int disabled = jcpuDisabled();
	
	require(jm != 0);

	if( !disabled )
		jcpuDisable();
	{
		require(jm->jm_Holder == jthread_current());

		jm->jm_Holder = jm->jm_Waiters;
		if( jm->jm_Waiters )
		{
			jm->jm_Waiters = jm->jm_Waiters->jt_Next;
			jm->jm_Holder->jt_Next = 0;
			jthreadStateSignal(&jm->jm_Holder->jt_State,
					   JTSIGB_LOCK);
		}

		ensure(jm->jm_Holder != jthread_current());
	}
	if( !disabled )
		jcpuEnable();
}

/** jcondvar hooks **/

static
void jcondvar_remove(jcondvar *jcv, jthread_t jt)
{
	jthread_t curr, *prev;

	require(jcv != 0);
	require(jt != 0);
	
	prev = &jcv->jcv_Waiters;
	curr = jcv->jcv_Waiters;
	while( curr && (curr != jt) )
	{
		prev = &curr->jt_Next;
		curr = curr->jt_Next;
	}
	if( curr )
	{
		*prev = jt->jt_Next;
		jt->jt_Next = 0;
	}
}

int jcondvar_wait(jcondvar *jcv,
		  jmutex *jm,
		  unsigned long long timeout)
{
	int disabled = jcpuDisabled();
	int retval = 0;
	
	require(jcv != 0);
	require(jm != 0);

	if( !disabled )
		jcpuDisable();
	{
		jsignal_set_t recvmask, waitmask = (JTSIG_LOCK |
						    JTSIG_INTERRUPT |
						    JTSIG_KILL);
		jthread_t curr = jthread_current();
		jtimer *jt = &curr->jt_TimerRequest;
		jsked *js = jskedCurrent();
		memResourceClient *mrc;
		
		require(jm->jm_Holder == curr);

		curr->jt_Next = jcv->jcv_Waiters;
		jcv->jcv_Waiters = curr;
		jmutex_unlock(jm);
		if( timeout )
		{
			unsigned long long ct;

			ct = jtimerCurrent();
			if( (ct + timeout) > ct )
			{
				jt->jt_Message.jm_Flags = 0;
				jt->jt_Flags = 0;
				jt->jt_AbsoluteTime = ct + timeout;
				jtimerAdd(&js->js_SubTimers, jt);
				waitmask |= JTSIG_TIMER;
			}
			else
			{
				/* Large timeout value, just ignore it. */
			}
		}
		if( (curr->jt_ThreadData.waiter &&
		     (mrc = gc_get_resource(curr->jt_ThreadData.waiter)) &&
		     (mrc->client.owner->state != JTS_ALIVE)) )
		{
			recvmask = JTSIG_INTERRUPT;
		}
		else
		{
			recvmask = jthreadStateWait(&curr->jt_State, waitmask);

			ensure(recvmask != 0);
		}
		if( timeout )
		{
			if( recvmask & JTSIG_TIMER )
			{
				jmessageGet(&curr->jt_TimerReplyPort);
				jt->jt_Flags = 0;
				jcondvar_remove(jcv, curr);
			}
			else
			{
				jtimerRem(&js->js_SubTimers, jt);
			}
		}
		if( (recvmask & JTSIG_INTERRUPT) || (recvmask & JTSIG_KILL) )
		{
			jcondvar_remove(jcv, curr);
		}
		if( recvmask & JTSIG_LOCK )
		{
			retval = 1;
		}
		else
		{
			jmutex_lock(jm);
		}
		
		ensure(jm->jm_Holder == curr);
	}
	if( !disabled )
		jcpuEnable();
	return( retval );
}

void jcondvar_signal(jcondvar *jcv, jmutex *jm)
{
	int disabled = jcpuDisabled();
	
	require(jcv != 0);
	require(jm != 0);

	if( !disabled )
		jcpuDisable();
	{
		jthread_t jt;
		
		require(jm->jm_Holder == jthread_current());

		if( (jt = jcv->jcv_Waiters) )
		{
			jcv->jcv_Waiters = jt->jt_Next;
			
			jt->jt_Next = jm->jm_Waiters;
			jm->jm_Waiters = jt;
		}
	}
	if( !disabled )
		jcpuEnable();
}

void jcondvar_broadcast(jcondvar *jcv, jmutex *jm)
{
	int disabled = jcpuDisabled();
	
	require(jcv != 0);
	require(jm != 0);

	if( !disabled )
		jcpuDisable();
	{
		jthread_t jt, tmp;
		
		require(jm->jm_Holder == jthread_current());

		if( (jt = jcv->jcv_Waiters) )
		{
			jcv->jcv_Waiters = 0;

			tmp = jt;
			while( tmp->jt_Next )
			{
				tmp = tmp->jt_Next;
			}
			tmp->jt_Next = jm->jm_Waiters;
			jm->jm_Waiters = jt;
		}
	}
	if( !disabled )
		jcpuEnable();
}

void jmutex_print(FILE *file, jmutex *jm)
{
	jthread_t jt;
	
	fprintf(file,
		"jmutex(%p)[jm_Holder=%p; "
		"jm_Waiters={ ",
		jm,
		jm->jm_Holder);
	jt = jm->jm_Waiters;
	while( jt )
	{
		fprintf(file, "%p ", jt);
		jt = jt->jt_Next;
	}
	fprintf(file, "}]");
}

void jcondvar_print(FILE *file, jcondvar *jcv)
{
	jthread_t jt;
	
	fprintf(file,
		"jcondvar(%p)[jcv_Waiters={ ",
		jcv);
	jt = jcv->jcv_Waiters;
	while( jt )
	{
		fprintf(file, "%p ", jt);
		jt = jt->jt_Next;
	}
	fprintf(file, "}]");
}

void jthread_sleep(unsigned long long ms)
{
	jthread_t jtid = jthread_current();
	jtimer *jt = &jtid->jt_TimerRequest;
	jsked *js = jskedCurrent();
	
	jcpuDisable();
	{
		jt->jt_Message.jm_Flags = 0;
		jt->jt_Flags = 0;
		jt->jt_AbsoluteTime = jtimerCurrent() + ms;
		jtimerAdd(&js->js_SubTimers, jt);
		jthreadStateWait(&jtid->jt_State,
				 JTSIG_TIMER |
				 JTSIG_INTERRUPT |
				 JTSIG_KILL);
		jtimerRem(&js->js_SubTimers, jt);
	}
	jcpuEnable();
}

void jthread_setpriority(jthread_t jt, int prio)
{
	jskedThreadMessage jtm;
	jmessagePort jmp;
	jsked *js;

	js = jt->jt_State.jts_OwnerSked;
	jmessagePortInit(&jmp, jactivationCurrentThread(), JTSIGB_ONCE);
	jtm.jtm_Message.jem_Message.jm_Link.ln_Pred = 0;
	jtm.jtm_Message.jem_Message.jm_Link.ln_Succ = 0;
	jtm.jtm_Message.jem_Message.jm_Flags = 0;
	jtm.jtm_Message.jem_Message.jm_ReplyPort = &jmp;
	jtm.jtm_Message.jem_Type = JSKED_EXTERNAL_CHANGE_SPEC;
	jtm.jtm_Target = &jt->jt_State;
	jtm.jtm_Payload.spec = (char *)&prio;
	jskedPutExternal(js, &jtm.jtm_Message);
	jcpuDisable();
	{
		jmessageWaitFor(&jmp);
	}
	jcpuEnable();
}

void jthread_dumpthreadinfo(jthread_t jt)
{
	jthreadStatePrint(stderr, &jt->jt_State);
	fprintf(stderr, "\n");
}
