/*
 * jthread.h
 *
 * Copyright (c) 2002, 2003 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

#ifndef _jthread_h
#define _jthread_h

#include "jio.h"
#include "jsked.h"
#include "jtimer_unix.h"
#include "jmessage.h"
#include "jthreadState.h"
#include "threadData.h"

#include "jit-md.h"
#include "exception.h"

#if !defined(UNIX_JTHREADS2)
#define UNIX_JTHREADS2
#endif

struct _sparseElement;

/*
 * jthread signals:
 *   JTSIGB_LOCK - Used for blocking/resuming threads using mutexes/condvars.
 *   JTSIGB_IO - Signal attached to the thread's I/O request reply port.
 *   JTSIGB_TIMER - Signal attached to the thread's timer request reply port.
 */
enum {
	JTSIGB_LOCK = JTSIGB_MAX,
	JTSIGB_IO,
	JTSIGB_TIMER,
	JTSIGB_KILL,
	JTSIGB_MAX2,
};

enum {
	JTSIG_LOCK = (1L << JTSIGB_LOCK),
	JTSIG_IO = (1L << JTSIGB_IO),
	JTSIG_TIMER = (1L << JTSIGB_TIMER),
	JTSIG_KILL = (1L << JTSIGB_KILL),
};

typedef enum {
	JTHREAD_STATUS_MIN,
	JTHREAD_STATUS_LIMBO,
	JTHREAD_STATUS_ALIVE,
	JTHREAD_STATUS_DYING,
	JTHREAD_STATUS_BURIED,
	JTHREAD_STATUS_MAX,
} jthread_status_t;

/*
 * A jthread is the standard kaffe thread object.  Its just a specialized
 * jthreadState object that has some extra fields for synchronously doing
 * I/O and timer requests.
 *
 * jt_State - The standard jthreadState stuff.
 * jt_Status - The current status of the thread.
 * jt_IOReplyPort - A jmessagePort used to receive reply jioRequest messages.
 * jt_IORequest - A preallocated jioRequest for doing file descriptor I/O.
 * jt_TimerReplyPort - A jmessagePort used to receive reply jtimer requests.
 * jt_TimerRequest - A preallocated jtimer for doing sleep/timeouts/etc...
 * jt_ExceptionStackTop - The top of the stack as recorded when a fault has
 *   occurred.
 * jt_ExceptionFrameTop - The top frame as recorded when a fault has occurred.
 * jt_ThreadData - Thread specific data used by kaffe.
 * jt_Next - Link to the next jthread in a singly linked list.  Used for
 *   jmutexes/jcondvars.
 */
typedef struct _jthread {
	jthreadState jt_State;

	jthread_status_t jt_Status;
	
	jmessagePort jt_IOReplyPort;
	jioRequest jt_IORequest;
	jmessagePort jt_TimerReplyPort;
	jtimer jt_TimerRequest;

	void *jt_ExceptionStackTop;
	void *jt_ExceptionFrameTop;

#if defined(LOCAL_THREAD_DATA)
	threadData jt_ThreadData;
#endif
	
	struct _jthread *jt_Next;
} jthread, *jthread_t;

/*
 * Initialize the global jthread state.
 *
 * preemptive - Activate preemptive threading.
 * maxpr - Maximum allowed priority.
 * minpr - Minimum allowed priority.
 * _destructor - Function to be called when a thread calls jthread_exit().
 */
void jthread_init(int preemptive,
		  int maxpr,
		  int minpr,

		  /* XXX Remove below */
		  void *(*_allocator)(size_t),
		  void (*_deallocator)(void *),
		  /* XXX Remove above */
		  
		  void (*_destructor)(void *),
		  void (*_onstop)(void),
		  void (*_ondeadlock)(void));

typedef void (*exchandler_t)(struct _exceptionFrame*);

void jthread_initexceptions(exchandler_t nullHandler,
			    exchandler_t floatingHandler,
			    exchandler_t stackOverflowHandler,
			    exchandler_t abortHandler);

void jthread_atexit(void (*f)(void));

void jthread_walkLiveThreads(void *arg,
			     void (*func)(void *arg,
					  jthread_t jtid,
					  void *jlThread));

void jthread_walk_jthreads(void *arg, void (*func)(void *, jthread_t));

jthread_t jthread_createfirst(size_t mainThreadStackSize,
			      unsigned char prio,
			      void *jlThread);
jthread_t jthread_create(unsigned char pri,
			 void (*func)(void *),
			 int daemon,
			 void *jlThread,
			 size_t threadStackSize);
void jthread_destroy(jthread_t jt);

bool jthread_alive(jthread_t jt);
bool jthread_killed(void);

extern void jthread_setpriority(jthread_t jtid, int prio);

void jthread_yield(void);
void jthread_sleep(unsigned long long ms);
void jthread_stop(jthread_t jt);
int jthread_slaughter(jthread_t jt);
void jthread_exit(void) __NORETURN__;

void jthread_interrupt(jthread_t jt);

void jthread_join(jthread_t jt, void **exit_value);

#if defined(LOCAL_THREAD_DATA)
static inline
threadData *jthread_get_data(jthread_t jt)
{
	return( &jt->jt_ThreadData );
}
#endif

// void jthread_migrate_root(struct _jteam *);

static inline
jthread_t jthread_current(void)
{
	jthread_t retval;

	retval = (jthread_t)jactivationCurrentThread();
	return( retval );
}

static inline
int jthread_on_stack(jthread_t jt, void *bp)
{
	int retval = 0;

	retval |= jthreadStackInRange(&jt->jt_State.jts_ExceptionStack, bp);
	retval |= jthreadStackInRange(&jt->jt_State.jts_Stack, bp);
	return( retval );
}

static inline
int jthread_on_current_stack(void *bp)
{
	jthread_t jt = jthread_current();

	return( jthread_on_stack(jt, bp) );
}

static inline
int jthread_stackcheck(int left)
{
	int retval;

#if defined(STACK_GROWS_UP)
        retval = jthread_on_current_stack((void *)(((char*)&retval) + left));
#else
        retval = jthread_on_current_stack((void *)(((char*)&retval) - left));
#endif
	return( retval );
}

#define	REDZONE	2048

static inline
void jthread_guardcheck(void)
{
#if defined(HAVE_MMAP)
#if defined(STACK_GROWS_UP)
	int left;
	
	*(((char *)&left) + REDZONE) = 0;
#else
#if defined(linux)
	char *rz = alloca(REDZONE);

	rz[REDZONE - 1] = 0;
#else
	int left;
	
	*(((char *)&left) - REDZONE) = 0;
#endif
#endif
#endif
}

static inline
void *jthread_stacklimit(void)
{
	jthread_t jt = jthread_current();
	
#if defined(STACK_GROWS_UP)
#error "Implement me"
#else
        return (void*)((uintp)jt->jt_State.jts_Stack.jts_Top + REDZONE);
#endif
}

static inline
void jthread_relaxstack(int yes)
{
	jthread_t jt = jthread_current();
	
	if( yes )
	{
#if defined(STACK_GROWS_UP)
#error "Implement me"
#else
		(uintp)jt->jt_State.jts_Stack.jts_Top -= REDZONE;
#endif
	}
	else
	{
#if defined(STACK_GROWS_UP)
#error "Implement me"
#else
		(uintp)jt->jt_State.jts_Stack.jts_Top += REDZONE;
#endif
	}
}

static inline
int jthread_on_alternate_stack(jthread_t jt)
{
	int retval = 0;
	
	if( jthreadStackInRange(&jt->jt_State.jts_ExceptionStack,
				jt->jt_State.jts_Context.jcc_StackTop) )
	{
		retval = 1;
	}
	return( retval );
}

static inline
int jthread_extract_stack(jthread_t jt,
			  int index,
			  void **used_from, unsigned *used_len,
			  void **free_from, unsigned *free_len)
{
	int on_ex = 0, retval = 1;
	jthreadStack *jts = 0;
	void *top = 0;

#if defined(STACK_GROWS_UP)
#error "Implement me"
#else
	if( jthreadStackInRange(&jt->jt_State.jts_ExceptionStack,
				jt->jt_State.jts_Context.jcc_StackTop) )
	{
		/* On the exception stack. */
		on_ex = 1;
		switch( index )
		{
		case 0:
			top = jt->jt_ExceptionStackTop;
			jts = &jt->jt_State.jts_Stack;
			break;
		case 1:
			top = jt->jt_State.jts_Context.jcc_StackTop;
			jts = &jt->jt_State.jts_ExceptionStack;
			break;
		default:
			retval = 0;
			break;
		}
	}
	else
	{
		switch( index )
		{
		case 0:
			top = jt->jt_State.jts_Context.jcc_StackTop;
			jts = &jt->jt_State.jts_Stack;
			break;
		default:
			retval = 0;
			break;
		}
	}

	if( retval )
	{
		/* 
		 * [ ..FREE.. | ..USED.. ]
		 */
		require(top <= jts->jts_Bottom);
		require(top >= jts->jts_Top);
		
		if (used_from != NULL)
		{
			*used_from = top;
		}
		if (used_len != NULL)
		{
			*used_len = jts->jts_Bottom - top - sizeof(void *);
		}
		if (free_from != NULL)
		{
			*free_from = jts->jts_Top;
		}
		if (free_len != NULL)
		{
			*free_len = top - jts->jts_Top;
		}
	}
#endif
	return( retval );
}

static inline
void jthread_disable_stop(void)
{
	jthread_t jt;

	jt = jthread_current();
	jt->jt_State.jts_Flags |= JTF_NO_STOP;
	jt->jt_State.jts_StopCount += 1;
}

static inline
void jthread_enable_stop(void)
{
	jthread_t jt;

	jt = jthread_current();
	jt->jt_State.jts_StopCount -= 1;
	if( jt->jt_State.jts_StopCount == 0 )
	{
		jt->jt_State.jts_Flags &= ~JTF_NO_STOP;
		/* XXX Fix for visits */
		if( jt->jt_Status == JTHREAD_STATUS_DYING )
		{
			jthread_slaughter(jt);
		}
	}
}

/* 
 * XXX this is supposed to count the number of stack frames 
 */
static inline
int jthread_frames(jthread *thrd)
{
        return (0);
}

static inline
void jthread_suspendall(void)
{
	jcpuDisable();
}

static inline
void jthread_unsuspendall(void)
{
	jcpuEnable();
}

static inline
void STOPWORLD(void)
{
	jcpuDisable();
}

static inline
void RESUMEWORLD(void)
{
	jcpuEnable();
}

static inline
void jthread_spinon(void *arg)
{
	jcpuDisable();
}

static inline
void jthread_spinoff(void *arg)
{
	jcpuEnable();
}

static inline
void *jthread_getcookie(jthread_t jt)
{
	return( jt->jt_State.jts_UserData );
}

static inline
void jthread_set_context_jsked(jthread_t jt, jsked *js)
{
	require(js != 0);
	
	jt->jt_State.jts_ContextSked = js;
}

static inline
void *jthread_get_frame_pointer(jthread_t jt, int idx)
{
	void *retval = 0;
	
	if( jthreadStackInRange(&jt->jt_State.jts_ExceptionStack,
				jt->jt_State.jts_Context.jcc_StackTop) )
	{
		/* On the exception stack. */
		switch( idx )
		{
		case 0:
			/* Main thread frame. */
			retval = jt->jt_ExceptionFrameTop;
			break;
		case 1:
			/* Exception thread frame. */
			retval = GET_BP(jt->jt_State.jts_Context.jcc_Env);
			break;
		default:
			break;
		}
	}
	else
	{
		retval = GET_BP(jt->jt_State.jts_Context.jcc_Env);
	}
	return( retval );
}

void jthread_dumpthreadinfo(jthread_t tid);

#define UNIX_JTHREADS2_CONFIG \
	UNIX_JTHREADS2_CONTEXT_CONFIG \
	UNIX_JTHREADS2_SIGNAL_CONFIG \
	UNIX_JTHREADS2_JCPU_CONFIG \
	UNIX_JTHREADS2_STACK_CONFIG

extern jthread metalJThreadState;

/*
 * The jskedReservation for the kernel jsked.  Expected to be defined
 * elsewhere.
 */
extern struct _jskedReservation jskedKernelReservation;

#endif
