/*
 * jthreadScheduler.c
 *
 * Copyright (c) 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

#include <stdarg.h>

#include "config.h"
#include "config-std.h"

#include "assert_pp.h"

#include "listNode-inline.h"

#include "jsked.h"

#include "jmutex.h"
#include "jthreadScheduler.h"

#include <janosvm/memResource.h>
#include <janosvm/jteam.h>

#include "jthread_config.h"

#if !defined(ALLOC_TYPE_JTHREAD)
#define ALLOC_TYPE_JTHREAD -1
#endif

typedef struct _jsthreadHeader {
	struct lnNode jst_Link;
} jsthreadHeader;

/*
 * A jsthread wraps the VM's jthread object.
 *
 * jst_Link - Link in the priority list.
 * jst_State - The VM's thread state.
 */
typedef struct _jsthread {
	struct lnNode jst_Link;
	jthread jst_State;
} jsthread, *jsthread_t;

/*
 * Convert a jthread object to a jsthread.
 */
#define JTHREAD_STATE_TO_JSTHREAD(jts) \
	((jsthread_t)(((struct lnNode *)jts) - 1))

/*
 * jsScheduler flags:
 *   JSSB_DISABLING - The scheduler is in the process of suspending all of
 *                    its threads.
 *   JSSB_DISABLED - The scheduler is disabled, so no threads are running and
 *                   it is only waiting for external control.
 */
enum {
	JSSB_ACTIVE,
	JSSB_DISABLING,
	JSSB_DISABLED,
};

enum {
	JSSF_ACTIVE = (1L << JSSB_ACTIVE),
	JSSF_DISABLING = (1L << JSSB_DISABLING),
	JSSF_DISABLED = (1L << JSSB_DISABLED),
};

/*
 * A jsScheduler object maintains any thread scheduler state.
 *
 * jss_Config - The configuration for this scheduler.
 * jss_Flags - Holds the JSSF flags.
 * jss_WaitMask - The current jsked wait mask.
 * jss_RunningThread - The currently executing thread.
 * jss_ReadyThreads - The list of threads that are ready to run.
 * jss_WaitThreads - The list of threads waiting for some input.
 * jss_DisableRequests - Holds the list of disable request messages.
 * jss_DisableReadyThreads - The list of disabled threads that are ready to
 *   run upon reenabling of the jsked.
 * jss_DisableWaitThreads - The list of disabled threads that will be waiting
 *   for an event upon reenabling of the jsked.
 * jss_LiveThreads - The number of live threads managed by this scheduler.
 * jss_DaemonThreads - The number of daemon threads managed by this scheduler.
 */
typedef struct _jsScheduler {
	jsthreadConfig *jss_Config;
	unsigned long jss_Flags;
	jsignal_set_t jss_WaitMask;
	jsthread *jss_RunningThread;
	struct lnList jss_ReadyThreads;
	struct lnList jss_WaitThreads;
	struct lnMinList jss_DeadThreads;
	jmessagePort jss_DisableRequests;
	struct lnList jss_DisabledReadyThreads;
	struct lnList jss_DisabledWaitThreads;
	int jss_LiveThreads;
	int jss_DaemonThreads;
} jsScheduler;

/* XXX Move this out and use the config! */
extern int noExit;

/*
 * Perform final processing for a dead thread.
 *
 * jss - The scheduler's jsScheduler object.
 * jst - The dead thread.
 * returns - True if there are only daemon threads left and (noExit == false).
 */
int jsthreadDeath(jsScheduler *jss, jsthread_t jst)
{
	int retval = 0, disabled;

	disabled = jcpuDisabled();

	/* Remove it from the jsked list. */
	if( !disabled )
		jcpuDisable();
	lnRemove(&jst->jst_State.jt_State.jts_Link);
	if( !disabled )
		jcpuEnable();

	jst->jst_State.jt_ThreadData.flags &= ~TDF_ZOMBIE;
	jss->jss_LiveThreads -= 1;
	if( jst->jst_State.jt_State.jts_Flags & JTF_DAEMON )
	{
		jss->jss_DaemonThreads -= 1;
	}
	if( (jss->jss_LiveThreads == jss->jss_DaemonThreads) && !noExit )
	{
		/* Nothing but daemon's left... */
		retval = 1;
	}
	return( retval );
}

/* XXX Stale? */
void jsthreadWalk(jsScheduler *jss,
		  struct lnMinList *list,
		  jactivationTrapEvent *jte)
{
	struct lnMinNode *curr, *succ;

	curr = list->lh_Head;
	while( curr->ln_Succ )
	{
		jthreadState *jts;
		
		succ = curr->ln_Succ;
		jts = (jthreadState *)curr;
		jte->jte_Function(jte, jts);
		if( jts->jts_Flags & JTF_STOPPED )
		{
			jsthread_t jst;

			jst = JTHREAD_STATE_TO_JSTHREAD(jts);
			lnRemove((struct lnMinNode *)&jst->jst_Link);
			jsthreadDeath(jss, jst);
		}
		curr = succ;
	}
}

/*
 * Pick the next thread to run.
 *
 * jss - The scheduler's jsScheduler object.
 * return - NULL if there are no threads to run or a valid jsthread_t object.
 */
jsthread_t jsthreadPickNext(jsScheduler *jss)
{
	jsthread_t retval = 0;

	while( (retval == 0) &&
	       !lnEmptyList((struct lnMinList *)&jss->jss_ReadyThreads) )
	{
		jsthread_t jst;

		jst = (jsthread_t)lnRemHead((struct lnMinList *)
					    &jss->jss_ReadyThreads);
		/* Check if we're in the process of disabling. */
		if( jss->jss_Flags & JSSF_DISABLING )
		{
			if( jst->jst_State.jt_State.jts_Flags & JTF_NO_STOP )
			{
				/* Can't disable this thread yet, let it run */
				retval = jst;
			}
			else
			{
				jthreadDisable(&jst->jst_State);
				lnAddTail((struct lnMinList *)
					  &jss->jss_DisabledReadyThreads,
					  (struct lnMinNode *)&jst->jst_Link);
			}
		}
		else
		{
			retval = jst;
		}
	}
	return( retval );
}

/*
 * Process a disable jskedExternalMessage.
 *
 * jss - The scheduler's jsScheduler object.
 * jem - The disable request.
 */
void jsthreadDisable(jsScheduler *jss, jskedExternalMessage *jem)
{
	if( jss->jss_Flags & JSSF_DISABLED )
	{
		/* Already disabled, immediately reply. */
		jcpuDisable();
		{
			jmessageReply(&jem->jem_Message);
		}
		jcpuEnable();
	}
	else
	{
		/* Queue it up until we're disabled. */
		jss->jss_Flags |= JSSF_DISABLING;
		jmessagePut(&jss->jss_DisableRequests, &jem->jem_Message);
	}
}

/*
 * Walk the WaitThreads list and disable any that are stoppable.
 *
 * jss - The scheduler's jsScheduler object.
 */
void jsthreadDisableWaiters(jsScheduler *jss)
{
	jsthread_t jst;
	
	while( (jst = (jsthread_t)lnRemHead((struct lnMinList *)
					    &jss->jss_WaitThreads)) )
	{
		if( jst->jst_State.jt_State.jts_Flags & JTF_NO_STOP )
		{
			lnAddHead((struct lnMinList *)&jss->jss_WaitThreads,
				  (struct lnMinNode *)&jst->jst_Link);
			break;
		}
		else
		{
			jthreadDisable(&jst->jst_State);
			lnAddTail((struct lnMinList *)
				  &jss->jss_DisabledWaitThreads,
				  (struct lnMinNode *)&jst->jst_Link);
		}
	}
}

/*
 * Reply to any disable requests.
 *
 * jss - The scheduler's jsScheduler object.
 */
void jsthreadDisableReply(jsScheduler *jss)
{
	jmessage *jm;
	
	require(jss != 0);

	jcpuDisable();
	while( (jm = jmessageGet(&jss->jss_DisableRequests)) )
	{
		jmessageReply(jm);
	}
	jcpuEnable();
}

void jsthreadChangeSpec(jsScheduler *jss, jskedExternalMessage *jem)
{
	jskedThreadMessage *jtm;
	jsthread_t jst;

	require(jss != 0);
	require(jem != 0);
	
	jtm = (jskedThreadMessage *)jem;
	jst = JTHREAD_STATE_TO_JSTHREAD(jtm->jtm_Target);
	jst->jst_Link.ln_Pri = *((int *)jtm->jtm_Payload.spec);
	lnRemove((struct lnMinNode *)&jst->jst_Link);
	if( jst->jst_State.jt_State.jts_Flags & JTF_BLOCKED )
	{
		lnAddTail((struct lnMinList *)&jss->jss_WaitThreads,
			  (struct lnMinNode *)&jst->jst_Link);
	}
	else
	{
		lnEnqueue(&jss->jss_ReadyThreads, &jst->jst_Link);
	}
	jcpuDisable();
	{
		jmessageReply(&jtm->jtm_Message.jem_Message);
	}
	jcpuEnable();
}

static
void jsthreadSwitch(jsked *js, jsthread_t jst)
{
	jcpuDisable();
	if( jactivationStall() )
	{
	}
	else
	{
		js->js_CurrentActivation->ja_Thread = &jst->jst_State.jt_State;
		jactivationResume(js->js_CurrentActivation);
	}
	jcpuEnable();
}

static
void jsthreadSchedulerSignalHandler(jsked *js,
				    jsScheduler *jss,
				    jsignal_set_t recvmask)
{
	int disabled = jcpuDisabled();
	jactivationEvent *jae;
	jsthread_t jst;

	require(jss->jss_Flags & JSSF_ACTIVE);
	
	if( recvmask & JSSIG_SKED_TRAP )
	{
		jactivationTrapEvent *jte;

		if( !disabled )
			jcpuDisable();
		{
			jte = js->js_ReceivedTrapEvents;
			js->js_ReceivedTrapEvents = 0;
		}
		if( !disabled )
			jcpuEnable();
		while( jte )
		{
			jst = JTHREAD_STATE_TO_JSTHREAD(
				jte->jte_Event.jae_Subject.activation->
				ja_Thread);
			switch( jte->jte_Event.jae_Type )
			{
			default:
				assert(0);
				break;
			}
			lnEnqueue(&jss->jss_ReadyThreads,
				  &jst->jst_Link);
			jactivationFree(&js->js_Activations,
					jte->jte_Event.jae_Subject.
					activation);
			jte = (jactivationTrapEvent *)
				jte->jte_Event.jae_Next;
		}
	}
	if( recvmask & JSSIG_PREEMPT )
	{
		/* Our thread was preempted. */
		jactivationFree(&js->js_Activations,
				js->js_PreemptEvent.jae_Subject.
				activation);
		js->js_PreemptEvent.jae_Subject.activation = 0;
	}
	if( recvmask & JSSIG_BLOCK )
	{
		jactivationEvent *jae_next;
		
		/* Our thread blocked. */
		if( !disabled )
			jcpuDisable();
		{
			jae = js->js_ReceivedBlockEvents;
			js->js_ReceivedBlockEvents = 0;
		}
		if( !disabled )
			jcpuEnable();
		while( jae )
		{
			jae_next = jae->jae_Next;
			jst = JTHREAD_STATE_TO_JSTHREAD(
				jae->jae_Subject.activation->ja_Thread);
			jactivationFree(&js->js_Activations,
					jae->jae_Subject.activation);
			lnRemove((struct lnMinNode *)&jst->jst_Link);
			switch( jae->jae_Type )
			{
			case JACTIVATION_EVENT_DEATH:
				if( jsthreadDeath(jss, jst) &&
				    !kernelTeam &&
				    (!jss->jss_Config ||
				     !(jss->jss_Config->jsc_Flags &
				       JSCF_NO_EXIT)) )
				{
					jss->jss_WaitMask = 0;
				}
				break;
			case JACTIVATION_EVENT_BLOCK:
				jst->jst_State.jt_State.jts_Flags |=
					JTF_BLOCKED;
				lnAddTail((struct lnMinList *)
					  &jss->jss_WaitThreads,
					  (struct lnMinNode *)
					  &jst->jst_Link);
				jae->jae_Type = JACTIVATION_EVENT_UNUSED;
				break;
			default:
				break;
			}
			jae = jae_next;
		}
	}
	if( recvmask & JSSIG_RESUME )
	{
		/* Someone resumed one of our threads. */
		if( !disabled )
			jcpuDisable();
		{
			jae = js->js_ReceivedResumeEvents;
			js->js_ReceivedResumeEvents = 0;
		}
		if( !disabled )
			jcpuEnable();
		while( jae )
		{
			jst = JTHREAD_STATE_TO_JSTHREAD(
				jae->jae_Subject.thread);
			switch( jae->jae_Type )
			{
			case JACTIVATION_EVENT_SPAWN:
				require(jst->jst_State.jt_Status ==
					JTHREAD_STATUS_LIMBO);
				require(jst->jst_State.jt_State.
					jts_OwnerSked == js);
				
				if( !disabled )
					jcpuDisable();
				lnAddTail(&js->js_Threads,
					  &jst->jst_State.jt_State.
					  jts_Link);
				if( !disabled )
					jcpuEnable();
					
				jst->jst_State.jt_Status =
					JTHREAD_STATUS_ALIVE;
				jss->jss_LiveThreads += 1;
				if( jst->jst_State.jt_State.jts_Flags &
				    JTF_DAEMON )
				{
					jss->jss_DaemonThreads += 1;
				}
				break;
			case JACTIVATION_EVENT_RESUME:
				require((jst->jst_State.jt_Status ==
					 JTHREAD_STATUS_ALIVE) ||
					(jst->jst_State.jt_Status ==
					 JTHREAD_STATUS_DYING));
				require(jst->jst_State.jt_State.
					jts_OwnerSked == js);

				jst->jst_State.jt_State.jts_Flags &=
					~JTF_BLOCKED;
				lnRemove((struct lnMinNode *)
					 &jst->jst_Link);
				break;
			default:
				fprintf(stderr,
					"Received activation event: "
					"%d\n",
					jae->jae_Type);
				ensure(0);
				break;
			}
			lnEnqueue(&jss->jss_ReadyThreads,
				  &jst->jst_Link);
			jae->jae_Type = JACTIVATION_EVENT_UNUSED;
			jae = jae->jae_Next;
		}
	}
	if( recvmask & JSSIG_EXTERNAL )
	{
		jskedExternalMessage *jem;
		
		/* Process the external messages. */
		while( (jem = jskedGetExternal(js)) )
		{
			switch( jem->jem_Type )
			{
			case JSKED_EXTERNAL_DISABLE:
				jsthreadDisable(jss, jem);
				break;
			case JSKED_EXTERNAL_ENABLE:
				/* XXX Check and stuff */
				jss->jss_Flags &= ~JSSF_DISABLED;
				break;
			case JSKED_EXTERNAL_CHANGE_SPEC:
				jsthreadChangeSpec(jss, jem);
				break;
			default:
				assert(0);
				break;
			}
		}
	}
	if( recvmask & JSSIG_RESET )
	{
		/* Time to go... */
		jss->jss_WaitMask = 0;
	}
	
	ensure(jss->jss_Flags & JSSF_ACTIVE);
}

static
jthreadState *jsthreadShortScheduler(jsked *js,
				     jsScheduler *jss,
				     jsignal_set_t recvmask)
{
	jthreadState *retval;
	jsthread_t next;

	require(js != 0);
	require(jss != 0);

	if( jss->jss_RunningThread )
	{
		lnEnqueue(&jss->jss_ReadyThreads,
			  &jss->jss_RunningThread->jst_Link);
		jss->jss_RunningThread = 0;
	}
	jsthreadSchedulerSignalHandler(js, jss, recvmask);
	if( (next = jsthreadPickNext(jss)) )
	{
		jss->jss_RunningThread = next;
		retval = &next->jst_State.jt_State;
	}
	else
	{
		retval = js->js_SchedulerThread;
	}

	ensure(retval != 0);
	
	return( retval );
}

static
void jsthreadSchedulerInternal(jsScheduler *jss)
{
	static jsignal_set_t ENABLED_WAIT_MASK = (JSSIG_RESET |
						  JSSIG_EXTERNAL |
						  JSSIG_PREEMPT |
						  JSSIG_BLOCK |
						  JSSIG_RESUME |
						  JSSIG_SKED_TRAP);

	static jsignal_set_t DISABLED_WAIT_MASK = JSSIG_EXTERNAL;

	jsked *js = jskedCurrent();
	jsignal_set_t recvmask;

	jss->jss_Flags = JSSF_ACTIVE;
	jss->jss_WaitMask = ENABLED_WAIT_MASK;
	jss->jss_RunningThread = 0;
	lnNewList((struct lnMinList *)&jss->jss_ReadyThreads);
	lnNewList((struct lnMinList *)&jss->jss_WaitThreads);
	lnNewList((struct lnMinList *)&jss->jss_DeadThreads);
	jmessagePortInit(&jss->jss_DisableRequests, 0, -1);
	lnNewList((struct lnMinList *)&jss->jss_DisabledReadyThreads);
	lnNewList((struct lnMinList *)&jss->jss_DisabledWaitThreads);
	jss->jss_LiveThreads = 0;
	jss->jss_DaemonThreads = 0;
	
	js->js_ShortSignalMask = (JSSIG_PREEMPT |
				  JSSIG_BLOCK |
				  JSSIG_RESUME);
	js->js_ShortArgument = jss;
	js->js_ShortFunction = (jthreadState *(*)(jsked *js, void *,
						  jsignal_set_t))
		jsthreadShortScheduler;
	
	while( jss->jss_WaitMask )
	{
		if( (jss->jss_RunningThread = jsthreadPickNext(jss)) )
		{
			/* Got a thread to run. */
			jsthreadSwitch(js, jss->jss_RunningThread);
			if( jss->jss_RunningThread )
			{
				lnEnqueue(&jss->jss_ReadyThreads,
					  &jss->jss_RunningThread->jst_Link);
				jss->jss_RunningThread = 0;
			}
		}
		else if( jss->jss_Flags & JSSF_DISABLING )
		{
			/* No unstoppable threads, try to finish disabling. */
			jsthreadDisableWaiters(jss);
			if( lnEmptyList((struct lnMinList *)
					&jss->jss_ReadyThreads) &&
			    lnEmptyList((struct lnMinList *)
					&jss->jss_WaitThreads) )
			{
				/* Nothing left to wait on... */
				jss->jss_Flags &= ~JSSF_DISABLING;
				jss->jss_Flags |= JSSF_DISABLED;
				jsthreadDisableReply(jss);
				jss->jss_WaitMask = DISABLED_WAIT_MASK;
			}
		}
		if( jss->jss_WaitMask )
		{
			/* Wait on/check the jsked's signals. */
			recvmask = jskedWait(js, jss->jss_WaitMask);
			jsthreadSchedulerSignalHandler(js, jss, recvmask);
		}
		
		ensure(jss->jss_Flags & JSSF_ACTIVE);
		ensure(js->js_Activations.jal_Unused != 0);
	}

	jss->jss_Flags &= ~JSSF_ACTIVE;
	jskedRem(js);

	ensure(0);
}

void jsthreadScheduler(void *arg)
{
	jsthreadConfig *jsc = arg;
	jsScheduler jss;

	memset(&jss, 0, sizeof(jss));
	jss.jss_Config = jsc;
	jsthreadSchedulerInternal(&jss);
}

static struct {
	jsthread jss_RootSchedulerThread;
	char jss_RootSchedulerStack[JTHREAD_STACK_DEFAULT_SIZE];
} jthread_scheduler_state;

jthreadState *jsthreadSchedulerRoot(void *arg)
{
	jsthread_t jst = &jthread_scheduler_state.jss_RootSchedulerThread;
	jthreadState *retval = 0;

	jthreadInit(&jst->jst_State, JTHREAD_TYPE_ROOT);
	if( jthreadStateInit(&jst->jst_State.jt_State,
			     JTSA_Stack, jthread_scheduler_state.
			     jss_RootSchedulerStack,
			     JTSA_StackSize, JTHREAD_STACK_DEFAULT_SIZE,
			     JTSA_Function, jsthreadScheduler,
			     JTSA_Argument, arg,
			     JTSA_TAG_DONE) )
	{
		retval = &jst->jst_State.jt_State;
	}
	else
	{
		require(0);
	}
	return( retval );
}

jthreadState *jsthreadSchedulerCreateThread(jsked *js,
					    char *spec,
					    size_t size,
					    int tag,
					    ...)
{
	jthreadState *retval = 0;
	jsthread_t jst;
	va_list args;

	require(spec != 0);
	
	require(size >= sizeof(struct _jthread)); // XXX

	va_start(args, tag);
	if( (jst = jcpuMalloc(sizeof(jsthreadHeader) + size,
			      ALLOC_TYPE_JTHREAD)) )
	{
		jst->jst_Link.ln_Pri = *((int *)spec);
		if( jthreadStateInitV(&jst->jst_State.jt_State, tag, args) )
		{
			jthread_type_t type;

			if( js == 0 )
				type = JTHREAD_TYPE_SCHEDULER;
			else
				type = JTHREAD_TYPE_NORMAL;
			if( jthreadInit(&jst->jst_State, type) )
			{
				if( js )
				{
					jst->jst_State.jt_State.
						jts_OwnerSked = js;
					jst->jst_State.jt_State.
						jts_ContextSked = js;
					jst->jst_State.jt_State.jts_CPUAccount.
						jca_Parent =
						&js->js_CPUAccount;
				}
				retval = &jst->jst_State.jt_State;
			}
			else
			{
				jthreadStateKill(&jst->jst_State.jt_State);
				jcpuFree(jst);
			}
		}
		else
		{
			jcpuFree(jst);
		}
	}
	va_end(args);
	return( retval );
}

void jsthreadSchedulerDeleteThread(jsked *js, jthreadState *jts)
{
	jsthread_t jst;

	require(js != 0);
	require(jts != 0);

	jst = JTHREAD_STATE_TO_JSTHREAD(jts);
	jthreadStateKill(&jst->jst_State.jt_State);
	jcpuFree(jst);
}
