/*
 * jthreadStack.c
 *
 * Copyright (c) 2002, 2003 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

#include "config.h"

#include <stdio.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <unistd.h>

#include "assert_pp.h"

#include "jthreadStack.h"

#if defined(HAVE_MMAP)
#include <janosvm/logResource.h>
#endif

void jthreadStackInit_process(jthreadStack *jts, int size)
{
	int stackObject;
	
	require(jts != 0);
	require(size >= JTHREAD_STACK_MIN_SIZE);

#if defined(STACK_GROWS_UP)
#error "Implement me"
#else
	/* Get a page-aligned stack pointer */
	jts->jts_Bottom = (void *)(((uintp)&stackObject + 0x1000 - 1)
				   & -0x1000);
	/*
	 * If we are _really_ close to the page boundary then push it up
	 * a bit.
	 *
	 * XXX The constants might be a little fishy.
	 */
	if( ((uintp)jts->jts_Bottom - (uintp)&stackObject) < (0x1000 / 2) )
	{
		jts->jts_Bottom += 0x1000;
	}

	{
		struct rlimit rl;

		getrlimit(RLIMIT_STACK, &rl);
		if( size > rl.rlim_max )
		{
			teamLogEntry(
				"jthreadStack",
				LEA_String, "Warning", "Process stack limit is smaller than expected.", -1,
				LEA_Integer, "expected-size", size,
				LEA_Integer, "max", rl.rlim_max,
				LEA_Integer, "errno", errno,
				LEA_String, "syserror", strerror(errno), -1,
				LEA_TAG_DONE);
			size = rl.rlim_max;
		}
		rl.rlim_cur = size;
		if( setrlimit(RLIMIT_STACK, &rl) < 0 )
		{
			teamLogEntry(
				"jthreadStack",
				LEA_String, "Warning", "Could not increase process stack limit.", -1,
				LEA_Integer, "errno", errno,
				LEA_String, "syserror", strerror(errno), -1,
				LEA_TAG_DONE);
		}
	}
	
	jts->jts_Top = jts->jts_Bottom - size;

#if defined(HAVE_MMAP)
	if( mprotect(jts->jts_Top - getpagesize(), getpagesize(), 0) == -1 )
	{
#if !defined(linux)
		teamLogEntry(
			"jthreadStack",
			LEA_String, "Warning", "Could not set guard page on process stack.", -1,
			LEA_Pointer, "top", jts->jts_Top - getpagesize(),
			LEA_Integer, "errno", errno,
			LEA_String, "syserror", strerror(errno), -1,
			LEA_TAG_DONE);
#endif
	}
#endif
	
	ensure(jts->jts_Bottom > jts->jts_Top);
#endif
}

int jthreadStackInit(jthreadStack *jts, int size)
{
	int retval = 0;

	require(jts != 0);
	require((size == -1) ||	(size >= JTHREAD_STACK_MIN_SIZE));
	
	if( size == -1 )
	{
		size = JTHREAD_STACK_DEFAULT_SIZE;
	}
#if defined(STACK_GROWS_UP)
#error "Implement me"
#else
	if( (jts->jts_Top = jcpuMalloc(size, ALLOC_TYPE_STACK)) )
	{
		jts->jts_Bottom = (void *)(((char *)jts->jts_Top) + size);
		retval = 1;
		
		ensure(jts->jts_Bottom > jts->jts_Top);
	}
#endif
	return( retval );
}

void jthreadStackPrint(FILE *file, jthreadStack *jts)
{
	require(file != 0);
	require(jts != 0);
	
	fprintf(file,
		"jthreadStack(%p)[jts_Bottom=%p; "
		"jts_Top=%p]",
		jts,
		jts->jts_Bottom,
		jts->jts_Top);
}
