/*
 * jsignalSet.c
 *
 * Copyright (c) 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

#ifndef _jthread_stack_h
#define _jthread_stack_h

#include "assert_pp.h"

#include "jcpu.h"

#include "jthread_config.h"

#if !defined(ALLOC_TYPE_STACK)
#define ALLOC_TYPE_STACK -1
#endif

/*
 * The "minimum" thread stack size.
 */
#define JTHREAD_STACK_MIN_SIZE 1024

/*
 * The "minimum" thread stack size.
 */
#define JTHREAD_EXCEPTION_STACK_SIZE 4096

/*
 * The default thread stack size.
 */
#define JTHREAD_STACK_DEFAULT_SIZE (16 * 1024)

/*
 * A jthreadStack tracks the top and bottom of a thread stack.  NOTE: The
 * values contained in this structure are dependent on the STACK_GROWS_UP
 * preprocessor variable.
 *
 * jts_Bottom - The bottom of the thread stack.
 * jts_Top - The top of the thread stack.
 */
typedef struct _jthreadStack {
	void *jts_Bottom;
	void *jts_Top;
} jthreadStack;

/*
 * Static initializer for the metal jthreadStack.
 */
#if defined(STACK_GROWS_UP)
#define METAL_THREAD_STACK_DATA \
{ \
	0, \
	((void *)~0) \
}
#else
#define METAL_THREAD_STACK_DATA \
{ \
	((void *)~0), \
	0 \
}
#endif

/*
 * Initialize a jthreadStack object using the process' stack.
 *
 * jts - A jthreadStack object.
 * size - The maximum size of the process' stack.
 */
void jthreadStackInit_process(jthreadStack *jts, int size);

/*
 * Print the value of the jthreadStack object.
 *
 * file - The file to print to.
 * jts - The object to print.
 */
void jthreadStackPrint(FILE *file, jthreadStack *jts);

/*
 * Initialize a jthreadStack object with jcpuMalloc'ed stack memory.
 *
 * jts - The object to initialize.
 * size - The size of stack to allocate.
 * return - True if the object was successfully initialized.
 */
int jthreadStackInit(jthreadStack *jts, int size);

/*
 * Initialize a jthreadStack object with static memory.
 *
 * jts - The object to initialize.
 * buf - Statically allocated memory to be used as the stack.
 * size - The size of "buf".
 */
static inline
void jthreadStackInit_static(jthreadStack *jts, void *buf, int size)
{
	require(jts != 0);
	require(buf != 0);
	require(size >= JTHREAD_STACK_MIN_SIZE);

#if defined(STACK_GROWS_UP)
#error "Implement me"
#else
	jts->jts_Top = buf;
	jts->jts_Bottom = (void *)(((char *)buf) + size);
	
	ensure(jts->jts_Bottom > jts->jts_Top);
#endif
}

/*
 * Finalize a jthreadInit()'ed jthreadStack object.  NOTE: This will jcpuFree()
 * the stack memory, so this function is not to be used with jthreadStack
 * objects initialized with static memory or the process' stack.
 *
 * jts - The object to finalize.
 */
static inline
void jthreadStackKill(jthreadStack *jts)
{
	require(jts != 0);
	
#if defined(STACK_GROWS_UP)
#error "Implement me"
#else
	
	jcpuFreeExplicit(jts->jts_Top,
			 jts->jts_Bottom - jts->jts_Top,
			 ALLOC_TYPE_STACK);
#endif
}

/*
 * Test if a pointer is within the range of a jthreadStack.
 *
 * jts - The jthreadStack object to test.
 * ptr - The pointer to check.
 * returns - True if "ptr" is within the memory allocated to the stack, false
 *   otherwise.
 */
static inline
int jthreadStackInRange(jthreadStack *jts, void *ptr)
{
	int retval = 0;

	require(jts != 0);

#if defined(STACK_GROWS_UP)
#error "Implement me"
#else
	require(jts->jts_Bottom > jts->jts_Top);
	
	if( (jts->jts_Bottom > ptr) && (ptr >= jts->jts_Top) )
	{
		retval = 1;
	}
#endif
	return( retval );
}

#if defined(HAVE_MMAP)
#define UNIX_JTHREADS2_STACK_CONFIG " (Guard pages)"
#else
#define UNIX_JTHREADS2_STACK_CONFIG
#endif

#endif
