/*
 * jthreadState.h
 *
 * Copyright (c) 2002, 2003 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

#include <assert.h>

#include "assert_pp.h"

#include "jthreadState.h"

int jthreadStateInit(jthreadState *jts, int tag, ...)
{
	va_list args;
	int retval;

	require(jts != 0);

	va_start(args, tag);
	retval = jthreadStateInitV(jts, tag, args);
	va_end(args);
	return( retval );
}

int jthreadStateInitV(jthreadState *jts, int tag, va_list args)
{
	int stackSize = -1, exceptionStackSize = JTHREAD_EXCEPTION_STACK_SIZE;
	void *stackBase = 0, *exceptionStackBase = 0;
	int retval = 1;
	int lpc;
	
	require(jts != 0);

	jsignalInit(&jts->jts_Signals);
	for( lpc = 0; lpc < JTSIGB_MAX; lpc++ )
	{
		int rc;

		rc = jsignalAlloc(&jts->jts_Signals, lpc);

		ensure(rc != -1);
	}
	jts->jts_ResumeEvent.jae_Type = JACTIVATION_EVENT_UNUSED;
	jts->jts_ResumeEvent.jae_Subject.thread = jts;
	while( retval && (tag != JTSA_TAG_DONE) )
	{
		switch( tag )
		{
		case JTSA_OwnerSked:
			jts->jts_OwnerSked = va_arg(args, struct _jsked *);
			jts->jts_ContextSked = jts->jts_OwnerSked;
			break;
		case JTSA_ProcessThread:
			if( va_arg(args, int) )
			{
				jts->jts_Flags |= JTF_PROCESS;
			}
			else
			{
				jts->jts_Flags &= ~JTF_PROCESS;
			}
			break;
		case JTSA_Stack:
			stackBase = va_arg(args, void *);
			break;
		case JTSA_StackSize:
			stackSize = va_arg(args, int);
			break;
		case JTSA_ExceptionStack:
			exceptionStackBase = va_arg(args, void *);
			break;
		case JTSA_ExceptionStackSize:
			exceptionStackSize = va_arg(args, int);
			break;
		case JTSA_Flags:
			jts->jts_Flags = va_arg(args, unsigned long);
			break;
		case JTSA_Daemon:
			if( va_arg(args, int) )
			{
				jts->jts_Flags |= JTF_DAEMON;
			}
			else
			{
				jts->jts_Flags &= ~JTF_DAEMON;
			}
			break;
		case JTSA_Function:
			jts->jts_Function = va_arg(args, void *);
			break;
		case JTSA_Argument:
			jts->jts_Argument = va_arg(args, void *);
			break;
		case JTSA_ParentAccount:
			jts->jts_CPUAccount.jca_Parent =
				va_arg(args, jcpuAccount *);
			break;
		case JTSA_AsyncSignalHandler:
			jts->jts_AsyncSignalHandler =
				va_arg(args, jcpuContextAsyncSignalHandler_t);
			break;
		case JTSA_UserData:
			jts->jts_UserData = va_arg(args, void *);
			break;
		default:
			assert(0);
			break;
		}
		tag = va_arg(args, int);
	}

	jcpuAccountingInit(&jts->jts_CPUAccount);
	if( jts->jts_Flags & JTF_PROCESS )
	{
		jthreadStackInit_process(&jts->jts_Stack, stackSize);
		jts->jts_Context.jcc_Flags |= JCF_ACTIVE;
	}
	else if( stackBase )
	{
		jthreadStackInit_static(&jts->jts_Stack, stackBase, stackSize);
		jcpuContextInit(&jts->jts_Context, jts->jts_Stack.jts_Bottom);
	}
	else if( jthreadStackInit(&jts->jts_Stack, stackSize) )
	{
		jcpuContextInit(&jts->jts_Context, jts->jts_Stack.jts_Bottom);
	}
	else
	{
		retval = 0;
	}
	
	if( !retval )
	{
	}
	else if( exceptionStackBase )
	{
		jthreadStackInit_static(&jts->jts_ExceptionStack,
					exceptionStackBase,
					exceptionStackSize);
	}
	else if( jthreadStackInit(&jts->jts_ExceptionStack,
				  exceptionStackSize) )
	{
		jcpuContext jcc;

		jcpuContextInit(&jcc, jts->jts_ExceptionStack.jts_Bottom);
	}
	else
	{
		retval = 0;
	}
	return( retval );
}

void jthreadStateKill(jthreadState *jts)
{
	require(jts != 0);
	
	jcpuAccountingKill(&jts->jts_CPUAccount);
	/* XXX Handle static stacks! */
	if( !(jts->jts_Flags & JTF_PROCESS) )
	{
		jthreadStackKill(&jts->jts_Stack);
		jthreadStackKill(&jts->jts_ExceptionStack);
	}
}

void jthreadStatePrint(FILE *file, jthreadState *jts)
{
	require(file != 0);
	require(jts != 0);
	
	fprintf(file,
		"jthreadState(%p)[jts_Flags=%lx; "
		"jts_Function=%p; "
		"jts_Argument=%p; "
		"jts_Stack={",
		jts,
		jts->jts_Flags,
		jts->jts_Function,
		jts->jts_Argument);
	jthreadStackPrint(file, &jts->jts_Stack);
	fprintf(file, "}; jts_Context={");
	jcpuContextPrint(file, &jts->jts_Context);
	fprintf(file, "}; jts_CPUAccount={");
	jcpuAccountingPrint(file, &jts->jts_CPUAccount);
	fprintf(file,
		"}; jts_UserData=%p]",
		jts->jts_UserData);
}
