/*
 * jthreadState.h
 *
 * Copyright (c) 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

#ifndef _jthread_state_h
#define _jthread_state_h

#include <stdarg.h>

#include "jcpu.h"
#include "jcpuAccounting.h"
#include "jioAccounting.h"
#include "jthreadStack.h"
#include "jactivationEvent.h"
#include "jsignalSet.h"

struct _jsked;
struct _jthreadState;

/*
 * jthreadState flags:
 *   JTF_PROCESS - The thread is using the process' stack.
 *   JTF_DAEMON - Indicates a daemon thread.
 *   JTF_NO_STOP - The thread is executing system code and cannot be stopped.
 *   JTF_EXCLUSIVE - The thread cannot be replaced by another in the same
 *                   jsked.
 *   JTF_BLOCKED - Indicates that the thread is blocked.
 */
enum {
	JTB_PROCESS,
	JTB_DAEMON,
	JTB_NO_STOP,
	JTB_STOPPED,
	JTB_EXCLUSIVE,
	JTB_BLOCKED,
};

enum {
	JTF_PROCESS = (1L << JTB_PROCESS),
	JTF_DAEMON = (1L << JTB_DAEMON),
	JTF_NO_STOP = (1L << JTB_NO_STOP),
	JTF_STOPPED = (1L << JTB_STOPPED),
	JTF_EXCLUSIVE = (1L << JTB_EXCLUSIVE),
	JTF_BLOCKED = (1L << JTB_BLOCKED),
};

/*
 * Preallocated jsignalSet signals:
 *   JTSIGB_ONCE - Used for short, one-off, signal/wait pairs.
 *   JTSIGB_INTERRUPT - Asynchronous thread interrupt.
 *   JTSIGB_CHILD - XXX
 */
enum {
	JTSIGB_ONCE,
	JTSIGB_INTERRUPT,
	JTSIGB_CHILD,
	JTSIGB_MAX,
};

enum {
	JTSIG_ONCE = (1L << JTSIGB_ONCE),
	JTSIG_INTERRUPT = (1L << JTSIGB_INTERRUPT),
	JTSIG_CHILD = (1L << JTSIGB_CHILD),
};

/*
 * Thread start function typedef.
 */
typedef void *(*jthread_function_t)(void *);

typedef void (*jcpuContextAsyncSignalHandler_t)(struct _jthreadState *jts,
						jsignal_set_t recvmask);

/*
 * A jthreadState object holds the major state information concerning a
 * jthread.
 *
 * jts_Link - Link to the other threads in owner jsked.
 * jts_OwnerSked - The jsked that owns this jthread.
 * jts_ContextSked - The jsked that this thread is executing within.
 * jts_Flags - Holds the JTF_* flags.
 * jts_StopCount - The number of times a thread has recursively disabled stop.
 * jts_Function - The thread's main function.
 * jts_Argument - The argument to pass to the thread's main function.
 * jts_ReturnValue - The value eventually returned by jts_Function.
 * jts_Stack - The jthread's stack.
 * jts_ExceptionStack - The jthread's exception handler stack.
 * jts_Context - The jthread's execution context.
 * jts_CPUAccount - The jthread's CPU account.
 * jts_Signals - The jthread's signal set.
 * jts_ResumeEvent - A preallocated jactivationEvent used for resuming the
 *   thread.
 * jts_AsyncSignalHandler - The asynchronous signal handler for this thread.
 * jts_UserData - XXX
 */
typedef struct _jthreadState {
	struct lnMinNode jts_Link;
	struct _jsked *jts_OwnerSked;
	struct _jsked *jts_ContextSked;
	unsigned long jts_Flags;
	unsigned long jts_StopCount;
	jthread_function_t jts_Function;
	void *jts_Argument;
	void *jts_ReturnValue;
	jthreadStack jts_Stack;
	jthreadStack jts_ExceptionStack;
	jcpuContext jts_Context;
	jcpuAccount jts_CPUAccount;

	jsignalSet jts_Signals;
	jactivationEvent jts_ResumeEvent;

	jcpuContextAsyncSignalHandler_t jts_AsyncSignalHandler;
	
	void *jts_UserData;
} jthreadState;

/*
 * Static initializer for a jthreadState.
 */
#define METAL_THREAD_STATE_DATA \
{ \
	{ 0 },				/* jts_Link */ \
	0,				/* jts_OwnerSked */ \
	0,				/* jts_ContextSked */ \
	0,				/* jts_Flags */ \
	0,				/* jts_StopCount */ \
	0,				/* jts_Function */ \
	0,				/* jts_Argument */ \
	0,				/* jts_ReturnValue */ \
	METAL_THREAD_STACK_DATA,	/* jts_Stack */ \
	METAL_THREAD_STACK_DATA,	/* jts_ExceptionStack */ \
	{ 0 },				/* jts_Context */ \
	{ 0 },				/* jts_CPUAccount */ \
	{ 0 },				/* jts_Signals */ \
	{ 0 },				/* jts_ResumeEvent */ \
	0,				/* jts_AsyncSignalHandler */ \
	0				/* jts_UserData */ \
}

enum {
	JTSA_TAG_DONE,		/* Terminate the tag list. */
	JTSA_ProcessThread,	/* (int) - Set the JTF_PROCESS flag. */
	JTSA_OwnerSked,		/* (jsked *) - The owner jsked. */
	JTSA_Stack,		/* (void *) - The stack base. */
	JTSA_StackSize,		/*
				 * (int) - The size of JTSA_Stack or the size
				 * to allocate.
				 */
	JTSA_ExceptionStack,	/* (void *) - The exception stack base. */
	JTSA_ExceptionStackSize,/*
				 * (int) - The size of JTSA_ExceptionStack or
				 * the size to allocate.
				 */
	JTSA_Flags,		/* (unsigned long) - The flags to set. */
	JTSA_Daemon,		/* (int) - Set the JTF_DAEMON flag. */
	JTSA_Function,		/*
				 * (jthread_function_t) - The thread's main
				 * function.
				 */
	JTSA_Argument,		/*
				 * (void *) - The argument to pass to the
				 * thread's main function.
				 */
	JTSA_ParentAccount,	/* (jcpuAccount *) - The parent CPU account. */
	JTSA_AsyncSignalHandler,/*
				 * (jcpuContextAsyncSignalHandler_t) - The
				 * async signal handler for this thread.
				 */
	JTSA_UserData,
};

/*
 * Initialize the jthreadState object with the given values.
 *
 * jts - The jthreadState object to initialize.
 * tag - The tag list used to initialize the object.
 */
int jthreadStateInit(jthreadState *jts, int tag, ...);
int jthreadStateInitV(jthreadState *jts, int tag, va_list args);

/*
 * Finalize a previously initialized jthreadState object.
 *
 * jts - The jthreadState object to finalize.
 */
void jthreadStateKill(jthreadState *jts);

/*
 * Print the value of the jthreadState object.
 *
 * file - The file to print to.
 * jts - The object to print.
 */
void jthreadStatePrint(FILE *file, jthreadState *jts);

/*
 * Callback used to wait on a set of signals.
 *
 * jts - The jthread that is waiting for the signals.
 * waitmask - The set of signals to wait for.
 * return - The set of waited on signals that were received during the wait.
 */
extern unsigned long jthreadStateWait(jthreadState *jts,
				      jsignal_set_t waitmask);

/*
 * Callback used to signal and possibly resume a blocked jthreadState object.
 *
 * jts - The jthread to signal.
 * signalNumber - The signal number to send to the jthread.
 */
extern void jthreadStateSignal(jthreadState *jts, int signalNumber);

#endif
