/*
 * jtimer_unix.c
 *
 * Copyright (c) 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

#include <stdio.h>
#include <stdarg.h>

#include "assert_pp.h"

#include "jcpu_unix.h"
#include "jtimer_unix.h"

jtimerList jtimer_state;

int jtimerInit(void)
{
	int retval = 1;
	
	jtimerListInit(&jtimer_state);
	return( retval );
}

void jtimerListInit(jtimerList *jtl)
{
	require(jtl != 0);
	
	memset(&jtl->jtl_CurrentTimer, 0, sizeof(jtimer));
	jtl->jtl_Parent = 0;
	lnNewList(&jtl->jtl_PendingTimers);
	lnNewList(&jtl->jtl_ReceivedTimers);
}

static
void jtimerAddWithTime(jtimerList *jtl,
		       jtimer *jt,
		       unsigned long long currTime);

static
void jtimerResetWithTime(jtimerList *jtl, unsigned long long curr)
{
	int done = 0;
	jtimer *jt;

	require(jtl != 0);
	
	/*
	 * Remove any jtimers at the head of the list that have expired
	 * and setup the new itimer when/if we find the first valid timer.
	 */
	jt = (jtimer *)jtl->jtl_PendingTimers.lh_Head;
	while( jt->jt_Message.jm_Link.ln_Succ && !done )
	{
		require(jt->jt_Flags & JTF_PENDING);
		require(!(jt->jt_Flags & JTF_RECEIVED));
		
		if( curr >= jt->jt_AbsoluteTime )
		{
			/* The time has passed. */
			lnRemove(&jt->jt_Message.jm_Link);
			jt->jt_Flags &= ~JTF_PENDING;
			jt->jt_Flags |= JTF_RECEIVED;
			if( jt->jt_Message.jm_ReplyPort )
			{
				/* Send reply to the thread. */
				jmessageReply(&jt->jt_Message);
			}
			else
			{
				/* No reply port, add to received list. */
				lnAddTail(&jtl->jtl_ReceivedTimers,
					  &jt->jt_Message.jm_Link);
			}
			jt = (jtimer *)jtl->jtl_PendingTimers.lh_Head;
		}
		else
		{
			/* jtimer is still valid. */
			if( jtl->jtl_CurrentTimer.jt_Flags )
			{
				require(jtl->jtl_Parent != 0);

				/* Remove ourself so we can be added later. */
				jtimerRem(jtl->jtl_Parent,
					  &jtl->jtl_CurrentTimer);
			}
			/* Compute new absolute time for this list. */
			jtl->jtl_CurrentTimer.jt_AbsoluteTime =
				jt->jt_AbsoluteTime;
			if( jtl->jtl_Parent )
			{
				/* Update parent. */
				jtimerAddWithTime(jtl->jtl_Parent,
						  &jtl->jtl_CurrentTimer,
						  curr);
			}
			else
			{
				unsigned long long diff;
				struct itimerval tv;

				require(jt->jt_AbsoluteTime > curr);
				
				/* Update process. */
				diff = jt->jt_AbsoluteTime - curr;
				tv.it_interval.tv_sec = 0;
				tv.it_interval.tv_usec = 0;
				tv.it_value.tv_sec = diff / 1000;
				tv.it_value.tv_usec = (diff % 1000) * 1000;
				switch( setitimer(ITIMER_REAL, &tv, 0) )
				{
				case -1:
					break;
				}
			}
			done = 1;
		}
	}
}

static
void jtimerAddWithTime(jtimerList *jtl,
		       jtimer *jt,
		       unsigned long long currTime)
{
	jtimer *curr;

	require(jtl != 0);
	require(jt != 0);
	require(jt->jt_Flags == 0);
	require(jcpuDisabled());

	jt->jt_Flags &= ~JTF_RECEIVED;
	jt->jt_Flags |= JTF_PENDING;
	curr = (jtimer *)jtl->jtl_PendingTimers.lh_Head;
	/* Find the correct position in the list. */
	while( curr->jt_Message.jm_Link.ln_Succ &&
	       (jt->jt_AbsoluteTime > curr->jt_AbsoluteTime) )
	{
		curr = (jtimer *)curr->jt_Message.jm_Link.ln_Succ;
	}
	lnInsert(curr->jt_Message.jm_Link.ln_Pred, &jt->jt_Message.jm_Link);
	if( jtl->jtl_PendingTimers.lh_Head == &jt->jt_Message.jm_Link )
	{
		/* New head, recompute the list. */
		jtimerResetWithTime(jtl, currTime);
	}
}

void jtimerReset(jtimerList *jtl)
{
	jtimerResetWithTime(jtl, jtimerCurrent());
}

void jtimerAdd(jtimerList *jtl, jtimer *jt)
{
	jtimerAddWithTime(jtl, jt, jtimerCurrent());
}

void jtimerRem(jtimerList *jtl, jtimer *jt)
{
	require(jtl != 0);
	require(jt != 0);
	require(jcpuDisabled());
	
	if( jt->jt_Flags & JTF_RECEIVED )
	{
		lnRemove(&jt->jt_Message.jm_Link);
	}
	else if( jt->jt_Flags & JTF_PENDING )
	{
		int reset = 0;
		
		if( jtl->jtl_PendingTimers.lh_Head == &jt->jt_Message.jm_Link )
		{
			reset = 1;
		}
		lnRemove(&jt->jt_Message.jm_Link);
		if( reset )
		{
			jtimerReset(jtl);
		}
	}
	else
	{
		/* The user picked it up already. */
	}
	jt->jt_Flags = 0;
}

void jtimerListPrint(FILE *file, jtimerList *jtl)
{
	jtimer *jt;
	
	require(file != 0);

	fprintf(file,
		"jtimerList(%p)[jtl_Parent=%p; "
		"jtl_PendingTimers={ ",
		jtl,
		jtl->jtl_Parent);
	jt = (jtimer *)jtl->jtl_PendingTimers.lh_Head;
	while( jt->jt_Message.jm_Link.ln_Succ )
	{
		jtimerPrint(file, jt);
		fprintf(file, " ");
		jt = (jtimer *)jt->jt_Message.jm_Link.ln_Succ;
	}
	fprintf(file, "}; jtl_ReceivedTimers={ ");
	jt = (jtimer *)jtl->jtl_ReceivedTimers.lh_Head;
	while( jt->jt_Message.jm_Link.ln_Succ )
	{
		jtimerPrint(file, jt);
		fprintf(file, " ");
		jt = (jtimer *)jt->jt_Message.jm_Link.ln_Succ;
	}
	fprintf(file, "}]");
}

void jtimerPrint(FILE *file, jtimer *jt)
{
	require(file != 0);
	require(jt != 0);
	
	fprintf(file,
		"jtimer(%p)[jt_Flags=%lx; jt_AbsoluteTime=%Ld]",
		jt,
		jt->jt_Flags,
		jt->jt_AbsoluteTime);
}
