/*
 * jtimer_unix.h
 *
 * Copyright (c) 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

#ifndef _jtimer_unix_h
#define _jtimer_unix_h

#include <stdio.h>
#include <sys/time.h>

#include "listNode.h"
#include "jmessage.h"

/*
 * jtimer flags:
 *   JTF_PENDING - The timer hasn't fired yet.
 *   JTF_RECEIVED - The timer has been received.
 */
enum {
	JTB_PENDING,
	JTB_RECEIVED,
};

enum {
	JTF_PENDING = (1L << JTB_PENDING),
	JTF_RECEIVED = (1L << JTB_RECEIVED),
};

/*
 * A jtimer message is a request for a timer notification.  The message will
 * be held by the timer handler until some absolute time has been reached, at
 * which point the handler will reply to the message.
 *
 * A jtimer has three states:
 *   Unused - The jtimer is held by the user code.  Indicated by JTF_PENDING
 *            and JTF_RECEIVED being unset.
 *   Pending - The jtimer is held by the system and waiting for its absolute
 *             time.  Indicated by the JTF_PENDING flag.
 *   Received - The jtimer has reached its absolute time and is held by a list
 *              for future processing.
 *
 * jt_Message - The message header.
 * jt_Flags - Holds the above JTF_* flags.
 * jt_AbsoluteTime - The absolute time in milliseconds to wait for before a
 *   reply is expected.
 * jt_UserData - XXX
 */
typedef struct _jtimer {
	jmessage jt_Message;
	unsigned long jt_Flags;
	unsigned long long jt_AbsoluteTime;
	void *jt_UserData;
} jtimer;

/*
 * A jtimerList is used as a node in the tree of timer requests.  Every node in
 * the tree maintains a list of child timers and uses the earliest absolute
 * time as it's own time.  The root of the tree is then used to determine the
 * value of the next timer interrupt for the entire process.
 *
 * The purpose of this hierarchy is to ease the burden when removing a jsked,
 * otherwise we would have to track every request from every thread.
 *
 * jtl_CurrentTimer - A jtimer message which can be used to make requests on
 *   the parent jtimerList.
 * jtl_Parent - An optional reference to a parent jtimerList.
 * jtl_PendingTimers - jtimer requests that are still pending.
 * jtl_ReceivedTimers - jtimer requests that have reached/passed their
 *   absolute time values and don't have a reply port.
 */
typedef struct _jtimerList {
	jtimer jtl_CurrentTimer;
	struct _jtimerList *jtl_Parent;
	struct lnMinList jtl_PendingTimers;
	struct lnMinList jtl_ReceivedTimers;
} jtimerList;

/*
 * The root jtimerList.
 */
extern jtimerList jtimer_state;

/*
 * Initialize the global jtimer state.
 */
int jtimerInit(void);

/*
 * Initialize a jtimerList object.
 *
 * jtl - The object to initialize.
 */
void jtimerListInit(jtimerList *jtl);

/*
 * Resort the jtimerList using the current time.  This involves transferring
 * and pending timers to the received list, updating the parent times, and
 * possibly resetting the process' timer interrupt.
 *
 * jtl - The jtimerList to update.
 */
void jtimerReset(jtimerList *jtl);

/*
 * Add a jtimer request to a jtimerList.
 *
 * jtl - The destination jtimerList.
 * jt - The jtimer to add.
 */
void jtimerAdd(jtimerList *jtl, jtimer *jt);

/*
 * Remove a jtimer request from a jtimerList.  If the timer has already been
 * received it is simply removed from whatever list it is on, the reply port,
 * or jtl->jtl_ReceivedTimers.  Otherwise the timer is removed from list of
 * pending timers and the tree is updated.
 *
 * jtl - The jtimerList that the jtimer request was previously added to.
 * jt - The jtimer to remove.
 */
void jtimerRem(jtimerList *jtl, jtimer *jt);

/*
 * file - The file to print to.
 * jtl - The jtimerList to print.
 */
void jtimerListPrint(FILE *file, jtimerList *jtl);

/*
 * file - The file to print to.
 * jtl - The jtimer to print.
 */
void jtimerPrint(FILE *file, jtimer *jt);

/*
 * return - The current time in milliseconds.
 */
static inline
unsigned long long jtimerCurrent(void)
{
	unsigned long long retval;
	struct timeval tv;
	
	gettimeofday(&tv, 0);
	retval = ((unsigned long long)tv.tv_usec / (unsigned long long)1000);
	retval += ((unsigned long long)tv.tv_sec * (unsigned long long)1000);
	return( retval );
}

#endif
