/*
 * verify.c
 * Perform stages 2 & 3 of class verification.  Stage 1 is performed
 *  when the class is being loaded (so isn't here) and stage 4 is performed
 *  as the method is being executed.
 *
 * Copyright (c) 1996, 1997
 *	Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */

#include "config.h"
#include "config-std.h"
#include "constants.h"
#include "classMethod.h"
#include "baseClasses.h"
#include "exception.h"
#include "errors.h"
#include "access.h"
#include "jni.h"
#include <janosvm/nameSpace.h>

#undef DBG
#define	DBG(s)

/*
 * Verify pass 2:  Check the internal consistency of the class file
 *  but do not check the bytecode.
 */
bool
verify2(Hjava_lang_Class* class, errorInfo *einfo)
{
	classView* cView;
	u2 class_ref, nat_ref, name_ref, sig_ref;
	const char *slot_name, *signature;
	constants* info;
	bool error;
	int i;

DBG(	dprintf("Verifing class %s\n", getClassTraceName(class));	);
	
	error = false;

	/* 1: Check final directives are obeyed */
		/* NOT YET */

	/* 2: Check class has a superclass or it java.lang.Object */
	cView = getClassView(class, einfo);
	if (cView == 0) /* Should be impossible? */
		return false; 

	if ((class->superclass == 0)
	    && strcmp(cView->name->data, OBJECTCLASS) != 0) {
		error = true;
	}

	/* 3: Check class constant pool is consistent */
	info = CLASS_CONSTANTS(class);
	/* Constant pool loaded - check it's integrity. */
	for (i = 1; i < info->size; i++) {
		switch (info->tags[i]) {
		case CONSTANT_Utf8:
		case CONSTANT_Integer:
		case CONSTANT_Float:
		case CONSTANT_ResolvedString:
		case CONSTANT_ResolvedClass:
			/* Always legal */
			break;

		case CONSTANT_Long:
		case CONSTANT_Double:
			/* Always legal */
			i++;
			break;

		case CONSTANT_Class:
		case CONSTANT_String: /* Same as Class */
			/* These have already been re-written to avoid the
			 * extra indirection added by Java.  This is to fit
			 * with the precompiled format.  Here we will not
			 * get an error.
			 */
			break;

		case CONSTANT_Methodref: /* Same as Fieldref */
		case CONSTANT_InterfaceMethodref: /* Same as Fieldref */
			class_ref = METHODREF_CLASS(i, info);
			if (! checkConstantType(class,
						class_ref,
						CONSTANT_Class,
						einfo))
				return (false);
			nat_ref = METHODREF_NAMEANDTYPE(i, info);
			if (! checkConstantType(class,
						nat_ref,
						CONSTANT_NameAndType,
						einfo))
				return (false);
			name_ref = NAMEANDTYPE_NAME(nat_ref, info);
			if (! checkConstantType(class,
						name_ref,
						CONSTANT_Utf8,
						einfo))
				return (false);
			slot_name = WORD2UTF(info->data[name_ref])->data;
			sig_ref = NAMEANDTYPE_SIGNATURE(nat_ref, info);
			if (! checkConstantType(class,
						sig_ref,
						CONSTANT_Utf8,
						einfo))
				return (false);
			signature = WORD2UTF(info->data[sig_ref])->data;
			if( countArgsInSignature(signature) < 0 )
			{
				/* XXX Better error message. */
				postExceptionMessage(
					einfo,
					JAVA_LANG(ClassFormatError),
					"%s "
					"(Method \"%s\" has illegal signature "
					"\"%s\")",
					CLASS_CNAME(class),
					slot_name,
					WORD2UTF(info->data[sig_ref])->data);
				return (false);
			}
			if( (slot_name[0] == '<') &&
			    (WORD2UTF(info->data[name_ref]) !=
			     constructor_name) )
			{
				postExceptionMessage(
					einfo,
					JAVA_LANG(VerifyError),
					"(class: %s, method: %s "
					"signature: %s) "
					"Illegal call to internal method",
					CLASS_CNAME(class),
					slot_name,
					signature);
				return (false);
			}
			if( (slot_name[0] == '<') &&
			    (WORD2UTF(info->data[name_ref]) ==
			     constructor_name) &&
			    (signature[strlen(signature) - 1] != 'V') )
			{
				postExceptionMessage(
					einfo,
					JAVA_LANG(VerifyError),
					"%s "
					"(Method \"%s\" has illegal signature "
					"\"%s\")",
					CLASS_CNAME(class),
					slot_name,
					signature);
				return (false);
			}
			break;
			
		case CONSTANT_Fieldref:
			class_ref = FIELDREF_CLASS(i, info);
			if (! checkConstantType(class,
						class_ref,
						CONSTANT_Class,
						einfo))
				return (false);
			nat_ref = FIELDREF_NAMEANDTYPE(i, info);
			if (! checkConstantType(class,
						nat_ref,
						CONSTANT_NameAndType,
						einfo))
				return (false);
			name_ref = NAMEANDTYPE_NAME(nat_ref, info);
			if (! checkConstantType(class,
						name_ref,
						CONSTANT_Utf8,
						einfo))
				return (false);
			slot_name = WORD2UTF(info->data[name_ref])->data;
			sig_ref = NAMEANDTYPE_SIGNATURE(nat_ref, info);
			if (! checkConstantType(class,
						sig_ref,
						CONSTANT_Utf8,
						einfo))
				return (false);
			signature = WORD2UTF(info->data[sig_ref])->data;
			if( (
			     /* Empty signature string. */
			     signature[0] == '\0' ||
			     /* Bad item. */
			     (sizeofSigItem(&signature, false) < 0) ||
			     /* Extra space. */
			     (signature[0] != '\0')) )
			{
				/* XXX Better error message. */
				postExceptionMessage(
					einfo,
					JAVA_LANG(ClassFormatError),
					"%s "
					"(Field \"%s\" has illegal signature "
					"\"%s\")",
					CLASS_CNAME(class),
					slot_name,
					WORD2UTF(info->data[sig_ref])->data);
				return (false);
			}
			break;
			
		case CONSTANT_NameAndType:
			break;
			
		default:
			error = true;
			break;
		}
	}

	for( i = 0; i < CLASS_NMETHODS(class); i++ )
	{
		Method *meth;

		meth = &CLASS_METHODS(class)[i];
		if( !(meth->accflags & ACC_ABSTRACT) &&
		    !(meth->accflags & ACC_NATIVE) &&
		    (meth->c.bcode.code == 0) )
		{
			postExceptionMessage(
				einfo,
				JAVA_LANG(VerifyError),
				"(class: %s, method: %s signature: %s) "
				"Empty code",
				meth->class->primaryView.name->data,
				meth->name->data,
				meth->class->primaryView.methods[i].
				parsed_sig->signature->data);
			return (false);
		}
	}

	/* If we found an inconsistency then throw an exception */
	if (error == true) {
		postException(einfo, JAVA_LANG(ClassFormatError));
		return (false);
	}
	return (true);
}

/*
 * Verify pass 3:  Check the consistency of the bytecode.
 */
bool
verify3(Hjava_lang_Class* class, errorInfo *einfo)
{
	if ((class->loader == 0 && (Kaffe_JavaVMArgs[0].verifyMode & 1) == 0) ||
	    (class->loader != 0 && (Kaffe_JavaVMArgs[0].verifyMode & 2) == 0)) {
		return (true);
	}

	/* Run bytecode verifier - for reasons only known to the Sun gods
	 * you are suppose to verify the bytecodes for each method at link
	 * time rather than when you run the damn thing.  It's much more
	 * useful to run the verifier on demand because it creates lots of
	 * info which helps the JIT code generator.  I suppose to conform
	 * to the spec we'll just have to run it twice - once here for
	 * compatibility reasons, and once on demand to get the JIT info.
	 * Hey ho ...
	 */

		/* NOT YET !!! */
	return (true);
}
