/*
 * Copyright (c) 1999, 2003 The University of Utah and
 * the Computer Systems Laboratory at the University of Utah (CSL).
 *
 * This file is part of Flick, the Flexible IDL Compiler Kit.
 *
 * Flick is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Flick is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Flick; see the file COPYING.  If not, write to
 * the Free Software Foundation, 59 Temple Place #330, Boston, MA 02111, USA.
 */

#include "config.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "fopen_search.h"

#include "gc.h"
#include "system.h"

/*
 * Open the file named by `path', in the mode specified by `mode', searching
 * the directories named in `dir_list'.
 *
 * If `path' is a relative pathname, search for the file in the directories
 * specified in `dir_list'.  `dir_list' is an array of strings, terminated by a
 * null string.
 *
 * If `path' is an absolute pathname, `dir_list' is not searched; we simply try
 * to open it.  Also, as a special case, if `dir_list' is null, we just try to
 * open `path' --- i.e., we act as if `dir_list' contained an empty string.
 *
 * If the named file is opened, `*out_path' is set to contain the path of the
 * opened file, and the open `FILE *' is returned.  Otherwise, `*out_path' is
 * set to null, and NULL is returned.
 */
FILE *fopen_search(const char *path,
		   struct lnList *dir_list,
		   const char *mode,
		   char **out_path)
{
	FILE *retval = 0;

	if( out_path )
		*out_path = 0;
	if( path )
	{
		/*
		 * If `path' is an an absolute pathname; don't search
		 * `dir_list'.  Also, if `dir_list' is null, just try to open
		 * `path'.
		 */
		if( (retval = fopen(path, mode)) == 0 )
		{
			pathEntry *pe;
			
			/* `path' is a relative path, so we search for it. */
			pe = (pathEntry *)dir_list->lh_Head;
			while( pe->pe_Link.ln_Succ && (retval == 0) )
			{
				char *full_path;
				
				if( (full_path =
				     gc_malloc(strlen(pe->pe_Path) +
					       strlen(file_separator) +
					       strlen(path) +
					       1,
					       GC_ALLOC_TEMP_STRING)) )
				{
					strcpy(full_path, pe->pe_Path);
					strcat(full_path, file_separator);
					strcat(full_path, path);
					retval = fopen(full_path, mode);
					if( (retval == 0) || (out_path == 0) )
					{
						gc_free(full_path);
					}
					if( retval && out_path )
					{
						*out_path = full_path;
					}
				}
			}
		}
	}
	return retval;
}

/* End of file. */

