/*
 * hwHandlers.c
 *
 * Copyright (c) 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

#include "gc.h"
#include "locks.h"

#include <jsi/jsi.h>
#include <janosvm/scripting.h>

#include <nodeos/an_chan.h>
#include <nodeos/an_info.h>
#include <nodeos/an_time.h>
#include <nodeos/an_event.h>
#include <nodeos/an_error.h>
#include <nodeos/an_thread.h>
#include <nodeos/an_demuxkey.h>

#include <nodeos/lib/utils.h>

static jsiHandlerFunctions hwNicFunctions;
static jsiNamedHandler hwNicHandler;

static char *hwHandleNicMessage(jsiHandlerContext *jhc,
				jsiHandler *jh,
				jsiMessage *jm);
static jsiHandler *hwResolveNicSpecifier(jsiHandler *jh,
					 jsiMessage *jm,
					 char *name,
					 char *data);
static int hwGetNicSuites(jsiHandler *jh, jsiMessage *jm);

static jsiHandlerFunctions hwCpuFunctions;
static jsiNamedHandler hwCpuHandler;

static char *hwHandleCpuMessage(jsiHandlerContext *jhc,
				jsiHandler *jh,
				jsiMessage *jm);
static jsiHandler *hwResolveCpuSpecifier(jsiHandler *jh,
					 jsiMessage *jm,
					 char *name,
					 char *data);
static int hwGetCpuSuites(jsiHandler *jh, jsiMessage *jm);

static jsiHandlerFunctions hwRootFunctions;
static jsiNamedHandler hwRootHandler;

static jsiHandler *hwResolveRootSpecifier(jsiHandler *jh,
					  jsiMessage *jm,
					  char *name,
					  char *data);
static int hwGetRootSuites(jsiHandler *jh, jsiMessage *jm);

static an_infocpu_t hwCPU;
#define MAX_IFS 16
static an_infoif_t hwNIC[MAX_IFS];
static int hwNICCount = MAX_IFS;

void installHWHandlers(void)
{
	an_info_getifs(hwNIC, &hwNICCount);
	an_info_getcpu(&hwCPU);
	
	jsiInitNamedHandler(
		&hwRootHandler,
		JNA_Name, "machine",
		JNA_Functions, &hwRootFunctions,
		JNA_ResolveSpecifier, hwResolveRootSpecifier,
		JNA_GetSuites, hwGetRootSuites,
		JNA_Child, jsiInitNamedHandler(
			&hwNicHandler,
			JNA_Name, "nic",
			JNA_Functions, &hwNicFunctions,
			JNA_HandleMessage, hwHandleNicMessage,
			JNA_ResolveSpecifier, hwResolveNicSpecifier,
			JNA_GetSuites, hwGetNicSuites,
			JNA_TAG_DONE),
		JNA_Child, jsiInitNamedHandler(
			&hwCpuHandler,
			JNA_Name, "cpu",
			JNA_Functions, &hwCpuFunctions,
			JNA_HandleMessage, hwHandleCpuMessage,
			JNA_ResolveSpecifier, hwResolveCpuSpecifier,
			JNA_GetSuites, hwGetCpuSuites,
			JNA_TAG_DONE),
		JNA_TAG_DONE);
	jsiInstallHandler(&jvmHandlerList, &hwRootHandler);
}

an_infoif_t *findNIC(char *name)
{
	an_infoif_t *retval = 0;
	int lpc, id;
	
	if( sscanf(name, "%d", &id) == 1 )
	{
		for( lpc = 0; lpc < hwNICCount; lpc++ )
		{
			if( hwNIC[lpc].id == id )
				retval = &hwNIC[lpc];
		}
	}
	else
	{
		for( lpc = 0; lpc < hwNICCount; lpc++ )
		{
			if( !strcmp(hwNIC[lpc].pname, name) )
				retval = &hwNIC[lpc];
		}
	}
	return( retval );
}

static char *moabIFTypes[] = {
	"ANY",
	"ETH",
	"ATM"
};

static char *hwAddress(char *dest, an_infoif_t *info)
{
	switch( info->type )
	{
	case ANIF_ETH:
		sprintf(dest,
			"%02x:%02x:%02x:%02x:%02x:%02x",
			info->addr.eth.daddr[0],
			info->addr.eth.daddr[1],
			info->addr.eth.daddr[2],
			info->addr.eth.daddr[3],
			info->addr.eth.daddr[4],
			info->addr.eth.daddr[5]);
		break;
	default:
		sprintf(dest, "(unknown)");
		break;
	}
	return( dest );
}

static void hwAddNic(jsiMessage *jm, an_infoif_t *info)
{
	char addrbuf[32];

	jsiMessageAdd(jm,
		      JMA_String, "type", moabIFTypes[info->type],
		      JMA_Integer, "id", info->id,
		      JMA_String, "name", info->pname,
		      JMA_Integer, "bandwidth", info->bandwidth,
		      JMA_String, "address", hwAddress(addrbuf, info),
		      JMA_TAG_DONE);
}

jsiHandler *hwResolveNicSpecifier(jsiHandler *jh,
				  jsiMessage *jm,
				  char *name,
				  char *data)
{
	return jh;
}

static jsiPropertyInfo hwSubNicProperties[] = {
	{ "type", (1L << JSI_GET_PROPERTY), "",
	  "The network interface type.\n" },
	{ "id", (1L << JSI_GET_PROPERTY), "",
	  "The network interface unique id.\n" },
	{ "name", (1L << JSI_GET_PROPERTY), "",
	  "The physical name.\n" },
	{ "bandwdith", (1L << JSI_GET_PROPERTY), "",
	  "The total bandwidth.\n" },
	{ "address", (1L << JSI_GET_PROPERTY), "",
	  "The MAC address.\n" },
	{ 0 }
};

char *hwHandleNicMessage(jsiHandlerContext *jhc,
			 jsiHandler *jh,
			 jsiMessage *jm)
{
	char *name, *value, *retval = 0;
	an_infoif_t *nic = 0;
	int lpc;
	
	switch( jm->what )
	{
	case JSI_LIST_PROPERTIES:
		jsiRestoreMessageSpecifier(jm);
		jsiPopMessageSpecifier(jm, &name, &value);
		if( (nic = findNIC(value)) )
		{
			jsiClearMessage(jm);
			jsiMessageAdd(jm,
				      JMA_What, JSI_REPLY,
				      JMA_String, "result", "type",
				      JMA_String, "result", "id",
				      JMA_String, "result", "name",
				      JMA_String, "result", "bandwidth",
				      JMA_String, "result", "address",
				      JMA_TAG_DONE);
		}
		else
		{
			jsiClearMessage(jm);
			jm->what = JSI_REPLY;
			for( lpc = 0; lpc < hwNICCount; lpc++ )
			{
				hwAddNic(jm, &hwNIC[lpc]);
			}
		}
		break;
	case JSI_GET_PROPERTY:
		jsiRestoreMessageSpecifier(jm);
		jsiPopMessageSpecifier(jm, &name, &value);
		if( (nic = findNIC(value)) )
		{
			if( jsiPopMessageSpecifier(jm, &name, &value) )
			{
				jsiClearMessage(jm);
				jm->what = JSI_REPLY;
				if( !strcmp(name, "type") )
				{
					jsiMessageAdd(
						jm,
						JMA_What, JSI_REPLY,
						JMA_String, "result", moabIFTypes[nic->type],
						JMA_TAG_DONE);
				}
				else if( !strcmp(name, "id") )
				{
					jsiMessageAdd(
						jm,
						JMA_What, JSI_REPLY,
						JMA_Integer, "result", nic->id,
						JMA_TAG_DONE);
				}
				else if( !strcmp(name, "name") )
				{
					jsiMessageAdd(
						jm,
						JMA_What, JSI_REPLY,
						JMA_String, "result", nic->pname,
						JMA_TAG_DONE);
				}
				else if( !strcmp(name, "bandwidth") )
				{
					jsiMessageAdd(
						jm,
						JMA_What, JSI_REPLY,
						JMA_Integer, "result", nic->bandwidth,
						JMA_TAG_DONE);
				}
				else if( !strcmp(name, "address") )
				{
					char addrbuf[32];
					
					jsiMessageAdd(
						jm,
						JMA_What, JSI_REPLY,
						JMA_String, "result", hwAddress(addrbuf, nic),
						JMA_TAG_DONE);
				}
				else
				{
					retval = "Unknown nic property";
				}
			}
			else
			{
				jsiClearMessage(jm);
				hwAddNic(jm, nic);
			}
		}
		else
		{
			retval = "Invalid NIC id";
		}
		break;
	case JSI_GET_SUITES:
		jsiClearMessage(jm);
		if( !jsiMessageAdd(
			jm,
			JMA_What, JSI_REPLY,
			JMA_PropertyList, "result", hwSubNicProperties,
			JMA_TAG_DONE) )
		{
			retval = "Out of memory";
		}
		break;
	default:
		retval = "Operation not supported";
		break;
	}
	return( retval );
}

static jsiPropertyInfo hwNicProperties[] = {
	{ "nic", (1L << JSI_LIST_PROPERTIES), "",
	  "List the Network Interface Cards in the machine.\n" },
	{ "nic", (1L << JSI_GET_PROPERTY), "nicID:int/string",
	  "Get info about a specific NIC.\n" },
	{ 0 }
};

int hwGetNicSuites(jsiHandler *jh, jsiMessage *jm)
{
	int retval;

	retval = jsiMessageAdd(jm,
			       JMA_PropertyList, "result", hwNicProperties,
			       JMA_TAG_DONE);
	return( retval );
}

jsiHandler *hwResolveCpuSpecifier(jsiHandler *jh,
				  jsiMessage *jm,
				  char *name,
				  char *data)
{
	return jh;
}

static jsiPropertyInfo hwSubCpuProperties[] = {
	{ "id", (1L << JSI_GET_PROPERTY), "",
	  "The CPU id.\n" },
	{ "cps", (1L << JSI_GET_PROPERTY), "",
	  "The CPU speed in cycles per second.\n" },
	{ 0 }
};

char *hwHandleCpuMessage(jsiHandlerContext *jhc,
			 jsiHandler *jh,
			 jsiMessage *jm)
{
	char *name, *value, *retval = 0;
	int id;

	switch( jm->what )
	{
	case JSI_LIST_PROPERTIES:
		jsiRestoreMessageSpecifier(jm);
		jsiPopMessageSpecifier(jm, &name, &value);
		if( value[0] )
		{
			jsiClearMessage(jm);
			jsiMessageAdd(jm,
				      JMA_String, "result", "id",
				      JMA_String, "result", "cps",
				      JMA_TAG_DONE);
		}
		else
		{
			int lpc;
			
			jsiClearMessage(jm);
			jm->what = JSI_REPLY;
			for( lpc = 0; lpc < hwCPU.ncpu; lpc++ )
			{
				jsiMessageAdd(jm,
					      JMA_Integer, "id", lpc,
					      JMA_Integer, "cps", hwCPU.cps,
					      JMA_TAG_DONE);
			}
		}
		break;
	case JSI_GET_PROPERTY:
		jsiRestoreMessageSpecifier(jm);
		jsiPopMessageSpecifier(jm, &name, &value);
		if( sscanf(value, "%d", &id) == 1 )
		{
			if( jsiPopMessageSpecifier(jm, &name, &value) )
			{
				jsiClearMessage(jm);
				if( !strcmp(name, "id") )
				{
					jsiMessageAdd(
						jm,
						JMA_What, JSI_REPLY,
						JMA_Integer, "result", id,
						JMA_TAG_DONE);
				}
				else if( !strcmp(name, "cps") )
				{
					jsiMessageAdd(
						jm,
						JMA_What, JSI_REPLY,
						JMA_Integer, "result", hwCPU.cps,
						JMA_TAG_DONE);
				}
				else
				{
					retval = "";
				}
			}
			else
			{
				jsiClearMessage(jm);
				jsiMessageAdd(jm,
					      JMA_What, JSI_REPLY,
					      JMA_Integer, "id", id,
					      JMA_Integer, "cps", hwCPU.cps,
					      JMA_TAG_DONE);
			}
		}
		else
		{
			retval = "Invalid CPU id";
		}
		break;
	case JSI_GET_SUITES:
		jsiClearMessage(jm);
		if( !jsiMessageAdd(
			jm,
			JMA_What, JSI_REPLY,
			JMA_PropertyList, "result", hwSubCpuProperties,
			JMA_TAG_DONE) )
		{
			retval = "Out of memory";
		}
		break;
	default:
		retval = "Operation not supported";
		break;
	}
	return( retval );
}

static jsiPropertyInfo hwCpuProperties[] = {
	{ "cpu", (1L << JSI_LIST_PROPERTIES), "",
	  "List the CPUs in the machine.\n" },
	{ "cpu", (1L << JSI_GET_PROPERTY), "cpuID:int",
	  "Get info about a specific CPU.\n" },
	{ 0 }
};

int hwGetCpuSuites(jsiHandler *jh, jsiMessage *jm)
{
	int retval;

	retval = jsiMessageAdd(jm,
			       JMA_PropertyList, "result", hwCpuProperties,
			       JMA_TAG_DONE);
	return( retval );
}

jsiHandler *hwResolveRootSpecifier(jsiHandler *jh,
				   jsiMessage *jm,
				   char *name,
				   char *data)
{
	jsiNamedHandler *jnh = (jsiNamedHandler *)jh;
	jsiHandler *retval;

	if( jsiPopMessageSpecifier(jm, &name, &data) )
	{
		if( (retval = (jsiHandler *)
		     jsiFindHandler(&jnh->children, name)) )
		{
			retval = retval->funcs->resolveSpecifier(retval,
								 jm,
								 name,
								 data);
		}
	}
	else
	{
		retval = jh;
	}
	return retval;
}

static jsiPropertyInfo hwRootProperties[] = {
	{ "machine", (1L << JSI_LIST_PROPERTIES), "",
	  "List the builtin properties of the machine.\n" },
	{ 0 }
};

int hwGetRootSuites(jsiHandler *jh, jsiMessage *jm)
{
	int retval;

	retval = jsiMessageAdd(jm,
			       JMA_PropertyList, "result", hwRootProperties,
			       JMA_TAG_DONE);
	return( retval );
}
