/*
 * moabperf.c
 *
 * Copyright (c) 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

#include "moabperf.h"

#if defined(PERF_TESTING)

/* Operations on timespecs */
#define	timespecclear(tvp)	((tvp)->tv_sec = (tvp)->tv_nsec = 0)
#define	timespecisset(tvp)	((tvp)->tv_sec || (tvp)->tv_nsec)
#define	timespeccmp(tvp, uvp, cmp)					\
	(((tvp)->tv_sec == (uvp)->tv_sec) ?				\
	    ((tvp)->tv_nsec cmp (uvp)->tv_nsec) :			\
	    ((tvp)->tv_sec cmp (uvp)->tv_sec))
#define timespecadd(vvp, uvp)						\
	do {								\
		(vvp)->tv_sec += (uvp)->tv_sec;				\
		(vvp)->tv_nsec += (uvp)->tv_nsec;			\
		if ((vvp)->tv_nsec >= 1000000000) {			\
			(vvp)->tv_sec++;				\
			(vvp)->tv_nsec -= 1000000000;			\
		}							\
	} while (0)
#define timespecsub(vvp, uvp)						\
	do {								\
		(vvp)->tv_sec -= (uvp)->tv_sec;				\
		(vvp)->tv_nsec -= (uvp)->tv_nsec;			\
		if ((vvp)->tv_nsec < 0) {				\
			(vvp)->tv_sec--;				\
			(vvp)->tv_nsec += 1000000000;			\
		}							\
	} while (0)

#include "gc.h"
#include "locks.h"

#include <jsi/jsi.h>
#include <janosvm/scripting.h>

#include <nodeos/an_chan.h>
#include <nodeos/an_info.h>
#include <nodeos/an_time.h>
#include <nodeos/an_event.h>
#include <nodeos/an_error.h>
#include <nodeos/an_thread.h>
#include <nodeos/an_demuxkey.h>

#include <nodeos/lib/utils.h>

static jsiHandlerFunctions mpSendFunctions;
static jsiNamedHandler mpSendHandler;

static char *mpHandleSendMessage(jsiHandlerContext *jhc,
				 jsiHandler *jh,
				 jsiMessage *jm);

static jsiHandlerFunctions mpRecvFunctions;
static jsiNamedHandler mpRecvHandler;

static char *mpHandleRecvMessage(jsiHandlerContext *jhc,
				 jsiHandler *jh,
				 jsiMessage *jm);

static jsiHandlerFunctions mpRootFunctions;
static jsiNamedHandler mpRootHandler;

static char *mpHandleRootMessage(jsiHandlerContext *jhc,
				 jsiHandler *jh,
				 jsiMessage *jm);

static iLock *initLock;

int initMoabPerfModule(int stage)
{
    static int init = 0;
    int iLockRoot;
    
    lockStaticMutex(&initLock);
    if( !init )
    {
	jsiInitNamedHandler(
		&mpRootHandler,
		JNA_Name, "moabperf",
		JNA_Functions, &mpRootFunctions,
		JNA_HandleMessage, mpHandleRootMessage,
		JNA_Child, jsiInitNamedHandler(
			&mpSendHandler,
			JNA_Name, "send",
			JNA_Functions, &mpSendFunctions,
			JNA_HandleMessage, mpHandleSendMessage,
			JNA_TAG_DONE),
		JNA_Child, jsiInitNamedHandler(
			&mpRecvHandler,
			JNA_Name, "recv",
			JNA_Functions, &mpRecvFunctions,
			JNA_HandleMessage, mpHandleRecvMessage,
			JNA_TAG_DONE),
		JNA_TAG_DONE);
	jsiInstallHandler(&jvmHandlerList, &mpRootHandler);
	init = 1;
    }
    unlockStaticMutex(&initLock);
    return( 1 );
}

void killMoabPerfModule(int stage)
{
}

struct recvData {
    iLock *lock;
    an_chan_t *chan;
    an_timespec_t start;
    an_timespec_t last;
    int count;
    int timeout;
    int *hist;
};

static
void drop_recv(an_pbuf_t *pbuf, void *arg, int msize)
{
    struct recvData *rd = (struct recvData *)arg;
    int iLockRoot;

    if( rd->start.tv_sec == 0 )
    {
	an_gettime(&rd->start);
    }
    {
	an_timespec_t curr;
	
	an_gettime(&curr);
	rd->last = curr;
	timespecsub(&curr, &rd->start);
	rd->hist[curr.tv_sec % rd->timeout] += 1;
    }
    rd->count -= 1;
    if( rd->count > 0 )
    {
#if 1
	an_pbuf_return(pbuf);
#else
	an_inchan_addpbuf(rd->chan, pbuf);
#endif
    }
    else
    {
	lockMutex(rd);
	signalCond(rd);
	unlockMutex(rd);
    }
    // an_thread_exit();
}

char *mpHandleRecvMessage(jsiHandlerContext *jhc,
			  jsiHandler *jh,
			  jsiMessage *jm)
{
    const char *retval = 0;
    char *name, *protocol;

    if( !(name = jsiMessageGet(jm,
			       JMA_String, "protocol", 0, &protocol,
			       JMA_TAG_DONE)) )
    {
	int timeout = 10, expecting = 0, runfor = 100000, pbuf_count = 32;
	struct {
	    an_chanaddr_t addr;
	    an_chanproto_t proto;
	    an_netspec_t net;
	} attr;
	an_pbuf_t **pbuf = 0, *pbuf_data = 0;
	an_demuxkey_t *dk, dk_data;
	struct recvData rd;

	memset(&rd, 0, sizeof(struct recvData));
	jsiMessageGet(jm,
		      JMA_Integer, "timeout", 0, &timeout,
		      JMA_Integer, "runfor", 0, &runfor,
		      JMA_Integer, "pbufcount", 0, &pbuf_count,
		      JMA_Integer, "expecting", 0, &expecting,
		      JMA_TAG_DONE);
	rd.count = runfor;
	rd.timeout = timeout;
	strcpy(attr.proto.data, protocol);
	attr.proto.len = strlen(protocol);
	memset(&attr.addr.eth.daddr, 0, sizeof(attr.addr.eth.daddr));
	memset(&attr.addr.eth.saddr, 0, sizeof(attr.addr.eth.saddr));
	attr.net.maxthreads = 1;
	dk = an_demuxkey_create(&dk_data, an_domain_current());
	assert(dk != 0);
	if( (pbuf = gc_malloc(pbuf_count * sizeof(an_pbuf_t *),
			      GC_ALLOC_FIXED)) &&
	    (pbuf_data = gc_malloc(pbuf_count * sizeof(an_pbuf_t),
				   GC_ALLOC_FIXED)) &&
	    (rd.hist = gc_malloc(timeout * sizeof(int), GC_ALLOC_FIXED)) )
	{
	    int lpc, error = 0;
	    
	    for( lpc = 0; (lpc < pbuf_count) && !error; lpc++ )
	    {
		if( !(pbuf[lpc] =
		      an_pbuf_create(&pbuf_data[lpc], AN_PBUF_ETH, 0, 0)) )
		{
		    retval = an_error_message();
		    error = 1;
		}
	    }
	    if( !error )
	    {
		an_chan_t *in, in_data;
		
		attr.net.buffers.npbufs = pbuf_count;
		attr.net.buffers.pbufs = pbuf;
		if( (in = an_inchan_create(&in_data,
					   an_domain_current(),
					   dk,
					   &attr.proto,
					   &attr.addr,
					   &attr.net,
					   drop_recv,
					   &rd)) )
		{
		    an_timespec_t end, diff;
		    an_chanstats_t stats;
		    int packets;
		    double secs, pps;
		    int iLockRoot;

		    rd.chan = in;
		    lockMutex(&rd);
		    waitCond(&rd, timeout * 1000);
		    unlockMutex(&rd);
		    an_gettime(&end);
		    an_inchan_stats(in, &stats);
		    an_inchan_destroy(in);
		    diff = end;
		    timespecsub(&diff, &rd.start);
		    secs = (double)diff.tv_sec +
			((double)diff.tv_nsec / (double)1000000000.0);
		    if( rd.count < 0 )
			packets = runfor;
		    else
			packets = runfor - rd.count;
		    rd.count = 0;
		    pps = packets / secs;
		    jsiClearMessage(jm);
		    if( !jsiMessageAdd(
			jm,
			JMA_What, JSI_REPLY,
			JMA_Integer, "expecting-pps", expecting,
			JMA_Integer, "packets", packets,
			JMA_Integer, "first", rd.start.tv_sec,
			JMA_Integer, "last", rd.last.tv_sec,
			JMA_Integer, "stats.droppackets", (int)stats.droppackets,
			JMA_Integer, "stats.dropnobufs", (int)stats.dropnobufs,
			JMA_Float, "seconds", secs,
			JMA_Float, "pps", pps,
			JMA_TAG_DONE) )
		    {
			retval = "No memory";
		    }
		    for( lpc = 0; lpc < timeout; lpc++ )
		    {
			jsiMessageAdd(jm,
				      JMA_Integer, "hist", rd.hist[lpc],
				      JMA_TAG_DONE);
		    }
		    /** XXX Silly wait to make sure drop_recv is done **/
		    lockMutex(&rd);
		    waitCond(&rd, 1000);
		    unlockMutex(&rd);
		}
		else
		{
		    retval = an_error_message();
		}
	    }
	    else
	    {
		pbuf_count = lpc;
	    }
	    for( lpc = 0; lpc < pbuf_count; lpc++ )
	    {
		an_pbuf_destroy(&pbuf_data[lpc]);
	    }
	}
	else
	{
	    retval = "No memory for packet buffers.";
	}
	gc_free(rd.hist);
	gc_free(pbuf_data);
	gc_free(pbuf);
	an_demuxkey_destroy(dk);
    }
    else
    {
	jsiClearMessage(jm);
	if( !jsiMessageAdd(jm,
			   JMA_What, JSI_ERROR,
			   JMA_String, "Why", "Missing value",
			   JMA_String, "Which", name,
			   JMA_TAG_DONE) )
	{
	    retval = "No memory";
	}
    }
    return( retval );
}

char *mpHandleSendMessage(jsiHandlerContext *jhc,
			  jsiHandler *jh,
			  jsiMessage *jm)
{
    char *name, *protocol, *data, *retval = 0;
    
    if( !(name = jsiMessageGet(jm,
			       JMA_String, "protocol", 0, &protocol,
			       JMA_String, "data", 0, &data,
			       JMA_TAG_DONE)) )
    {
	int runfor = 100000, pbuf_count = 1024;
	struct {
	    an_chanaddr_t addr;
	    an_chanproto_t proto;
	    an_netspec_t net;
	} attr;
	an_chan_t *out, out_data;

	jsiMessageGet(jm,
		      JMA_Integer, "runfor", 0, &runfor,
		      JMA_Integer, "pbufcount", 0, &pbuf_count,
		      JMA_TAG_DONE);
	strcpy(attr.proto.data, protocol);
	attr.proto.len = strlen(protocol);
	memset(&attr.addr.eth.daddr, 0, sizeof(attr.addr.eth.daddr));
	memset(&attr.addr.eth.saddr, 0, sizeof(attr.addr.eth.saddr));
	attr.net.bandwidth = 100 * 1000000;
	if( (out = an_outchan_create(&out_data,
				     an_domain_current(),
				     &attr.proto,
				     &attr.addr,
				     &attr.net)) )
	{
	    an_pbuf_t *pbuf_data;

	    if( (pbuf_data = gc_malloc(pbuf_count * sizeof(an_pbuf_t),
				       GC_ALLOC_FIXED)) )
	    {
		int lpc, error = 0;

		for( lpc = 0; (lpc < pbuf_count) && !error; lpc++ )
		{
		    if( !an_pbuf_create(&pbuf_data[lpc], AN_PBUF_ETH, 0, 0) )
		    {
			retval = an_error_message();
			error = 1;
		    }
		}
		if( !error )
		{
		    an_timespec_t start, end, diff;
		    an_error_t err = 0;
		    double pps, secs;
		    char *packet;
		    int len;

		    len = strlen(data);
		    for( lpc = 0; lpc < pbuf_count; lpc++ )
		    {
			int lpc2;
						
			packet = an_pbuf_data(&pbuf_data[lpc]);
			for( lpc2 = 0; lpc2 < len; lpc2 += 2 )
			{
			    int value;
							
			    sscanf(&data[lpc2], "%2x", &value);
			    packet[lpc2 / 2] = (char)(value & 0xff);
			}
			err = an_pbuf_setsize(&pbuf_data[lpc], len / 2, 0);
			assert(err == 0);
		    }
		    an_gettime(&start);
		    for( lpc = 0; (lpc < runfor) && (err == 0); lpc++ )
		    {
			int slot;

			slot = lpc % pbuf_count;
			while( an_pbuf_isbusy(&pbuf_data[slot]) )
			{
			    an_thread_yield();
			}
			err = an_outchan_send(out, &pbuf_data[slot]);
		    }
		    an_gettime(&end);
		    diff = end;
		    timespecsub(&diff, &start);
		    secs = (double)diff.tv_sec +
			((double)diff.tv_nsec / (double)1000000000.0);
		    pps = lpc / secs;
		    jsiClearMessage(jm);
		    jsiMessageAdd(jm,
				  JMA_What, JSI_REPLY,
				  JMA_Integer, "Packets", lpc,
				  JMA_Float, "Seconds", secs,
				  JMA_Float, "PPS", pps,
				  JMA_TAG_DONE);
		}
		else
		{
		    pbuf_count = lpc;
		}
		for( lpc = 0; lpc < pbuf_count; lpc++ )
		{
		    an_pbuf_destroy(&pbuf_data[lpc]);
		}
		gc_free(pbuf_data);
	    }
	    else
	    {
		retval = "Not enough memory for pbufs.";
	    }
	    an_outchan_destroy(out);
	}
	else
	{
	    retval = an_error_message();
	}
    }
    else
    {
	jsiClearMessage(jm);
	if( !jsiMessageAdd(jm,
			   JMA_What, JSI_ERROR,
			   JMA_String, "Why", "Missing value",
			   JMA_String, "Which", name,
			   JMA_TAG_DONE) )
	{
	    retval = "No memory";
	}
    }
    return( retval );
}

char *mpHandleRootMessage(jsiHandlerContext *jhc,
			  jsiHandler *jh,
			  jsiMessage *jm)
{
    char *retval = 0;

    switch( jm->what )
    {
    default:
	retval = "Unsupported operation";
	break;
    }
    return( retval );
}

int mpGetSuites(jsiHandler *jh, jsiMessage *jm)
{
    int retval = 1;

    return( retval );
}

#endif
