/*
 * Sensor.c
 * Native implementation of kaffe.sensors.Sensor functions.
 *
 * Copyright (c) 2002 University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.
 * See the file "license.terms" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * Contributed by the Flux Research Group, Department of Computer Science,
 * University of Utah, http://www.cs.utah.edu/flux/
 */

#include "config.h"
#include "object.h"
#include <native.h>
#include <files.h>
#include "support.h"

#include "gc.h"

#include "lm_sensors.h"

#include "kaffe_sensors_Sensor.h"
#include "kaffe_sensors_SensorFeature.h"

void
kaffe_sensors_Sensor_nativeInit(void)
{
	static char *defaultConfigFileName = "/etc/sensors.conf";
	
	char *configFileName;
	errorInfo einfo;
	int error = 0;
	FILE *file;

	configFileName = getProperty(getSystemProperties(),
				     "kaffe.sensors.lm.config",
				     defaultConfigFileName);
	if( (file = fopen(configFileName, "r")) )
	{
		if( sensors_init(file) == 0 )
		{
			/* Success */
		}
		else
		{
			postExceptionMessage(&einfo,
					     JAVA_IO(FileNotFoundException),
					     "%s: Bad format",
					     configFileName);
			error = 1;
		}
		fclose(file);
	}
	else
	{
		postExceptionMessage(&einfo,
				     JAVA_IO(FileNotFoundException),
				     "%s: %s",
				     configFileName,
				     SYS_ERROR(errno));
		error = 1;
	}
	if( configFileName != defaultConfigFileName )
	{
		gc_free(configFileName);
	}
	if( error )
	{
		throwError(&einfo);
	}
}


struct Hkaffe_sensors_Sensor *
kaffe_sensors_Sensor_nativeDetect(struct Hjava_lang_Class *cls)
{
	struct Hkaffe_sensors_Sensor *sensor, *retval = 0;
	const sensors_chip_name *scn;
	int sensorIndex = 0;
	
	require(cls);

	while( (scn = sensors_get_detected_chips(&sensorIndex)) )
	{
		sensor = (struct Hkaffe_sensors_Sensor *)
		    execute_java_constructor(0, 0, cls, "()V");
		sensor->next = retval;
		retval = sensor;
		sensor->internal = (struct Hkaffe_util_Ptr *)scn;
		sensor->name = stringC2Java(scn->prefix);
	}
	return( retval );
}

struct Hkaffe_sensors_SensorFeature *
kaffe_sensors_Sensor_detectFeatures(struct Hkaffe_sensors_Sensor *sensor,
				    struct Hjava_lang_Class *cls)
{
	struct Hkaffe_sensors_SensorFeature *feature, *retval = 0;
	const sensors_feature_data *sfd;
	sensors_chip_name *scn;
	int nr1 = 0, nr2 = 0;

	scn = (sensors_chip_name *)sensor->internal;
	while( (sfd = sensors_get_all_features(*scn, &nr1, &nr2)) )
	{
		feature = (struct Hkaffe_sensors_SensorFeature *)
		    execute_java_constructor(0, 0, cls, "()V");
		feature->next = retval;
		retval = feature;
		feature->sensor = sensor;
		feature->name = stringC2Java(sfd->name);
		feature->mode = sfd->mode;
		feature->internal = (struct Hkaffe_util_Ptr *)sfd;
	}
	return( retval );
}
