/*
 * Sensor.java
 * Native class for accessing motherboard sensor information.
 *
 * Copyright (c) 2002, 2003 University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.
 * See the file "license.terms" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * Contributed by the Flux Research Group, Department of Computer Science,
 * University of Utah, http://www.cs.utah.edu/flux/
 */

package kaffe.sensors;

import java.util.Hashtable;
import java.util.Enumeration;

import java.io.FileNotFoundException;

/**
 * A Sensor corresponds to a chip that contains several features that can be
 * read from or written to.  This class is relatively primitive and should
 * probably be extended by more capable classes.
 */
public class Sensor
{
    /**
     * Cached reference to this class.
     */
    private static final Class THIS_CLASS = Sensor.class;
    
    /**
     * Cached reference to the feature class.
     */
    private static final Class SENSOR_FEATURE_CLASS = SensorFeature.class;

    /**
     * Initialize any native data.
     */
    private static native void nativeInit()
	throws Exception;

    /**
     * Detect and return a list sensors in the computer.
     *
     * @param cls The Sensor subclass that should be used when creating new
     * objects.
     * @return A linked list of Sensors that were detected.
     */
    private static native Sensor nativeDetect(Class cls);

    static {
	try
	{
	    /* Load our native library. */
	    System.loadLibrary("kaffesensors");
	    
	    /* Initialize the native library. */
	    Sensor.nativeInit();
	}
	catch(RuntimeException e)
	{
	    throw e;
	}
	catch(Exception e)
	{
	    throw new ExceptionInInitializerError(e);
	}
    }

    /**
     * Detect and return a table of sensors.
     *
     * @param sensorClass The class to use when creating Sensor objects.
     * @param featureClass The class to use when creating SensorFeature
     * objects.
     * @return A Hashtable that maps sensor chip names to Sensor objects.
     */
    public static Hashtable detect(Class sensorClass, Class featureClass)
    {
	final Hashtable retval = new Hashtable();
	Sensor sensor;

	/* Make sure its a valid subclass. */
	if( !THIS_CLASS.isAssignableFrom(sensorClass) )
	{
	    throw new IllegalArgumentException("Sensor class must be a "
					       + "subclass of Sensor");
	}
	
	if( !SENSOR_FEATURE_CLASS.isAssignableFrom(featureClass) )
	{
	    throw new IllegalArgumentException("Feature class must be a "
					       + "subclass of Sensor");
	}
	
	sensor = Sensor.nativeDetect(sensorClass);
	while( sensor != null )
	{
	    SensorFeature curr, features;
	    
	    features = sensor.detectFeatures(featureClass);
	    curr = features;
	    while( curr != null )
	    {
		sensor.features.put(curr.getName(), curr);
		curr = curr.next;
	    }
	    retval.put(sensor.name, sensor);
	    sensor = sensor.next;
	}
	return retval;
    }

    /**
     * Detect and return a table of sensors.
     * 
     * @return A Hashtable that maps sensor chip names to Sensor objects.
     */
    public static Hashtable detect()
    {
	return detect(THIS_CLASS, SENSOR_FEATURE_CLASS);
    }

    /**
     * Link to next node in the list.  XXX This is only used during creation.
     */
    private Sensor next;

    /**
     * Link to any internal data structures used by the native library.
     */
    private kaffe.util.Ptr internal;

    /**
     * The name of the sensor.
     */
    private String name;

    /**
     * The set of sensor features.
     */
    private final Hashtable features = new Hashtable();

    /**
     * Construct an empty Sensor object.
     */
    protected Sensor()
    {
    }

    /**
     * Detect the set of sensor features.
     *
     * @param featureClass The class to use when creating SensorFeature
     * objects.
     * @return A list of SensorFeatures that are supported by this Sensor.
     */
    private native SensorFeature detectFeatures(Class featureClass);

    /**
     * @return The name of the sensor.
     */
    public String getName()
    {
	return this.name;
    }

    /**
     * @return An Enumeration of all the sensor's features.
     */
    public Enumeration getFeatures()
    {
	return this.features.elements();
    }

    /**
     * @param name The specific name of a sensor feature.
     * @return The SensorFeature corresponding to the given name or null if it
     * couldn't be found.
     */
    public SensorFeature getFeature(String name)
    {
	return (SensorFeature)this.features.get(name);
    }

    public int hashCode()
    {
	return this.name.hashCode();
    }

    public boolean equals(Object obj)
    {
	boolean retval = false;

	if( obj instanceof Sensor )
	{
	    Sensor sensor;

	    sensor = (Sensor)obj;
	    retval = this.name.equals(sensor.name);
	}
	return retval;
    }
    
    public String toString()
    {
	return "Sensor[name="
	    + this.name
	    + "; features="
	    + this.features
	    + "]";
    }
}
