/*
 * SensorFeature.java
 * Native class for accessing motherboard sensor information.
 *
 * Copyright (c) 2002 University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.
 * See the file "license.terms" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * Contributed by the Flux Research Group, Department of Computer Science,
 * University of Utah, http://www.cs.utah.edu/flux/
 */

package kaffe.sensors;

import java.io.IOException;

/**
 * A SensorFeature corresponds to a specific feature of a sensor chip and is
 * used to read and write data.
 */
public class SensorFeature
{
    /**
     * SensorFeature mode indicating that no reading or writing is allowed.
     */
    public static final int MODE_NO_RW = 0;
    
    /**
     * SensorFeature mode indicating that only reading is allowed.
     */
    public static final int MODE_R = 1;
    
    /**
     * SensorFeature mode indicating that only writing is allowed.
     */
    public static final int MODE_W = 2;
    
    /**
     * SensorFeature mode indicating that reading and writing is allowed.
     */
    public static final int MODE_RW = (MODE_R | MODE_W);

    /**
     * Link to next node in the list.  XXX This is only used during creation.
     */
    SensorFeature next;

    /**
     * The sensor that this feature is a part of.
     */
    private Sensor sensor;

    /**
     * The name of this feature.
     */
    private String name;

    /**
     * The access modes for this feature.
     */
    private int mode;

    /**
     * The internal data for this feature.
     */
    private kaffe.util.Ptr internal;

    /**
     * Construct an empty SensorFeature.
     */
    protected SensorFeature()
    {
    }

    /**
     * @return The sensor that this feature is a part of.
     */
    public Sensor getSensor()
    {
	return this.sensor;
    }

    /**
     * @return The name of this feature.
     */
    public String getName()
    {
	return this.name;
    }

    /**
     * @param mode The access modes for this feature.
     */
    protected void setMode(int mode)
    {
	this.mode = mode;
    }

    /**
     * @return The access modes for this feature.
     */
    public int getMode()
    {
	return this.mode;
    }

    /**
     * @return The current value from this sensor feature.
     * @throws IOException if there was a problem reading the sensor.
     */
    public native double read()
	throws IOException;

    /**
     * @param value The value to write to this sensor feature.
     * @throws IOException if there was a problem writing to the sensor.
     */
    public native void write(double value)
	throws IOException;

    public int hashCode()
    {
	return this.name.hashCode() + this.sensor.hashCode();
    }

    public boolean equals(Object obj)
    {
	boolean retval = false;

	if( obj instanceof SensorFeature )
	{
	    SensorFeature sf;

	    sf = (SensorFeature)obj;
	    retval = (this.sensor.equals(sf.sensor) &&
		      this.name.equals(sf.name));
	}
	return retval;
    }
    
    public String toString()
    {
	return "SensorFeature[sensor="
	    + this.sensor.getName()
	    + "; name="
	    + this.name
	    + "; mode="
	    + this.mode
	    + "]";
    }
}
