/*
 * ClassPathSpec.java
 * Resource specification for a team's class path.
 *
 * Copyright (c) 2000, 2001, 2002, 2003 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Lesser 
 * General Public License.
 * 
 * See the file "license-lesser.terms" for information on 
 * usage and redistribution of this file.
 */

package edu.utah.janosvm.resources;

/**
 * Specification for the class path resource. Used to specify the
 * classpath for a new team.  The given classpath is used to load new
 * classes in the team, but is not used for shared classes (those
 * specified via the {@link NameSpaceSpec}).
 *
 * <p>A <code>ClassPathSpec</code> is comprised of
 * <code>String</code>s, one for each element.  Only globally valid
 * class path elements (see {@link ClassPathResource}) can be used
 * (otherwise an ResourceException will be thrown by the team
 * constructor).  A null or empty ClassPathSpec is valid (it implies
 * the new team cannot load any classes and can only use imported
 * classes).
 *
 * <p>If not specified, the default class path for a new Team is the
 * full class path of the JVM.
 *
 * @see ClassPathResource
 * @see edu.utah.janosvm.sys.Team#create()
 * @author Tim Stack, <a href="mailto:janos-discuss@flux.cs.utah.edu">Utah Janos Team</a>
 */
public final class ClassPathSpec
    extends ResourceSpec
    implements Cloneable
{
    /**
     * Constructs the native part of the class.
     */
    private native void nativeConstructor(String entries[]);

    /**
     * Finalizes the native part of the class.
     */
    private native void nativeFinalizer();
    
    /**
     * Construct a class path specification with the given paths.  Entries
     * must be valid ones defined via {@link ClassPathResource}.
     *
     * @param entries An array of strings specifying each part of the
     * classpath or null if the ClassPath is empty.  Any null elements
     * are ignored.
     */
    public ClassPathSpec(String entries[])
    {
	this.nativeConstructor(entries);
    }
    
    protected void finalize()
    {
	this.nativeFinalizer();
    }

    /**
     * Get the number of components in this classpath.
     *
     * @return The number of paths in the spec.
     */
    public native int getPathCount();

    /**
     * Get a specific component of this classpath.
     *
     * @param index The index of the class path to get.
     * @return The class path at the given index.
     */
    public native String getPath(int index);

    /**
     * Return a new String array of the components of the classpath.
     *
     * @return An array of Strings, one for each component of the classpath.
     */
    public String[] getPaths()
    {
	String[] s = new String[getPathCount()];
	for (int i = 0; i < s.length; i++)
	{
	    s[i] = getPath(i);
	}
	return s;
    }

    /**
     * Compare the given object with this ClassPathSpec for equality.  Returns true
     * if the given object is a ClassPathSpec that contains the same
     * number of equivalent components, in the same order as this.
     */
    public boolean equals(Object obj)
    {
	boolean retval = false;

	if( obj instanceof ClassPathSpec )
	{
	    ClassPathSpec cps;
	    
	    cps = (ClassPathSpec)obj;
	    if( this.getPathCount() == cps.getPathCount() )
	    {
		int lpc, count;

		retval = true;
		count = this.getPathCount();
		for( lpc = 0; retval && (lpc < count); lpc++ )
		{
		    retval = this.getPath(lpc).equals(cps.getPath(lpc));
		}
	    }
	}
	return retval;
    }

    public String toString()
    {
	int lpc, count = this.getPathCount();
	String retval =
	    "ClassPathSpec[paths["
	    + count
	    + "]={ ";

	count = this.getPathCount();
	for( lpc = 0; lpc < count; lpc++ )
	{
	    if( lpc > 0 )
		retval += ", ";
	    retval += this.getPath(lpc);
	}
	retval += " }]";
	return retval;
    }
    
    public native Object clone();

    /**
     * Get the ClassPathSpec associated with the current team.
     *
     * @return The ClassPathSpec associated with the current team.
     */
    public static native ClassPathSpec current();
}
