/*
 * EventQueue.java
 * A queue for Events
 *
 * Copyright (c) 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Lesser 
 * General Public License.
 * 
 * See the file "license-lesser.terms" for information on 
 * usage and redistribution of this file.
 */

package edu.utah.janosvm.sys;

/**
 * An internal queue/dispatcher for events.
 *
 * @see EventHandler
 * @see Event
 * @author Tim Stack, <a href="mailto:janos-discuss@flux.cs.utah.edu">Utah Janos Team</a>
 */
/*package*/ final class EventQueue
    extends List
    implements Runnable
{
    EventQueue()
    {
    }

    /**
     * Enqueue an event.  Because the handler and type are
     * stored in the event, do not enqueue the same event
     * more than once.
     *
     * @param event An appropriately typed <code>Event</code> object to queue.
     * @param eh The target of the event.
     * @param type The type of event.
     */
    public synchronized void enqueue(Event event, EventHandler eh, int type)
    {
	event.setEvent(type);
	event.setTarget(eh);
	super.addTail(event);
	this.notifyAll();
    }

    /**
     * Fetch (and remove) the head event from this queue.
     *
     * @return An <code>Event</code> from the front of the queue.
     */
    synchronized Event dequeue()
    {
	Event retval;

	while( (retval = (Event)super.remHead()) == null )
	{
	    try
	    {
		this.wait();
	    }
	    catch(InterruptedException e)
	    {
	    }
	}
	return( retval );
    }

    /** 
     * Run the handler thread for this event queue.  The handler
     * thread will dequeue each event and run the event's associated
     * {@linkplain Event#getHandler handler}.
     */
    public void run()
    {
	boolean done = false;

	try
	{
	    while( !done )
	    {
		Event ev;
		
		ev = this.dequeue();
		ev.getTarget().handleEvent(ev);
	    }
	}
	catch(Throwable th)
	{
	    VMInstance.panic("Exception in EventQueue", th);
	}
    }

    public String toString()
    {
	return "EventQueue["
	    + super.toString()
	    + "]";
    }
}
