/*
 * EventTeamHandle.java
 * An extended handle for accessing TeamBackEnd objects and receiving events.
 *
 * Copyright (c) 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Lesser 
 * General Public License.
 * 
 * See the file "license-lesser.terms" for information on 
 * usage and redistribution of this file.
 */

package edu.utah.janosvm.sys;

/**
 * An extended handle for accessing a team and receiving events from
 * it.
 *
 * <p>To get events for a new team, create a new <code>EventTeamHandle</code>, then
 * pass it to {@link Team#create(TeamHandle,String,ResourceSpec[]) Team.create}
 * when you create the team.  Or, to get events for an existing Team,
 * pass a {@link TeamHandle} to the <code>EventTeamHandle</code> constructor.
 *
 * <p>The same {@link TeamEventHandler} may be registered with more
 * than one <code>EventTeamHandle</code>.  
 *
 * @see TeamEvent
 * @see TeamEventHandler
 * @see Team#create(TeamHandle,String,ResourceSpec[])
 * @author Tim Stack, <a href="mailto:janos-discuss@flux.cs.utah.edu">Utah Janos Team</a>
 */
public class EventTeamHandle
    extends TeamHandle
{
    /**
     * Mask of events we're interested in.
     */
    private int eventMask = 0xffffffff;

    /**
     * The user defined event handler.
     */
    private TeamEventHandler teh;

    /**
     * Construct an <code>EventTeamHandle</code> with the given mask and
     * handler for the team associated with the given <code>TeamHandle</code>.
     * The given <code>TeamHandle</code> is copied.
     *
     * @param eventMask Mask of events we're interested in.
     * @param teh The user defined event handler.
     * @param th Handle on the team to associate with.
     */
    public EventTeamHandle(int eventMask, TeamEventHandler teh, TeamHandle th)
    {
	super(th);
	this.eventMask = eventMask;
	this.teh = teh;
    }


    /**
     * Construct an <code>EventTeamHandle</code> with the given mask and handler.
     * The handle will not be attached to any <code>TeamBackEnd</code>, and can 
     * only be attached implicitly by passing the new handle
     * to {@link Team#create}.
     *
     * @param eventMask Mask of events we're interested in.
     * @param teh The user defined event handler.
     */
    public EventTeamHandle(int eventMask, TeamEventHandler teh)
    {
	super();
	this.eventMask = eventMask;
	this.teh = teh;
    }

    /**
     * Construct an <code>EventTeamHandle</code> with the given handler.  The
     * mask is set match to all events, and the handle is
     * not associated with any <code>TeamBackEnd</code>.
     *
     * @param teh The user defined event handler.
     */
    public EventTeamHandle(TeamEventHandler teh)
    {
	this(0xffffffff, teh);
    }

    /**
     * Construct an <code>EventTeamHandle</code> with the given handler
     * and attach it to the given team.
     *
     * @param teh The user defined event handler.
     * @param th Handle on the <code>TeamBackEnd</code> to attach this to.
     */
    public EventTeamHandle(TeamEventHandler teh, TeamHandle th)
    {
	this(0xffffffff, teh, th);
    }

    /**
     * Construct an empty EventTeamHandle.
     */
    public EventTeamHandle()
    {
	this(null);
    }

    /**
     * Set the event handler associated with this handle, overwrites
     * previous handler, if any.
     *
     * @param teh The user defined event handler.
     */
    public synchronized void setEventHandler(TeamEventHandler teh)
    {
	this.teh = teh;
    }

    /**
     * Get the event handler associated with this handle.
     *
     * @return The user defined event handler.
     */
    public synchronized TeamEventHandler getEventHandler()
    {
	return this.teh;
    }

    /**
     * Invoke the callback associated with this handler for the
     * given <code>TeamEvent</code>
     *
     * @param tr A reservation on the team owning this handle.
     * @param te The event.
     */
    final synchronized void event(TeamReservation tr, TeamEvent te)
    {
	if( (this.teh != null) && ((this.eventMask & te.getEvent()) != 0) )
	{
	    this.teh.event(this, tr, te);
	}
    }

    /**
     * Clone this handle, the cloned handle gets a <code>null</code> handler.
     */
    public Object clone()
    {
	Object retval = super.clone();
	EventTeamHandle eth;

	eth = (EventTeamHandle)retval;
	eth.teh = null;
	return retval;
    }
    
    public String toString()
    {
	return "EventTeamHandle["
	    + super.toString()
	    + "; eventMask="
	    + Integer.toHexString(this.eventMask)
	    + "; teh="
	    + this.teh
	    + "]";
    }
}
