/*
 * Team.java
 * Peer object for a team.
 *
 * Copyright (c) 2000, 2001, 2002, 2003 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Lesser 
 * General Public License.
 * 
 * See the file "license-lesser.terms" for information on 
 * usage and redistribution of this file.
 */

package edu.utah.janosvm.sys;

import java.util.Hashtable;

import edu.utah.janosvm.resources.ResourceSpec;
import edu.utah.janosvm.resources.ResourceSpecException;
import edu.utah.janosvm.resources.DuplicateResourceSpecException;

/**
 * Peer object for a TeamBackEnd.  An instance of a Team
 * is created inside each team to contain that team's local
 * data.  Specifically, this class keeps track of the cross-team
 * references
 * coming from its team.  Team objects are never created
 * directly, they are created implicitly as a side-effect 
 * of the static method {@link Team#create}.
 *
 * <p>The association with the TeamBackEnd in the kernel team is
 * implicit (through TeamBackEnd.getCurrentTeam()).
 *
 * <p>Because this class extends TeamHandle, it is importable.
 *
 * <p>A new Team is entirely passive.  No threads are created in the
 * team, nor are any threads required to be created.  Threads from
 * some other team must {@linkplain TeamHandle#switchTo visit} the new
 * team and create threads in it, if an active Team is desired.
 * Because a passive Team is perfectly legitimate, Teams do not
 * automatically exit when all non-daemon threads have exited.  See 
 * {@link Team#notifyOnFinish}.
 *
 * @see TeamBackEnd
 * @see TeamHandle
 * @author Tim Stack, <a href="mailto:janos-discuss@flux.cs.utah.edu">Utah Janos Team</a>
 */
public final class Team
    extends EventTeamHandle
{
    private static TeamHandle create(TeamHandle handle,
				     String name,
				     ResourceSpec[] specs,
				     kaffe.util.Ptr nativeSpecs)
	throws DuplicateExportException,
	       ResourceSpecException,
	       DeadTeamException
    {
	TeamBackEnd kernel = TeamBackEnd.getKernelTeam();
	TeamHandle retval = handle;
	TeamReservation tr;

	try
	{
	    ExportManager em = null;
	    
	    tr = kernel.pushReservation();
	    tr.visit();
	    {
		PortAuthority pa = kernel.getPeer().getPortAuthority();
		
		if( (name != null) && !name.equals("") )
		{
		    name = CloneFactory.clone(name);
		    em = pa.reserveExportID(name);
		}

		TeamBackEnd be;

		try
		{
		    be = new TeamBackEnd(name, specs, nativeSpecs);
		}
		catch(Throwable th)
		{
		    if( em != null )
		    {
			try
			{
			    pa.revokeObject(name);
			}
			catch(NoSuchExportException nsee)
			{
			}
		    }
		    if( th instanceof IndexedResourceSpecException )
		    {
			throw (IndexedResourceSpecException)th;
		    }
		    else if( th instanceof DuplicateResourceSpecException )
		    {
			throw (DuplicateResourceSpecException)th;
		    }
		    else if( th instanceof DeadTeamException )
		    {
			throw (DeadTeamException)th;
		    }
		    else if( th instanceof Error )
		    {
			throw (Error)th;
		    }
		    else if( th instanceof RuntimeException )
		    {
			throw (RuntimeException)th;
		    }
		    else
		    {
			VMInstance.panic("Throw while making " + name, th);
			throw new InternalError(th);
		    }
		}
		
		Team t = be.getPeer();
		
		if( em == null )
		    em = pa.exportAnonymousObject(be);
		be.setExportManager(em);
		em.setExport(be);
		em.addImporter(t);
	    }
	    tr.leave();
	    if( retval == null )
		retval = new TeamHandle();
	    em.addImporter(retval);
	    tr.pop();
	}
	catch(TeamEjectionException e)
	{
	    VMInstance.panic("Ejected from kernel?", e);
	}
	catch(IndexedResourceSpecException e)
	{
	    if (specs == null)
	    {
		throw new ResourceSpecException("null specs "
						+ e.getIndex());
	    }

	    int index = e.getIndex();
	    if (index == specs.length)
		throw new ResourceSpecException("Internal resource allocation failure (see messages)");

	    if ( index < 0 || index > specs.length )
		throw new ResourceSpecException("(error in Team.java) spec index (" +index+ ") out of range...");
	    else
		throw new ResourceSpecException(specs[index]);
	}
	return retval;
    }

    /**
     * Create a new team.  Use the given TeamHandle for the return
     * value, rebinding the handle if necessary.  Give the team the
     * given name (in its export table), and setup the given resource
     * specifications in the new team.  Any of the parameters may be
     * null.  A new TeamHandle will be allocated if null is passed for
     * <code>th</code>.
     *
     * @return a TeamHandle on the new team.
     */
    public static TeamHandle create(TeamHandle th,
				    String name,
				    ResourceSpec[] specs)
	throws DuplicateExportException,
	       ResourceSpecException,
	       DeadTeamException
    {
	return Team.create(th, name, specs, null);
    }

    /**
     * Create a new team. 
     *
     * Equivalent to <code>{@link Team#create(TeamHandle,String,ResourceSpec[]) Team.create}(null, name, specs)</code>.
     */
    public static TeamHandle create(String name, ResourceSpec[] specs)
	throws DuplicateExportException,
	       ResourceSpecException,
	       DeadTeamException
    {
	return Team.create(null, name, specs, null);
    }

    /**
     * Create a new team. 
     *
     * Equivalent to <code>{@link Team#create(TeamHandle,String,ResourceSpec[]) Team.create}(th, null, specs)</code>.
     */
    public static TeamHandle create(TeamHandle th, ResourceSpec[] specs)
	throws ResourceSpecException,
	       DeadTeamException
    {
	try
	{
	    return Team.create(th, null, specs);
	}
	catch(DuplicateExportException e)
	{
	    VMInstance.panic("Duplicate export for an anonymous object?", e);
	    return null;
	}
    }
    
    /**
     * Create a new team. 
     *
     * Equivalent to <code>{@link Team#create(TeamHandle,String,ResourceSpec[]) Team.create}(null, name, null)</code>.
     */
    public static TeamHandle create(String name)
	throws DuplicateExportException,
	       ResourceSpecException,
	       DeadTeamException
    {
	return Team.create(null, name, null);
    }

    /**
     * Create a new team. 
     *
     * Equivalent to <code>{@link Team#create(TeamHandle,String,ResourceSpec[]) Team.create}(null, null, specs)</code>.
     */
    public static TeamHandle create(ResourceSpec[] specs)
	throws ResourceSpecException,
	       DeadTeamException
    {
	try
	{
	    return Team.create(null, null, specs);
	}
	catch(DuplicateExportException e)
	{
	    VMInstance.panic("Duplicate export for an anonymous object?", e);
	    return null;
	}
    }
    
    /**
     * Create a new team. 
     *
     * Equivalent to <code>{@link Team#create(TeamHandle,String,ResourceSpec[]) Team.create}(null, null, null)</code>.
     */
    public static TeamHandle create()
	throws ResourceSpecException,
	       DeadTeamException
    {
	try
	{
	    return Team.create(null, null, null);
	}
	catch(DuplicateExportException e)
	{
	    VMInstance.panic("Duplicate export for an anonymous object?", e);
	    return null;
	}
    }
    
    /**
     * The manager of any cross team references.
     */
    private final PortAuthority pa = new PortAuthority();

    /**
     * Hashtable for TeamLocals.
     */
    private final Hashtable locals = new Hashtable();

    /**
     * The cached kernel team handle.
     */
    private final TeamHandle kernel = new TeamHandle();

    /**
     * A description of the contents of the team.
     */
    private String description;
    
    /**
     * @return The Team object for the current team.
     */
    public static Team current()
    {
	return TeamBackEnd.getCurrentTeam().getPeer();
    }

    /**
     * Get a handle on the Team with the given name.  TeamHandles with the <code>null</code>
     * name cannot be found this way.
     *
     * @param th The {@link TeamHandle} to initialize.
     * @param name The name of the TeamHandle to locate.
     * @return The given <code>TeamHandle</code> <code>th</code>, or <code>null</code>
     * @throws NoSuchExportException if name is <code>null</code> or empty
     */
    public static TeamHandle find(TeamHandle th, String name)
	throws DeadTeamException,
	       TeamEjectionException,
	       NoSuchExportException
    {
	TeamHandle retval = null;

	if( (name != null) && !name.equals("") )
	{
	    Team t = Team.current();
	    PortAuthority pa = t.getPortAuthority();

	    pa.importObject(th, t.getKernelTeam(), name);
	    retval = th;
	}
	else
	{
	    throw new NoSuchExportException("null or empty name");
	}
	return( retval );
    }

    /**
     * Get a handle on the Team with the given name.  TeamHandles with the <code>null</code>
     * name cannot be found this way.
     *
     * @param name The name of the TeamHandle to locate.
     * @return A {@link TeamHandle} object that imports the TeamBackEnd with
     * the given name.
     * @throws NoSuchExportException if name is <code>null</code> or empty
     */
    public static TeamHandle find(String name)
	throws DeadTeamException,
	       TeamEjectionException,
	       NoSuchExportException
    {
	return find(new TeamHandle(), name);
    }

    static TeamHandle find(TeamHandle th, TeamBackEnd be)
    {
	TeamHandle retval = th;
	ExportManager em;

	if( retval == null )
	{
	    retval = new TeamHandle();
	}
	em = be.getExportManager();
	em.addImporter(retval);
	return retval;
    }
    
    Team()
    {
    }
    
    /**
     * Construct a Team object.
     * <p>
     * NOTE: For kernel use only; there is only one of these objects
     * per team.
     */
    Team(ExportManager kernelManager)
    {
	kernelManager.addImporter(this.kernel);
    }

    void setKernelHandle(ExportManager kernelManager)
    {
	kernelManager.addImporter(this.kernel);
    }

    /**
     * @param str A string that describes this team.
     */
    public synchronized void setDescription(String str)
    {
	this.description = CloneFactory.clone(str);
	this.be.setDescription(this.description);
    }

    /**
     * @return A string that describes this team.
     */
    public synchronized String getDescription()
    {
	return this.description;
    }
    
    /**
     * Get a handle on the kernel team.
     *
     * @return The cached kernel Team handle.
     */
    public TeamHandle getKernelTeam()
    {
	return this.kernel;
    }

    /**
     * Get the {@link PortAuthority} for the current team.
     *
     * @return The {@link PortAuthority} for the current team.
     */
    public PortAuthority getPortAuthority()
    {
	return this.pa;
    }

    /** @see PortAuthority#reserveExportID */
    public ExportManager reserveExportID(Object id)
	throws DuplicateExportException
    {
	return this.pa.reserveExportID(id);
    }
    
    /** @see PortAuthority#exportObject */
    public ExportManager exportObject(Object id, Exportable ex)
	throws DuplicateExportException
    {
	return this.pa.exportObject(id, ex);
    }

    /**
     * XXX rename to revokeExport
     */
    public void revokeObject(Object id)
	throws NoSuchExportException
    {
	this.pa.revokeObject(id);
    }
    
    public ExportManager exportAnonymousObject(Exportable ex)
    {
	return this.pa.exportAnonymousObject(ex);
    }
    
    public void revokeAnonymousObject(Exportable ex)
	throws NoSuchExportException
    {
	this.pa.revokeAnonymousObject(ex);
    }
    
    /**
     * @return The local data for the current team.
     */
    Hashtable getLocals()
    {
	return this.locals;
    }

    /**
     * Associate the given object with the given class in this team.
     *
     * @param cl The class the object should be stored with.
     * @param obj The object to store with the class.
     * @see #getClassLocal
     * @see #setCurrentClassLocal
     */
    public synchronized void setClassLocal(Class cl, Object obj)
    {
	if( this.be != null )
	    this.be.setClassLocal(cl, obj);
	else
	    throw new NullPointerException();
    }

    /**
     * Associate the given object with the given class in the current team.
     *
     * @param cl The class the object should be stored with.
     * @param obj The object to store with the class.
     * @see #getCurrentClassLocal
     * @see #setClassLocal
     */
    public static void setCurrentClassLocal(Class cl, Object obj)
    {
	TeamBackEnd.setCurrentClassLocal(cl, obj);
    }
    
    /**
     * Get the team-local object associated with the given class in
     * this team.
     *
     * @param cl The class to retrieve the local object for.
     * @return The object associated with the class or <code>null</code>.
     * @see #getCurrentClassLocal
     * @see #setClassLocal
     */
    public synchronized Object getClassLocal(Class cl)
    {
	if( this.be != null )
	    return this.be.getClassLocal(cl);
	else
	    throw new NullPointerException();
    }

    /**
     * Get the team-local object associated with the given class in
     * the current team.
     *
     * @param cl The class to retrieve the local object for.
     * @return The object associated with the class or <code>null</code>.
     * @see #getClassLocal
     * @see #setCurrentClassLocal
     */
    public static Object getCurrentClassLocal(Class cl)
    {
	return TeamBackEnd.getCurrentClassLocal(cl);
    }

    /**
     * Add a dependency to this team.
     *
     * @param th The team this one should depend on.
     * @return True if the addition succeeded, false otherwise.
     */
    public synchronized boolean addDependency(TeamHandle th)
    {
	boolean retval = false;

	if( this.be != null )
	{
	    retval = this.be.addDependency(th.be);
	}
	return retval;
    }

    /**
     * Set the flag on the team object that indicates that the last
     * non-daemon thread in the team should send out a
     * TeamEvent.FINISHED event when it dies.  By registering an
     * {@link TeamEventHandler} with this Team,
     * termination-on-last-non-daemon-thread-exiting semantics can be
     * achived by terminating the team when that event is received.
     *
     * @see TeamEvent
     * @see EventTeamHandle
     */
    public synchronized void notifyOnFinish()
    {
	if( this.be != null )
	{
	    this.be.notifyOnFinish();
	}
    }

    /**
     * Send a user-defined event to all EventTeamHandles that are
     * registered to listen for user events.
     *
     * @param type must be one of {@link TeamEvent#USER1}, {@link
     * TeamEvent#USER2}, or {@link TeamEvent#USER3}
     */
    public synchronized void sendEvent(int type)
    {
	if( (type & TeamEvent.USER_MASK) == 0 )
	{
	    throw new IllegalArgumentException(
		"Must pass a user-event flag to sendUserEvent()");
	}
	if (this.be != null)
	{
	    try
	    {
		this.kernel.switchTo();
		{
		    this.be.event(type);
		}
		this.kernel.returnFrom();
	    }
	    catch(DeadTeamException e)
	    {
	    }
	    catch(TeamEjectionException e)
	    {
	    }
	}
    }

    public String toString()
    {
	StringBuffer sbuf = new StringBuffer(512);

	sbuf.append("Team[")
	    .append(super.toString())
	    .append("; pa=");
	try
	{
	    sbuf.append(this.pa.toString());
	}
	catch (SegmentationFaultError sfe)
	{
	    sbuf.append("<unviewable>");
	}

	sbuf.append("; locals=");
	try
	{
	    sbuf.append(this.locals.toString());
	}
	catch (SegmentationFaultError sfe)
	{
	    sbuf.append("<unviewable>");
	}
	
	sbuf.append("; kernel=");
	try
	{
	sbuf.append(this.kernel.toString());
	}
	catch (SegmentationFaultError sfe)
	{
	    sbuf.append("<unviewable>");
	}

	sbuf.append("; description=");
	try
	{
	    if (this.description == null)
		sbuf.append("<null>");
	    else
		sbuf.append(this.description.toString());
	}
	catch (SegmentationFaultError sfe)
	{
	    sbuf.append("<unviewable>");
	}

	sbuf.append("]");

	return sbuf.toString();
    }
}
