/*
 * TeamBackEnd.java
 * The class used to represent a team
 *
 * Copyright (c) 2000, 2001, 2002, 2003 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Lesser 
 * General Public License.
 * 
 * See the file "license-lesser.terms" for information on 
 * usage and redistribution of this file.
 */

package edu.utah.janosvm.sys;

import java.util.Stack;
import java.util.Vector;
import java.util.EmptyStackException;

import edu.utah.janosvm.resources.ResourceSpec;
import edu.utah.janosvm.resources.DuplicateResourceSpecException;

/**
 * Kernel class used to represent a team, this is then extended internally
 * with a nativeClass.  The class is an Exportable since all instances
 * of TeamBackEnd are allocated in the kernel team.  The team they represent
 * has a Team instance associated with the TeamBackEnd.
 *
 * @see Team
 * @see TeamHandle
 * @author Tim Stack, <a href="mailto:janos-discuss@flux.cs.utah.edu">Utah Janos Team</a>
 */
class TeamBackEnd
    extends Exportable
    implements EventHandler
{
    /**
     * Version info.
     */
    private static final String _versionName = "JanosVM";
    private static final int _versionMajor = 0;
    private static final int _versionMinor = 8;

    /**
     * Reference to the kernel team.
     */
    private static TeamBackEnd kernelTeam;
    static final Team kernelTeamPeer = new Team();
    private static ExportManager kernelManager;
    
    public static final String kernelName = "kernel";

    /**
     * Queue for events handled by the kernel.
     */
    static final EventQueue eq = new EventQueue();

    /**
     * Fixed-size pool of pre-allocated TeamEvent objects.
     */
    static final Stack eventPool = new Stack();

    static native void staticNativeInit(EventQueue eq, Thread th);

    /**
     * @return An a TeamEvent object from the pool.
     */
    static TeamEvent allocEvent()
    {
	TeamEvent retval = null;
	
	synchronized( eventPool )
	{
	    while( retval == null )
	    {
		try
		{
		    retval = (TeamEvent)eventPool.pop();
		}
		catch(EmptyStackException e)
		{
		    retval = new TeamEvent();
		}
	    }
	}
	return retval;
    }

    /**
     * @param te A TeamEvent allocated with allocEvent.
     */
    static void freeEvent(TeamEvent te)
    {
	te.setTarget(null);
	synchronized( eventPool )
	{
	    eventPool.push(te);
	}
    }

    /**
     * @return The kernel TeamBackEnd object.
     */
    static TeamBackEnd getKernelTeam()
    {
	return TeamBackEnd.kernelTeam;
    }

    /**
     * Indicates whether or not the team is being terminated.
     */
    private boolean dying = false;

    /**
     * The ImportEnumerator for our ExportManager.
     */
    private ExportManager.ImportEnumerator ie;

    /**
     * The ExportManager for our object.
     */
    private ExportManager em;

    /**
     * The object local to the team's heap that contains
     * per-team state specific to that team.  Specifically,
     * it contains the tables of items exported from the team.
     */
    private Team peer;

    private TeamEvent cpuOverdrawEvent = new TeamEvent();
    
    /**
     * Gets the Team peer associated with this TeamBackEnd.
     *
     * @return The peer object.
     */
    Team getPeer()
    {
	return this.peer;
    }

    private String name;

    /**
     * @return The name of the team.
     */
    String getName()
    {
	return this.name;
    }

    /**
     * Constructs the native part of the class.
     *
     * @param name The name of this team.
     * @param specs The resources specifications for the team.
     * @return The index of the ResourceSpec that couldn't be satisfied.
     */
    private native int nativeConstructor(String name,
					 ResourceSpec specs[],
					 kaffe.util.Ptr nativeSpecs)
	throws DuplicateResourceSpecException;

    /**
     * Tests to see if an object is owned by this team.
     * (Ownership is defined as paying for the memory used by
     * the object instance.)
     *
     * @param obj The object to test for ownership.
     * @return True if this TeamBackEnd owns the given object.
     */
    native boolean owns(Object obj);

    static native TeamBackEnd getObjectOwner(Object obj);
    
    /**
     * Start a visit.  A visit is completed with a matching call
     * to <code>returnFrom()</code>.
     *
     * @param wrapCount Indicates the number of wrapper functions around this
     * one.  This is required so that we can record where visit take place and
     * can properly back out of them if an exception is thrown.
     * @return Whether the visit was successful or not.
     * @see #returnFrom()
     */
    native boolean switchTo(int wrapCount) throws TeamEjectionException;

    /**
     * Basic visit wrapper provided for code that operates directly on
     * TeamBackEnds.
     *
     * @return Whether the visit was successful or not.
     */
    boolean switchTo()
	throws TeamEjectionException
    {
	return this.switchTo(1);
    }
    
    /**
     * End a visit and return to the previous team.
     * <p>
     * XXX this shouldn't be static.  Just to make the symmetry a bit
     * more apparent, if nothing else.  Also would be a bit easier
     * to add sanity checking code with a this pointer.
     */
    static native void returnFrom();

    /**
     * Get the TeamBackEnd associated with the current thread.
     * This returns whatever team the thread is currently executing
     * in (i.e., during a visit), and is not necessairly the home
     * team of the thread.
     *
     * @return The team that this thread is currently running in.
     */
    static native TeamBackEnd getCurrentTeam();

    /**
     * Get the unique integer identifier for this team.
     * Identifiers are unique among currently executing
     * teams, but may be re-used when a team is destroyed.
     * (Analogous to a "PID" in unix.)
     *
     * @return The unique integer identifier for this team.
     */
    native int getID();

    /**
     * @return The number of threads in this team.
     */
    native int getThreadCount();
    
    /**
     * @return The number of daemon threads in this team.
     */
    native int getDaemonCount();
    
    /**
     * @param desc A description of the contents of this team.
     */
    native void setDescription(String desc);
    
    /**
     * Anchor an object so the collector doesn't free it.  The
     * object becomes a "root" in the collector so the collector
     * will never free it.  The object must be explicitly
     * un-anchored before it will be freed.
     *
     * @see #floatObject(Object)
     */
    private native void anchorObject(Object obj);

    /**
     * Float a previously anchored object.  The object
     * will be reclaimed when it is unreachable.
     *
     * @see #anchorObject(Object)
     */
    private native void floatObject(Object obj);
    
    /**
     * Constructor used for the kernel team.
     */
    private TeamBackEnd()
    {
	super(0);
	
	TeamBackEnd.kernelTeam = this;
	/*
	 * The native structure will setup the resources and migrate us
	 * to this team.
	 */
	try
	{
	    PortAuthority pa;
	    Thread reaper;
	    Thread eventThread;

	    this.name = kernelName;
	    this.nativeConstructor(kernelName, null, null);
	    this.peer = kernelTeamPeer;

	    pa = this.peer.getPortAuthority();
	    this.em = pa.exportObject(kernelName, this);
	    this.em.addImporter(this.peer);

	    TeamBackEnd.kernelManager = em;
	    this.peer.setKernelHandle(em);

	    eventThread = new Thread(eq, "Kernel EventThread Thread");
	    eventThread.setDaemon(true);

	    staticNativeInit(eq, eventThread);
	    
	    eventThread.start();
	}
	catch(Throwable th)
	{
	    VMInstance.panic("Caught exception while trying to create the "
			     + "kernel team:",
			     th);
	}
    }

    /**
     * Native function that collects and returns the team's resources to the
     * VM.
     */
    private native void doTheReap();

    /**
     * Trigger an event notification for this team.
     *
     * @param event The TeamEvent number that indicates what event this is.
     */
    void event(int event)
    {
	eq.enqueue(allocEvent(), this, event);
    }

    /**
     * Send a TeamEvent.DEAD event.
     */
    void deadTeamEvent()
    {
	this.event(TeamEvent.DEAD);
    }
    
    /**
     * Send a TeamEvent.FINISHED event.
     */
    void finishTeamEvent()
    {
	this.event(TeamEvent.FINISHED);
    }
    
    /**
     * Create a user team with the given name and resources.  This is the
     * internal version that can be used by the native code to construct a
     * team with native resource specifications.
     *
     * @param name The name of the team.
     * @param specs The specifications required by the team.
     * @param nativeSpecs Native array of resource specifications.
     * @exception DuplicateResourceSpecException if the specs array
     * contains more than one ResourceSpec for the same resource.
     * @exception IndexedResourceSpecException if the one of the
     * ResourceSpec's couldn't be satisfied.
     */
    TeamBackEnd(String name,
		ResourceSpec specs[],
		kaffe.util.Ptr nativeSpecs)
	throws DuplicateResourceSpecException,
	       IndexedResourceSpecException,
	       DeadTeamException
    {
	int index;

	if( name != null )
	    name = (String)name.clone();
	if( (index = this.nativeConstructor(name, specs, nativeSpecs)) < 0 )
	{
	    try
	    {
		Vector local = null;
		Team team = null;
		
		if( this.switchTo() )
		{
		    /*
		     * NOTE: We are now in the new team, but the
		     * team object is held in kernel space, so any
		     * memory allocated and referenced by the
		     * TeamBackEnd object will not be traced by the
		     * new team's collector.  Therefore, any
		     * objects created must be anchored or they
		     * might be freed.
		     */
		    if( (name != null) && !name.equals("") )
		    {
			this.name = name;
		    }
		    
		    /* Create the peer object */
		    this.peer = new Team(kernelManager);
		    this.anchorObject(this.peer);

		    this.cpuOverdrawEvent.setTarget(this);
		    this.cpuOverdrawEvent.setEvent(TeamEvent.CPU_OVERDRAW);

		    try { System.loadLibrary("native"); }
		    catch(Throwable th) { }
		    try { System.loadLibrary("kaffevm"); }
		    catch(Throwable th) { }

		    TeamBackEnd.returnFrom();
		}
		else
		{
		    /* Ooph, we got nailed pretty quickly... */
		    throw new DeadTeamException();
		}
	    }
	    catch(InstantiationError e)
	    {
		throw new DeadTeamException();
	    }
	    catch(TeamEjectionException e)
	    {
		throw new DeadTeamException();
	    }
	    catch(OutOfMemoryError e)
	    {
		throw e;
	    }
	    catch(Throwable th)
	    {
		VMInstance.panic("Throw while making " + name, th);
	    }
	}
	else
	{
	    /*
	     * XXX Actually say what went wrong instead of just throwing
	     * this.
	     */
	    throw new IndexedResourceSpecException(index);
	}
    }

    /**
     * Native level code to mark the start of termination.
     */
    private native void nativeTerminating();

    /**
     * Native level cage code.
     */
    private native void nativeCage();

    /**
     * Terminate this team, revoke all its exports, kill its threads and
     * return its resources to the system.
     * <p>
     * If the calling thread is part of the team being terminated
     * then this function will not return.
     * <p>
     * Note that termination is delayed by any active visits from
     * other teams, so to provide quick termination all visits
     * must be "short".
     */
    void terminate()
    {
	/* Need to be in the kernel to do this */
	try
	{
	    if( this == TeamBackEnd.kernelTeam )
	    {
		VMInstance.shutdown();
	    }
	    else if( TeamBackEnd.kernelTeam.switchTo() )
	    {
		/*
		 * Only the first time the Team is terminated is anyone
		 * responsible for cleaning up the external references.
		 */
		synchronized(this)
		{
		    if( this.dying == false )
		    {
			this.dying = true;
			this.nativeTerminating();
			this.event(TeamEvent.TERMINATED);
		    }
		}
		TeamBackEnd.returnFrom();
	    }
	}
	catch(TeamEjectionException dte)
	{
	    /* Our team is dead, but not buried */
	}
	catch(Throwable th)
	{
	    System.debug("TeamBackEnd.terminate: Caught unexpected exception:");
	    System.debugE(th);
	}
    }

    /**
     * @param wrapCount Indicates the number of wrapper functions around this
     * one.  This is required so that we can record where pushes take place and
     * can properly back out of them if an exception is thrown.
     * @return A TeamReservation object that reserves this team.
     */
    native TeamReservation pushReservation(int wrapCount)
	throws DeadTeamException;

    TeamReservation pushReservation()
	throws DeadTeamException
    {
	return this.pushReservation(1);
    }
    
    /**
     * Pop the last TeamReservation off the stack.
     *
     * <b>NOTE:</b> the given <code>tr</code> is just used as a sanity
     * check to make sure an intervening reservation hasn't been
     * skipped.
     */
    static native void popReservation(TeamReservation tr);

    /**
     * Store a per-team class local object.
     *
     * @param cl The class the object should be stored with.
     * @param obj The object to store with the class.
     */
    native void setClassLocal(Class cl, Object obj);
    
    static native void setCurrentClassLocal(Class cl, Object obj);
    
    /**
     * @param cl The class to retrieve the local object for.
     * @return The Object associated with the class or null.
     */
    native Object getClassLocal(Class cl);

    static native Object getCurrentClassLocal(Class cl);

    /**
     * Add a dependency to this team.
     *
     * @param th The team this one should depend on.
     * @return True if the addition succeeded, false otherwise.
     */
    native boolean addDependency(TeamBackEnd dep);

    /**
     * Set the flag on the team object that indicates that the last non-daemon
     * thread in the team should send out a TeamEvent.FINISHED event.
     */
    native void notifyOnFinish();

    /**
     * @param em The object thats managing this export.
     */
    synchronized void setExportManager(ExportManager em)
    {
	this.em = em;
	this.ie = em.getEnumerator();
    }

    /**
     * @return The object thats managing this export.
     */
    synchronized ExportManager getExportManager()
    {
	return this.em;
    }

    /**
     * Broadcast a TeamEvent to the list of objects importing this object.
     *
     * @param te The event to broadcast.
     */
    private void broadcastTeamEvent(TeamEvent te)
    {
	ExportManager man;

	synchronized( this )
	{
	    man = this.em;
	}
	if( man != null )
	{
	    synchronized( this.ie )
	    {
		/* Put the Enumerator on the imports list */
		man.imports(this.ie);
		/* Traverse the list of Importable's */
		while( this.ie.nextImport() )
		{
		    try
		    {
			Importable im = this.ie.getImportable();
			
			if( im instanceof EventTeamHandle )
			{
			    EventTeamHandle eth = (EventTeamHandle)im;
			    
			    /* Do the call back */
			    eth.event(this.ie.getReservation(), te);
			}
		    }
		    catch(Throwable th)
		    {
			/*
			 * XXX Its a call back, i don't care, well log it i
			 * guess
			 */
			System.err.println("Error in TeamEvent call back:");
			th.printStackTrace(System.err);
		    }
		}
	    }
	}
    }

    public void handleEvent(Event ev)
    {
	TeamEvent te;

	te = (TeamEvent)ev;
	this.broadcastTeamEvent(te);
	switch( ev.getEvent() )
	{
	case TeamEvent.TERMINATED:
	    /* Visit the team and revoke its imports and exports */
	    try
	    {
		if( this.switchTo() )
		{
		    try
		    {
			PortAuthority pa;
			
			pa = this.peer.getPortAuthority();
			/* Cut external refs */
			PortAuthority.revokeAllImports();
			pa.revokeAllExports();
			this.peer = null;
		    }
		    catch(Throwable th)
		    {
			TeamBackEnd.returnFrom();
			// XXX:
			VMInstance.panic("Throw during termination of:"
					 + this,
					 th);
		    }
		    
		    /*
		     * Return from the team being terminated.  If there are no
		     * other active visits into (and no absent threads from)
		     * the team, it will start dying in this call.
		     */
		    TeamBackEnd.returnFrom();
		}
		else
		{
		    VMInstance.panic("Unable to visit team I'm terminating?");
		}
	    }
	    catch(TeamEjectionException e)
	    {
		VMInstance.panic("Ejected during revocation?", e);
	    }
	    /* Let the native code do the rest */
	    kernelTeam.anchorObject(this);
	    try
	    {
		this.nativeCage();
	    }
	    catch(Throwable th)
	    {
		VMInstance.panic("Throw during nativeCage", th);
	    }
	    break;
	case TeamEvent.DEAD:
	    /* Clean up the resources. */
	    this.doTheReap();
	    /* Notify the importers that the team is now buried. */
	    ev.setEvent(TeamEvent.BURIED);
	    this.broadcastTeamEvent(te);
	    /* Revoke any references to us. */
	    try
	    {
		if( this.name != null )
		{
		    kernelTeam.peer.getPortAuthority().revokeObject(this.name);
		    this.name = null;
		}
		else
		{
		    kernelTeam.peer.getPortAuthority().
			revokeAnonymousObject(this);
		}
	    }
	    catch(NoSuchExportException e)
	    {
	    }
	    break;
	case TeamEvent.CPU_OVERDRAW:
	    this.cpuOverdrawEvent = te;
	    te = null;
	    break;
	}
	if( te != null )
	{
	    freeEvent(te);
	}
    }

    public String toString()
    {
	return "TeamBackEnd["
	    // + super.toString()
	    + "; dying="
	    + this.dying
	    + "; peer="
	    + System.identityHashCode(this.peer)
	    + "; name="
	    + this.name
	    + "]";
    }
}
