/*
 * TeamEvent.java
 * Simple class that describes a team related event.
 *
 * Copyright (c) 2001, 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Lesser 
 * General Public License.
 * 
 * See the file "license-lesser.terms" for information on 
 * usage and redistribution of this file.
 */

package edu.utah.janosvm.sys;

/**
 * Simple class that describes a team related event.  Events are
 * delivered to registered {@link TeamEventHandler}s.  A <tt>TeamEventHandler</tt>
 * is registered with an {@link EventTeamHandle}, which is a special
 * TeamHandle.
 *
 * @see TeamEventHandler#event
 * @author Tim Stack, <a href="mailto:janos-discuss@flux.cs.utah.edu">Utah Janos Team</a>
 */
public final class TeamEvent
    extends Event
{
    private static final int MIN				=  0;
    private static final int B_UNCAUGHT_EXCEPTION		=  1;
    private static final int B_ADD_DEPENDENCY			=  2;
    private static final int B_REM_DEPENDENCY			=  3;
    private static final int B_FINISHED				=  4;
    private static final int B_TERMINATED			=  5;
    private static final int B_DEAD				=  6;
    private static final int B_BURIED				=  7;
    private static final int B_OVERDRAW				=  8;
    private static final int B_CPU				=  9;
    private static final int B_MEMORY				= 10;
    private static final int B_NETWORK				= 11;
    private static final int B_DISK				= 12;
    private static final int B_IPC				= 13;
    // ...
    private static final int B_USER3				= 29;
    private static final int B_USER2				= 30;
    private static final int B_USER1				= 31;

    /**
     * Default event state.
     */
    public static final int NONE         	= 0;

    /**
     * Not yet used.
     */
    public static final int UNCAUGHT_EXCEPTION	= (1 << B_UNCAUGHT_EXCEPTION);

    /**
     * A dependency on this Team has been added.
     *
     * @see Team#addDependency
     */
    public static final int ADD_DEPENDENCY	= (1 << B_ADD_DEPENDENCY);

    /**
     * A Dependency on this Team has been removed. 
     *
     * <p>XXX never used.
     */
    public static final int REM_DEPENDENCY	= (1 << B_REM_DEPENDENCY);

    /**
     * Sent when a Team has no more runnable threads in it.
     */
    public static final int FINISHED		= (1 << B_FINISHED);

    /**
     * Sent when a Team is being terminated.  The sequence of events is:
     *   1. Send event and wait for all call backs to finish.
     *   2. Revoke all existing imports/exports and disallow new ones.
     *   3. Drop the cage on the team.
     */
    public static final int TERMINATED		= (1 << B_TERMINATED);

    /**
     * Sent when a Team is dead, all visiting threads have left, and
     * all absent threads have returned.
     */
    public static final int DEAD		= (1 << B_DEAD);

    public static final int CPU_OVERDRAW	= ((1 << B_OVERDRAW) |
						   (1 << B_CPU));

    /**
     * Sent when a Team is buried.  When this has been sent, all of
     * the Team's threads are gone, all its resources have been
     * released, etc.
     */
    public static final int BURIED		= (1 << B_BURIED);

    public static final int USER1		= (1 << B_USER1);
    public static final int USER2		= (1 << B_USER2);
    public static final int USER3		= (1 << B_USER3);

    static final int USER_MASK = (USER1 | USER2 | USER3);
    
    public TeamEvent()
    {
	super.event = NONE;
    }

    /** 
     * Override getEvent so that non-kernel teams can invoke it
     * (Event superclass is not exported).
     */
    public int getEvent()
    {
	int rc = super.getEvent();
	return rc;
    }

    public String toString()
    {
	String evStr;

	switch(super.event)
	{
	case NONE:
	    evStr = "None";
	    break;
	case UNCAUGHT_EXCEPTION:
	    evStr = "Except";
	    break;
	case ADD_DEPENDENCY:
	    evStr = "Add Dep";
	    break;
	case REM_DEPENDENCY:
	    evStr = "Rem Dep";
	    break;
	case FINISHED:
	    evStr = "Finish";
	    break;
	case TERMINATED:
	    evStr = "Term";
	    break;
	case DEAD:
	    evStr = "Dead";
	    break;
	case BURIED:
	    evStr = "Buried";
	    break;
	case USER1:
	    evStr = "User1";
	    break;
	case USER2:
	    evStr = "User2";
	    break;
	case USER3:
	    evStr = "User3";
	    break;
	default:
	    evStr = "<UNDEFINED: " +super.event+ ">";
	    break;
	}

	return "TeamEvent["
	    + super.toString()
	    + "; event="
	    + evStr
	    + "]";
    }
}
