/*
 * TeamReservation.java
 * Manages team reservation/migration.
 *
 * Copyright (c) 2000, 2001, 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Lesser 
 * General Public License.
 * 
 * See the file "license-lesser.terms" for information on 
 * usage and redistribution of this file.
 */

package edu.utah.janosvm.sys;

/**
 * Record a reservation with another team.  A TeamReservation is a
 * link between two teams that prevents either from being terminated.
 * A TeamReservation is more flexible than {@link
 * TeamHandle#switchTo}, as the termination-delaying time can be
 * expanded to include more than the context switch to the other team.
 *
 * <p>While an active reservation delays termination of two teams,
 * there are two conditions that can cause a {@link TeamEjectionException} 
 * during a visit:
 * <ul>
 *   <li>A termination request while blocked in {@link Object#wait}.
 *   <li>A failed memory allocation, or failed exception dispatch.
 * </ul>
 *
 * <p>If a thread blocks on <tt>Object.wait</tt>, that will re-enable
 * termination of the Teams involved.  If the team is terminated
 * during a <tt>wait</tt>, an exception will be thrown to back the
 * thread out of the visit.  (Most likely a {@link
 * TeamEjectionException}.)
 *
 * <p>If a thread tries to allocate memory during a visit, and the
 * current Team has reached its memory limits, the allocation will
 * throw a <tt>TeamEjectionException</tt>.  This exception may also
 * occur during a <tt>throw</tt> statement.  In all cases, such an
 * exception implies that the visited Team will be terminated.
 * 
 * <p>XXX need something about inter-Team pointers and how to be
 * careful {@link Object#clone}, etc...
 *
 * <p><b>CAUTION:</b> TeamReservation's must be created and popped at
 * the same scope.  Ideally, there would be language support for
 * TeamReservations that works like <code>synchronized</code> does for
 * restricting monitor lock/unlock to the same scope.  Of course, you should
 * always use visit/leave in pairs, inside the push/pop, and you shouldn't
 * use the TeamReservation after a pop.
 *
 * <p>Here's an example:<br>
 * <pre>
 *      try {
 *          TeamReservation resv;
 *          Object obj;
 *          resv = {@link TeamHandle#pushReservation otherTeam.pushReservation}();
 *          
 *          resv.{@link #visit visit}();
 *          // Do things in the context of 'otherTeam'
 *          obj = lookupSimpleFoo();
 *          resv.{@link #leave leave}();
 *          
 *          // Do things in original context, but guarantee that
 *          // otherTeam will not go away ... Reservation still active. 
 *          
 *          Object localObj = new Foo(obj);
 *          
 *          // End the Reservation, as we're done with all pointers into otherTeam
 *          // (Assuming the Foo constructor didn't do anything stupid.)
 *          resv.{@link #pop pop}();
 *      } catch ({@link DeadTeamException} e) {
 *          // pushReservation failed
 *      } catch ({@link TeamEjectionException} e) {
 *          // allocation/throw failed during visit
 *      }
 * </pre>
 *
 *
 * @see Team#pushReservation()
 * @see Team#popReservation()
 * @see TeamHandle#switchTo()
 * @author Tim Stack, <a href="mailto:janos-discuss@flux.cs.utah.edu">Utah Janos Team</a>
 */
public class TeamReservation
{
    /**
     * Hide the constructor.  No one can allocate a TeamReservation,
     * as the JVM manufactures them in special memory in response to a
     * <code>Team.pushReservation()</code>.
     *
     * @see Team#pushReservation()
     */
    private TeamReservation()
    {
    }

    /**
     * Switch the current thread to the team encoded in this Reservation.
     *
     * @see #leave
     */
    public native void visit() throws TeamEjectionException;

    /**
     * Return to the previous team.
     *
     * @see #visit
     */
    public static native void leave();

    /**
     * Pop this TeamReservation off the stack.  This is no longer a
     * useful object after being popped.
     *
     * <b>NOTE:</b> TeamReservations must be popped in the order they
     * were {@linkplain Team#pushReservation pushed}.  More over, they
     * should be popped in the same scope as they were pushed.
     */
    public void pop()
    {
	// XXX pass 'this' as a sanity check
	TeamBackEnd.popReservation(this);
    }

    public String toString()
    {
	return "TeamReservation["
	    + "]";
    }
}
