/*
 * VMInstance.java
 * Functions specific to the VM as a whole
 *
 * Copyright (c) 2001, 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Lesser 
 * General Public License.
 * 
 * See the file "license-lesser.terms" for information on 
 * usage and redistribution of this file.
 */

package edu.utah.janosvm.sys;

/**
 * Functions specific to the VM as a whole.
 *
 * @author Tim Stack, <a href="mailto:janos-discuss@flux.cs.utah.edu">Utah Janos Team</a>
 */
public final class VMInstance
{
    /**
     * Crash into the team boundary, this is basically a violent exit from
     * a visit and results in a DeadTeamException being thrown in the
     * previous team context.  NOTE: This is only used for testing, its
     * possible for the VM to do something similar so we needed a way to
     * explicitly cause it.
     */
    public static native void crash();
    
    public static void panic(String msg)
    {
	VMInstance.panic(msg, null);
    }

    public static void panic(String msg, Throwable th)
    {
	String exceptionMsg = th.toString();
	TeamBackEnd kernel = TeamBackEnd.getKernelTeam();
	try
	{
	    kernel.switchTo();
	    System.err.println("Panic: " + msg);
	    if (th != null)
	    {
		try
		{
		    th.printStackTrace(System.err);
		}
		catch (SegmentationFaultError sfe)
		{
		    System.err.println("<sfe printing backtrace: msg=" +exceptionMsg+ ">");
		}
	    }
	    kernel.returnFrom();
	}
	catch (TeamEjectionException tee)
	{
	    //... I hate it when panic fails...
	    throw new InternalError("VMInstance.panic() failed.", th);
	}
	shutdown();
    }
    
    /**
     * Print a message to the VM's console.  This is really only for
     * debugging or for panic-type messages.
     */
    public static void println(String s)
    {
	TeamBackEnd kernel = TeamBackEnd.getKernelTeam();
	try
	{
	    kernel.switchTo();
	    System.err.println(s);
	    kernel.returnFrom();
	}
	catch (TeamEjectionException tee)
	{
	    VMInstance.panic("VMInstance.println failed", tee);
	}
    }
    
    /**
     * Print a message to the VM's console.  This is really only for
     * debugging or for panic-type messages.
     */
    public static void println(Throwable th)
    {
	TeamBackEnd kernel = TeamBackEnd.getKernelTeam();
	try
	{
	    kernel.switchTo();
	    th.printStackTrace(System.err);
	    kernel.returnFrom();
	}
	catch (TeamEjectionException tee)
	{
	    VMInstance.panic("VMInstance.println failed", tee);
	}
    }
    
    /**
     * Print a message to the VM's console.  This is really only for
     * debugging or for panic-type messages.
     */
    public static void print(String s)
    {
	TeamBackEnd kernel = TeamBackEnd.getKernelTeam();
	try
	{
	    kernel.switchTo();
	    System.err.print(s);
	    kernel.returnFrom();
	}
	catch (TeamEjectionException tee)
	{
	    VMInstance.panic("VMInstance.print failed", tee);
	}
    }

    /**
     * Shutdown the VM.
     */
    public static native void shutdown();

    /**
     * Get the name of this VM instance.
     */
    public static native String getName();

    /**
     * This is a static only object.
     */
    private VMInstance()
    {
    }
}
