/*
 * Java core library component.
 *
 * Copyright (c) 1997, 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 */

package java.lang;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.FileDescriptor;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.PrintStream;
import java.util.Properties;
import java.util.Enumeration;
import java.util.PropertyPermission;

import edu.utah.janosvm.sys.Team;
import edu.utah.janosvm.sys.TeamLocal;
import edu.utah.janosvm.sys.TeamReservation;
import edu.utah.janosvm.sys.DeadTeamException;
import edu.utah.janosvm.sys.TeamEjectionException;

import edu.utah.janosvm.io.StdOutputStream;
import edu.utah.janosvm.io.StdinInputStream;

/**
 * @janosvm Per-team properties.  Copied on first use, so you may have to force first-use
 * if you want the copy to happen earlier.
 * @janosvm Exit (via {@link Runtime#exit} is per-team.
 */
public final class System {
	final public static InputStream in;
	final public static PrintStream out;
	final public static PrintStream err;

	/*
	 * What follows is some getTeamLocal_*() stuff.  Basically, it allows
	 * us to make the above global variables per-team variables.
	 */
	private static final TeamLocal stdin = new TeamLocal() {
		protected Object initialValue()
		{
		    return new BufferedInputStream(new StdinInputStream(),
						   128);
		}
	    };
	
	private static InputStream getTeamLocal_in()
	{
		return (InputStream)stdin.get();
	}

	private static final TeamLocal stdout = new TeamLocal() {
		protected Object initialValue()
		{
		    return new PrintStream(new StdOutputStream(
			StdOutputStream.STDOUT), true);
		}
	    };
	
	private static PrintStream getTeamLocal_out()
	{
		PrintStream retval;

		retval = (PrintStream)stdout.get();
		return retval;
	}

	private static final TeamLocal stderr = new TeamLocal() {
		protected Object initialValue()
		{
		    return new PrintStream(new StdOutputStream(
			StdOutputStream.STDERR), true);
		}
	    };
	
	private static PrintStream getTeamLocal_err()
	{
		PrintStream retval;

		retval = (PrintStream)stderr.get();
		return retval;
	}

	private static final TeamLocal security = new TeamLocal();

	/**
	 * Type for storing per-Team properties.  defaultProperties
	 * are used for the kernel team, and per-team props are
	 * copied from the kernel at "first use".
	 */
	private static class LocalProperties
		extends TeamLocal
	{
		private Properties defaultProperties;
		
		LocalProperties(Properties defaultProperties)
		{
			this.defaultProperties = defaultProperties;
		}

		protected Object initialValue()
		{
			Properties newP = new Properties();
			try
			{
				TeamReservation kres;
				// DEEP clone of kernel props
				// Actually switch into kernel mode because the 'propertyNames'
				// method may cache an object in defaultProperties (and that has
				// to be done in kernel mode.) XXX lame.
				kres = Team.current().getKernelTeam().pushReservation();
				kres.visit();
				synchronized(defaultProperties)
				{
					Enumeration e = defaultProperties.propertyNames();
					while (e.hasMoreElements())
					{
						String k = (String)(e.nextElement());
						String p = defaultProperties.getProperty(k);
						kres.leave();
						newP.put(k.clone(), p.clone());
						kres.visit();
					}
				}
				kres.leave();
				kres.pop();
			}
			catch (TeamEjectionException tee)
			{
				// Nothing to do as we're toast if the kernel dies.
			}
			catch (DeadTeamException dte)
			{
				// Nothing to do as we're toast if the kernel dies.
			}
			return newP;
		}

	}
	
	private static LocalProperties internalProps;

// When trying to debug Java code that gets executed early on during
// JVM initialization, eg, before System.err is initialized, debugging
// println() statements don't work. In these cases, the following routines
// are very handy. Simply uncomment the following two lines to enable them.

public static native void debug(String s);	// print s to stderr, then \n
public static native void debugNoLF(String s);
public static native void debugE(Throwable t);	// print stack trace to stderr
public static native String getName();

private static final Class systemClass = java.lang.System.class;

static {
	Properties kernelProps;

	// XXX what are the constraints on the initialization order in here?

	kernelProps = initProperties(new Properties());
	// kernelProps.propertyNames();	// TLM -- prevents segfault in apiWrapper
	internalProps = new LocalProperties(kernelProps);

	// Initialise the I/O
	if (kernelProps.getProperty("kaffe.embedded", "false").equals("false")) {

		in = new BufferedInputStream(new StdinInputStream(), 128);
		out = new PrintStream(new StdOutputStream(StdOutputStream.STDOUT), true);
		err = new PrintStream(new StdOutputStream(StdOutputStream.STDERR), true);
	}
	else {
		in = new BufferedInputStream(new kaffe.io.StdInputStream(), 128);
		//out = new PrintStream(new BufferedOutputStream(new kaffe.io.StdOutputStream(), 128), true);
		out = new PrintStream(new BufferedOutputStream(new kaffe.io.StdErrorStream(), 128), true);
		err = new PrintStream(new BufferedOutputStream(new kaffe.io.StdErrorStream(), 128), true);
	}

	// Initialize the system class loader
	try {
		Class.forName("kaffe.lang.SystemClassLoader");
	}
	catch (ClassNotFoundException _)
	{
		// Kaffe won't let exceptions be thrown this early in
		// the init process, anyway...
	}
}

private System() { }

native public static void arraycopy(Object src,
	int src_position, Object dst, int dst_position, int length);

private static void checkPropertiesAccess() {
	SecurityManager sm = getSecurityManager();
	if (sm != null)
		sm.checkPropertiesAccess();
}

private static void checkPropertyAccess(String key) {
	SecurityManager sm = getSecurityManager();
	if (sm != null)
		sm.checkPropertyAccess(key);
}

native public static long currentTimeMillis();

public static void exit (int status) {
	Runtime.getRuntime().exit(status);
}

public static void gc() {
	Runtime.getRuntime().gc();
}

private static Properties getPropertiesInternal()
{
	return (Properties)(internalProps.get());
}

public static Properties getProperties() {
	checkPropertiesAccess();

	return getPropertiesInternal();
}

public static String getProperty(String key) {
	checkPropertyAccess(key);

	return getPropertiesInternal().getProperty(key);
}

public static String getProperty(String key, String def) {
	checkPropertyAccess(key);
	
	return getPropertiesInternal().getProperty(key, def);
}

public static SecurityManager getSecurityManager() {
	return (SecurityManager)security.get();
}

public static String getenv(String name) {
	throw new Error("System.getenv no longer supported");
}

native public static int identityHashCode(Object x);

native private static Properties initProperties(Properties props);

public static void load(String filename) {
	Runtime.getRuntime().load(filename, Class.getStackClass(1).getClassLoader());
}

public static void loadLibrary(String libname) {
	Runtime.getRuntime().loadLibrary(libname,
	    Class.getStackClass(1).getClassLoader());
}

public static String mapLibraryName(String fn) {
	return NativeLibrary.getLibPrefix() + fn + NativeLibrary.getLibSuffix();
}

public static void runFinalization() {
	Runtime.getRuntime().runFinalization();
}

public static void runFinalizersOnExit(boolean value) {
	Runtime.getRuntime().runFinalizersOnExit(value);
}

public static void setErr(PrintStream err) {
	// XXX call security manager for RuntimePermission("SetIO")
	setErr0(err);
}

native private static void setErr0(PrintStream err);

public static void setIn(InputStream in) {
	// XXX call security manager for RuntimePermission("SetIO")
	setIn0(in);
}

native private static void setIn0(InputStream in);

public static void setOut(PrintStream out) {
	// XXX call security manager for RuntimePermission("SetIO")
	setOut0(out);
}

native private static void setOut0(PrintStream out);

public static String setProperty(String key, String value) {
	if (key.length() == 0)
		throw new IllegalArgumentException("key can't be empty");
	SecurityManager sm = getSecurityManager();
	if (sm != null)
		sm.checkPermission(new PropertyPermission(key, "write"));
	return (String)getPropertiesInternal().setProperty(key, value);
}

public static void setProperties(Properties newProps) {
	checkPropertiesAccess();
	if (newProps == null) {
		getProperties().clear();
		return;
	}
	internalProps.set(newProps);
}

public static void setSecurityManager(SecurityManager s) {
	SecurityManager sm = getSecurityManager();
	if (sm != null) {
		sm.checkPermission(
			new RuntimePermission("setSecurityManager"));
	}
	security.set(s);
}
}
