
/*
 * Java core library component.
 *
 * Copyright (c) 1997, 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 */

package java.util;

import java.io.File;
import java.io.IOException;
import java.io.Serializable;
import kaffe.util.UNIXTimeZone;

import edu.utah.janosvm.sys.Team;
import edu.utah.janosvm.sys.TeamLocal;
import edu.utah.janosvm.sys.TeamHandle;
import edu.utah.janosvm.sys.DeadTeamException;
import edu.utah.janosvm.sys.TeamEjectionException;

public abstract class TimeZone
    implements Serializable,
	       Cloneable
{
    private static final long serialVersionUID = 3581463369166924961L;
    
    private static final String[] zoneDirs = {
	"/usr/share/zoneinfo",
	"/usr/share/lib/zoneinfo"
	// XXX others to try??
    };
    
    /**
     * zones maps timezone IDs (of type String) to either
     * TimeZone objects or to File objects.
     *
     * @see loadTimeZone
     */
    private static final HashMap zones = new HashMap();
    
    private static TeamLocal DEFAULT_TIME_ZONE = new TeamLocal() {
	    protected Object initialValue()
	    {
		TimeZone retval;
		String zne;

		zne = System.getProperty("user.timezone", "GMT");
		retval = getTimeZone(zne);
		if( retval == null )
		{
		    retval = getTimeZone("GMT");
		}
		if( retval == null )
		{
		    throw new InternalError("Cannot intialize timezone."
					    + " GMT & "
					    + zne
					    + " zones are undefined.");
		}
		return retval;
	    }
	};

    private static TeamLocal CUSTOM_ZONES = new TeamLocal() {
	    protected Object initialValue()
	    {
		return new HashMap();
	    }
	};

    // Load the system timezones
    static {
	// Install some standard SimpleTimeZones
	addSimple(-11*60*60*1000, "MIT");
	addSimple(-10*60*60*1000, "HST");
	addSimple(-9*60*60*1000, "AST");
	addSimple(-8*60*60*1000, "PST", Calendar.APRIL, 1, Calendar.SUNDAY, 2*60*60*1000, Calendar.OCTOBER, -1, Calendar.SUNDAY, 2*60*60*1000);
	addSimple(-8*60*60*1000, "PDT", Calendar.APRIL, 1, Calendar.SUNDAY, 2*60*60*1000, Calendar.OCTOBER, -1, Calendar.SUNDAY, 2*60*60*1000);
	addSimple(-7*60*60*1000, "PNT");
	addSimple(-7*60*60*1000, "MST", Calendar.APRIL, 1, Calendar.SUNDAY, 2*60*60*1000, Calendar.OCTOBER, -1, Calendar.SUNDAY, 2*60*60*1000);
	addSimple(-7*60*60*1000, "MDT", Calendar.APRIL, 1, Calendar.SUNDAY, 2*60*60*1000, Calendar.OCTOBER, -1, Calendar.SUNDAY, 2*60*60*1000);
	addSimple(-6*60*60*1000, "CST", Calendar.APRIL, 1, Calendar.SUNDAY, 2*60*60*1000, Calendar.OCTOBER, -1, Calendar.SUNDAY, 2*60*60*1000);
	addSimple(-6*60*60*1000, "CDT", Calendar.APRIL, 1, Calendar.SUNDAY, 2*60*60*1000, Calendar.OCTOBER, -1, Calendar.SUNDAY, 2*60*60*1000);
	addSimple(-5*60*60*1000, "EST", Calendar.APRIL, 1, Calendar.SUNDAY, 2*60*60*1000, Calendar.OCTOBER, -1, Calendar.SUNDAY, 2*60*60*1000);
	addSimple(-5*60*60*1000, "EDT", Calendar.APRIL, 1, Calendar.SUNDAY, 2*60*60*1000, Calendar.OCTOBER, -1, Calendar.SUNDAY, 2*60*60*1000);
	addSimple(-5*60*60*1000, "IET");
	addSimple(-4*60*60*1000, "PRT");
	addSimple(-3500*60*60, "CNT");
	addSimple(-3*60*60*1000, "AGT");
	addSimple(-1*60*60*1000, "CAT");
	addSimple(0, "GMT");
	addSimple(0, "UTC");
	addSimple(1*60*60*1000, "ECT");
	addSimple(1*60*60*1000, "EET");
	addSimple(2*60*60*1000, "ART");
	addSimple(3*60*60*1000, "EAT");
	addSimple(3500*60*60, "MET");
	addSimple(4*60*60*1000, "NET");
	addSimple(5*60*60*1000, "PLT");
	addSimple(5500*60*60, "IST");
	addSimple(6*60*60*1000, "BST");
	addSimple(7*60*60*1000, "VST");
	addSimple(8*60*60*1000, "CTT");
	addSimple(9*60*60*1000, "JST");
	addSimple(9500*60*60, "ACT");
	addSimple(10*60*60*1000, "AET");
	addSimple(11*60*60*1000, "SST");
	addSimple(12*60*60*1000, "NST");
	
	// Now read in any 'TZif' timezone files we find
	for( int i = 0; i < zoneDirs.length; i++ )
	{
	    File dir = new File(zoneDirs[i]);
	    
	    if( dir.isDirectory() )
	    {
		addZoneFiles(null, dir);
	    }
	}
    }

    private static void addSimple(int rawOffset, String id)
    {
	// assert Team.current() .equals( Team.current().getKernelTeam());
	zones.put(id, new SimpleTimeZone(rawOffset, id));
    }
    
    private static void addSimple(int off, String id, int i1,
				  int i2, int i3, int i4,
				  int i5, int i6, int i7, int i8)
    {
	// assert Team.current() .equals( Team.current().getKernelTeam());
	zones.put(id, new SimpleTimeZone(off, id, i1, i2, i3, i4,
					 i5, i6, i7, i8));
    }

    private static void addZoneFile(String ID, File file)
    {
	TimeZone tz = null;

	try
	{
	    tz = new UNIXTimeZone(ID, file);
	    zones.put(ID, tz);
	}
	catch (IOException e)
	{
	    /* If file cannot parsed properly, simply fall back to GMT
	     * timezone.  This is what the getTimeZone() API doc demands.
	     */
	    tz = getTimeZone("GMT");
	    zones.put(ID, tz);
	}
    }
    
    // Recurse through a directory tree adding any UNIX zone files found
    private static void addZoneFiles(String prefix, File dir)
    {
	// assert Team.current() .equals( Team.current().getKernelTeam());
	
	String[] files = dir.list();
	
	for( int i = 0; i < files.length; i++ )
	{
	    File file = new File(dir, files[i]);
	    String filePrefix = (prefix == null) ?
		files[i] : prefix + "/" + files[i];
	    
	    if( file.isFile() )
	    {
		addZoneFile(filePrefix, file);
	    }
	    else if( file.isDirectory() )
	    {
		addZoneFiles(filePrefix, file);
	    }
	}
    }
    
    private static TimeZone getCustomTimeZone(String ID)
    {
	TimeZone retval;
	HashMap hm;

	hm = (HashMap)CUSTOM_ZONES.get();
	synchronized( hm )
	{
	    if( (retval = (TimeZone)hm.get(ID)) == null )
	    {
		retval = (TimeZone)zones.get("GMT");
		if( ID.startsWith("GMT") )
		{
		    String sign = ID.substring(3,4);
		    String hh;
		    String mm;
		    int i = ID.indexOf(":");
		
		    if( i >= 0 )
		    {
			hh = ID.substring(4,i);
			mm = ID.substring(i+1);
		    }
		    else
		    {
			hh = ID.substring(4,6);
			mm = ID.substring(6,8);
		    }
		    
		    int hhi = Integer.parseInt(hh);
		    int mmi = Integer.parseInt(mm);
		    int rawOffset = (hhi * 60 + mmi) * 60 * 1000;
		    
		    if( sign.equals("+") )
		    {
			retval = new SimpleTimeZone(rawOffset, ID);
		    }
		    else if( sign.equals("-") )
		    {
			retval = new SimpleTimeZone(-rawOffset, ID);
		    }
		}
		hm.put(ID, retval);
	    }
	}
	return retval;
    }

    public static String[] getAvailableIDs()
    {
	HashMap hm;

	hm = (HashMap)CUSTOM_ZONES.get();
	synchronized( hm )
	{
	    String retval[];

	    retval = new String[zones.size() + hm.size()];
	    zones.keySet().toArray(retval);
	    if( hm.size() > 0 )
	    {
		System.arraycopy(hm.keySet().toArray(),
				 0,
				 retval,
				 zones.size(),
				 hm.size());
	    }
	    return retval;
	}
    }

    private static void fillZoneSet(HashSet ids, HashMap zones, int rawOffset)
    {
	for( Iterator i = zones.entrySet().iterator(); i.hasNext(); )
	{
	    Map.Entry ent = (Map.Entry)i.next();
	    TimeZone tz = (TimeZone)zones.get((String)ent.getKey());
	    
	    if( tz.getRawOffset() == rawOffset )
	    {
		ids.add(ent.getKey());
	    }
	}
    }
    
    public static String[] getAvailableIDs(int rawOffset)
    {
	HashSet ids = new HashSet();
	HashMap hm;
	
	fillZoneSet(ids, zones, rawOffset);
	hm = (HashMap)CUSTOM_ZONES.get();
	synchronized( hm )
	{
	    fillZoneSet(ids, zones, rawOffset);
	}
	return (String[])ids.toArray(new String[ids.size()]);
    }
    
    public static TimeZone getDefault()
    {
	return (TimeZone)DEFAULT_TIME_ZONE.get();
    }
    
    public static void setDefault(TimeZone zone)
    {
	DEFAULT_TIME_ZONE.set(zone);
    }

    public static TimeZone getTimeZone(String ID)
    {
	TimeZone retval;

	if( (retval = (TimeZone)zones.get(ID)) == null )
	{
	    retval = getCustomTimeZone(ID);
        }
	return retval;
    }
    
    private String timezoneID;
    
    public TimeZone()
    {
    }
    
    public void setID(String ID)
    {
	timezoneID = ID;
    }
    
    public String getID()
    {
	return (timezoneID);
    }
    
    abstract public void setRawOffset(int offsetMillis);
    
    abstract public boolean useDaylightTime();
    
    abstract public int getOffset(int era, int year,
				  int month, int day, int dayOfWeek,
				  int milliseconds);
    
    abstract public int getRawOffset();
    
    abstract public boolean inDaylightTime(Date date);
    
    public String toString()
    {
	return (timezoneID);
    }
    
    public Object clone()
    {
	try
	{
	    return (TimeZone)super.clone();
	}
	catch (CloneNotSupportedException e)
	{
	    throw new InternalError("TimeZone not cloneable?");
	}
    }
}
