/*
 * HTTPHeader.java
 *
 * Copyright (c) 2003 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Lesser 
 * General Public License.
 * 
 * See the file "license-lesser.terms" for information on 
 * usage and redistribution of this file.
 */

package edu.utah.janosvm.apps.isoserv;

import java.util.Hashtable;
import java.util.Enumeration;

import java.io.IOException;
import java.io.BufferedReader;

/**
 * An HTTPHeader is used to parse an HTTP request header and build a reply
 * header.
 */
public class HTTPHeader
    extends Hashtable
    implements HTTPConstants
{
    /**
     * The response code.
     */
    private int response;

    /**
     * Construct an HTTPHeader with the given value.
     *
     * @param response The response code.
     */
    public HTTPHeader(int response)
    {
	this.response = response;
    }
    
    /**
     * Construct an HTTPHeader with the default response code
     * (HTTP_INTERNAL_ERROR).
     */
    public HTTPHeader()
    {
	this(HTTP_INTERNAL_ERROR);
    }

    /**
     * @param response The response code.
     */
    public void setResponse(int response)
    {
	this.response = response;
    }

    /**
     * @return The response code.
     */
    public int getResponse()
    {
	return this.response;
    }

    /**
     * @return A description of the response code.
     */
    public String getResponseString()
    {
	String retval;
	
	switch( this.response / 100 )
	{
	case 2:
	    retval = HTTP_2XX_STRINGS[response % 100];
	    break;
	case 3:
	    retval = HTTP_3XX_STRINGS[response % 100];
	    break;
	case 4:
	    retval = HTTP_4XX_STRINGS[response % 100];
	    break;
	case 5:
	    retval = HTTP_5XX_STRINGS[response % 100];
	    break;
	default:
	    retval = "Internal Error";
	    break;
	}
	return retval;
    }

    /**
     * Parse an HTTP header line and store the result in this object.
     *
     * @param str A line containing an HTTP header value.
     * @throws IllegalArgumentException if "str" does not contain a properly
     * formatted line.
     */
    public synchronized void fromString(String str)
	throws IllegalArgumentException
    {
	int index;

	if( (index = str.indexOf(':')) > 0 )
	{
	    super.put(str.substring(0, index),
		      str.substring(index + 1).trim());
	}
	else
	{
	    throw new IllegalArgumentException("Invalid HTTP header field: "
					       + str);
	}
    }

    /**
     * Parse an HTTP header and store the results in this object.  Parsing
     * stops at the first blank line read from "br".
     *
     * @param br The source of the HTTP header.
     * @throws IllegalArgumentException if the header is improperly formatted.
     * @throws IOException if there was a problem reading from "br".
     */
    public synchronized void fromReader(BufferedReader br)
	throws IllegalArgumentException,
	       IOException
    {
	String line;

	while( ((line = br.readLine()) != null) && !line.equals("") )
	{
	    this.fromString(line);
	}
    }
    
    /**
     * @see java.lang.Object#toString()
     */
    public synchronized String toString()
    {
	Enumeration enum;
	String retval;
	
	retval = HTTP_VERSION
	    + " "
	    + this.response
	    + " "
	    + this.getResponseString()
	    + HTTP_EOL;
	enum = this.keys();
	while( enum.hasMoreElements() )
	{
	    String str;

	    str = (String)enum.nextElement();
	    retval += str + ": " + this.get(str) + HTTP_EOL;
	}
	retval += HTTP_EOL;
	return retval;
    }
}
