/*
 * HelperMonitor.java
 *
 * Copyright (c) 2003 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Lesser 
 * General Public License.
 * 
 * See the file "license-lesser.terms" for information on 
 * usage and redistribution of this file.
 */

package edu.utah.janosvm.apps.isoserv;

import java.lang.isolate.Link;
import java.lang.isolate.Isolate;
import java.lang.isolate.IsolateEvent;
import java.lang.isolate.IsolateMessage;
import java.lang.isolate.ClosedLinkException;
import java.lang.isolate.IsolateStartupException;

import java.util.prefs.Preferences;
import java.util.prefs.TransientPreferences;

import java.io.IOException;
import java.io.PrintStream;
import java.io.FileOutputStream;

/**
 * HelperMonitor is a Runnable that can be used to monitor helper Isolates.
 */
public class HelperMonitor
    implements Runnable
{
    /**
     * The I/O stream bindings for the created Isolate.
     */
    private final IsolateMessage io[];

    /**
     * The PrintStream associated with the log file.
     */
    private final PrintStream log;

    /**
     * The preferences for the created Isolate.
     */
    private final TransientPreferences tp;

    /**
     * The name of the helper class, sans package prefix.
     */
    private final String helperName;

    /**
     * The Isolate handle.
     */
    private Isolate isolate;

    /**
     * The link to use when sending messages to the helper.
     */
    private Link toHelperLink;

    /**
     * The link used to monitor the events generated by the created Isolate.
     */
    private Link eventLink;

    /**
     * Create a HelperMonitor object.
     *
     * @param logFile The log file to use as the Isolate's standard output and
     * error.
     * @param tp The preferences for the created Isolate.
     * @param helperName The name of the helper class, sans package prefix.
     */
    public HelperMonitor(FileOutputStream logFile,
			 TransientPreferences tp,
			 String helperName)
    {
	this.io = new IsolateMessage[] {
	    null,
	    IsolateMessage.newFileOutputStreamMessage(logFile),
	    IsolateMessage.newFileOutputStreamMessage(logFile),
	};
	this.log = new PrintStream(logFile);
	this.tp = tp;
	this.helperName = helperName;
    }

    /**
     * Start the Isolate.
     *
     * @throws IsolateStartupException if there was a problem starting the
     * Isolate.
     * @throws ClosedLinkException If the event link could not be created.
     */
    private synchronized void start()
	throws IsolateStartupException,
	       ClosedLinkException
    {
	IsolateMessage im[];
	
	this.isolate =
	    new Isolate("edu.utah.janosvm.apps.isoserv." + helperName,
			new String[0],
			this.tp,
			this.io[0],
			this.io[1],
			this.io[2]);
	/* Create the link this Isolate will use to talk to the helper. */
	this.toHelperLink = Link.newLink(Isolate.currentIsolate(),
					 this.isolate);
	im = new IsolateMessage[] {
	    IsolateMessage.newLinkMessage(this.toHelperLink),
	};
	this.eventLink = this.isolate.newEventLink();
	this.isolate.start(im);
    }

    /**
     * @return The link to use when sending messages to the helper.
     */
    public synchronized Link getToHelperLink()
    {
	return this.toHelperLink;
    }

    /**
     * Thread loop used to monitor the helper isolate and automatically restart
     * it in case it dies.
     *
     * @see java.lang.Runnable#run()
     */
    public void run()
    {
	try
	{
	    this.start();
	    while( true )
	    {
		try
		{
		    IsolateMessage im;
		    IsolateEvent ie;
		    
		    im = this.eventLink.receive();
		    ie = im.getIsolateEvent();
		    if( ie.getType() == IsolateEvent.Type.TERMINATED )
		    {
			this.log.println("Helper ("
					 + this.helperName
					 + ") died: "
					 + ie);
			this.start();
		    }
		}
		catch(IOException e)
		{
		    e.printStackTrace();
		}
	    }
	}
	catch(IsolateStartupException e)
	{
	    this.toHelperLink = null;
	    this.eventLink = null;
	    e.printStackTrace(this.log);
	}
	catch(ClosedLinkException e)
	{
	    e.printStackTrace(this.log);
	}
    }

    /**
     * 
     * @see java.lang.Object#toString()
     */
    public String toString()
    {
	return "HelperMonitor["
	    + "]";
    }
}
