/*
 * ServletDaemon.java
 *
 * Copyright (c) 2003 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Lesser 
 * General Public License.
 * 
 * See the file "license-lesser.terms" for information on 
 * usage and redistribution of this file.
 */

package edu.utah.janosvm.apps.isoserv;

import java.lang.isolate.Link;
import java.lang.isolate.Isolate;
import java.lang.isolate.IsolateMessage;
import java.lang.isolate.ClosedLinkException;
import java.lang.isolate.IsolateResourceError;
import java.lang.isolate.IsolateStartupException;
import java.lang.isolate.LinkSerializationException;

import java.util.Properties;

import java.util.prefs.Preferences;
import java.util.prefs.TransientPreferences;

import java.io.IOException;
import java.io.PrintStream;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.BufferedReader;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.InputStreamReader;

import java.net.Socket;
import java.net.ServerSocket;

/**
 * ServletDaemon is the main Isolate for the simple isolated servlet example.
 */
public class ServletDaemon
    implements HTTPConstants,
	       Runnable
{
    /**
     * The default properties file name (servlet.properties).
     */
    public static final String DEFAULT_PROPERTIES_FILE = "servlet.properties";

    /**
     * The default root directory for serving static pages (the current
     * directory).
     */
    public static final String DEFAULT_ROOT = System.getProperty("user.dir");

    /**
     * The default log file name (servlet.log).
     */
    public static final String DEFAULT_LOG = "servlet.log";

    /**
     * The default user name (nobody).
     */
    public static final String DEFAULT_USER = "nobody";

    /**
     * The default port number (8080).
     */
    public static final int DEFAULT_PORT = 8080;
    
    public static final int DEFAULT_TIMEOUT = 0;

    public static final int DEFAULT_HELPER_COUNT = 0;
    
    public static void main(String args[])
	throws Throwable
    {
	String propFile = DEFAULT_PROPERTIES_FILE;
	IsolateMessage startIM[];
	ServletDaemon sd;
	int lpc;
	
	if( args.length > 0 )
	{
	    propFile = args[0];
	}
	sd = new ServletDaemon();
	try
	{
	    sd.loadProperties(propFile);
	}
	catch(IOException e)
	{
	    System.err.println("Unable to load properties file: " + propFile);
	}
	sd.configure();
	sd.startHelpers();
	startIM = Isolate.currentIsolateStartMessages();
	if( startIM.length > 0 )
	{
	    startIM[0].getLink().send(IsolateMessage.newStringMessage("sync"));
	}
	for( lpc = 0; lpc < DEFAULT_HELPER_COUNT; lpc++ )
	{
	    new Thread(sd).start();
	}
	sd.run();
    }

    /**
     * The ServerSocket attached to the server port.
     */
    private ServerSocket ss;
    private final Object ssLock = new Object();
    
    protected final IsolateMessage io[] = new IsolateMessage[3];
    protected final Properties props = new Properties();
    protected final TransientPreferences tp = new TransientPreferences();

    protected String rootDirectory;
    protected FileOutputStream logFile;
    protected PrintStream log;
    protected int port;
    protected int timeout;

    protected HelperMonitor errorHelper;
    protected HelperMonitor staticHelper;
    
    public ServletDaemon()
    {
    }

    public void loadProperties(String propFile)
	throws IOException
    {
	FileInputStream fis;

	fis = new FileInputStream(propFile);
	this.props.load(fis);
    }

    public void configure()
	throws IOException
    {
	Preferences props;
	
	this.rootDirectory = this.props.getProperty("root", DEFAULT_ROOT);
	this.logFile = new FileOutputStream(
		this.props.getProperty("log", DEFAULT_LOG));
	this.io[1] = IsolateMessage.newFileOutputStreamMessage(this.logFile);
	this.io[2] = IsolateMessage.newFileOutputStreamMessage(this.logFile);
	this.log = new PrintStream(this.logFile);
	this.port = Integer.parseInt(
		this.props.getProperty("port",
				       Integer.toString(DEFAULT_PORT)));
	this.timeout = Integer.parseInt(
		this.props.getProperty("timeout",
				       Integer.toString(DEFAULT_TIMEOUT)));

	props = this.tp.node("java.properties");
	props.put("java.class.path", System.getProperty("java.class.path"));
	props.put("user.name", DEFAULT_USER);
	props.put("servlet.root", this.rootDirectory);
	
	/* A JanosVM extension, used to chroot() the team. */
	//this.tp.put("edu.utah.janosvm.resources.FileSystem.root",
	//this.rootDirectory);

	this.ss = new ServerSocket(this.port);
    }

    public void startHelpers()
    {
	this.errorHelper = new HelperMonitor(this.logFile,
					     this.tp,
					     "ErrorHelper");
	new Thread(this.errorHelper).start();
	this.staticHelper = new HelperMonitor(this.logFile,
					      this.tp,
					      "StaticHelper");
	new Thread(this.staticHelper).start();
    }

    public void run()
    {
	try
	{
	    HTTPHeader hh = new HTTPHeader();
	    
	    while( true )
	    {
		Socket socket = null;
		
		try
		{
		    int method = HTTP_METHOD_UNSUPPORTED;
		    BufferedReader br;
		    PrintStream ps;
		    String line;

		    synchronized( this.ssLock )
		    {
			socket = this.ss.accept();
		    }
		    try
		    {
			br = new BufferedReader(
				new InputStreamReader(socket.
						      getInputStream()));
			line = br.readLine();
			if( line == null )
			{
			}
			else if( line.startsWith("GET ") )
			{
			    method = HTTP_METHOD_GET;
			}
			else if( line.startsWith("HEAD ") )
			{
			    method = HTTP_METHOD_HEAD;
			}
			if( method != HTTP_METHOD_UNSUPPORTED )
			{
			    int fstart, fend;
			    
			    fstart = line.indexOf(' ');
			    fend = line.lastIndexOf(' ');
			    if( (fstart > 0) && (fend > fstart) )
			    {
				String version;

				version = line.substring(fend + 1);
				if( version.equals(HTTP_VERSION) ||
				    version.equals(HTTP_ALTERNATE_VERSION) )
				{
				    try
				    {
					String file;
					
					hh.clear();
					hh.fromReader(br);
					
					file = line.substring(fstart + 1,
							      fend);
					this.dispatch(socket,
						      method,
						      file,
						      hh);
				    }
				    catch(IllegalArgumentException e)
				    {
					this.dispatchError(socket,
							   method,
							   "",
							   HTTP_BAD_REQUEST);
				    }
				}
				else
				{
				    this.dispatchError(
					socket,
					method,
					"",
					HTTP_VERSION_NOT_SUPPORTED);
				}
			    }
			    else
			    {
				this.dispatchError(socket,
						   method,
						   "",
						   HTTP_BAD_REQUEST);
			    }
			}
			else
			{
			    this.dispatchError(socket,
					       method,
					       "",
					       HTTP_BAD_METHOD);
			}
		    }
		    catch(IOException e)
		    {
			/* Ignore. */
		    }
		}
		finally
		{
		    if( socket != null )
		    {
			socket.close();
		    }
		}
	    }
	}
	catch(IOException e)
	{
	    e.printStackTrace();
	}
	catch(Throwable th)
	{
	    th.printStackTrace();
	}
	System.exit(0);
    }

    public void dispatch(Socket socket, int method, String file, HTTPHeader hh)
    {
	if( file.endsWith(".jsp") )
	{
	    this.dispatchServlet(socket, method, file, hh);
	}
	else
	{
	    this.dispatchFile(socket, method, file, hh);
	}
    }

    public void dispatchServlet(Socket socket,
				int method,
				String file,
				HTTPHeader hh)
    {
	try
	{
	    IsolateMessage im[];
	    Isolate child;
	    
	    child = new Isolate("edu.utah.janosvm.apps.isoserv.ServletHelper",
				new String[] {
				    file.substring(1, file.length() - 4).
				    replace('/', '.'),
				},
				this.tp,
				this.io[0],
				this.io[1],
				this.io[2]);
	    im = new IsolateMessage[] {
		IsolateMessage.newSocketMessage(socket),
		IsolateMessage.newSerializableMessage(hh),
	    };
	    child.start(im);
	}
	catch(IsolateResourceError e)
	{
	    e.printStackTrace(this.log);
	    this.dispatchError(socket,
			       method,
			       file,
			       HTTP_UNAVAILABLE);
	}
	catch(IsolateStartupException e)
	{
	    this.dispatchError(socket,
			       method,
			       file,
			       HTTP_INTERNAL_ERROR);
	}
    }

    public void dispatchFile(Socket socket,
			     int method,
			     String file,
			     HTTPHeader hh)
    {
	try
	{
	    IsolateMessage im;

	    im = IsolateMessage.newCompositeMessage(new IsolateMessage[] {
		IsolateMessage.newSocketMessage(socket),
		IsolateMessage.newStringMessage(Integer.toString(method)),
		IsolateMessage.newStringMessage(file),
		IsolateMessage.newSerializableMessage(hh),
	    });
	    this.staticHelper.getToHelperLink().send(im);
	}
	catch(IOException e)
	{
	    e.printStackTrace();
	}
    }

    public void dispatchError(Socket socket,
			      int method,
			      String file,
			      int response)
    {
	try
	{
	    IsolateMessage im;
	    
	    im = IsolateMessage.newCompositeMessage(new IsolateMessage[] {
		IsolateMessage.newSocketMessage(socket),
		IsolateMessage.newStringMessage(Integer.toString(method)),
		IsolateMessage.newStringMessage(file),
		IsolateMessage.newStringMessage(Integer.toString(response)),
	    });
	    this.errorHelper.getToHelperLink().send(im);
	}
	catch(IOException e)
	{
	    e.printStackTrace();
	}
    }

    public String toString()
    {
	return "ServletDaemon["
	    + "]";
    }
}
