/*
 * StaticHelper.java
 *
 * Copyright (c) 2003 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Lesser 
 * General Public License.
 * 
 * See the file "license-lesser.terms" for information on 
 * usage and redistribution of this file.
 */

package edu.utah.janosvm.apps.isoserv;

import java.lang.isolate.Link;
import java.lang.isolate.Isolate;
import java.lang.isolate.IsolateMessage;

import java.io.File;
import java.io.IOException;
import java.io.PrintStream;
import java.io.OutputStream;
import java.io.BufferedReader;
import java.io.FileInputStream;
import java.io.InputStreamReader;
import java.io.FileNotFoundException;

import java.text.SimpleDateFormat;

import java.net.Socket;

import java.util.Date;

/**
 * StaticHelper is an Isolate that will handle replying to requests for static
 * pages.
 */
public class StaticHelper
    implements HTTPConstants,
	       Runnable
{
    /**
     * The default number of helper threads (not including the Isolate's main
     * thread).
     */
    public static final int DEFAULT_THREAD_COUNT = 4;

    /**
     * The file name extension to MIME type mapping to use.
     */
    public static final MIMETypes MIME_TYPES = new MIMETypes();

    private static final SimpleDateFormat DATE_FORMAT =
	new SimpleDateFormat("dd-MMM-yyyy hh:mm");

    /**
     * Main function for the StaticHelper isolate.  First, it will create a
     * StaticHelper instance with the Link stored in the first element of the
     * isolate's start messages.  Then, it will start DEFAULT_THREAD_COUNT
     * worker threads that will execute StaticHelper.run().  Finally, it will
     * also execute StaticHelper.run().
     */
    public static void main(String args[])
    {
	IsolateMessage im[];

	System.out.println("StaticHelper is alive.");
	
	im = Isolate.currentIsolateStartMessages();
	if( im.length > 0 )
	{
	    StaticHelper sh;
	    int lpc;
	    
	    sh = new StaticHelper(im[0].getLink(),
				  new File(System.getProperty("servlet.root",
							      "/")));
	    for( lpc = 0; lpc < DEFAULT_THREAD_COUNT; lpc++ )
	    {
		new Thread(sh).start();
	    }
	    sh.run();
	}
	else
	{
	    System.err.println("Isolate Usage:");
	    System.err.println("\tArguments: (none)");
	    System.err.println("\tMessages: <Request link>");
	}
    }
    
    /**
     * The link used by the parent to send requests to us.
     */
    protected Link inputLink;

    /**
     * The root directory for static pages.
     */
    protected File rootDirectory;

    /**
     * Construct a StaticHelper with the given value.
     *
     * @param inputLink The link used by the parent to send requests to us.
     */
    public StaticHelper(Link inputLink, File rootDirectory)
    {
	this.inputLink = inputLink;
	this.rootDirectory = rootDirectory;
    }
    
    /**
     * Thread loop used to answer requests for static pages.  The method will
     * synchronize internally so that multiple threads will serialize correctly
     * when reading from the same Link.
     *
     * @see java.lang.Runnable#run()
     */
    public void run()
    {
	try
	{
	    HTTPHeader hh = new HTTPHeader();
	    byte buf[] = new byte[4096];
	    IsolateMessage im[];
	    HTTPHeader rh;
	    
	    while( true )
	    {
		String fileName;
		PrintStream ps;
		Socket socket;
		int method;

		/* Get the messages from the server. */
		im = this.inputLink.receive().getComposite();

		/* Convert IsolateMessages to regular primitives/objects */
		socket = im[0].getSocket();
		try
		{
		    OutputStream os;
		    
		    method = Integer.parseInt(im[1].getString());
		    fileName = im[2].getString();
		    rh = (HTTPHeader)im[3].getSerializable();

		    /* Special case used to test isolate restarting... */
		    if( fileName.endsWith("exit.now") )
		    {
			System.out.println("StaticHelper exiting...");
			System.exit(0);
		    }

		    os = socket.getOutputStream();
		    ps = new PrintStream(os);
		    
		    /* Build up the response */
		    try
		    {
			File file;
			int len;

			file = new File(this.rootDirectory, fileName);
			if( file.isDirectory() )
			{
			    String list[];
			    int lpc;
			    
			    if( (list = file.list()) == null )
			    {
				throw new FileNotFoundException(fileName);
			    }
			    
			    hh.setResponse(HTTP_OK);
			    hh.put("Content-type", "text/html");
			    ps.print(hh.toString());

			    ps.println("<html>");
			    ps.println(" <head>");
			    ps.println("  <title>Index of "
				       + fileName
				       + "</title>");
			    ps.println(" </head>");
			    ps.println(" <body>");
			    ps.println("<h1>Index of "
				       + fileName
				       + "</h1>");
			    ps.println("<table>");
			    ps.println("<tr>");
			    ps.println("<td>Name</td>");
			    ps.println("<td>Last Modified</td>");
			    ps.println("<td>Size</td>");
			    ps.println("</tr>");
			    ps.println("<tr>");
			    ps.println("<td colspan=3><hr></td>");
			    ps.println("</tr>");
			    ps.println("<tr><td>");
			    ps.println("<a href=\"..\">Parent Directory</a>");
			    ps.println("</td></tr>");

			    for( lpc = 0; lpc < list.length; lpc++ )
			    {
				File sub = new File(file, list[lpc]);

				ps.println("<tr>");
				if( sub.isDirectory() )
				{
				    ps.print("<td><a href=\""
					     + list[lpc]
					     + "/\">"
					     + list[lpc]
					     + "/</a></td>");
				}
				else if( sub.isFile() )
				{
				    ps.print("<td><a href=\""
					     + list[lpc]
					     + "\">"
					     + list[lpc]
					     + "</a></td>");
				}
				ps.print("<td>"
					 + DATE_FORMAT.format(
						new Date(sub.lastModified()))
					 + "</td><td>"
					 + sub.length()
					 + "</td>");
				ps.println("</tr>");
			    }
			    ps.println("<tr>");
			    ps.println("<td colspan=3><hr></td>");
			    ps.println("</tr>");
			    ps.println("</table>");
			    ps.println("</body>");
			    ps.println("</html>");
			}
			else
			{
			    FileInputStream fis;
			    
			    fis = new FileInputStream(file);

			    try
			    {
				hh.setResponse(HTTP_OK);
				hh.put("Content-type",
				       MIME_TYPES.mapFileName(fileName));
				ps.print(hh.toString());
				
				while( (len = fis.read(buf)) > 0 )
				{
				    os.write(buf, 0, len);
				}
			    }
			    finally
			    {
				fis.close();
			    }
			}
		    }
		    catch(FileNotFoundException e)
		    {
			String responseString;
			
			hh.setResponse(HTTP_NOT_FOUND);
			hh.put("Content-type", "text/html");

			responseString = hh.getResponseString();
			
			ps.print(hh.toString());
			ps.println("<html>");
			ps.println("<head>");
			ps.println("<title>" + responseString + "</title>");
			ps.println("</head>");
			ps.println("<body>");
			ps.println("<h1>" + responseString + "</h1>");
			ps.println("The requested URL "
				   + fileName
				   + " was not found on this server.");
			ps.println("<hr>");
			ps.println("</body>");
			ps.println("</html>");
		    }
		}
		catch(Throwable th)
		{
		    System.err.println("Exception in StaticHelper:");
		    th.printStackTrace(System.err);
		}
		finally
		{
		    /* We're done with them. */
		    socket.close();
		}
	    }
	}
	catch(IOException e)
	{
	    /* Lost connection to the server. */
	    e.printStackTrace(System.err);
	}
	catch(Throwable th)
	{
	    th.printStackTrace(System.err);
	}
    }

    /**
     * @see java.lang.Object#toString()
     */
    public String toString()
    {
	return "StaticHelper["
	    + "]";
    }
}
