/*
 * IsoShell.java
 *
 * Copyright (c) 2001-2003 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Lesser 
 * General Public License.
 * 
 * See the file "license-lesser.terms" for information on 
 * usage and redistribution of this file.
 */

package edu.utah.janosvm.apps.isoshell;

import java.io.BufferedReader;
import java.io.InputStreamReader;
import java.util.ArrayList;
import java.util.LinkedList;
import java.util.StringTokenizer;
import java.util.prefs.TransientPreferences;
//import java.util.HashSet;
//import java.util.HashMap;
//import java.util.Iterator;
import java.lang.isolate.Isolate;
import java.lang.isolate.IsolateMessage;
import java.lang.isolate.IsolateStartupException;
import java.lang.isolate.ClosedLinkException;
import java.lang.isolate.Link;

/**
 * Isolate-based command shell main.
 *
 * @author Patrick Tullmann, University of Utah &lt;<a href="mailto:tullmann@cs.utah.edu">tullmann@cs.utah.edu</a>&gt;
 */
public class IsoShell
{
    private static final String prompt = "IsoShell 0.5> ";

    private static final ParsedCommand[] EXIT_COMMAND = new ParsedCommand[]{ParsedCommand.newSimpleInstance("exit")};

    /**
     * Parse a string into an array of commands.  May return null.
     * Usually returns a 1-element array.
     */
    private static ParsedCommand[] parseCommandLine(String line)
    {
	ArrayList commands = new ArrayList();
	LinkedList args = new LinkedList();
	boolean background = false;
	String delims = "&; \n\r\t\f";

	Trace.t(Trace.PARSE, "Got line: " +line);

	StringTokenizer tokline = new StringTokenizer(line, delims, true);

	while (tokline.hasMoreTokens())
	{
	    String tok = tokline.nextToken();
			
	    if (tok.equals("&") || tok.equals(";"))
	    {
		if (tok.equals("&"))
		{
		    Trace.t(Trace.PARSE, "Got backgrounder token.");
		    background = true;
		    // Fall through to end-of-command stuff
		    tok = ";";
		}
		else
		{
		    Trace.t(Trace.PARSE, "Got command-ender token.");
		}
				
		if (args.size() > 0)
		{
		    // command copies contents of args.
		    commands.add(ParsedCommand.newInstance(args, background));
		}
	
		// reset state
		background = false;
		args.clear();
	    }
	    else if ((tok.length() == 1)
		     && (delims.indexOf(tok) != -1))
	    {
		// some other delim we don't care about....
	    }
	    else 
	    {
		Trace.t(Trace.PARSE, "Got token: `" +tok+ "'");
		args.add(tok);
	    }
	}

	Trace.t(Trace.PARSE, "Ran out of tokens.");
	if (args.size() > 0)
	    commands.add(ParsedCommand.newInstance(args, background));
		
	if (commands.size() == 0)
	    return null;
		
	ParsedCommand[] cmdArray = new ParsedCommand[commands.size()];
	commands.toArray(cmdArray);
		
	return cmdArray;
    }

    /**
     * Get a command from the user.  Uses given stream to retrive a line
     * and parses it into a new ParsedCommand object.
     */
    private static String getCommandLine(BufferedReader in)
	throws java.io.IOException
    {
	System.out.print(prompt);

	String line = in.readLine();
	if ((line == null) 
	    || ((line.length() > 0) && (line.charAt(0) == (char)0x03)))
	{
	    Trace.t(Trace.PARSE, "Got <null>/^D from " +in);
	    return null;
	}
    
	return line;
    }

    private static TransientPreferences context = null;

    /**
     * Do the given command.
     */
    private static boolean doCommand(JobManager jobManager, ParsedCommand command)
    {
	Isolate child;
	IsolateMessage[] initialMessages = null;
	Link eventLink;
	IsolateMessage stdin  = Isolate.currentIsolateIOBindings()[0];
	IsolateMessage stdout = Isolate.currentIsolateIOBindings()[1];
	IsolateMessage stderr = Isolate.currentIsolateIOBindings()[2];

	/* Initialize default context if necessary. */
	if (context == null)
	{
	    context = new TransientPreferences();
	    context.node("java.properties").put("java.class.path", System.getProperty("java.class.path", "."));
	}

	/* Start a new Isolate for command */
	String name = command.name();
	try
	{
	    Trace.t(Trace.FORK, "Creating a new Isolate for command " +name);

	    child = new Isolate(name, command.args(), 
				context, stdin, stdout, stderr);

	    eventLink = child.newEventLink();

	    Trace.t(Trace.FORK, "Starting child");
	    child.start(initialMessages);
	}
	catch (ClosedLinkException cle)
	{
	    // XXX this is stupid
	    System.err.println("Closed link exception creating child for command `" +name+ "': " +cle);
	    cle.printStackTrace(System.err);
	    return false;
	}
	catch (IsolateStartupException ise)
	{
	    System.err.println("Exception creating/starting child for command `" +name+ "': " +ise);
	    ise.printStackTrace(System.err);
	    return false;
	}
		
	if (command.background())
	{
	    jobManager.newBackgroundJob(name, child, eventLink);
	} 
	else 
	{
	    jobManager.waitForEnd(name, child, eventLink);
	}
	
	return true;
    }


    static void defineInternalCommands(InternalCommandManager iCmdManager, final JobManager jobManager)
    {
	iCmdManager.addCommand(
	    "exit",
	    new InternalCommandManager.Action() {
		public void doAction(ParsedCommand cmd)
		{
		    // assert cmd.name().equals("exit");
		    int rc = 0;
		    if (cmd.argCount() > 0)
		    {
			try 
			{
			    rc = Integer.parseInt(cmd.args()[0]);
			}
			catch (NumberFormatException _)
			{
			    // ignored
			}
		    }
		    jobManager.terminateAll();
		    Trace.t(Trace.MAIN, "System.exit(" +rc+ ")");
		    System.exit(rc);
		}
	    });

	iCmdManager.addCommand(
	    ".jobs", 
	    new InternalCommandManager.Action() {
		public void doAction(ParsedCommand cmd)
		{
		    // assert cmd.name().equals(".jobs");
		    jobManager.listJobs(System.out);
		}
	    });

	iCmdManager.addCommand(
	    ".set", 
	    new InternalCommandManager.Action() {
		public void doAction(ParsedCommand cmd)
		{
		    String[] args = cmd.args();
		    System.getProperties().setProperty(args[0], args[1]);
		}
	    });

	iCmdManager.addCommand(
	    ".trace", 
	    new InternalCommandManager.Action() {
		public void doAction(ParsedCommand cmd)
		{
		    // assert cmd.name().equals(".trace");
		    try
		    {
			Trace.setMask(cmd.args());
		    }
		    catch (IllegalArgumentException iae)
		    {
			iae.printStackTrace(System.err);
		    }
		}
	    });
    }

    public static void main(String[] args)
    {
	final JobManager jobManager = new JobManager();
	final InternalCommandManager iCmdManager = new InternalCommandManager();
	PackagePath packagePath = new PackagePath(".:edu.utah.janosvm.apps");
	final BufferedReader in = new BufferedReader(new InputStreamReader(System.in));
	boolean exited = false;
	int rc = 0;

	Trace.t(Trace.MAIN, "Current java.class.path=" 
		+ System.getProperty("java.class.path","<undefined>"));

	defineInternalCommands(iCmdManager, jobManager);
		
	while (!exited)
	{
	    ParsedCommand[] commands = null;

	    try
	    {
		String line = getCommandLine(in);
		if (line == null)
		    commands = EXIT_COMMAND;
		else
		    commands = parseCommandLine(line);
				
	    }
	    catch (java.io.IOException ioe)
	    {
		System.err.println("Exception getting command line: " + ioe);
		ioe.printStackTrace(System.err);
	    }

	    if (commands != null)
	    {
		// Tracing
		for (int i = 0; i < commands.length; i++)
		    Trace.t(Trace.MAIN, "Got command[" +i+ "]:" 
			    +commands[i]);

		// Actual work
		for (int c = 0; c < commands.length; c++)
		{
		    ParsedCommand cmd = commands[c];

		    // silenty skip empty commands
		    if (cmd != null)
		    {
			// Expand aliases
			

			// Either internal command or external command
			if (iCmdManager.isInternalCommand(cmd.name()))
			{
			    iCmdManager.doInternalCommand(cmd);
			}
			else 
			{
			    String oname = cmd.name();
			    
			    cmd.setName(packagePath.find(oname));
			    if (cmd.name() == null)
			    {
				System.err.println(oname+ ": Command not found.");
			    }
			    else
				doCommand(jobManager, cmd);
			}
		    }
		}
	    }
	}

	jobManager.terminateAll();
	Trace.t(Trace.MAIN, "System.exit(" +rc+ ")");
	System.exit(rc);
    }
}
