/*
 * JobManager.java
 *
 * Copyright (c) 2001-2003 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Lesser 
 * General Public License.
 * 
 * See the file "license-lesser.terms" for information on 
 * usage and redistribution of this file.
 */

package edu.utah.janosvm.apps.isoshell;

import java.util.HashSet;
import java.io.PrintStream;
import java.util.Iterator;
import java.lang.isolate.Isolate;
import java.lang.isolate.IsolateMessage;
import java.lang.isolate.IsolateMessageDispatcher;
import java.lang.isolate.IsolateEvent;
import java.lang.isolate.ClosedLinkException;
import java.lang.isolate.Link;

/**
 * Manager for background tasks, also implements waiting for end of a job.
 *
 * @author Patrick Tullmann, University of Utah &lt;<a href="mailto:tullmann@cs.utah.edu">tullmann@cs.utah.edu</a>&gt;
 */
class JobManager
{
    private final HashSet jobs;
    private final IsolateMessageDispatcher dispatcher;

    JobManager()
    {
	this.jobs = new HashSet();
	this.dispatcher = new IsolateMessageDispatcher();
	Thread t = new Thread(this.dispatcher, "Jobs Message Dispatcher");
	t.start();
    }

    private static class JobListEntry
    	implements IsolateMessageDispatcher.Listener
    {
	private final String name;
	private final int id;
	private final Isolate isolate;
	private static int currentId = 11;
	private IsolateEvent state = null;
    	
	JobListEntry(String name, Isolate isolate)
	{
	    this.name = name;
	    this.isolate = isolate;
	    synchronized(JobListEntry.class)
	    {
		this.id = currentId++;
	    }
	}

	void terminate()
	{
	    this.isolate.halt(1);
	}

	public String toString()
	{
	    return name +"[" + isolate+ "]";
	}

	private boolean handleEv(IsolateEvent ev)
	{
	    boolean remove = false;

	    if (this.state == null)
	    {
		if (ev.getType() != IsolateEvent.Type.STARTING)
		    throw new InternalError("Expected STARTING, got " +ev);
	    }
	    else if (this.state.getType() == IsolateEvent.Type.STARTING)
	    {
		if (ev.getType() != IsolateEvent.Type.EXITING)
		    throw new InternalError("Expected EXITING, got " +ev);
	    }
	    else if (this.state.getType() == IsolateEvent.Type.EXITING)
	    {
		if (ev.getType() != IsolateEvent.Type.TERMINATED)
		    throw new InternalError("Expected TERMINATED, got " +ev);
		
		remove = true;
	    }
	    else if (this.state.getType() == IsolateEvent.Type.TERMINATED)
	    {
		throw new InternalError("Unexpected event: " +ev);
	    }
		
	    Trace.t(Trace.JOBS, "Job " +this+ " -> " +ev);
	    this.state = ev;

	    return remove;
	}

	// -- IsolateMessageDispatcher.Listener interfaces:

	public void messageReceived(IsolateMessageDispatcher dispatcher, Link link, IsolateMessage message)
	{
	    boolean remove = handleEv(message.getIsolateEvent());
	    if (remove)
		dispatcher.remove(link);
	}
	
	public void receiveFailed(IsolateMessageDispatcher dispatcher, Link link, Throwable throwable)
	{
	    // We'll get this on an event

	    if (this.state.getType() != IsolateEvent.Type.TERMINATED)
		// XXX use chaining
		throw new InternalError("Recieve failed on " +this+ " -> " +throwable);
	}
    }

    void listJobs(PrintStream ps)
    {
        Iterator i = jobs.iterator();
        while(i.hasNext())
        {
	    JobListEntry jle = (JobListEntry)(i.next());
	    ps.println("  " + jle.name);
        }
    }

    void terminateAll()
    {
	Trace.t(Trace.JOBS, "Terminate all.");

        Iterator i = jobs.iterator();
        while(i.hasNext())
        {
	    JobListEntry jle = (JobListEntry)(i.next());
	    Trace.t(Trace.JOBS, "   terminating " +jle);
	    jle.terminate();
        }
    }

    /**
     * Wait for the given isolate to die.
     */
    static void waitForEnd(String name, Isolate isolate, Link eventLink)
    {
	Trace.t(Trace.FORK, "Waiting for " +name+ "...");

	try
	{
	    IsolateEvent ev = eventLink.receive().getIsolateEvent();
	    if (ev.getType() != IsolateEvent.Type.STARTING)
	    {
		throw new InternalError("Expected starting event:" +isolate);
	    }
	    
	    ev = eventLink.receive().getIsolateEvent();
	    if (ev.getType() != IsolateEvent.Type.EXITING)
	    {
		throw new InternalError("Expected exiting event:" +isolate);
	    }
	    
	    ev = eventLink.receive().getIsolateEvent();
	    if (ev.getType() != IsolateEvent.Type.TERMINATED)
	    {
		throw new InternalError("Expected terminated event:" +isolate);
	    }
	}
	catch (ClosedLinkException cle)
	{
	    throw new InternalError("Event link closed early? " +cle);
	}

	Trace.t(Trace.FORK, "Join with " +name+ " completed...");
    }

    void newBackgroundJob(final String name, 
			  final Isolate isolate,
			  final Link eventLink)
    {
        JobListEntry jl = new JobListEntry(name, isolate);
        this.jobs.add(jl);
	this.dispatcher.add(eventLink, jl);
    }
}

