/*
 * Trace.java
 *
 * Copyright (c) 2001-2003 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Lesser 
 * General Public License.
 * 
 * See the file "license-lesser.terms" for information on 
 * usage and redistribution of this file.
 */

package edu.utah.janosvm.apps.isoshell;

/** 
 * A class for run-time tracing.  See the '.trace' internal command.
 *
 * @author Patrick Tullmann, University of Utah &lt;<a href="mailto:tullmann@cs.utah.edu">tullmann@cs.utah.edu</a>&gt;
 */
final class Trace
{
    static final Flag FORK    = new Flag("FORK",    0x0001);
    static final Flag MAIN    = new Flag("MAIN",    0x0002);
    static final Flag PARSE   = new Flag("PARSE",   0x0004);
    static final Flag CLOOKUP = new Flag("CLOOKUP", 0x0008);
    static final Flag JOBS    = new Flag("JOBS",    0x0010);
  
    private static long currentMask = 0xFF;

    private static final class Flag
    {
	final String name;
	final long mask;
    
	static final Flag[] flags = new Flag[64];
	static int fi = 0;

	Flag(String name, long mask)
	{
	    this.name = name;
	    this.mask = mask;

	    synchronized(flags)
	    {
		flags[fi++] = this;
	    }
	}

	static Flag find(String name)
	{
	    synchronized(flags)
	    {
		for (int i = 0; i < flags.length; i++)
		{
		    if ((flags[i] != null) && (flags[i].name.equals(name)))
			return flags[i];
		}
		throw new IllegalArgumentException("No trace flag with name " +name);
	    }
	}

	public String toString()
	{
	    return this.name;
	}
    }

    static boolean isSet(Flag f)
    {
	return (f.mask & currentMask) != 0;
    }

    static void t(Flag f, String msg)
    {
	if (isSet(f))
	    System.err.println(f.name+ ":: " +msg);
    }
  
    static void setMask(String[] args)
    {
	if ((args == null) 
	    || (args.length < 1) 
	    || (args[0].equalsIgnoreCase("list")))
	{
	    for (int i = 0; i < Flag.fi; i++)
	    {
		Flag f = Flag.flags[i];
		if (isSet(f))
		    System.out.print("* ");
		else
		    System.out.print("  ");
		System.out.println(Flag.flags[i]);
	    }
	}
	else if (args[0].equalsIgnoreCase("all"))
	{
	    currentMask = 0xFFFFFFFFL;
	}
	else if (args[0].equalsIgnoreCase("none") ||
		  args[0].equalsIgnoreCase("clear"))
	{
	    currentMask = 0x0L;
	    System.err.println("Trace mask cleared.");
	}
	else
	{
	    String s = args[0];
	    //System.err.println("s=" +s+ "; s[0]=" +s.charAt(0));
	    if (s.charAt(0) == '+')
	    {
		//System.err.println("adding " +s);
		s = s.substring(1);
		currentMask |= Flag.find(s).mask;
	    }
	    else if (s.charAt(0) == '-')
	    {
		//System.err.println("removing " +s);
		s = s.substring(1);
		currentMask &= (~(Flag.find(s).mask));
	    }
	    else
	    {
		//System.err.println("setting to " +s);
		currentMask = Flag.find(s).mask;
	    }
	}
    }
}
