/*
 * Loader.java
 *
 * Copyright (c) 2003 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Lesser 
 * General Public License.
 * 
 * See the file "license-lesser.terms" for information on 
 * usage and redistribution of this file.
 */

package edu.utah.janosvm.kit.loader;

import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.lang.reflect.InvocationTargetException;

import java.util.Hashtable;
import java.util.Properties;

import java.io.IOException;
import java.io.ByteArrayInputStream;
import java.io.FileNotFoundException;

import edu.utah.janosvm.sys.Team;
import edu.utah.janosvm.sys.TeamHandle;
import edu.utah.janosvm.sys.TeamReservation;
import edu.utah.janosvm.sys.DeadTeamException;
import edu.utah.janosvm.sys.TeamEjectionException;
import edu.utah.janosvm.sys.DuplicateExportException;

import edu.utah.janosvm.resources.NameSpaceSpec;

import edu.utah.janosvm.resources.StdioSpec;
import edu.utah.janosvm.resources.NativeSpec;
import edu.utah.janosvm.resources.ResourceSpec;
import edu.utah.janosvm.resources.NameSpaceSpec;
import edu.utah.janosvm.resources.NameSpaceGroup;
import edu.utah.janosvm.resources.ClassPathResource;
import edu.utah.janosvm.resources.NameSpaceResource;
import edu.utah.janosvm.resources.ResourceSpecException;
import edu.utah.janosvm.resources.InvalidClassGroupException;
import edu.utah.janosvm.resources.ViewGroupNotFoundException;
import edu.utah.janosvm.resources.ClassGroupNotFoundException;

public class Loader
    extends ClassLoader
{
    public static final String SERVER_NAME_PROPERTY = "server.name";

    public static final String SERVER_GROUP_PROPERTY = "server.group";

    public static final String SERVER_RESOURCES_PROPERTY = "server.resource";
    public static final String SERVER_RESOURCES_DEFAULT =
	"edu.utah.janosvm.kit.loader.ServerResources";

    public static final String SYSTEM_PACKAGE[] = {
	"java.lang",
	"java.lang.ref",
	"java.lang.reflect",
	"java.io",
	"java.math",
	"java.net",
	"java.security",
	"java.security.cert",
	"java.security.interfaces",
	"java.security.spec",
	"java.util",
	"java.util.jar",
	"java.util.mime",
	"java.util.prefs",
	"java.util.zip",
	"kaffe.io",
	"kaffe.jar",
	"kaffe.lang",
	"kaffe.management",
	"kaffe.net",
	"kaffe.security",
	"kaffe.security.provider",
	"kaffe.util",
	"kaffe.util.prefs",
	"kaffe.util.zip",
	"edu.utah.janosvm.sys",
	"edu.utah.janosvm.resources",
	"edu.utah.janosvm.io",
	"edu.utah.janosvm.kit.loader",
    };

    public static void main(String args[])
	throws Throwable
    {
	if( args.length >= 1 )
	{
	    TeamHandle th = null;
	    Team curr;

	    curr = Team.current();
	    if( curr.equals(curr.getKernelTeam()) )
	    {
		String oldArgs[];
		int lpc;
		
		NameSpaceResource.exportClassGroup("janosvm");
		NameSpaceResource.exportClassGroup("kit-loader");

		th = Team.create(new ResourceSpec[] {
		    StdioSpec.currentFD(),
		    NativeSpec.current(),
		    new NameSpaceSpec(new NameSpaceGroup[] {
			new NameSpaceGroup(curr, "kernel"),
			new NameSpaceGroup(curr, "kit-loader"),
		    }),
		});
		th.switchTo();

		oldArgs = args;
		args = new String[args.length];
		for( lpc = 0; lpc < args.length; lpc++ )
		{
		    args[lpc] = CloneFactory.clone(oldArgs[lpc]);
		}
	    }
	    try
	    {
		final Class[] mainSig = { args.getClass() };
		Loader loader = new Loader();
		final String subargs[];
		Class mainClass;
		final Method m;

		subargs = new String[args.length - 1];
		System.arraycopy(args,
				 1,
				 subargs,
				 0, subargs.length);
		mainClass = loader.loadClass(args[0]);
		m = mainClass.getMethod("main", mainSig);

		if( !Modifier.isStatic(m.getModifiers()) )
		{
		    throw new NoSuchMethodException("static main");
		}

		new Thread(new Runnable() {
			public void run()
			{
			    try
			    {
				m.invoke(null, new Object[] { subargs } );
			    }
			    catch(InvocationTargetException e)
			    {
				e.getTargetException().printStackTrace();
			    }
			    catch(Throwable e)
			    {
				e.printStackTrace();
			    }
			}
		    }).start();
	    }
	    catch(Throwable e)
	    {
		if( th != null )
		{
		    th.returnFrom();
		    th = null;
		}
		e.printStackTrace();
	    }
	    if( th != null )
	    {
		th.returnFrom();
	    }
	}
	else
	{
	    System.err.println("Usage: Loader <main class> [arg1, arg2, ...]");
	}
    }

    private final Hashtable servers = new Hashtable(); 
    
    public Loader()
    {
    }

    public boolean isSystemPackage(String pack)
    {
	boolean retval = false;
	int lpc;

	for( lpc = 0; (lpc < SYSTEM_PACKAGE.length) && !retval; lpc++ )
	{
	    retval = pack.equals(SYSTEM_PACKAGE[lpc]);
	}
	if( !retval )
	{
	    retval = this.servers.containsKey(pack);
	}
	return retval;
    }

    protected void startServer(Properties props)
	throws ClassNotFoundException
    {
	Team current = Team.current();
	String name, group;
	TeamHandle kernel;

	kernel = current.getKernelTeam();
	if( ((name = props.getProperty(SERVER_NAME_PROPERTY)) != null) &&
	    ((group = props.getProperty(SERVER_GROUP_PROPERTY)) != null) )
	{
	    try
	    {
		String resourceClass;
		ServerResources sr;
		TeamReservation tr;
		ResourceSpec rs[];
		Class cl;

		resourceClass = props.getProperty(SERVER_RESOURCES_PROPERTY,
						  SERVER_RESOURCES_DEFAULT);
		cl = super.loadClass(resourceClass);
		sr = (ServerResources)cl.newInstance();
		tr = kernel.pushReservation();
		rs = sr.getResourceSpecs();
		{
		    TeamHandle th;

		    try
		    {
			th = Team.create(name, rs);
			th.switchTo();
			{
			    NameSpaceResource.exportClassGroup(group);
			}
			th.returnFrom();
		    }
		    catch(DuplicateExportException e)
		    {
			th = Team.find(name);
		    }
		    NameSpaceResource.importClassGroup(
			new NameSpaceGroup(th, group));
		}
		tr.pop();
	    }
	    catch(Throwable th)
	    {
		throw new ClassNotFoundException(th.toString());
	    }
	}
    }

    public Class loadClass(String name, boolean resolve)
	throws ClassNotFoundException
    {
	Class retval;

	if( (retval = this.findLoadedClass(name)) == null )
	{
	    String pack = null, packpath = null;
	    byte bits[];
	    int offset;
	    
	    if( (offset = name.lastIndexOf('.')) > 0 )
	    {
		pack = name.substring(0, offset);
		packpath = pack.replace('.', '/');
	    }
	    try
	    {
		synchronized( this.servers )
		{
		    if( (pack != null) && !this.servers.containsKey(pack) )
		    {
			Properties props;
			
			bits = ClassPathResource.
			    loadFile(packpath + "/namespace.properties");
			props = new Properties();
			props.load(new ByteArrayInputStream(bits));
			this.startServer(props);
			
			this.servers.put(pack, props);
		    }
		}
	    }
	    catch(FileNotFoundException e)
	    {
		/* Ignore... */
	    }
	    catch(DeadTeamException e)
	    {
		throw new ClassNotFoundException(e.toString());
	    }
	    catch(IOException e)
	    {
		throw new ClassNotFoundException(e.toString());
	    }
	    
	    try
	    {
		if( (pack == null) || !this.isSystemPackage(pack) )
		{
		    try
		    {
			String path;
			
			path = name.replace('.', '/').concat(".class");
			bits = ClassPathResource.loadFile(path);
			retval = super.defineClass(name,
						   bits,
						   0,
						   bits.length);
		    }
		    catch(DeadTeamException e)
		    {
			throw new ThreadDeath();
		    }
		    catch(IOException e)
		    {
		    }
		}
		if( retval == null )
		{
		    ClassLoader parent;
		    
		    if( (parent = this.getParent()) == null )
		    {
			parent = ClassLoader.getSystemClassLoader();
		    }
		    retval = parent.loadClass(name);
		}
	    }
	    catch(ClassNotFoundException e)
	    {
		retval = super.findClass(name);
	    }
	}
	if( resolve )
	{
	    this.resolveClass(retval);
	}
	return retval;
    }

    public String toString()
    {
	return "Loader["
	    + super.toString()
	    + "]";
    }
}
