/*
 * ResourceMonitor.java
 *
 * Copyright (c) 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Lesser 
 * General Public License.
 * 
 * See the file "license-lesser.terms" for information on 
 * usage and redistribution of this file.
 */

package edu.utah.janosvm.kit.management.team;

import java.util.Vector;

import edu.utah.janosvm.sys.Team;
import edu.utah.janosvm.sys.TeamHandle;
import edu.utah.janosvm.sys.DeadTeamException;
import edu.utah.janosvm.sys.NoSuchExportException;
import edu.utah.janosvm.sys.TeamEjectionException;

/**
 * Base class for applications that want to poll team's about their resource
 * consumption.  An application is expected to start a thread and execute
 * the run() method to actually perform the polling.
 */
public class ResourceMonitor
    implements Runnable
{
    /**
     * The number of times to poll.
     */
    private final int iterations;

    /**
     * The millisecond interval between polls.
     */
    private final long msInterval;

    /**
     * The class used to sample the state of a team.
     */
    private final Class sampler;

    /**
     * The set of ResourceSamplers held by this monitor.
     */
    private final Vector samplers = new Vector();

    /**
     * Construct a ResourceMonitor with the given values.
     *
     * @param iterations The number of times to poll.
     * @param msInterval The millisecond interval between polls.
     * @param sampler The class used to sample the state of a team.
     */
    public ResourceMonitor(int iterations, long msInterval, Class sampler)
    {
	this.iterations = iterations;
	this.msInterval = msInterval;
	this.sampler = sampler;
    }

    /**
     * Construct a ResourceMonitor with the given values.  The default
     * iteration count is infinite and the default interval is one second.
     *
     * @param sampler The class used to sample the state of a team.
     */
    public ResourceMonitor(Class sampler)
    {
	this(Integer.MAX_VALUE, 1000, sampler);
    }

    /**
     * Add a number of teams that should be monitored.
     *
     * @param names The names of the teams to monitor.
     * @return this
     */
    public ResourceMonitor addTeams(String names[])
	throws InstantiationException,
	       IllegalAccessException
    {
	int lpc;

	for( lpc = 0; lpc < names.length; lpc++ )
	{
	    try
	    {
		ResourceSampler rs;

		rs = (ResourceSampler)this.sampler.newInstance();
		rs.setTeam(Team.find(names[lpc]));
		this.samplers.addElement(rs);
	    }
	    catch(DeadTeamException e)
	    {
	    }
	    catch(TeamEjectionException e)
	    {
	    }
	    catch(NoSuchExportException e)
	    {
	    }
	}
	return this;
    }

    /**
     * Poll the teams.
     */
    public synchronized void run()
    {
	int iter;
	
	for( iter = 0; iter < this.iterations; iter++ )
	{
	    int lpc, size;

	    this.beginSampling();
	    size = this.samplers.size();
	    for( lpc = 0; lpc < size; lpc++ )
	    {
		ResourceSampler rs;
		TeamHandle th;

		rs = (ResourceSampler)this.samplers.elementAt(lpc);
		if( (th = rs.getTeam()) != null )
		{
		    try
		    {
			th.switchTo();
			{
			    rs.update();
			}
			th.returnFrom();
		    }
		    catch(DeadTeamException e)
		    {
			rs.setTeam(null);
			rs.empty();
		    }
		    catch(TeamEjectionException e)
		    {
			rs.setTeam(null);
			rs.empty();
		    }
		}
		else
		{
		    rs.empty();
		}
		rs.output();
	    }
	    this.endSampling();
	    try { this.wait(this.msInterval); }
	    catch(InterruptedException e) { }
	}
    }
    
    public void beginSampling()
    {
    }
    
    public void endSampling()
    {
    }
    
    public String toString()
    {
	return "ResourceMonitor[iterations="
	    + this.iterations
	    + "; msInterval="
	    + this.msInterval
	    + "; samplers="
	    + this.samplers
	    + "]";
    }
}
