/*
 * AggregateStateManager.java
 *
 * Copyright (c) 2002, 2003 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Lesser 
 * General Public License.
 * 
 * See the file "license-lesser.terms" for information on 
 * usage and redistribution of this file.
 */

package java.lang.isolate;

import edu.utah.janosvm.sys.TeamReservation;
import edu.utah.janosvm.sys.Team;
import edu.utah.janosvm.sys.TeamHandle;
import edu.utah.janosvm.sys.Exportable;
import edu.utah.janosvm.sys.ExportManager;
import edu.utah.janosvm.sys.Importable;
import edu.utah.janosvm.sys.TemporaryExport;
import edu.utah.janosvm.sys.TeamEjectionException;
import edu.utah.janosvm.sys.DeadTeamException;
import edu.utah.janosvm.sys.DuplicateExportException;
import edu.utah.janosvm.sys.NoSuchExportException;
import edu.utah.janosvm.sys.VMInstance;

/**
 * A global map of aggregate-wide Isolate state.
 *
 * <p>All globally useful state related to an isolate is kept here, in
 * <code>IsolateStatus</code> objects.  The <code>IsolateStatus</code>
 * is {@link Exportable} (exported under the isolate's GUID).  All
 * this state may continue to exist after an isolate has terminated
 * (any active handle on the isolate may query this or may create an
 * EventLink to the isolate).
 *
 * <p>The parent isolate will advance a new isolate from 'INVALID' to
 * 'NEWBORN' to 'STARTING'.  The main thread of the isolate will
 * advance it to 'RUNNING'.  The IsolateBackEnd in the isolate will
 * advance it to 'DYING' and 'DEAD'.  Thus we avoid race conditions.
 *
 * <p>XXX: Kernel entry/exit on each method is really verbose and ugly.
 */
/*package*/ class AggregateStateManager
{
  private static final String EXIT_ON_LAST =
    System.getProperty("java.lang.isolate.exit-on-last", "yes");
  
  /**
   * Global state for a single isolate.  The associated isolate
   * is implicit in the name this object is exported under.
   */
  private static class IsolateStatus
  extends Exportable
  implements TemporaryExport
  {
    /** The isolate guid this status represents. */
    private final Long guid;

    private IsolateState state = IsolateState.INVALID;

    /** 
     * Number of active EventLink references to this Isolate.  (An
     * optimization used to eliminate tracking starter/exit info for
     * most isolates.)
     */
    private int eventLinkCount;

    /**
     * Handle on the isolate that started the isolate this Isolate
     * represents (or -1).
     */
    private long starterGUID;

    /** Exit reason code for this isolate (or <code>null</code>) */
    private IsolateEvent.ExitReason exitReason;

    /** Exit status code for this isolate (or -1) */
    private int exitStatus;

    IsolateStatus(Long guid)
    {
      this.guid = guid;
      this.state = IsolateState.INVALID;
      this.eventLinkCount = 0;
      this.starterGUID = -1;
      // Default exit reason and exit status:
      this.exitReason = IsolateEvent.ExitReason.IMPLICIT_EXIT;
      this.exitStatus = 0;
    }

    /** 
     * Invoked by the {@link ExportManager} when there are no more
     * importables to this Exportable.  We're allowed to decide if
     * this Exportable should be dropped from the export table (and
     * thus be available for garbage collection) or if it should hang
     * around some more.
     *
     * Currently just used to get some extra asserts in.
     */
    public boolean expired()
    {
      Assert.require(eventLinkCount == 0, "expected link count 0, but got " +eventLinkCount);
// NOTE: no one bothers setting the aggregate-wide state to "dead"
//      if (state != IsolateState.DEAD)
//      {
//	VMInstance.println("WARNING: dropping IsolateState for not-yet-dead isolate: " +this.guid);
//      }
      return true;
    }

    public Object getName()
    {
      return this.guid;
    }

    void addEventLink()
    {
      this.eventLinkCount++;
    }
    
    void removeEventLink()
    {
      this.eventLinkCount--;
    }

    boolean hasEventLinks()
    {
      //return true; // disable event handle optimization
      return this.eventLinkCount > 0;
    }

    IsolateState getState()
    {
      synchronized(this)
      {
	return this.state;
      }
    }

    void advanceState(IsolateState newState)
    {
      synchronized(this)
      {
	if (this.state.comesAfter(newState))
	  throw new InternalError("Bogus state change (new=" +newState+ "; old=" +this.state+ ")");
	this.state = newState; // no need to copy, all states are global objects
      }
    }

    void setStarterGUID(long starterGUID)
    {
      Assert.require(this.starterGUID == -1);
      // Doesn't need to be synchronized
      this.starterGUID = starterGUID;
    }
    
    long getStarterGUID()
    {
      // Doesn't need to be synchronized
      Assert.require(this.starterGUID != -1);
      return this.starterGUID;
    }

    void setExitInfo(int s, IsolateEvent.ExitReason r)
    {
      synchronized(this)
      {
	Assert.require(this.starterGUID != -1);
	Assert.require(r != null);
	this.exitStatus = s;
	this.exitReason = r; // global objects, okay to not clone
      }
    }

    IsolateEvent.ExitReason getExitReason()
    {
      synchronized(this)
      {
	Assert.require(this.starterGUID != -1);
	return this.exitReason;
      }
    }

    int getExitStatus()
    {
      synchronized(this)
      {
	Assert.require(this.starterGUID != -1);
	return this.exitStatus;
      }
    }
    
    public String toString()
    {
      return "IsolateStatus[hash=0x" +Integer.toHexString(this.hashCode())+
	"; guid=" +this.guid+ 
	"; state=" +this.state+ 
//	"; evct=" +this.eventLinkCount+ 
//	"; starter=" +this.starterGUID+ 
	"; er=" +this.exitReason+ 
//	"; es=" +this.exitStatus+ 
	"]";
    }
  }

  /** 
   * A handle on an IsolateStatus for use by an Isolate object.
   * The handle assumes that it is *always* valid.  It should
   * only be invalidated when its owner drops it or dies.
   */
  static class IsolateStatusHandle
  extends Importable
  {
    private IsolateStatus isolateStatus = null;

    IsolateStatusHandle()
    {
    }

    protected void bindTo(Exportable ex)
    {
      Assert.requireLock(this);
      Assert.require(this.isolateStatus == null);
      Assert.require(ex != null);
      Assert.require(ex instanceof IsolateStatus);

      this.isolateStatus = (IsolateStatus)ex;
    }

    protected void unbindFrom()
    {
      Assert.requireLock(this);
      Assert.require(this.isolateStatus != null);
      this.isolateStatus = null;
    }

    void advanceState(IsolateState newState)
    {
      Trace.t(Trace.AGGSTATE, "ASM("+this+").advanceState(" +newState+ ")");
      synchronized(this)
      {
	if (this.isolateStatus != null)
	  this.isolateStatus.advanceState(newState);
      }
    }

    IsolateState getState()
    {
      Trace.t(Trace.AGGSTATEREAD, "ASM("+this+").getState()");
      synchronized(this)
      {
	if (this.isolateStatus != null)
	  return this.isolateStatus.getState(); // no need to copy, all IsolateState objects are global
	return IsolateState.DEAD;
      }
    }

    void setStarter(Isolate starter)
    {
      Assert.require(starter != null);

      Trace.t(Trace.AGGSTATE, "ASM("+this+").setStarter(" +starter+ ")");
      synchronized(this)
      {
	if (this.isolateStatus != null)
	{
	  Assert.require(starter != null);
	  if (this.isolateStatus.hasEventLinks())
	    this.isolateStatus.setStarterGUID(starter.getGUID().longValue());
	}
      }
    }

    long getStarterGUID()
    {
      Trace.t(Trace.AGGSTATEREAD, "ASM("+this+").getStarter()");
      synchronized(this)
      {
	if(this.isolateStatus != null)
	  return this.isolateStatus.getStarterGUID();
	return 1L;
      }
    }

    void setExitInfo(int exitStatus, IsolateEvent.ExitReason exitReason)
    {
      Trace.t(Trace.AGGSTATE, "ASM("+this+").setExitInfo(" +exitStatus+ ", " +exitReason+ ")");
      synchronized(this)
      {
	if (this.isolateStatus != null)
	{
	  // Just skip if isolate is already dead.
	  if (this.isolateStatus.getState() != IsolateState.DEAD)
	    if (this.isolateStatus.hasEventLinks())
	      this.isolateStatus.setExitInfo(exitStatus, exitReason);
	}
      }
    }
    
    IsolateEvent.ExitReason getExitReason()
    {
      Trace.t(Trace.AGGSTATEREAD, "ASM("+this+").getExitReason()");
      synchronized(this)
      {
	if (this.isolateStatus != null)
	  return this.isolateStatus.getExitReason();
	return IsolateEvent.ExitReason.IMPLICIT_EXIT;
      }
    }

    int getExitStatus()
    {
      Trace.t(Trace.AGGSTATEREAD, "ASM("+this+").getExitStatus()");
      synchronized(this)
      {
	if(this.isolateStatus != null)
	  return this.isolateStatus.getExitStatus();
	return 0;
      }
    }

    public String toString()
    {
      synchronized(this)
      {
	return "IsolateStatusHandle[" +this.isolateStatus+ "]";
      }
    }
  }
  
  /** 
   * A handle on an IsolateStatus used by an EventLink (enables state
   * tracking in the IsolateStatus). 
   */
  static class EventStatusHandle
  extends IsolateStatusHandle
  {
    EventStatusHandle()
    {
      super();
    }

    protected void bindTo(Exportable ex)
    {
      super.bindTo(ex);
      super.isolateStatus.addEventLink();
    }

    protected void unbindFrom()
    {
      IsolateStatus is = super.isolateStatus;
      super.unbindFrom();
      is.removeEventLink();
    }

    public String toString()
    {
      synchronized(this)
      {
	return "EventStatusHandle[" +super.isolateStatus+ "]";
      }
    }
  }

  private final Object lock = new Object();
  private int isolateCount = 0;

  IsolateStatusHandle registerNewIsolate(Long targetGUID)
  {
    return registerNewIsolate(targetGUID,
			      IsolateBackEnd.current().getIsolateServer());
  }

  /** 
   * Register a new isolate with the aggregate.  Invoked exactly
   * once for each isolate (see {@linkIsolate#Isolate}).
   */
  IsolateStatusHandle registerNewIsolate(Long targetGUID, IsolateServer isoServer)
  {
    Assert.require(targetGUID != null);
    Assert.require(isoServer != null);

    TeamReservation kres;
    IsolateStatusHandle retval = new IsolateStatusHandle();

    try
    {
      // ENTER KERNEL
      kres = isoServer.pushReservation();
      kres.visit();

      Trace.t(Trace.AGGSTATE, "ASM.registerNewIsolate(" +targetGUID+ ")");

      Long guid = new Long(targetGUID.longValue()); // clone the Long 
      IsolateStatus is = new IsolateStatus(guid);
      try
      {
	ExportManager em = Team.current().exportObject(guid, is);
	em.addImporter(retval);
      }
      catch (DuplicateExportException dee)
      {
	throw new InternalError("Already have a team exported with id " +guid, dee);
      }
      
      synchronized(lock)
      {
	isolateCount++;
      }

      // EXIT KERNEL
      kres.leave();
      kres.pop();

      return retval; //newHandleOn(targetGUID, isoServer);
    }
    catch (TeamEjectionException tee)
    {
      Trace.t(Trace.AGGSTATE, " tee.");
      // Nothing to do as we're toast if the kernel dies.
      throw new InternalError("Fatal error" +tee);
    }
    catch (DeadTeamException dte)
    {
      Trace.t(Trace.AGGSTATE, " dte.");
      // Nothing to do as we're toast if the kernel dies.
      throw new InternalError("Fatal error" +dte);
    }
  }

  /** Create a new handle on the given Isolate's IsolateState */ 
  IsolateStatusHandle newHandleOn(Long targetGUID)
  {
    return newHandleOn(targetGUID, 
		       IsolateBackEnd.current().getIsolateServer());
  }

  /** Create a new handle on the given Isolate's IsolateState */ 
  IsolateStatusHandle newHandleOn(Long targetGUID, IsolateServer isoServer)
  {
    Assert.require(targetGUID != null);
    Assert.require(isoServer != null);

    try
    {
      Trace.t(Trace.AGGSTATEREAD, "ASM.newHandleOn(" +targetGUID+ ")");
      IsolateStatusHandle ish = new IsolateStatusHandle();
      isoServer.importObject(ish, targetGUID);
      return ish;
    }
    catch (DeadTeamException dte)
    {
      throw new InternalError("Fatal error", dte);
    }
    catch (TeamEjectionException tee)
    {
      throw new InternalError("Fatal error", tee);
    }
    catch (NoSuchExportException nsee)
    {
      throw new InternalError("Expected an IsolateState for " +targetGUID+ " but it does not exist.", nsee);
    }
  }

  EventStatusHandle newEventHandleOn(Long targetGUID)
  {
    Assert.require(targetGUID != null);

    try
    {
      IsolateServer isoServer = IsolateBackEnd.current().getIsolateServer();
      Trace.t(Trace.AGGSTATEREAD, "ASM.newEventHandleOn(" +targetGUID+ ")");
      EventStatusHandle esh = new EventStatusHandle();
      isoServer.importObject(esh, targetGUID);
      return esh;
    }
    catch (DeadTeamException dte)
    {
      throw new InternalError("Fatal error", dte);
    }
    catch (TeamEjectionException tee)
    {
      throw new InternalError("Fatal error", tee);
    }
    catch (NoSuchExportException nsee)
    {
      throw new InternalError("Expected an IsolateState for " +targetGUID+ " but it does not exist.", nsee);
    }
  }

  /** 
   * Invoked by {@link IsolateBackEnd#event} to indicate when
   * the given isolate has died.
   *
   * <p>Warning: Invoked in the context of an arbitrary team
   */
  void isolateDied(IsolateBackEnd ibe)
  {
    synchronized(lock)
    {
      isolateCount--;
      Trace.t(Trace.AGGSTATE, "isolateDied(" +ibe+ ") totalleft=" +isolateCount);
      if (isolateCount == 0 && EXIT_ON_LAST.equals("yes") )
      {
	System.out.println("No more isolates to manage.  Shutting down."); // XXX
	VMInstance.shutdown();
      }
    }
  }
}
