/*
 * EventLink.java
 *
 * Copyright (c) 2002,2003 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Lesser 
 * General Public License.
 * 
 * See the file "license-lesser.terms" for information on 
 * usage and redistribution of this file.
 */

package java.lang.isolate;

import edu.utah.janosvm.sys.Team;
import edu.utah.janosvm.sys.TeamHandle;

/*
 * NOTES:
 *
 * EventLink has two subclasses: EventLink_Attached and
 * EventLink_Unattached.  Attached is when the EventLink is in the
 * receiver Isolate, Unattached is for otherwise.  The base EventLink
 * class tracks the receiver and target isolates, and provides
 * wrappers for receiveEvent.  Subclasses implement constructor,
 * close, isOpen, and receiveEvent.
 */
/*package*/ abstract class EventLink
  extends Link 
{ 
  private static final IsolateGUID guidManager = new IsolateGUID(0);

  /** Globally unique name for EventLink rendezvous point. */
  /*package*/ final String name;
  /*package*/ final Isolate target;
  /*package*/ final Isolate receiver;

  /*package*/ EventLink(Isolate target, Isolate receiver, String name)
  {
    super("Event");
    Assert.require(name != null);
    Assert.require(target != null);
    Assert.require(receiver != null);
    Assert.require(Team.current().owns(target));
    Assert.require(Team.current().owns(receiver));
    Assert.require(Team.current().owns(name));

    this.target = target;
    this.receiver = receiver;
    this.name = name;
  }

  /*package*/ static EventLink newInstance(Isolate target, Isolate receiver, String name)
  {
    Isolate curr = Isolate.currentIsolate();

    if (target == null)
      throw new NullPointerException("target is null");

    if (receiver == null)
      throw new NullPointerException("receiver is null");

    // If this is a new unique rendezvous point, create a unique name for it
    if (name == null)
      name = "EventLink#" + guidManager.nextGUID();

    if (receiver.equals(curr))
      return EventLink_Attached.findEventLink(target, receiver, name);
    
    return new EventLink_Unattached(target, receiver, name);
  }

  /**
   * Returns false.  No user code can ever invoke <tt>send</tt> on 
   * an EventLink.
   */
  public boolean isSender(Isolate i) 
  { 
    return false;
  }

  public boolean isReceiver(Isolate i)
  {
    return i.equals(this.receiver);
  }

  /**
   * Only *receiver* Isolate can close, otherwise <tt>IllegalStateException</tt>.
   */
  public abstract void close() 
    throws IllegalStateException;


  public abstract boolean isOpen() 
    throws IllegalStateException;

  /**
   * Throws <tt>IllegalStateException</tt>, as no user code is allowed
   * to send on an EventLink.
   */
  public boolean send(IsolateMessage lm, long millis, int nanos)
    throws ClosedLinkException, 
    IllegalStateException
  {
    throw new IllegalStateException("An EventLink is a receive-only Link.");
  }
  
  public IsolateMessage receive(long millis, int nanos)
    throws ClosedLinkException, IllegalStateException
  {
    IsolateMessage lm = null;
    lm = IsolateMessage.newIsolateEventMessage(receiveEvent(millis, nanos));
    return lm;
  }

  public IsolateEvent receiveEvent()
    throws ClosedLinkException, IllegalStateException
  {
    IsolateEvent rc;
    rc = receiveEvent(-1, 0);
    Assert.ensure(rc != null);
    return rc;
  }

  public abstract IsolateEvent receiveEvent(long millis, int nanos)
    throws ClosedLinkException, IllegalStateException;

  public boolean equals(Object o)
  {
    if (o instanceof EventLink)
    {
      EventLink other = (EventLink)o;
      return this.target.equals(other.target)
	&& this.receiver.equals(other.receiver);
    }
    return false;
  }

  /*package*/ Link clone(TeamHandle sourceTeam, TeamHandle currentTeam)
    throws ClosedLinkException
  {
    // XXX assert(!sourceTeam.equals(currentTeam))

    // 'currentTeam' is the destination Team for the new Link.
    return EventLink.newInstance(this.target.clone(sourceTeam, currentTeam),
				 this.receiver.clone(sourceTeam, currentTeam),
				 Isolate.cloneString(this.name));
  }

  public String toString()
  {
    return "EventLink[" +this.name+ "; receiver=" +this.receiver+"; target=" +this.target+ "]";
  }

}
