/*
 * EventLink_Attached.java
 *
 * Copyright (c) 2002,2003 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Lesser 
 * General Public License.
 * 
 * See the file "license-lesser.terms" for information on 
 * usage and redistribution of this file.
 */

package java.lang.isolate;

import edu.utah.janosvm.sys.Team;
import edu.utah.janosvm.sys.TeamHandle;
import edu.utah.janosvm.sys.EventTeamHandle;
import edu.utah.janosvm.sys.TeamEjectionException;
import edu.utah.janosvm.sys.TeamEventHandler;
import edu.utah.janosvm.sys.TeamReservation;
import edu.utah.janosvm.sys.TeamEvent;
import java.util.HashMap;

/*package*/ class EventLink_Attached
	extends EventLink 
{ 
  private static final Class eventLink_AttachedClass = EventLink_Attached.class;

  static HashMap currentIsolateLinkTable()
  {
    HashMap m;
    synchronized(eventLink_AttachedClass)
    {
      m = (HashMap)Team.getCurrentClassLocal(eventLink_AttachedClass);
      if (m == null)
      { 
	m = new HashMap();
	Team.setCurrentClassLocal(eventLink_AttachedClass, m);
      }
    }
    return m;
  }

  static EventLink_Attached findEventLink(Isolate target, Isolate receiver, String name)
  {
    HashMap table = currentIsolateLinkTable();
    EventLink_Attached el;

    /**
     * Find the pre-existing event link with the given name (if any).
     */
    synchronized(table)
    {
      if (table.containsKey(name))
	el = (EventLink_Attached)table.get(name);
      else
      {
	el = new EventLink_Attached(target, receiver, name);
	table.put(name, el);
      }
    }

    Assert.ensure(el != null);

    return el;
  }
  
  private boolean closed;
  private final Object lock = new Object();
  private final EventTeamHandle eth;

  private final AggregateStateManager.EventStatusHandle statusHandle;
  private final IsolateEvent[] buf;
  private int bufHead;
  private int bufTail;

  /*package*/ EventLink_Attached(Isolate target, Isolate receiver, String name)
  {
    super(target, receiver, name);

    this.closed = false;
    this.buf = new IsolateEvent[3]; // STARTING, EXITING, TERMINATED
    this.bufHead = 0;
    this.bufTail = 0;


    TeamHandle targetTeamHandle = target.getTeamHandle();

    if (targetTeamHandle != null)
    {
      //Register target with global status map (enabling status updates to be recoreded)
      //System.out.println("Adding listener " +target+ " because of link " +this);
      this.statusHandle = Isolate.aggregateStateManager.newEventHandleOn(target.getGUID());

      // Attach an event handler to target isolate's team
      this.eth = new EventTeamHandle(new TeamEventHandler() {
	  public void event(EventTeamHandle eth, TeamReservation tr, TeamEvent ev)
	  {
	    EventLink_Attached.this.handleEvent(tr, ev);
	  }
	}, targetTeamHandle);
    }
    else
    {
      this.closed = true;
      this.statusHandle = null;
      this.eth = null;
    }
    
  }

  private void handleIsoEvent(TeamReservation tr, IsolateEvent.Type isoEvType)
  {
    // Starts execution in kernel context

    try
    {
      tr.visit();
      synchronized(this.lock)
      {
	// Silently skip if link is closed
	if (!this.closed)
	{
	  Isolate target = super.target;
	  IsolateEvent ev;

	  if (isoEvType == IsolateEvent.Type.STARTING)
	  {
	    Isolate starter = Isolate.newHandle(new Long(this.statusHandle.getStarterGUID()));
	    ev = new IsolateEvent(target, IsolateEvent.Type.STARTING,
				  starter, null, 0);
	  }
	  else if (isoEvType == IsolateEvent.Type.EXITING)
	  {
	    ev = new IsolateEvent(target, IsolateEvent.Type.EXITING,
				  null, null, 0);
	  }
	  else if (isoEvType == IsolateEvent.Type.TERMINATED)
	  {
	    IsolateEvent.ExitReason exitReason = this.statusHandle.getExitReason();
	    int exitStatus = this.statusHandle.getExitStatus();

	    // defaults:
	    if (exitReason == null)
	      exitReason = IsolateEvent.ExitReason.IMPLICIT_EXIT;
	    if (exitStatus == -1)
	      exitStatus = 0;

	    ev = new IsolateEvent(target, IsolateEvent.Type.TERMINATED,
				  null, exitReason, exitStatus);

	    // Revoke the status handle, we no longer need it
	    this.statusHandle.revoke();
	  }
	  else
	  {
	    throw new InternalError("unexpected event type " +isoEvType);
	  }

	  Assert.require(bufHead <= bufTail);
	  Assert.require(bufHead < buf.length);
	  Assert.require(bufTail < buf.length);
	  buf[bufTail] = ev;
	  bufTail++;
	  this.lock.notifyAll();

	  Trace.t(Trace.ISO_EVENTS, "Stored " +ev+ " in " +this);

	  Assert.ensure(bufTail <= buf.length);
	}
      }
      tr.leave();
    }
    catch (TeamEjectionException tee)
    {
      // booted from the listener back into the kernel... oh well.
    }
    
  }

  private void handleEvent(TeamReservation tr, TeamEvent ev)
  {
    // Still executing in kernel context.

    Trace.t(Trace.JANOS_EVENTS, "EventLink_Attached got event" + ev);
    
    switch (ev.getEvent())
    {
    case TeamEvent.UNCAUGHT_EXCEPTION:
    case TeamEvent.ADD_DEPENDENCY:
    case TeamEvent.REM_DEPENDENCY:
    case TeamEvent.FINISHED:
    case TeamEvent.DEAD:
      // Don't care
      // XXX Use the mask on the EventTeamHandle constructor to prevent these
      break;
    case IsolateEvent.Type.EXITING_JANOSVM:
      // XXX need to send this event on Runtime.exit and Runtime.halt (use a shutdownHook)
      break;
    case IsolateEvent.Type.STARTING_JANOSVM:
      handleIsoEvent(tr, IsolateEvent.Type.STARTING);
      break;
    case TeamEvent.TERMINATED:
      // XXX This isn't quite right (a bit late), but will do for now.
      handleIsoEvent(tr, IsolateEvent.Type.EXITING);
      break;
    case TeamEvent.BURIED:
      handleIsoEvent(tr, IsolateEvent.Type.TERMINATED);
      break;
    default:
      // XXX
      throw new InternalError("EventLink_Attached.handleEvent(): unknown event " +ev);
    }
  }

  public void close() 
  {
    Assert.ensure(Isolate.currentIsolate().equals(super.receiver));

    synchronized(this.lock)
    {
      if (!this.closed)
      {
	this.closed = true;
	for (int i = 0; i < buf.length; i++)
	  this.buf[i] = null;

	this.statusHandle.revoke();
	Trace.t(Trace.ISO_EVENTS, "Closed " +this);
      }
    }
  }

  public boolean isOpen() 
  {
    synchronized (this.lock)
    {
      return !this.closed;
    }
  }

  public IsolateEvent receiveEvent(long millis, int nanos)
    throws ClosedLinkException
  {
    Assert.ensure(Isolate.currentIsolate().equals(super.receiver));

    synchronized(this.lock)
    {
      if (this.closed)
	throw new ClosedLinkException();

      Assert.require(bufHead <= bufTail);
      Assert.require(bufHead <= buf.length);
      Assert.require(bufTail <= buf.length);

      // if just polling, just do a poll
      if ((millis == 0)
	  && (bufHead == bufTail))
	return null;

      // Wait for something to be in the queue...
      while (bufHead == bufTail)
      {
	try
	{
	  if (millis < 0) // negative values mean wait indefinitely
	    this.lock.wait();
	  else 
	    this.lock.wait(millis); // XXX nanos
	}
	catch (InterruptedException ie)
	{
	  Thread.currentThread().interrupt(); // XXX set interrupt flag
	}
      }

      IsolateEvent ev = buf[bufHead];
      bufHead++;

      Assert.ensure(ev != null);
      Assert.ensure(bufHead <= bufTail);
      Assert.ensure(bufHead <= buf.length);

      Trace.t(Trace.ISO_EVENTS, "Fetched " +ev+ " from " +this);

      // Automagically close the link if the Terminated event is delivered...
      if (ev.getType() == IsolateEvent.Type.TERMINATED)
	this.close();

      return ev;
    }
  }
}
