/*
 * Isolate.java
 *
 * Copyright (c) 2002, 2003 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Lesser 
 * General Public License.
 * 
 * See the file "license-lesser.terms" for information on 
 * usage and redistribution of this file.
 */

package java.lang.isolate;

import java.util.Vector;
import java.util.HashMap;
import java.util.StringTokenizer;
import edu.utah.janosvm.resources.*; // XXX
import edu.utah.janosvm.sys.*; // XXX
import java.io.FileDescriptor;
import java.io.IOException;
import java.util.Properties;
import java.util.prefs.Preferences;
import java.util.prefs.TransientPreferences;
import java.util.prefs.BackingStoreException;


/**
 * @author Pat Tullmann <tullmann@cs.utah.edu>
 */
public final class Isolate
{
  /** Global Isolate GUID manager. */
  private static final IsolateGUID guidManager = new IsolateGUID(0);

  /** Global status (mapping isolate handles to starter/exit status) */
  /*package*/ static final AggregateStateManager aggregateStateManager = new AggregateStateManager();

  // cache the lookup for the class objects.  These are kernel class
  // objects, btw.
  private static final Class isolateBackEndClass = IsolateBackEnd.class;
  private static final Class isolateClass = Isolate.class;

  private static final String bootclassPath = stringifyStrArray(ClassPathSpec.current().getPaths());

  private static final IsolatePermission CREATE_PERMISSION =
    new IsolatePermission("create");
  
  private static final IsolatePermission CONTROL_PERMISSION =
    new IsolatePermission("control");
  
  private static final IsolatePermission CONTEXT_PERMISSION =
    new IsolatePermission("context");
  
  private final Long guid;

  /** 
   * Handle on the global isolate status.  This handle is always valid
   * until the owner (not the target) is dead, or the owner GC's this
   * Isolate object.
   */
  private final AggregateStateManager.IsolateStatusHandle statusHandle;

  private static String stringifyStrArray(String[] paths)
  {
    if (paths.length == 0)
      return "";
    else
    {
      StringBuffer p = new StringBuffer(200);
      for (int i = 0; i < paths.length-1; i++)
	p = p.append(paths[i]).append(":"); // XXX use 'path.separator'
      p.append(paths[paths.length-1]);
      return p.toString();
    }
  }

  /** Default class path for new Isolates. */
  private static final String[] defaultClassPath = { "." };

  class IsolateEventHandler
    implements TeamEventHandler
  {
    private Vector cleaners = new Vector();
    
    public IsolateEventHandler()
    {
    }
    
    boolean cleanersUsed()
    {
      // assert synchronized(this)
      return (cleaners == null);
    }

    void addShutdownHook(Runnable r)
    {
      // assert synchronized(this)
      this.cleaners.addElement(r);
    }
    
    public void event(final EventTeamHandle eth, TeamReservation tr, TeamEvent ev)
    {
      Trace.t(Trace.JANOS_EVENTS, "Isolate got event" + ev);

      switch(ev.getEvent())
      {
      case TeamEvent.FINISHED:
	/* 
	 * Last non-daemon thread exited, Team should be dying soon... 
	 * Actual terminate handler is installed in the Team itself
	 * (see IsolateBackEnd)
	 *
	 * Will get a TERMINATED event shortly, and that will
	 * do the dying bit.
	 */
	break;
      case TeamEvent.TERMINATED:
	// Note: The target isolate's IsolateBackEnd will take care of
	// advancing the IsolateStatus state...

	Isolate.this.remoteState = IsolateState.DYING;
	try
	{
	  tr.visit();
	  synchronized( this )
	  {
	    Runnable cleaner;
	    int lpc, size;

	    size = this.cleaners.size();
	    for( lpc = 0; lpc < size; lpc++ )
	    {
	      cleaner = (Runnable)this.cleaners.elementAt(lpc);
	      cleaner.run();
	    }
	    this.cleaners = null;
	  }
	  tr.leave();
	}
	catch(TeamEjectionException e)
	{
	}
	break;
      case TeamEvent.BURIED:
	Isolate.this.remoteState = IsolateState.DEAD;
	Isolate.this.teamH = null;
	break;
      default:
	// ignored
      }
    }
  }
  
  final IsolateEventHandler handler = new IsolateEventHandler();

  /**
   * State of remote Isolate.  This is not always 
   * correct, though the remote Isolate should never
   * be in an "earlier" life state than this state.
   *
   * XXX is this at all useful?  I guess if its "DEAD", then you
   * can do everything locally...
   */
  private IsolateState remoteState = IsolateState.INVALID;

  /**
   * Handle on the Team associated with this Isolate (local or remote).
   */
  private EventTeamHandle teamH;

  private static Long cloneLong(Long l)
  {
    Assert.require(l != null);
    return new Long(l.longValue());
  }

  /*package*/ static String cloneString(String s)
  {
    if (s == null)
      return null;
    
    // DEEP clone:
    char[] chars = new char[s.length()];
    s.getChars(0, chars.length, chars, 0);
    return new String(chars);
  }

  private static String[] cloneStringArray(String[] s)
  {
    if (s == null)
      return null;

    String[] ns = new String[s.length];
    for (int i = 0; i < s.length; i++)
    {
      ns[i] = cloneString(s[i]);
    }
    return ns;
  }

  private void installDefaultPrefs(TransientPreferences childPrefs)
  {
  }

  private void fixupImmutablePrefs(TransientPreferences childPrefs)
  {
    childPrefs.putBoolean("edu.utah.janosvm", true);
    childPrefs.putBoolean("java.class.allow-caching", false);
    childPrefs.putBoolean("java.lang.isolate.jni-isolation", false);
    childPrefs.putBoolean("java.lang.isolate.vm-isolation", false);
  }


  private void cloneContextNode(TeamReservation inittr,
				TransientPreferences parentPrefs,
				TransientPreferences childPrefs)
    throws TeamEjectionException
  {
    // Copy all the keys from the parentPrefs (in the parent team)
    inittr.leave();
    String[] skeys = parentPrefs.keys();
    Trace.t(Trace.ISO_CONTEXT, "non-null parentprefs has " +skeys.length+ " keys");
    inittr.visit();

    for (int i = 0; i < skeys.length; i++)
    {
      String val = parentPrefs.get(skeys[i], null);
      Trace.t(Trace.ISO_CONTEXT, "copying key " +skeys[i]+ " = " +val);
      if (val != null)
      {
	childPrefs.put(cloneString(skeys[i]), cloneString(val));
      }
    }
    
    inittr.leave();
    // Merge all the children
    String[] pchildren = parentPrefs.childrenNames();
    // Trace.t(Trace.ISO_CONTEXT, "Supplied prefs has " +pchildren.length+ " children");
    inittr.visit();

    for (int i = 0; i < pchildren.length; i++)
    {
      TransientPreferences pchild = (TransientPreferences)parentPrefs.node(pchildren[i]);
      TransientPreferences cchild = (TransientPreferences)childPrefs.node(cloneString(pchildren[i]));
      
      // Recursive call
      Trace.t(Trace.ISO_CONTEXT, "Cloning pref sub-tree " +pchildren[i]);
      cloneContextNode(inittr, pchild, cchild);
    }
  }

  /**
   * Deeply clone the context.  Also installs default values (before the
   * clone) and fixes up immutable values (after the clone).  Does *not*
   * remove the java.properties node (that is removed inside the new
   * isolate initialization code after it defines the system settings.)
   */
  private TransientPreferences cloneContext(TeamReservation inittr, 
					    TransientPreferences parentPrefs)
    throws TeamEjectionException
  {
    TransientPreferences childPrefs = new TransientPreferences();

    // Install default prefs that may be overridden
    installDefaultPrefs(childPrefs);

    // Copy parent-specified prefs into child context
    if (parentPrefs != null)
      cloneContextNode(inittr, parentPrefs, childPrefs);
    
    // Install immutable prefs (perhaps overwriting illegal parent prefs)
    fixupImmutablePrefs(childPrefs);

    return childPrefs;
  }

  private static IsolateMessage cloneMessage(IsolateMessage m, TeamHandle source, TeamHandle current)
  {
    try
    {
      return m.translate(source, current);
    }
    catch (ClosedLinkException cle)
    {
      throw new IsolateResourceError("Closed link exception passing an IsolateMessage: " +cle.getMessage(), cle);
    }
    catch (IOException ioe) // XXX
    {
      throw new IsolateResourceError("IOError passing an IsolateMessage: " +ioe.getMessage(), ioe);
    }
  }

  /**
   * XXX race condition between copying an FD out of the isolate message
   * and then that fd being closed and something completely unrelated being
   * re-opened on that FD before we get to the resource subscription.
   */ 
  private static StdioSpec makeStdioSpec(IsolateMessage stdin, 
					 IsolateMessage stdout, 
					 IsolateMessage stderr)
    throws IsolateStartupException
  {
    String stdinURI;
    String stdoutURI;
    String stderrURI;
    
    try
    {
      if (stdin == null)
	stdinURI = "builtin://none";
      else if (! stdin.isIOMessage() ||
	       ! ((IsolateMessage_IOMessage)stdin).isReadable())
	throw new IllegalArgumentException("stdin message is not suitable for input redirection");
      else
	stdinURI = "fd://" + ((IsolateMessage_IOMessage)stdin).getFD();
      
      if (stdout == null)
	stdoutURI = "builtin://none";
      else if (! stdout.isIOMessage() ||
	       ! ((IsolateMessage_IOMessage)stdout).isWritable())
	throw new IllegalArgumentException("stdout message is not suitable for output redirection");
      else
	stdoutURI = "fd://" + ((IsolateMessage_IOMessage)stdout).getFD();
      
      if (stderr == null)
	stderrURI = "builtin://none";
      else if (! stderr.isIOMessage() ||
	       ! ((IsolateMessage_IOMessage)stderr).isWritable())
	throw new IllegalArgumentException("stderr message is not suitable for output redirection");
      else
	stderrURI = "fd://" + ((IsolateMessage_IOMessage)stderr).getFD();
      
      return new StdioSpec(stdinURI, stdoutURI, stderrURI);
    }
    catch (IOException ioe)
    {
      throw new IsolateStartupException("stdio message is broken", ioe);
    }
  }

  private static FileSystemSpec makeFileSystemSpec(TransientPreferences tp)
  {
    FileSystemSpec retval = null;
    String root;
    
    if( (tp != null) &&
	((root = tp.get("edu.utah.janosvm.resources.FileSystem.root", null))
	 != null) )
    {
      retval = new FileSystemSpec(root);
    }
    return retval;
  }

  private static String[] getClassPath(TransientPreferences context, String[] defcp)
  {
    if (context != null)
    {
      Preferences props = context.node("java.properties");
      String cpstr = props.get("java.class.path", null);
      if (cpstr != null)
      {
	StringTokenizer st;
	String result[];
	int lpc;
	
	st = new StringTokenizer(cpstr, java.io.File.pathSeparator);
	result = new String[st.countTokens()];
	for( lpc = 0; lpc < result.length; lpc++ )
	{
	  result[lpc] = st.nextToken();
	}
	return result;
      }
    }
    return defcp;
  }

  /*package*/ static Isolate newHandle(Long guid)
  {
    EventTeamHandle t = new EventTeamHandle();
    try
    {
      Object rc = Team.find(t, isolateName(guid));
      if (rc == null)
	throw new InternalError("Cannot find team with name " +isolateName(guid));
      Isolate i = new Isolate(guid, t);
      t.setEventHandler(i.handler);
      return i;
    }
    catch (DeadTeamException dte)
    {
      // Must be that remote team is dead:
      return new Isolate(guid, null);
    }
    catch (TeamEjectionException tee)
    {
      throw new InternalError("Don't know how to progress.");
    }
    catch (NoSuchExportException nsee)
    {
      // Must have died a long time ago?
      return new Isolate(guid, null);
    }
  }

  /** 
   * Copy given array so that mutations to it will not work around
   * safety and security checks on the contents.
   */
  private Object[] checkArray(Object arr[])
  {
    Object retval[] = null;
    
    if( arr != null )
    {
      int lpc;

      retval = (Object[])arr.clone();
      for( lpc = 0; lpc < retval.length; lpc++ )
      {
	if( retval[lpc] == null )
	{
	  throw new IllegalArgumentException("Null element at index: "
					     + lpc);
	}
      }
    }
    return retval;
  }

  /**
   * We must create a clone of the given type of IsolateMessage but as
   * a wrapper for the fd in the given spec.
   */
  private IsolateMessage makeStdioMessage(String spec, IsolateMessage rawSourceMessage)
    throws IOException
  {
    if (rawSourceMessage == null)
      return null;

    final int skipLen = 5; // "fd://".length();
    Assert.require(spec.startsWith("fd://"));
    Assert.require(rawSourceMessage instanceof IsolateMessage_IOMessage);

    IsolateMessage_IOMessage sourceMessage = (IsolateMessage_IOMessage)rawSourceMessage;
    int fd = Integer.parseInt(spec.substring(skipLen));
    return sourceMessage.cloneTypeAroundFD(fd);
  }

  public Isolate(String mainClass, String[] mainArgs) 
    throws IsolateStartupException
  {
    this(mainClass, mainArgs,
	 null, // Context
	 // XXX ArrayIndexOutOfBoundsException:
	 currentIsolateIOBindings()[0],
	 currentIsolateIOBindings()[1],
	 currentIsolateIOBindings()[2]);
  }

  public Isolate(String mainClass, String[] mainArgs, 
		 TransientPreferences context,
		 IsolateMessage stdin, 
		 IsolateMessage stdout, 
		 IsolateMessage stderr)
    throws IsolateStartupException
  {
    Team parent = Team.current();
    SecurityManager sm;
    String constructionFailure = null;
    
    this.remoteState = IsolateState.INVALID;

    Trace.t(Trace.ISO_START, "Creating new Isolate from parent " +Isolate.currentIsolate());

    if( (sm = System.getSecurityManager()) != null )
    {
      sm.checkPermission(CREATE_PERMISSION);
    }
    if( mainClass == null )
    {
      throw new NullPointerException("Null main class name");
    }
    if( mainArgs == null )
      mainArgs = new String[0];
    mainArgs = (String[])checkArray(mainArgs);

    // XXX A hack to get the cached keyset loaded in the owner's team!
    if (context != null)
    {
      context.keys();
      context.childrenNames();
    }

    this.guid = new Long(guidManager.nextGUID());

    try
    {
      // Throws IsolateStartupException if any io message is bogus
      StdioSpec stdioSpec = makeStdioSpec(stdin, stdout, stderr);
      ClassPathSpec classPathSpec = new ClassPathSpec(getClassPath(context, defaultClassPath));

      Trace.t(Trace.ISO_START, "  NameSpaceSpec=" +NameSpaceSpec.current());
      Trace.t(Trace.ISO_START, "  NativeSpec=" +NativeSpec.current());
      Trace.t(Trace.ISO_STARTFD, "  StdioSpec=" +stdioSpec);
      Trace.t(Trace.ISO_START, "  ClassPathSpec=" +classPathSpec);

      ResourceSpec resources[] = new ResourceSpec[] {
	classPathSpec,
	NameSpaceSpec.current(),
	NativeSpec.current(), /* XXX for loading of native libraries. */
	stdioSpec,
	makeFileSystemSpec(context),
      };

      TeamHandle rc;
      this.teamH = new EventTeamHandle(this.handler);
      rc = Team.create(this.teamH, 
		       "Isolate-" + this.guid,
		       resources);
      if (rc == null)
 	throw new DeadTeamException("Mustadied"); // XXX message
    }
    catch (DuplicateExportException dee)
    {
      // XXX perhaps we shouldn't name teams?
      throw new InternalError("Already have an Isolate named 'Isolate-" +this.guid+ "'??");
    }
    catch (DeadTeamException dte)
    {
      // XXX why would this happen?
      throw new IsolateStartupException("New team terminated: "
					+ dte.getMessage());
    }
    catch (ResourceSpecException rse)
    {
      throw new IsolateResourceError("Resource Alloation failure: "
				     + rse.getUnsatisfiedSpec());
    }

    // Register this isolate with the aggregate state manager.
    this.statusHandle = aggregateStateManager.registerNewIsolate(this.guid);

    // Just create a minimal child.  Copy in some initial objects (no
    // need to keep them here), but don't start any new threads.
    // The start() method will set those up.
    TeamReservation inittr = null;
    try
    {
      inittr = this.teamH.pushReservation();
      try
      {
	// Switch into the new child context
	inittr.visit();
	Team child = Team.current();
	
	child.notifyOnFinish();

	// Copy state into the new child.
	child.setDescription(mainClass);
	StdioSpec currStdioSpec = StdioSpec.currentFD();
	Trace.t(Trace.ISO_STARTFD, "    New Team self.StdioSpec(fd)=" +currStdioSpec);
	try
	{
	  IsolateBackEnd childIsoBE = new IsolateBackEnd(child,
							 cloneLong(this.guid),
							 cloneString(mainClass),
							 cloneStringArray(mainArgs),
							 cloneContext(inittr, context),
							 //stdin == null ? null : cloneMessage(stdin, parent, child),
							 //stdout == null ? null : cloneMessage(stdout, parent, child),
							 //stderr == null ? null : cloneMessage(stderr, parent, child));
							 makeStdioMessage(currStdioSpec.getInput(), stdin),
							 makeStdioMessage(currStdioSpec.getOutput(), stdout),
							 makeStdioMessage(currStdioSpec.getError(), stderr));
	  child.setClassLocal(isolateBackEndClass, childIsoBE);
	  Isolate childIsolate = new Isolate(cloneLong(this.guid), child);
	  child.setClassLocal(isolateClass, childIsolate);
	
	  // Fixup the child's view of its context/prefs, set system properties, etc.
	  context = childIsolate.currentIsolateContext();
	  Assert.ensure(context != null);
	  childSidePrefsFixup(context);

	  // Now the child Isolate is "real".  Advance state from INVALID
	  // to NEWBORN.
	  childIsolate.statusHandle.advanceState(IsolateState.NEWBORN);
	}
	// XXX catch other stuff too?
	catch (IOException ioe)
	{
	  constructionFailure = ioe.getMessage();
	}
	
	inittr.leave();

	if (constructionFailure != null)
	  constructionFailure = cloneString(constructionFailure);
      }
      catch (TeamEjectionException tee)
      {
	this.remoteState = IsolateState.DEAD;
	this.teamH = null;
	// XXX fix exception type
	throw new IsolateStartupException("Ba-da-boom" + tee);
      }

      if (constructionFailure != null)
	throw new IsolateStartupException("Some sort of construction time problem:" +constructionFailure); // XXX fixup message
	
      // Advance state of this handle
      this.remoteState = IsolateState.NEWBORN;
    }
    catch (DeadTeamException dte)
    {
      // XXX move code into remoteIsDead()

      this.remoteState = IsolateState.DEAD;
      this.teamH = null;
      // XXX fix exception type
      throw new IsolateStartupException("Ba-da-boom" +dte);
    }
    finally
    {
      if (inittr != null)
	inittr.pop();
    }
  }

  /**
   * Constructor for creating an Isolate object that refers to the
   * given Team (and implicitly that Team's IsolateBackEnd).
   */
  /*package*/ Isolate(Long guid, EventTeamHandle t)
  {
    Assert.require(Team.current().owns(guid));
    Assert.require(Team.current().owns(t));

    this.guid = guid;
    this.teamH = t;
    this.remoteState = IsolateState.INVALID;
    this.statusHandle = aggregateStateManager.newHandleOn(this.guid);
  }


  private static void childSidePrefsFixup(TransientPreferences context)
  {
    TransientPreferences propsPrefs = (TransientPreferences)(context.node("java.properties"));
    String[] keys = propsPrefs.keys();
    Properties sysprops = System.getProperties();

    Trace.t(Trace.ISO_CONTEXT, "Fixup " +keys.length+ " props from parent");

    for (int i = 0; i < keys.length; i++)
    {
      String val = propsPrefs.get(keys[i], null);
      Trace.t(Trace.ISO_CONTEXT, "Fixup " +keys[i]+ " -> " +val);
      if (val != null) // shouldn't ever be null
	sysprops.setProperty(keys[i], val);
    }

    // Remove the java.properties node, as the child must go through
    // Properties API to get properties
    propsPrefs.removeNode();

    // Overwrite a couple critical system properties with relevant values.
    sysprops.setProperty("java.bootclass.path", bootclassPath);
    sysprops.setProperty("java.class.path", stringifyStrArray(ClassPathSpec.current().getPaths()));
  }

  public void start(IsolateMessage[] messages)
    throws IllegalStateException, IsolateStartupException
  {
    TeamHandle teamH = this.teamH;

    // This check is not sufficient (multiple threads may get here at
    // the same time), but the check at the end will catch that.  This
    // check prevents dereferencing this.teamH on a dead team (its
    // null).
    if (teamH == null)
    {
      Assert.require(this.remoteState == IsolateState.DEAD);
      throw new IllegalStateException("starting an isolate that has been terminated (DEAD)");
    }
    
    try
    {
      TeamHandle parentTeam = Team.current();
      Isolate parentIsolate = Isolate.currentIsolate();

      Trace.t(Trace.ISO_START, "Starting " +this+ " from parent " +parentIsolate);

      // XXX check that this is the creator Isolate.

      // XXX Need to prime some message types.
      if( messages != null )
      {
	messages = (IsolateMessage[])this.checkArray(messages);
	for (int i = 0; i < messages.length; i++)
	{
	  try
	  {
	    messages[i].pretranslate();
	  }
	  catch(IOException e)
	  {
	      throw new IsolateStartupException(
		"Exception while encoding message at index: "
		+ i,
		e);
	  }
	}
      }

      // Jump into the child, copy the links and start the main thread.

      teamH.switchTo();
      IsolateBackEnd childBE = IsolateBackEnd.current();
      if (childBE == null)
	throw new IllegalStateException("child initialization was not completed."); // XXX really broken

      // copy the messages.  All allocations are done in child context.
      IsolateMessage[] initialMessages;
      if (messages == null)
      {
	Trace.t(Trace.ISO_START, "making empty initial message array for child");
	initialMessages = new IsolateMessage[0];
      }
      else
      {
	Trace.t(Trace.ISO_START, "copying " +messages.length+ " messages into child.");

	initialMessages = new IsolateMessage[messages.length];
	for (int i = 0; i < messages.length; i++)
	{
	  IsolateMessage im = messages[i];
	  if (im == null)
	  {
	    throw new IllegalArgumentException("Null entry (at " +i+ ") in initial messages array.");
	  }
	  else
	  {
	    try
	    {
	      // Trace.t(Trace.ISO_START, "Copying messages[" +i+ "] " +messages[i]);
	      initialMessages[i] = messages[i].translate(parentTeam, this.teamH);
	    }
	    catch (ClosedLinkException cle)
	    {
	      throw new IsolateStartupException("Cloning message " +i, cle);
	    }
	    catch (IOException ioe)
	    {
	      throw new IsolateStartupException("Cloning message " +i, ioe);
	    }
	  }
	}
      }

      // Synchronize in case multiple parent thread invoke 'start' simultaneous-like
      synchronized(childBE)
      {
	IsolateState actualState = this.statusHandle.getState();

	// XXX is it okay to throw exceptions during the visit?

	if (actualState == IsolateState.INVALID)
	  throw new IllegalStateException("child never finished initialization!");
      
	if (actualState == IsolateState.DEAD)
	  throw new IllegalStateException("starting an isolate that has been terminated (DEAD)");
	
	if (actualState != IsolateState.NEWBORN)
	  throw new IllegalStateException("starting an isolate that is already started (state=" +actualState+ ")");
      
	this.statusHandle.advanceState(IsolateState.STARTING);
	this.statusHandle.setStarter(parentIsolate);
      }

      this.remoteState = IsolateState.STARTING;
	
      // Start the main thread
      childBE.startMainThread(initialMessages);
      this.teamH.returnFrom();

    }
    catch (DeadTeamException dte)
    {
      // XXX remoteIsDead();

      teamH.terminate();
      this.remoteState = IsolateState.DEAD;
      this.teamH = null;
      throw new IllegalStateException("starting an isolate that has been terminated");
    }
    catch (TeamEjectionException tee)
    {
      // XXX remoteIsDead();

      teamH.terminate();
      this.remoteState = IsolateState.DEAD;
      this.teamH = null;
      throw new IllegalStateException("starting an isolate that has been terminated");
    }
  }

  public static IsolateMessage[] currentIsolateIOBindings()
  {
    Isolate.currentIsolate(); // XXX
    
    SecurityManager sm = System.getSecurityManager();
    if( sm != null )
      sm.checkPermission(CONTEXT_PERMISSION);
    
    IsolateBackEnd currentIsoBE = IsolateBackEnd.current();
    Assert.require(currentIsoBE != null);

    IsolateMessage[] rc = currentIsoBE.getIOBindings();

    Assert.ensure(rc != null);
    return (IsolateMessage[])rc.clone();
  }

  /**
   * Retrieve the available IsolateMessage objects for the currently running isolate.
   * @returns the array of zero or more IsolateMessage objects
   */
  public static IsolateMessage[] currentIsolateStartMessages()
  {
    Isolate.currentIsolate(); // XXX
      
    SecurityManager sm = System.getSecurityManager();
    if( sm != null )
      sm.checkPermission(CONTEXT_PERMISSION);
    
    IsolateBackEnd currentIsoBE = IsolateBackEnd.current();
    Assert.require(currentIsoBE != null);

    IsolateMessage[] rc = currentIsoBE.getIsolateMessages();

    Assert.ensure(rc != null);
    return (IsolateMessage[])rc.clone();
  }

  private void doShutdown(boolean doHalt, int status)
    throws SecurityException
  {
    IsolateEvent.ExitReason er;

    this.updateRemoteState();

    // Ignore invocations on a dead Isolate.
    if (this.remoteState == IsolateState.DEAD)
      return;

    if( this.isCurrent() )
    {
      if( doHalt )
	er = IsolateEvent.ExitReason.SELF_HALT;
      else
	er = IsolateEvent.ExitReason.SELF_EXIT;
    }
    else
    {
      if( doHalt )
	er = IsolateEvent.ExitReason.OTHER_HALT;
      else
	er = IsolateEvent.ExitReason.OTHER_EXIT;
    }

    TeamHandle teamH = this.teamH;

    if (teamH != null)
    {
      this.statusHandle.setExitInfo(status, er);
      teamH.terminate(); // XXX this is a bit harsh for an exit (vs. halt)
    }
    else
    {
      Assert.require(this.remoteState == IsolateState.DEAD);
    }
  }

  /**
   * Equivalent to this Isolate calling <tt>Runtime.exit()</tt> or <tt>System.exit</tt>.
   * @todo Rename
   * @param Termination status. By convention, a nonzero status code indicates abnormal termination.
   * @throws SecurityException if security policy does not permit
   * exiting this Isolate.
   * @throws IllegalStateException if this isolate is terminated
   */
  public void exit(int status)
    throws SecurityException, IllegalStateException
  {
    SecurityManager sm = System.getSecurityManager();
    if( sm != null )
      sm.checkPermission(CONTROL_PERMISSION);
    doShutdown(false, status);
  }
  

  public void halt(int status)
    throws SecurityException, IllegalStateException
  {
    SecurityManager sm = System.getSecurityManager();
    if( sm != null )
      sm.checkPermission(CONTROL_PERMISSION);
    doShutdown(true, status);
  }

  private void updateRemoteState()
  {
    Trace.t(Trace.ISO_STATE, "UpdateRemoteState: curr=" +this.remoteState);

    TeamHandle teamH = this.teamH;
    if (teamH == null)
    {
      Assert.require(this.remoteState == IsolateState.DEAD);
    }

    if (this.remoteState == IsolateState.DEAD)
      return;

    this.remoteState = this.statusHandle.getState();

    if (this.remoteState == IsolateState.DEAD)
      this.teamH = null;

    Trace.t(Trace.ISO_STATE, "UpdateRemoteState: new=" +this.remoteState);
  }

  public Link newEventLink()
    throws ClosedLinkException
  {
    return Link.newEventLink(this, Isolate.currentIsolate());
  }
  
  public boolean hasStarted()
  {
    updateRemoteState();
    return (this.remoteState.atLeast(IsolateState.STARTING));
  }

  public boolean hasExited()
  {
    updateRemoteState();
    return (this.remoteState.atLeast(IsolateState.DYING));
  }

  public boolean hasTerminated()
  {
    updateRemoteState();
    return (this.remoteState == IsolateState.DEAD);
  }

  public static Isolate currentIsolate()
  {
    Team curr = Team.current();
    Isolate iso;
    try
    {
      iso = (Isolate)curr.getClassLocal(isolateClass);
    }
    catch (NullPointerException npe)
    {
      // XXX XXX Implies I'm dying (why not a DTE then?)
      return null;
    }
    if (iso == null)
    {
      Team rootTeam = curr;

      synchronized( rootTeam )
      {
	if( (iso = (Isolate)curr.getClassLocal(isolateClass)) == null )
	{
	  Trace.t(Trace.ISO_START, "Creating root isolate in team " +rootTeam);

	  /*
	   * We assume that if the current team doesn't have an
	   * isolate back end, then it must be the root one, as all
	   * subsequent isolates are initialized with a back end.
	   *
	   * Invocation order is important here, otherwise we'll 
	   * re-enter this function while creating the root isolate
	   * infrastructure.
	   */
	  IsolateServer is = IsolateServer.findServer();
	  Long rootGUID = new Long(guidManager.nextGUID());
	  aggregateStateManager.registerNewIsolate(rootGUID, is);
	  rootTeam.setClassLocal(isolateBackEndClass,
				 new IsolateBackEnd(rootGUID, is));
	  iso = new Isolate(rootGUID, rootTeam);
	  rootTeam.setClassLocal(isolateClass, iso);
	  rootTeam.notifyOnFinish();
	  iso.statusHandle.setStarter(iso); // root has self as parent...
	  iso.statusHandle.advanceState(IsolateState.RUNNING); // root is obviously running...
	}
      }
    }
    return iso;
  }

  /*package*/ boolean isCurrent()
  {
    return this.equals(currentIsolate());
  }
  
  public static TransientPreferences currentIsolateContext()
  {
    Isolate.currentIsolate(); // XXX
    
    SecurityManager sm = System.getSecurityManager();
    if( sm != null )
      sm.checkPermission(CONTEXT_PERMISSION);
    
    IsolateBackEnd currentIsoBE = IsolateBackEnd.current();
    Assert.require(currentIsoBE != null);

    TransientPreferences rc = currentIsoBE.getContext();

    Assert.ensure(rc != null);
    return rc;
  }

  /**
   * Returns the hashcode for this Isolate 
   * @return the hashcode
   */
  public int hashCode() {
    return this.guid.hashCode();
  }

  /*package*/ Long getGUID()
  {
    return this.guid;
  }

  /** May return null. */
  /*package*/ TeamHandle getTeamHandle()
  {
    return this.teamH;
  }

  /*package*/ Isolate clone(TeamHandle source, TeamHandle current)
  {
    // source is the source from which the Isolate handle is being
    // copied.

    // current is current context we're running in, its also the
    // destination of the cloned object.

    // this.teamH is the handle on the isolate which 'source' has
    // and 'current' will get a handle to.

    TeamHandle teamH = this.teamH;

    EventTeamHandle copyOfHandle = null;
    if (teamH != null)
      copyOfHandle = new EventTeamHandle(null, teamH);

    // XXX special case this.remoteState == IsolateState.DEAD??

    Isolate clonedIso = new Isolate(cloneLong(this.guid), copyOfHandle);
    copyOfHandle.setEventHandler(clonedIso.handler);

    return clonedIso;
  }

  /*package*/ static String isolateName(Long guid)
  {
    return "Isolate-" + guid.longValue();
  }

  public String toString() {
    return isolateName(this.guid);
  }

  public boolean equals(Object other) {
    if (other instanceof Isolate)
    {
      Isolate oiso = (Isolate)other;
      return oiso.guid.equals(this.guid);
    }
    return false;
  }

  /**
   * Drop the ref count in the global status map when the
   * Isolate handle is collected.
   *
   * If finalizer is invoked on a never-started isolate, then
   * clean up the team.
   */
  protected void finalize() {
    this.statusHandle.revoke();

    // No races, as if its finalized there are no references in this
    // team, and only this team can advance this isolate beyond
    // NEWBORN.
    if (this.remoteState.atLeast(IsolateState.NEWBORN))
      return;

    // We have a handle on an isolate that was never started.  Clean
    // up the refuse.
    TeamHandle teamH = this.teamH;
    if (teamH != null)
    {
      teamH.terminate();
      teamH = null;
      this.remoteState = IsolateState.DEAD; // useless, I know
    }
  }
}

  
