/*
 * IsolateBackEnd.java
 *
 * Copyright (c) 2002,2003 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Lesser 
 * General Public License.
 * 
 * See the file "license-lesser.terms" for information on 
 * usage and redistribution of this file.
 */

package java.lang.isolate;

import edu.utah.janosvm.sys.*; // XXX
import edu.utah.janosvm.sys.Exportable;
import edu.utah.janosvm.resources.ClassPathSpec;
import edu.utah.janosvm.resources.NameSpaceResource;
import edu.utah.janosvm.resources.ClassGroupNotFoundException;
import edu.utah.janosvm.resources.StdioSpec;
import java.io.IODupFactory;
import java.io.IOException;
import java.util.List;
import java.util.Iterator;
import java.util.Properties;
import java.util.LinkedList;
import java.util.prefs.TransientPreferences;


/**
 * Unique object (one per Isolate) that stores the state
 * for that isolate (initial io bindings, context, etc.)
 *
 * <p>This object is allocated on the target Isolate's heap.
 *
 * <p>Normally this should be an "Exportable" and the Isolate class
 * should be an "Importable".  However, since the IsolateBackEnd
 * is one-to-one with each Team, I just look up the IsolateBackEnd
 * via Team.getClassLocal().  I just need to make sure I manipulate
 * the IsolateBackEnd in the correct context...
 */
/*package*/ class IsolateBackEnd
	implements TeamEventHandler
{
  private static final Class isolateBackEndClass = IsolateBackEnd.class;
  private static final Class isolateClass = Isolate.class;

  /**
   * Initial IO messages
   */
  private final IsolateMessage[] ioMessages;

  /**
   * Initial messages passed in start method.
   */
  private IsolateMessage[] initialMessages;

  /**
   * Preferences context object
   */
  private final TransientPreferences context;

  /**
   * Handle on the main thread for this Isolate.  Between INVALID and NEWBORN
   * states, this thread is dormant, and just holds the initialization
   * state.  This thread is started when the Isolate moves into NEWBORN state.
   * Then, this reference is nulled, so the thread can be collected if it
   * exits.
   */
  private MainThread mainThread;

  /** XXX still need this?? */
  private final EventTeamHandle eventHandleKeptLive;

  /** Cached handle on our "kernel" team. */
  private final IsolateServer is;

  /** Handle on the aggregate-level state for this team. */
  private final AggregateStateManager.IsolateStatusHandle statusHandle;

  /** Get the current IsolateBackEnd. */
  static IsolateBackEnd current()
  {
    // XXX race condition here: team could be killed (no back end, means getclasslocal throws npe)

    Object ibe = Team.getCurrentClassLocal(isolateBackEndClass);
    if (ibe == null)
      throw new InternalError("No IsolateBackEnd associated with current team: " + Team.current());
    return (IsolateBackEnd)ibe;
  }

  private static IsolateMessage[] rootIOMessages()
    throws IOException
  {
    StdioSpec ioSpec = StdioSpec.currentFD();

    Trace.t(Trace.ISO_START, "Root ioSpec=" +ioSpec+ " (" +StdioSpec.current() + ")");

    final int skipLen = 5; // "fd://".length();
    int infd = -1, outfd = -1, errfd = -1;

    if( ioSpec.getInput().startsWith("fd://") )
    {
      infd = Integer.parseInt(ioSpec.getInput().substring(skipLen));
    }
    if( ioSpec.getOutput().startsWith("fd://") )
    {
      outfd = Integer.parseInt(ioSpec.getOutput().substring(skipLen));
    }
    if( ioSpec.getError().startsWith("fd://") )
    {
      errfd = Integer.parseInt(ioSpec.getError().substring(skipLen));
    }

    IODupFactory dupFactory = IODupFactory.getFactory();

    return new IsolateMessage[] {
      infd >= 0 ? IsolateMessage.newFileInputStreamMessage(dupFactory.makeFileInputStream(infd)) : null,
	outfd >= 0 ? IsolateMessage.newFileOutputStreamMessage(dupFactory.makeFileOutputStream(outfd)) : null,
	errfd >= 0 ? IsolateMessage.newFileOutputStreamMessage(dupFactory.makeFileOutputStream(errfd)) : null,
	};
  }

  private static final TransientPreferences defaultRootContext()
  {
    TransientPreferences tp = new TransientPreferences();

    // XXX share this stuff better with code in Isolate()
    tp.putBoolean("edu.utah.janosvm", true);
    tp.putBoolean("java.class.allow-caching", false);
    tp.putBoolean("java.lang.isolate.jni-isolation", false);
    tp.putBoolean("java.lang.isolate.vm-isolation", false);

    return tp;
  }

  /**
   * For constructing the root IsolateBackEnd...
   */
  IsolateBackEnd(Long guid, IsolateServer is)
  {
    Assert.require(guid != null);
    Assert.require(is != null);

    try
    {
      this.ioMessages = rootIOMessages();
    }
    catch (IOException ioe)
    {
      throw new InternalError("Failure establishing root IOMessages", ioe);
    }

    this.initialMessages = new IsolateMessage[0];
    this.mainThread = null; 
    this.eventHandleKeptLive = new EventTeamHandle(this, Team.current()); // XXX
    this.context = defaultRootContext();
    this.is = is;
    this.statusHandle = Isolate.aggregateStateManager.newHandleOn(guid, this.is);

    Assert.ensure(this.is != null);
    Assert.ensure(this.statusHandle != null);
    Assert.ensure(this.initialMessages != null);
  }

  IsolateBackEnd(Team team,
		 Long guid,
		 String mainClass,
		 String[] mainArgs,
		 TransientPreferences ctx,
		 IsolateMessage stdin,
		 IsolateMessage stdout,
		 IsolateMessage stderr)
  {
    Assert.require(team.equals(Team.current()));
    Assert.require(mainClass != null);
    Assert.require(mainArgs != null);
    Assert.require(ctx != null);
    Assert.require(team.owns(mainClass));
    Assert.require(team.owns(mainArgs));
    Assert.require(team.owns(ctx));

    this.initialMessages = null;
    this.ioMessages = new IsolateMessage[] { stdin, stdout, stderr };
    this.is = IsolateServer.findServer();
    this.statusHandle = Isolate.aggregateStateManager.newHandleOn(guid, this.is);
    this.context = ctx;

    // NOTE: System properties are initialized from context in MainThread

    this.mainThread = new MainThread(mainClass, mainArgs);

    // Install a handler that will terminate this Team when the last
    // non-daemon thread in the Team exits.
    this.eventHandleKeptLive = new EventTeamHandle(this, team);
    // XXX
    // this.eventHandleKeptLive = null;

    Assert.ensure(this.is != null);
  }

  /** Event handler for JanosVM events.  This is mostly to get the FINISHED
   * event (which is sent when there are no more non-daemon threads in
   * the team).
   */
  public void event(EventTeamHandle eth, TeamReservation tr, TeamEvent ev)
  {
    /* Handler is executed in kernel context. */
    Trace.t(Trace.JANOS_EVENTS, "IsolateBackEnd got event" + ev);
    try
    {
      switch(ev.getEvent())
      {
      case TeamEvent.FINISHED:
	tr.visit(); // switch to owner of "this", so thread is running in right context
	this.statusHandle.advanceState(IsolateState.DYING);
	// Terminate team when no non-daemon threads are left
	Team.current().terminate();
	tr.leave();
	break;
      case TeamEvent.TERMINATED:
	// XXX is this really going to work here?  Isn't this team dead?
	//System.out.println("Marking my status dead.");
	this.statusHandle.advanceState(IsolateState.DEAD);
	this.statusHandle.revoke();
	Isolate.aggregateStateManager.isolateDied(this);
	break;
      default:
	// ignored
      }
    }
    catch (TeamEjectionException tee)
    {
      // no matter
    }
  }

  /*package*/ void startMainThread(IsolateMessage[] initialMessages)
  {
    Assert.require(this == current());
    Assert.require(this.statusHandle.getState() == IsolateState.STARTING);
    Assert.require(this.mainThread != null);
    Assert.require(initialMessages != null);
    Assert.require(Team.current().owns(initialMessages));

    this.initialMessages = initialMessages;

    Team.current().sendEvent(IsolateEvent.Type.STARTING_JANOSVM);

    this.mainThread.start();
  }

  /** Invoked by the main thread when it actually starts running. */
  void mainThreadRunning()
  {
    this.statusHandle.advanceState(IsolateState.RUNNING);
  }

  /** Invoked by the main thread when it exits or blows up. */
  void setExitInfo(int status, IsolateEvent.ExitReason reason)
  {
    this.statusHandle.setExitInfo(status, reason);
  }

  IsolateMessage[] getIsolateMessages()
  {
    Assert.require(this.initialMessages != null);
    Assert.require(Team.current().owns(this.initialMessages));

    IsolateMessage[] rc = this.initialMessages;

    Assert.ensure(rc != null);
    return rc;
  }

  IsolateMessage[] getIOBindings()
  {
    Assert.require(this.initialMessages != null);
    Assert.require(Team.current().owns(this.ioMessages));

    IsolateMessage[] rc = this.ioMessages;

    Assert.ensure(rc != null);
    return rc;
  }

  TransientPreferences getContext()
  {
    Assert.require(this.context != null);
    Assert.require(Team.current().owns(this.context));
    return this.context;
  }

  IsolateServer getIsolateServer()
  {
    Assert.require(this.is != null);
    return this.is;
  }
}
