/*
 * IsolateEvent.java
 *
 * Copyright (c) 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Lesser 
 * General Public License.
 * 
 * See the file "license-lesser.terms" for information on 
 * usage and redistribution of this file.
 */

package java.lang.isolate;

import edu.utah.janosvm.sys.TeamEvent;
import edu.utah.janosvm.sys.TeamHandle;

/**
 * XXX Use a static global exiting event, since it has no associated
 * mutable state.
 */
public class IsolateEvent 
{
  private final Type type;
  private final Isolate isolate;
  private final Isolate starter;
  private final ExitReason exitReason;
  private final int exitStatus;

  // Globally shared ExitReason objects
  public static class ExitReason
  {
    public static final ExitReason IMPLICIT_EXIT  = new ExitReason("IMPLICIT_EXIT");
    public static final ExitReason SELF_EXIT  = new ExitReason("SELF_EXIT");
    public static final ExitReason SELF_HALT = new ExitReason("SELF_HALT");
    public static final ExitReason OTHER_EXIT = new ExitReason("OTHER_EXIT");
    public static final ExitReason OTHER_HALT = new ExitReason("OTHER_HALT");
    public static final ExitReason UNCAUGHT_EXCEPTION = new ExitReason("UNCAUGHT_EXCEPTION");

    public String toString()
    {
      return "IsolateEvent.ExitReason." +this.type;
    }

    private final String type;

    private ExitReason(String s) {
      this.type = s;
    }
  }

  // Globally shared Type objects
  public static class Type
  {
    /*package*/ static final int STARTING_JANOSVM = TeamEvent.USER1;
    /*package*/ static final int EXITING_JANOSVM = TeamEvent.USER2;

    public static final Type STARTING = new Type("STARTING");
    public static final Type EXITING = new Type("EXITING");
    public static final Type TERMINATED = new Type("TERMINATED");

    public String toString()
    {
      return "IsolateEvent.Type." +this.s;
    }

    private final String s;

    private Type(String s)
    {
      this.s = s;
    }
  }

  /*package*/ IsolateEvent(Isolate isolate, Type type, Isolate starter, ExitReason exitReason, int exitStatus) 
  { 
    if (isolate == null)
      throw new NullPointerException("Isolate parameter must not be null.");

    if (type == null)
      throw new NullPointerException("IsolateEvent.Type parameter must not be null.");

    // XXX check type and other parameters

    this.isolate = isolate;
    this.type = type;
    this.starter = starter;
    this.exitReason = exitReason;
    this.exitStatus = exitStatus;
  }

  public Type getType() { 
    return this.type; 
  }

  public ExitReason getExitReason()
  {
    if (this.type != Type.TERMINATED)
      throw new IllegalStateException("not a TERMINATED event");

    return this.exitReason;
  }

  public int getExitStatus() { 
    if (this.type != Type.TERMINATED)
      throw new IllegalStateException("not a TERMINATED event");
    return this.exitStatus;
  }

  public Isolate getIsolate()
  {
    return this.isolate;
  }

  public Isolate getStarter()
  {
    if (this.type != Type.STARTING)
      throw new IllegalStateException("not a STARTING event");

    return this.starter;
  }

  public String toString()
  {
    if (this.type == Type.TERMINATED)
      return "IsolateEvent[TERMINATED; reason=" +this.getExitReason()
	+ "; status=" +this.getExitStatus()+ "]";

    if (this.type == Type.EXITING)
      return "IsolateEvent[EXITING]";

    if (this.type == Type.STARTING)
      return "IsolateEvent[STARTING; starter=" +this.getStarter()+ "]";

    return "IsolateEvent[<illegal>]";
  }

  /*package*/ IsolateEvent clone(TeamHandle sourceTeam, TeamHandle currentTeam)
  {
    return new IsolateEvent(this.isolate.clone(sourceTeam, currentTeam),
			    this.type, 
			    this.starter.clone(sourceTeam, currentTeam),
			    this.exitReason, this.exitStatus);
  }
}
