/*
 * IsolateMessageDispatcher.java
 *
 * Copyright (c) 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Lesser 
 * General Public License.
 * 
 * See the file "license-lesser.terms" for information on 
 * usage and redistribution of this file.
 */

package java.lang.isolate;

import java.util.HashMap;
import java.util.Vector;
import java.util.List;
import java.util.Enumeration;

/**
 * XXX Horrible polling implementation.
 */
public final class IsolateMessageDispatcher
  implements Runnable
{
  static public interface Listener
  {
    void messageReceived(IsolateMessageDispatcher dispatcher, Link link, IsolateMessage message);

    void receiveFailed(IsolateMessageDispatcher dispatcher, Link link, Throwable throwable);
  }

  // Map Each Link to a vector of Listeners
  private final HashMap linkToListenersMap = new HashMap();
  private boolean started = false;
  private boolean done = false;

  // A lame optimization for backing off the polling when nothing's happening...
  private boolean didDispatch;

  public IsolateMessageDispatcher()
  {
  }

  public void add(Link link, Listener listener)
  {
    Trace.t(Trace.IMD_MAINT, "imd.add(" +link+ ", " +listener+ ")");

    synchronized(this.linkToListenersMap)
    {
      // Don't do anything if the dispatcher is toast (the map was cleared)
      if (this.done)
	return;

      Vector listenerList;
      listenerList = (Vector)(linkToListenersMap.get(link));
      if (listenerList == null)
      {
	listenerList = new Vector();
	Trace.t(Trace.IMD_MAINT, "  new link ->" +listenerList);
	listenerList.add(listener);
	linkToListenersMap.put(link, listenerList);
      }
      else
      {
	Trace.t(Trace.IMD_MAINT, "  prior listeners ->" +listenerList);
	listenerList.add(listener);
      }
    }
  }

  public void remove(Link link)
  {
    if (link == null)
      throw new NullPointerException("Link parameter to remove must not be null");

    synchronized(this.linkToListenersMap)
    {
      if (this.done)
	return;

      linkToListenersMap.remove(link);
    }
  }

  /** Invoke the given listener's <code>receiveFailed</code>. */
  private void dispatchFailure(Object[] listeners, Link link, Throwable th)
  {
    Trace.t(Trace.IMD_RUN, "  Failure " +th+ " on link " +link);

    // remove the link from this dispatcher before invoking failure callbacks
    this.remove(link);

    for (int i = 0; i < listeners.length; i++)
      ((Listener)listeners[i]).receiveFailed(this, link, th);

    this.didDispatch = true;
  }

  /** Invoke the given listener's <code>messageReceived</code>. */
  private void dispatchMessage(Object[] listeners, Link link, IsolateMessage im)
  {
    Trace.t(Trace.IMD_RUN, "  Got message " +im+ " on link " +link);

    for (int i = 0; i < listeners.length; i++)
      ((Listener)listeners[i]).messageReceived(this, link, im);

    this.didDispatch = true;
  }

  /** 
   * Do a pass through all the registered links, handling any that
   * have something to do.  Return true if the dispatcher should
   * shutdown, false if it should continue plugging away.
   */
  private boolean processMessages()
  {
    Object[] links = null;

    // Copy the link objects to allow concurrent modifications to
    // the listener map
    synchronized(this.linkToListenersMap)
    {
      if (this.done)
	return true;

      links = linkToListenersMap.keySet().toArray();
    }
    
    if ((links == null) || (links.length == 0))
    {
      Trace.t(Trace.IMD_RUN, "  Have no links to check.");
      sleep(1000);
      return false;
    }

    Trace.t(Trace.IMD_RUN, "  Have " +links.length+ " links to check.");

    this.didDispatch = false;
    for (int i = 0; i < links.length; i++)
    {
      Link link = (Link)links[i];
      Object[] listeners = null;
      
      synchronized(this.linkToListenersMap)
      {
	// Only check for bailing between messages, not between listener invocations.
	if (this.done)
	  return true;

	Vector listenersv = (Vector)linkToListenersMap.get(link);
	if ((listenersv != null) && (! listenersv.isEmpty()))
	{
	  listeners = listenersv.toArray();
	}
      }
      
      // May have been removed
      if ((listeners == null) || (listeners.length == 0))
      {
	Trace.t(Trace.IMD_RUN, "  No listeners installed with link " +link);
	continue;
      }

      // Have a valid link and a set of listeners, try it.
      try
      {
	Trace.t(Trace.IMD_RUN, "  Trying link " +link);
	IsolateMessage lm = link.tryReceive();
	if (lm != null)
	{
	  dispatchMessage(listeners, link, lm);
	}
	else
	{
	  Trace.t(Trace.IMD_RUN, "  Nuthin happening on link " +link);
	}
      }
      catch (ClosedLinkException cle)
      {
	dispatchFailure(listeners, link, cle);
      }
      catch (IllegalStateException ise)
      {
	dispatchFailure(listeners, link, ise);
      }
      catch (UnsupportedOperationException uoe)
      {
	dispatchFailure(listeners, link, uoe);
      }
      catch (IsolateResourceError ire)
      {
	dispatchFailure(listeners, link, ire);
      }
      catch (SecurityException se)
      {
	dispatchFailure(listeners, link, se);
      }
    }

    if (! this.didDispatch)
      sleep(500);

    return false;
  }

  /** Weak sleep wrapper that doesn't throw exceptions. */
  private static void sleep(int millis)
  {
    try
    {
      Thread.sleep(millis);
    }
    catch (InterruptedException _)
    {
    }
  }

  public void run()
  {
    synchronized(this.linkToListenersMap)
    {
      if (this.started)
	throw new IllegalStateException("IsolateMessageDispatcher may only be started once.");
      this.started = true;
    }

    try
    {
      boolean done = false;
      
      while(! done)
	done = processMessages();
    }
    finally
    {
      // Drop all the tracking info on the floor.
      synchronized(this.linkToListenersMap)
      {
	this.linkToListenersMap.clear();
      }
    }
  }

  public void shutdown()
  {
    synchronized(this.linkToListenersMap)
    {
      this.done = true;
    }
  }
    
}
