/*
 * IsolateMessage_Serializable.java
 *
 * Copyright (c) 2002, 2003 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Lesser 
 * General Public License.
 * 
 * See the file "license-lesser.terms" for information on 
 * usage and redistribution of this file.
 */

package java.lang.isolate;

import java.io.InvalidClassException;
import java.io.StreamCorruptedException;
import java.io.OptionalDataException;
import java.io.NotSerializableException;
import java.io.ByteArrayOutputStream;
import java.io.ByteArrayInputStream;
import java.io.ObjectOutputStream;
import java.io.ObjectInputStream;

import edu.utah.janosvm.sys.TeamHandle;
import edu.utah.janosvm.sys.Team;

class IsolateMessage_Serializable
  extends IsolateMessage_DataMessage
{
  private final Object lock = new Object();
  private byte[] serForm;  // only used for received serializables
  private Object objCache;

  /*package*/ IsolateMessage_Serializable(Object o)
  {
    if (o == null)
      throw new NullPointerException("Link Message cannot contain <null>");
    
    this.objCache = o;
    this.serForm = null;
  }

  /** Used for creating a IsolateMessage_Serializable from a byte-array containing a serialized object. */
  /*package*/ IsolateMessage_Serializable(byte[] b)
  {
    this.serForm = b;
    this.objCache = null;
  }
  
  /*package*/ void encodeObject()
    throws LinkSerializationException
  {
    synchronized(this.lock)
    {
      // If its a recieved serializable that's never been deserialized, cannot reserialize
      if (this.objCache != null)
	this.serForm = encodeObject(this.objCache);
    }
    
    // Must always have a non-null serForm (even if encoding didn't just happen).
    Assert.ensure(this.serForm != null);
  }

  private static byte[] encodeObject(Object o)
    throws LinkSerializationException
  {
    ByteArrayOutputStream bs = null;
    try
    {
      bs = new ByteArrayOutputStream();
      ObjectOutputStream oos = new ObjectOutputStream(bs);
      oos.writeObject(o);
      oos.flush();
      oos.close();
      return bs.toByteArray();
    }
    catch (NotSerializableException nse)
    {
      throw new LinkSerializationException("during encoding", nse);
    }
    catch (InvalidClassException ice)
    {
      throw new LinkSerializationException("during encoding", ice);
    }
    catch (java.io.IOException ioe)
    {
      // XXX use chaining
      throw new Error("WTF:" +ioe);// XXX
    }
  }

  /*package*/ static Object decodeObject(byte[] b)
    throws LinkSerializationException
  {
    try
    {
      ObjectInputStream ois = new ObjectInputStream(new ByteArrayInputStream(b));
      return ois.readObject();
    }
    catch (IllegalStateException _)    { throw new LinkSerializationException("during readObject", _); }
    catch (ClassNotFoundException _)   { throw new LinkSerializationException("during readObject", _); }
    catch (InvalidClassException _)    { throw new LinkSerializationException("during readObject", _); }
    catch (StreamCorruptedException _) { throw new LinkSerializationException("during readObject", _); }
    catch (OptionalDataException _)    { throw new LinkSerializationException("during readObject", _); }
    catch (java.io.IOException ioe)
    {
      throw new Error("WTF: " +ioe);
    }
  }
  
  /**
   * Returns the decode version of the object this serializable
   * contains.  Returns cached pointer if one is available.
   */
  private Object decodeObject()
    throws LinkSerializationException
  {
    synchronized(this.lock)
    {
      if (this.objCache == null)
	this.objCache = decodeObject(this.serForm);
    }

    return this.objCache;
  }

  public void acceptVisitor(IsolateMessageVisitor lmv)
    throws LinkSerializationException
  {
    lmv.visitSerializableMessage(this);
  }

  public Object getSerializable()
    throws LinkSerializationException
  {
    return decodeObject();
  }

  public boolean equals(Object o)
  {
    if (o instanceof IsolateMessage_Serializable)
    {
      IsolateMessage_Serializable other = (IsolateMessage_Serializable)o;
      // XXX
      return this.objCache == other.objCache;
    }
    return false;
  }

  void pretranslate()
    throws LinkSerializationException
  {
    this.encodeObject();
  }

  /** 
   * INTERNAL translation of IsolateMessage
   */
  IsolateMessage translate(TeamHandle sourceTeam, TeamHandle currentTeam)
    throws ClosedLinkException
  {
    Assert.require(sourceTeam != null);
    Assert.require(currentTeam != null);
    Assert.require(currentTeam.equals(Team.current()));
    // Assert.require(this.serForm != null, "Null serForm in " +this);

    // XXX synchronize?
    int len = this.serForm.length;
    byte[] rc = new byte[len];
    System.arraycopy(this.serForm, 0, rc, 0, len);
    return new IsolateMessage_Serializable(rc);
  }

  public String toString()
  {
    return "IsolateMessage[Serializable, obj=" +this.objCache+ "; serForm=" +this.serForm+ "]";
  }
}
