/*
 * Link.java
 *
 * Copyright (c) 2002, 2003 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Lesser 
 * General Public License.
 * 
 * See the file "license-lesser.terms" for information on 
 * usage and redistribution of this file.
 */

package java.lang.isolate;

import java.io.IOException;

import edu.utah.janosvm.resources.*; // XXX
import edu.utah.janosvm.sys.*; // XXX
import edu.utah.janosvm.sys.TeamHandle;

public abstract class Link {
  /** A Hack to avoid toString() invocations on non-shared types (like BidiLink_Neither) */
  private final String typeName;

  protected Link(String typeName) 
  {
    this.typeName = typeName;
  }

  public static Link newLink(Isolate sender, Isolate receiver) 
    throws ClosedLinkException
  {
    // Name is null to indicate we're creating a whole new link
    return OneWayLink.newInstance(sender, receiver, null);
  }

/*
  public static Link newBidirectionalLink(Isolate sender, Isolate receiver) 
    throws ClosedLinkException
  {
    // Name is null to indicate we're creating a whole new link
    return BidirectionalLink.newInstance(sender, receiver, null);
  }
*/

  public static Link newEventLink(Isolate sender, Isolate receiver) 
    throws ClosedLinkException
  {
    // Name is null to indicate we're creating a whole new link
    return EventLink.newInstance(sender, receiver, null);
  }

  public abstract boolean isSender(Isolate i);
    
  public abstract boolean isReceiver(Isolate i);

  public abstract void close() 
    throws IllegalStateException;

  public abstract boolean isOpen()
    throws IllegalStateException;

  /**
   * Clone this link (from the sourceTeam context) in the (current,
   * destintion) Team context.
   *
   * @param sourceTeam the team the link is coming from
   * @param currentTeam the team the link is going to (also the current team context).
   * @return the cloned link object
   */
  /*package*/ abstract Link clone(TeamHandle sourceTeam, TeamHandle currentTeam)
    throws ClosedLinkException;

  // Send

  /*package*/ abstract boolean send(IsolateMessage x, long millis, int nanos)
    throws ClosedLinkException, IOException;

  public void send(IsolateMessage x)
    throws ClosedLinkException, IOException
  {
    if( x == null )
    {
      throw new NullPointerException("Null message argument");
    }
    boolean rc = send(x, -1, 0);
    Assert.ensure(rc == true);
  }

  /**
   * Non-blocking send, only succeeds if a receiver is waiting.
   */
  /*package*/ boolean trySend(IsolateMessage x)
    throws ClosedLinkException, IOException
  {
    return send(x, 0, 0);
  }

  // Receive 

  /*package*/ abstract IsolateMessage receive(long millis, int nanos)
    throws ClosedLinkException;

  /**
   * Blocking receive.  Blocks until a recieve is completed (or 
   * an error or close occurs).
   */
  public IsolateMessage receive()
    throws ClosedLinkException
  {
    IsolateMessage rc;
    rc = receive(-1, 0);
    Assert.ensure(rc != null);
    return rc;
  }

  /*package*/ IsolateMessage tryReceive()
    throws ClosedLinkException
  {
    return receive(0, 0);
  }


  public String toString()
  {
    return this.typeName + "Link[]";
  }
}

