/*
 * OneWayLink.java
 *
 * Copyright (c) 2002, 2003 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Lesser 
 * General Public License.
 * 
 * See the file "license-lesser.terms" for information on 
 * usage and redistribution of this file.
 */

package java.lang.isolate;

import java.io.IOException;

import edu.utah.janosvm.sys.TeamHandle;
import edu.utah.janosvm.sys.Team;

/**
 * A unidirectional connection from one Isolate to another.
 *
 * <p>JanosVM Implementation Notes: OneWayLink class just contains the
 * sender/reciever, and global name information.  Subclasses contain
 * the details.
 *
 * @since 1.5
 **/
/*package*/ abstract class OneWayLink 
  extends Link 
{ 
  private static final IsolateGUID guidManager = new IsolateGUID(0);

  private final Isolate sender;
  private final Isolate receiver;
  private final String name;

  /**
   * Construct a new Link with the given name.  If the name is
   * null a unique one will be generated.  If the name is non-null
   * it must be the name of an associated link.
   *
   * XXX If the subclasses were inner, I could make this method "private".
   */
  /*package*/ OneWayLink(Isolate sender, Isolate receiver, String name)
    throws ClosedLinkException
  {
    super("OneWay");
    Assert.require(Team.current().owns(sender));
    Assert.require(Team.current().owns(receiver));
    Assert.require(Team.current().owns(name));
    
    this.sender = sender; 
    this.receiver = receiver;
    //this.amSender = (sender.equals(Isolate.currentIsolate()));
    //this.amReceiver = (receiver.equals(Isolate.currentIsolate()));
    if (name == null)
      this.name = "Global_OneWayLink#" + guidManager.nextGUID();
    else
      this.name = name;

    Assert.ensure(this.name != null);
  }

  /**
   * Create a link with the given sender/receiver and name.
   */
  /*package*/ static OneWayLink newInstance(Isolate sender, Isolate receiver, String name) 
    throws ClosedLinkException
  {
    if (sender == null)
      throw new NullPointerException("sender is null");

    if (receiver == null)
      throw new NullPointerException("sender is null");

    if (sender.hasTerminated())
      throw new ClosedLinkException("Sender is terminated.");

    if (receiver.hasTerminated())
      throw new ClosedLinkException("Sender is terminated.");

    Isolate current = Isolate.currentIsolate();
    boolean amSender = sender.equals(current);
    boolean amReceiver = receiver.equals(current);

    if (amSender && amReceiver)
      return new OneWayLink_Both(sender, receiver, name);

    if (amSender)
      return new OneWayLink_Sender(sender, receiver, name);

    if (amReceiver)
      return new OneWayLink_Receiver(sender, receiver, name);

    return new OneWayLink_Neither(sender, receiver, name);
  }

  /*package*/ String name()
  {
    Assert.require(this.name != null);
    return this.name;
  }

  // Bidi* needs access to sender
  /*package*/ Isolate sender()
  {
    Assert.require(this.sender != null);
    return this.sender;
  }
  
  // Bidi* needs access to receiver
  /*package*/ Isolate receiver()
  {
    Assert.require(this.receiver != null);
    return this.receiver;
  }

  public boolean isSender(Isolate i) 
  { 
    return this.sender.equals(i);  
  }

  public boolean isReceiver(Isolate i) 
  { 
    return this.receiver.equals(i); 
  }

  public boolean equals(Object x)
  {
    if (x instanceof OneWayLink)
    {
      OneWayLink other = (OneWayLink)x;
      return other.name.equals(this.name);
    }
    return false;
  }

  public String toString()
  {
    Isolate current = Isolate.currentIsolate();
    boolean amSender = this.sender.equals(current);
    boolean amReceiver = this.receiver.equals(current);

    return "Link["
//      + this.getClass().getName()
//      + ": "
      + (amSender?"*":"")
      + sender
      + ":"
      + (amReceiver?"*":"")
      + receiver
      + "; n=" + this.name
      + "]";
  }

  public abstract boolean isOpen() 
    throws IllegalStateException;

  public abstract void close() 
    throws IllegalStateException;

  public abstract boolean send(IsolateMessage lm, long millis, int nanos)
    throws ClosedLinkException, IllegalStateException, IOException;

  public abstract IsolateMessage receive(long millis, int nanos)
    throws ClosedLinkException, IllegalStateException;

  /*package*/ Link clone(TeamHandle sourceTeam, TeamHandle currentTeam)
    throws ClosedLinkException
  {
    // XXX assert(!sourceTeam.equals(currentTeam))

    // 'currentTeam' is the destination Team for the new Link.

    Assert.require(this.name != null);

    return OneWayLink.newInstance(this.sender.clone(sourceTeam, currentTeam),
				  this.receiver.clone(sourceTeam, currentTeam),
				  Isolate.cloneString(this.name));
  }
}
