/*
 * OneWayLink_Both.java
 *
 * Copyright (c) 2002, 2003 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Lesser 
 * General Public License.
 * 
 * See the file "license-lesser.terms" for information on 
 * usage and redistribution of this file.
 */

package java.lang.isolate;

import edu.utah.janosvm.sys.TeamHandle;
import edu.utah.janosvm.sys.Team;
import edu.utah.janosvm.sys.TeamLocal;
import java.util.HashMap;

/**
 * Implementation of a OneWayLink for the case where the same
 * Isolate is both the sender and the receiver.
 *
 * <p>XXX For a sender-heavy or receiver-heavy link, too many folks
 * are getting woken up on operation.
 *
 * <p>XXX there is no way for a sender to transmit an object to a
 * waiting reciever without blocking.  The sender always waits at
 * least once.  Should be relatively easy for the OneWayLink_Both
 * because only a single isolate is involved.
 */
/*package*/ class OneWayLink_Both
      	extends OneWayLink
{
  private static final class Timeout
  {
    long millis;
    int nanos;
    int pollCt;

    Timeout(long millis, int nanos)
    {
      if ((millis == 0) && (nanos < 0))
	millis = -1;
      this.millis = millis;
      this.nanos = nanos;
      this.pollCt = 4;
    }

    /** 
     * Wait on the given sync object with a timeout.  Object
     * must be locked before this is invoked.
     */
    void waitOn(Object syncobj)
      throws InterruptedException
    {
      if (justPoll())  // XXX Don't wait long if just polling...  Maybe just yield?
	syncobj.wait(1, 0);
      else if (this.millis < 0)
	syncobj.wait(); // indefinite wait
      else
	syncobj.wait(this.millis /*, this.nanos*/);
    }

    boolean justPoll()
    {
      return (millis == 0) && (nanos == 0);
    }
  }

  /**
   * The LinkPoint is the common rendezvous point for different Link
   * objects that all point to the same Link.  E.g., a link is
   * constructed within a single isolate, sent to a second isolate
   * which then sends the link back.  The received link object's
   * rendezvous point needs to be equivalent to original rendezvous
   * point.
   */
  private static final class LinkPoint
  {
    private int receiversWaiting;
    private int sendersWaiting;
    
    private IsolateMessage sendersMessage;
    private boolean closed;

    private LinkPoint()
    {
      this.sendersWaiting = 0;
      this.receiversWaiting = 0;
      this.sendersMessage = null;
      this.closed = false;
    }

    public boolean isOpen()
    {
      synchronized (this)
      {
        return !this.closed;
      }
    }
    
    public void close()
    {
      // flag the list as closed, then notify anyone who might be
      // waiting.
      
      synchronized (this)
      {
        // Double-close is silently ignored (and thread-safe).
        if (this.closed)
          return;
        this.closed = true;
        this.notifyAll();
      }
    }
    
    
    public boolean send(IsolateMessage lm, Timeout timeout)
      throws ClosedLinkException, LinkSerializationException
    {
      Assert.require(lm != null);
      Assert.require(timeout != null);

      while (true)
      {
        synchronized(this)
        {
          if (this.closed)
          {
            Trace.t(Trace.OWLINK_SEND, "  Link is closed.");
            throw new ClosedLinkException("Link is closed");
          }
          
          if (this.receiversWaiting > 0)
          {
            Trace.t(Trace.OWLINK_SEND, "  Found waiting recievers, kicking them.");
            this.notifyAll();
          }
          else
          {
            // No receivers; if just polling then give up.
            if (timeout.justPoll())
            {
              Trace.t(Trace.OWLINK_SEND, "  No recievers, just polling -> Bailing.");
              return false;
            }
          }
          
	  try
	  {
	    if (this.sendersMessage == null)
	    {
	      Trace.t(Trace.OWLINK_SEND, "  Front of send queue, installed my object.");
	      // I am the front of the send queue
	      this.sendersMessage = lm;
	      timeout.waitOn(this);
	      if (this.sendersMessage != lm) // its gone! Woot, success
	      {
		Trace.t(Trace.OWLINK_SEND, "  My object is gone.  Success.");
		return true;
	      }
	      
	      // Bah, i woke up but no one took my message.  Try again..
	      Trace.t(Trace.OWLINK_SEND, "  My object is still here.  Trying again.");
	    }
	    else
	    {
	      // Bah a sender already at the front...

	      Trace.t(Trace.OWLINK_SEND, "  SenderMessage already taken.  Joining " +sendersWaiting+ " waiting senders.");
	      try
	      {
		this.sendersWaiting++;
		timeout.waitOn(this);
	      }
	      finally
	      {
		this.sendersWaiting--;
	      }
	    }
	  }
	  catch(InterruptedException e)
	  {
	    this.close();
	    throw new ClosedLinkException("Sender thread interrupted");
	  }
	} // synchronized(this)
      }
    }
    
    
    public IsolateMessage receive(Timeout timeout)
      throws ClosedLinkException
    {
      Assert.require(timeout != null);

      while (true)
      {
        synchronized (this)
        {
          if (this.closed)
          {
            Trace.t(Trace.OWLINK_RECV, "  Link is closed.");
            throw new ClosedLinkException("Link is closed");
          }
    
          if (this.sendersMessage != null)
          {
            IsolateMessage result = null;
	    
            Trace.t(Trace.OWLINK_RECV, "  Found sender message, grabbing it.  Success.");
            // Great, take the object and reset.
            result = this.sendersMessage;
            this.sendersMessage = null;
            
            // Everyone needs to try again
            this.notifyAll();
            
            // ALL DONE
            return result;
          }
          
          if (this.sendersWaiting > 0)
          {
            Trace.t(Trace.OWLINK_RECV, "  Found waiting senders.  Kicking them.");
            // Wake the senders so one will fill the message box
            this.notifyAll();
          }
          else
          {
            // No senders; if just polling then bail
            if (timeout.justPoll())
            {
              Trace.t(Trace.OWLINK_RECV, "  No waiting senders, just polling -> Bailing.");
              return null;
            }
          }
          
          Trace.t(Trace.OWLINK_RECV, "  Joining " +receiversWaiting+ " waiting receivers.");
	  try
	  {
	    this.receiversWaiting++;
	    timeout.waitOn(this);
	  }
	  catch(InterruptedException e)
	  {
	    this.close();
	    throw new ClosedLinkException("Receiver thread interrupted");
	  }
	  finally
	  {
	    this.receiversWaiting--;
	  }
        }
      }
    }
  }

  /**
   * Per-isolate mapping of link names to link points.  By definition
   * OneWayLink_Both instances are only usable in a single isolate, so
   * the table can be per-isolate.
   */
  private static class LinkMapAccessor
  extends TeamLocal
  {
    LinkMapAccessor()
    {
    }
    
    protected Object initialValue()
    {
      return new HashMap();
    }
  }
  
  private static final LinkMapAccessor linkMapAccessor = new LinkMapAccessor();


  private static LinkPoint lookupLinkPoint(String name)
  {
    // No need to synchronize between get/put because the names are
    // unique, and the first call will create the entry before the
    // link could possibly be duplicated.

    HashMap map = (HashMap)(linkMapAccessor.get());
    LinkPoint p;
    synchronized(map)
    {
      p = (LinkPoint)map.get(name);
    }

    if (p == null)
    {
      p = new LinkPoint();
      synchronized(p)
      {
	map.put(name, p);
      }
    }
    return p;
  }

  private final LinkPoint linkPoint;

  /*package*/ OneWayLink_Both(Isolate sender, Isolate receiver, String name)
    throws ClosedLinkException
  {
    super(sender, receiver, name);

    Assert.require(sender.equals(Isolate.currentIsolate()));
    Assert.require(receiver.equals(Isolate.currentIsolate()));

    /*
     * Find the linkpoint this link should correspond to.  Creates
     * a new linkpoint if this is a new link.
     */
    this.linkPoint = lookupLinkPoint(this.name());
  }
  
  public boolean isOpen()
  {
    return this.linkPoint.isOpen();
  }

  public void close()
  {
    this.linkPoint.close();
  }

  public boolean send(IsolateMessage lm, long millis, int nanos)
    throws ClosedLinkException, LinkSerializationException
  {
    Trace.t(Trace.OWLINK_SEND, "Trying to send '" +lm+ "' (t/o=" +millis+ ")");

    if (lm == null)
      throw new NullPointerException("Cannot send null!");

    return this.linkPoint.send(lm, new Timeout(millis, nanos));
  }

  public IsolateMessage receive(long millis, int nanos)
    throws ClosedLinkException
  {
    Trace.t(Trace.OWLINK_RECV, "Trying to receive (t/o=" +millis+ ")");

    return this.linkPoint.receive(new Timeout(millis, nanos));
  }
}
