/*
 * OneWayLink_Receiver.java
 *
 * Copyright (c) 2002, 2003 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Lesser 
 * General Public License.
 * 
 * See the file "license-lesser.terms" for information on 
 * usage and redistribution of this file.
 */

package java.lang.isolate;

import edu.utah.janosvm.sys.TeamHandle;

/**
 * OneWayLink instance for the receiver end of the link.  All the
 * interesting state of a onewaylink is kept in the
 * OneWayLink_RendezvousPoint which is kept in the sender team.
 */
/*package*/ class OneWayLink_Receiver
      	extends OneWayLink
{
  private static final String exceptionMsg = "Not the sender for this Link";

  // Always valid (never null), but may be a cancelled import, of course.
  private final OneWayLink_RendezvousImport rendezvousImport;

  /*package*/ OneWayLink_Receiver(Isolate sender, Isolate receiver, String nameguess)
    throws ClosedLinkException
  {
    super(sender, receiver, nameguess);
    Assert.require(!sender.equals(Isolate.currentIsolate()));
    Assert.require(receiver.equals(Isolate.currentIsolate()));

    TeamHandle senderTH = sender.getTeamHandle();
    
    // Make sure the rendezvous point exists in the sender:
    OneWayLink_RendezvousPoint.receiverDoesSetup(senderTH, this.name());

    // Import it.
    this.rendezvousImport = OneWayLink_RendezvousImport.newInstance(senderTH, this.name());

    Trace.t(Trace.OWLINK_OP, "Created " +this+ " in receiver.");

    sender.handler.addShutdownHook(new Runnable() { 
	public void run() { OneWayLink_Receiver.this.close(); } });
  }

  public void close()
  {
    Trace.t(Trace.OWLINK_OP, this+ ".close()");
    this.rendezvousImport.close();
  }

  public boolean isOpen()
  {
    Trace.t(Trace.OWLINK_OP, this+ ".isOpen()");

    return this.rendezvousImport.isOpen();
  }

  public boolean send(IsolateMessage lm, long millis, int nanos)
    throws IllegalStateException
  {
    throw new IllegalStateException(exceptionMsg);
  }

  public IsolateMessage receive(long millis, int nanos)
    throws ClosedLinkException, IllegalStateException
  {
    // XXX only support indefinite blocking send/recv.
    Assert.require(millis == -1);
    Assert.require(nanos == 0);

    Trace.t(Trace.OWLINK_RECV, this+ ".receive()");
    try
    {
      IsolateMessage rc;
      rc = this.rendezvousImport.receive();
      Trace.t(Trace.OWLINK_RECV, this+ ".receive() => im=" +rc);
      return rc;
    }
    catch (ClosedLinkException cle)
    {
      Trace.t(Trace.OWLINK_RECV, this+ ".receive() => THREW " +cle);
      throw cle;
    }
    catch (IllegalStateException ise)
    {
      Trace.t(Trace.OWLINK_RECV, this+ ".receive() => THREW " +ise);
      throw ise;
    }
    // Not reached.
  }
}
