/*
 * OneWayLink_RendezvousImport.java
 *
 * Copyright (c) 2003 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Lesser 
 * General Public License.
 * 
 * See the file "license-lesser.terms" for information on 
 * usage and redistribution of this file.
 */

package java.lang.isolate;

import edu.utah.janosvm.sys.Team;
import edu.utah.janosvm.sys.TeamHandle;
import edu.utah.janosvm.sys.Importable;
import edu.utah.janosvm.sys.Exportable;
import edu.utah.janosvm.sys.DeadTeamException;
import edu.utah.janosvm.sys.TeamHandle;
import edu.utah.janosvm.sys.TeamEjectionException;
import edu.utah.janosvm.sys.NoSuchExportException;
import edu.utah.janosvm.sys.TeamReservation;

/**
 * Importable reference to a OneWayLink_RendezvousPoint.  Only created
 * inside reciever teams, the OneWayLink_RendezvousPoint is always
 * created inside sender teams.
 */
/*package*/ class OneWayLink_RendezvousImport
      	extends Importable
{
  /** 
   * Create a new OneWayLink_RendezvousImport that points to the given
   * Team's OneWayLink_RendezvousPoint.
   */
  static OneWayLink_RendezvousImport newInstance(TeamHandle sender, String name)
  {
    OneWayLink_RendezvousImport ri = new OneWayLink_RendezvousImport(sender);

    try
    {
      // Bind the OneWayLink_RendezvousImport to the exportable exported under 'name'
      sender.importObject(ri, name);
    }
    catch (DeadTeamException dte)
    {
      // return the unbound OneWayLink_RendezvousImport
    }
    catch (TeamEjectionException tee)
    {
      // return the unbound OneWayLink_RendezvousImport
    }
    catch (NoSuchExportException nsee)
    {
      throw new InternalError("Expected export '" +name+ "' from team " +sender+ ".");
    }

    return ri;
  }

  /** 
   * Reference to the rendezvous point object in the sender team.  If
   * this is null, then the connection is closed.
   */
  private OneWayLink_RendezvousPoint rendezvousPoint;
  private final TeamHandle sender;

  private OneWayLink_RendezvousImport(TeamHandle sender)
  {
    this.sender = sender;
  }

  /** 
   * Invoked when this importable is bound to a
   * OneWayLink_RendezvousPoint.  Should only happen once. 
   */
  protected void bindTo(Exportable ex)
  {
    Assert.require(this.rendezvousPoint == null);
    Assert.require(ex != null);
    
    // assert synchronized(this)
    this.rendezvousPoint = (OneWayLink_RendezvousPoint)ex;
  }

  /** Invoked when this Importable is being broken. */
  protected void unbindFrom()
  {
    Assert.require(this.rendezvousPoint != null);

    // assert synchronized(this)
    this.rendezvousPoint = null;
  }
  
  void close()
  {
    try
    {
      OneWayLink_RendezvousPoint rp;

      this.sender.switchTo();
      rp = this.rendezvousPoint;
      if (rp != null)
	rp.close();
      this.sender.returnFrom();
    }
    catch (DeadTeamException dte)
    {
      // Someone else will cleanup (via unbindFrom)
    }
    catch (TeamEjectionException tee)
    {
      // Someone else will cleanup (via unbindFrom)
    }
  }

  boolean isOpen()
  {
    boolean rc = false; // assume it closed
    try
    {
      OneWayLink_RendezvousPoint rp;

      this.sender.switchTo();
      rp = this.rendezvousPoint;
      if (rp != null)
	rc = rp.isOpen();
      this.sender.returnFrom();
    }
    catch (DeadTeamException dte)
    {
      // Someone else will cleanup (via unbindFrom)
    }
    catch (TeamEjectionException tee)
    {
      // Someone else will cleanup (via unbindFrom)
    }
    return rc;
  }
  
  // See receive, below
  private Object translateFetchedMessage(Object rc)
    throws ClosedLinkException
  {
    // Translate the message into current team's heap
    if (rc instanceof IsolateMessage)
    {
      IsolateMessage im = (IsolateMessage)rc;
      try
      {
	// XXX 
	rc = im.translate(this.sender, Team.current());
      }
      catch (java.lang.OutOfMemoryError oom)
      {
	throw new IsolateResourceError("Ran out of memory copying IsolateMessage", oom);
      }
      catch (java.io.IOException ioe)
      {
	throw new IsolateResourceError("translating object", ioe);
      }
      Assert.ensure(rc != null);
    }
    else if (rc == OneWayLink_RendezvousPoint.CLOSED_TOKEN)
    {
      rc = new ClosedLinkException();
    }
    else if (rc == OneWayLink_RendezvousPoint.INTERRUPTED_TOKEN)
    {
      rc = new ClosedLinkException("Receiver thread interrupted");
    }
    else 
    {
      throw new InternalError("Dunno what to do with " +rc);
    }

    return rc;
  }

  private static IsolateMessage returnWithMessage(Object rc)
    throws ClosedLinkException
  {
    if (rc instanceof IsolateMessage)
      return (IsolateMessage)rc;
    else if (rc instanceof ClosedLinkException)
      throw (ClosedLinkException)rc;
    else
      throw new InternalError("Now I'm really confused.");
  }

  IsolateMessage receive()
    throws ClosedLinkException
  {
    // assert Team.current() == receiver;

    Object remote_msg = null;
    Object local_msg = null;

    try
    {
      OneWayLink_RendezvousPoint rp;

      /*
       * Switch to sender to do recieve (and wait for senders if necessary).
       * This internal recieve returns either a *remote* IsolateMessage
       * or a *remote* ClosedLinkException object.  So, I need to
       * make sure this importable is valid, then do the receive,
       * decode the result, switch back, and return (or throw)
       * the result.
       */

      TeamReservation senderRes = this.sender.pushReservation();
      senderRes.visit();
      rp = this.rendezvousPoint;
      if (rp != null)
      {
	remote_msg = rp.fetchMessage();
      }
      senderRes.leave();

      // Sender still locked down because remote_msg probably points
      // to a sender object.

      if (remote_msg == null)
	local_msg = new ClosedLinkException();
      else
	local_msg = translateFetchedMessage(remote_msg);
      
      // Let the sender go
      senderRes.pop();
    }
    catch (DeadTeamException dte)
    {
      // Someone else will cleanup (via unbindFrom)
      throw new ClosedLinkException();
    }
    catch (TeamEjectionException tee)
    {
      // Someone else will cleanup (via unbindFrom)
      throw new ClosedLinkException();
    }
    
    // Either return IsolateMessage or throw a ClosedLinkException
    return returnWithMessage(local_msg);
  }

  public String toString()
  {
    return "OneWayLink_RendezvousPoint[]";
  }
}
