/*
 * OneWayLink_RendezvousPoint.java
 *
 * Copyright (c) 2003 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Lesser 
 * General Public License.
 * 
 * See the file "license-lesser.terms" for information on 
 * usage and redistribution of this file.
 */

package java.lang.isolate;

import java.io.IOException;

import edu.utah.janosvm.sys.Team;
import edu.utah.janosvm.sys.TeamHandle;
import edu.utah.janosvm.sys.TeamReservation;
import edu.utah.janosvm.sys.Exportable;
import edu.utah.janosvm.sys.PortAuthority;
import edu.utah.janosvm.sys.ExportManager;
import edu.utah.janosvm.sys.ManagedExportable;
import edu.utah.janosvm.sys.TemporaryExport;
import edu.utah.janosvm.sys.DuplicateExportException;
import edu.utah.janosvm.sys.DeadTeamException;
import edu.utah.janosvm.sys.TeamEjectionException;
import edu.utah.janosvm.sys.NoSuchExportException;

/**
 * A rendezvous point for a one-way link.  Rendezvous points are
 * always allocated by and installed as an export in the sending team.
 */
/*package*/ class OneWayLink_RendezvousPoint
	extends Exportable
	implements TemporaryExport, ManagedExportable
{

  private static OneWayLink_RendezvousPoint rpFromEm(ExportManager em)
  {
    Assert.require(em != null);
    Exportable e = em.getExport();
    if (e == null)
      throw new InternalError("null export in manager");
    if (! (e instanceof OneWayLink_RendezvousPoint))
      throw new InternalError("wrong type of export, expected OneWayLink_RendezvousPoint, got " +e);
    return (OneWayLink_RendezvousPoint)e;
  }
  

  /** 
   * Find existing rendezvous point with the given name, or
   * install a new one if there is no such export.
   */
  static OneWayLink_RendezvousPoint senderDoesSetup(String name)
  {
    OneWayLink_RendezvousPoint rp;
    Team curr = Team.current();

    PortAuthority pa = curr.getPortAuthority();
    Assert.require(pa != null);
    
    // See if we can find an export with the given name
    ExportManager em = pa.findExport(name);
    if (em == null)
    {
      // If not, try to create such an export
      try
      {
	rp = new OneWayLink_RendezvousPoint(name);
	pa.exportObject(name, rp);
      }
      catch (DuplicateExportException dee)
      {
	// Must have been created between last check and create
	em = pa.findExport(name);
	rp = rpFromEm(em);
      }
    }
    else
    {
      // Found it, someone already created it
      rp = rpFromEm(em);
    }
    
    Assert.ensure(rp != null);
    return rp;
  }

  /**
   * Find existing rendezvous point with the given name in the 
   * sender team.  Install a new rendezvous point in the sender
   * if one does not exist.  Object must be imported to be 
   * used by receiver, of course.  Returns false if the sender
   * team is dead.
   *
   * @return true if rendezvouspoint was set up, false if not (i.e.,
   * if link should be closed).
   */
  static boolean receiverDoesSetup(TeamHandle sender, String xname)
  {
    try
    {
      sender.switchTo();
      {
	String name = Isolate.cloneString(xname);
	senderDoesSetup(name); // ignore return value
	// receiver must import the exportable (we just make sure it exists)
      }
      sender.returnFrom();
      return true;
    }
    catch (DeadTeamException dte)
    {
      return false;
    }
    catch (TeamEjectionException tee)
    {
      return false;
    }
  }

  // ---

  private final Object lock = new Object();
  private final String name;
  private boolean closed;
  private Object waitingSender;
  private IsolateMessage waitingMessage;
  private ExportManager.ImportEnumerator ienum;
  private ExportManager em;
  
  private OneWayLink_RendezvousPoint(String name)
  {
    this.name = name;
    this.closed = false;
    this.waitingMessage = null;
    this.em = null;
    this.ienum = null;
  }

  private void notifyAllSenders()
  {
    // assert Team.current() == sender;
    // assert synchronized(this.lock);
    this.lock.notifyAll();
  }
  
  private void notifyAllReceivers()
  {
    // assert Team.current() == sender;
    // assert synchronized(this.lcok);
    this.lock.notifyAll();
  }
  
  void close()
  {
    // assert Team.current() == sender 

    synchronized(this.lock)
    {
      if (this.closed)
	return;

      this.closed = true;

      try
      {
	// Revoke the importables
	Team.current().revokeObject(this.name);

	// Any waiting senders 
      }
      catch(NoSuchExportException e)
      {
      }

      // Do not clear 'waitingMessage' as the sender thread
      // needs to know if the message was taken by a reciever
      // or if it should throw a ClosedLinkException.

      this.lock.notifyAll();
    }
  }
  
  boolean isOpen()
  {
    // assert Team.current() == sender 
    synchronized(this.lock)
    {
      return !this.closed;
    }
  }

  /** 
   * To satisfy the TemporaryExport interface.  (Invoked when there
   * are no more imports of this exportable.) 
   */
  public boolean expired()
  {
    return true;
  }

  /** 
   * To satisfy the TemporaryExport interface.
   */
  public Object getName()
  {
    return this.name;
  }

  /** 
   * To satisfy the ManagedExportable interface.  (Invoked when the
   * exportable being reclaimed.)
   */
  public void release()
  {
    this.close();
  }

  void waitWithCLE()
    throws ClosedLinkException
  {
    Assert.requireLock(this);
    try
    {
      this.lock.wait();
    }
    catch (InterruptedException ie)
    {
      this.close();
      throw new ClosedLinkException("Interrupted: " +ie);
    }
  }

  /**
   *  Either returns cleanly (indicating the message was successfully
   *  sent) or throws an exception.
   */
  void send(IsolateMessage im)
    throws ClosedLinkException, LinkSerializationException
  {
    // assert Team.current() == sender;

    // Prepare message for translation (i.e., serialize unserialized serializables)
    im.pretranslate();

    // A unique token that identifies this invocation of send from
    // other invocations of send.
    Object sendToken = Thread.currentThread();
    
    synchronized(this.lock)
    {
      Assert.ensure(this.waitingSender != sendToken);

      // Loop until our a spot is available
      while (this.waitingSender != null)
      {
	if (this.closed)
	  throw new ClosedLinkException();

	Trace.t(Trace.OWLINK_SEND, this+ ".send(tok, " +im+ ") => prior sender...");
	waitWithCLE();
      }
      
      Trace.t(Trace.OWLINK_SEND, this+ ".send(tok, " +im+ ") => installing...");
      this.waitingSender = sendToken;
      this.waitingMessage = im;
      this.notifyAllReceivers();

      // Loop until our message is picked up
      while (this.waitingSender == sendToken)
      {
	Trace.t(Trace.OWLINK_SEND, this+ ".send(tok, " +im+ ") => waiting for pickup...");

	if (this.closed)
	  throw new ClosedLinkException();

	waitWithCLE();
      }

      Trace.t(Trace.OWLINK_SEND, this+ ".send(tok, " +im+ ") => success!");
      //this.notifyAllSenders();
    }
    // Success!
  }
  
  /** Global constant. */
  static final Object CLOSED_TOKEN = new ClosedLinkException();

  /** Global constant. */
  static final Object INTERRUPTED_TOKEN = new ClosedLinkException();

  /** Returns a sender's IsolateMessage, or a ClosedLinkException, or null. */
  Object fetchMessage()
  {
    Object rc = null;
    // assert Team.current() == sender 

    Trace.t(Trace.OWLINK_RECV, this+ ".fetchMessage()");

    synchronized(this.lock)
    {
      while (rc == null)
      {
	if (this.closed)
	  rc = CLOSED_TOKEN;
	else if (this.waitingSender != null)
	{
	  rc = this.waitingMessage;
	  this.waitingMessage = null;
	  this.waitingSender = null;
	  this.notifyAllSenders();
	}
	else
	{
	  try
	  {
	    Trace.t(Trace.OWLINK_RECV, "  " +this+ ".fetchMessage() -> waiting");
	    this.lock.wait();
	  }
	  catch (InterruptedException ie)
	  {
	    this.close();
	    rc = INTERRUPTED_TOKEN;
	  }
	}
      }
    }

    Trace.t(Trace.OWLINK_RECV, "  " +this+ ".fetchMessage() -> " +rc);
    return rc;
  }

  public String toString()
  {
    return "OneWayLink_RendezvousPoint[" +this.name+ "]";
  }
}
