/*
 * OneWayLink_Sender.java
 *
 * Copyright (c) 2002, 2003 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Lesser 
 * General Public License.
 * 
 * See the file "license-lesser.terms" for information on 
 * usage and redistribution of this file.
 */

package java.lang.isolate;

import java.io.IOException;

import edu.utah.janosvm.sys.TeamHandle;
import edu.utah.janosvm.sys.Team;

/*package*/ class OneWayLink_Sender
      	extends OneWayLink
{
  private static final String exceptionMsg = "Not the receiver for this Link";

  /** Null if the link is closed. */
  private OneWayLink_RendezvousPoint rendezvousPoint;

  /*package*/ OneWayLink_Sender(Isolate sender, Isolate receiver, String guessname)
    throws ClosedLinkException
  {
    super(sender, receiver, guessname);
    Assert.require(!sender.equals(receiver));
    Assert.require(sender.equals(Isolate.currentIsolate()));
    Assert.require(!receiver.equals(Isolate.currentIsolate()));

    // note, guessname may or may not equal this.name()

    this.rendezvousPoint = OneWayLink_RendezvousPoint.senderDoesSetup(this.name());

    Trace.t(Trace.OWLINK_OP, "Created " +this+ " in sender.");

    receiver.handler.addShutdownHook(new Runnable() { 
	public void run() { OneWayLink_Sender.this.close(); } });
  }

  public void close()
  {
    Trace.t(Trace.OWLINK_OP, this+ ".close()");

    OneWayLink_RendezvousPoint rp = this.rendezvousPoint;

    if (rp != null)
    {
	this.rendezvousPoint.close();
	this.rendezvousPoint = null;
    }
  }

  public boolean isOpen()
  {
    Trace.t(Trace.OWLINK_OP, this+ ".isOpen()");

    OneWayLink_RendezvousPoint rp = this.rendezvousPoint;

    if (rp != null)
      return rp.isOpen();
    else
      return false;
  }

  public boolean send(IsolateMessage im, long millis, int nanos)
    throws ClosedLinkException, IllegalStateException, IOException
  {
    // XXX only support indefinite blocking send/recv.
    Assert.require(millis == -1);
    Assert.require(nanos == 0);

    if (im == null)
      throw new NullPointerException("null IsolateMessage");
    
    Trace.t(Trace.OWLINK_SEND, this+ ".send(" +im+ ")");
    try
    {
      OneWayLink_RendezvousPoint rp = this.rendezvousPoint;

      if (rp != null)
	rp.send(im); // blocks until send completes or throws
      else
	throw new ClosedLinkException();
    }
    catch (ClosedLinkException cle)
    {
      Trace.t(Trace.OWLINK_SEND, this+ ".send(" +im+ " => THREW " +cle);
      throw cle;
    }
    catch (IllegalStateException ise)
    {
      Trace.t(Trace.OWLINK_SEND, this+ ".send(" +im+ " => THREW " +ise);
      throw ise;
    }
    catch (LinkSerializationException lse)
    {
      Trace.t(Trace.OWLINK_SEND, this+ ".send(" +im+ " => THREW " +lse);
      throw lse;
    }

    Trace.t(Trace.OWLINK_SEND, this+ ".send(" +im+ ") => (success)");
    return true;
  }

  public IsolateMessage receive(long millis, int nanos)
    throws IllegalStateException
  {
    throw new IllegalStateException(exceptionMsg);
  }

}
