/*
 * Trace.java
 *
 * Copyright (c) 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Lesser 
 * General Public License.
 * 
 * See the file "license-lesser.terms" for information on 
 * usage and redistribution of this file.
 */

package java.lang.isolate;

import java.util.StringTokenizer;
import java.io.PrintStream;

import edu.utah.janosvm.sys.VMInstance;

/**
 * Use by setting property "isolate.trace" to a string of ",| " separated
 * string constants from the table below.
 */
/*package*/ class Trace
{
  private static class Flag
  {
    final String name;
    final long mask;
    
    static final Flag[] flags = new Flag[64];
    static int fi = 0;

    Flag(String name, long mask)
    {
      this.name = name;
      this.mask = mask;

      synchronized(flags)
      {
	flags[fi++] = this;
      }
    }

    static Flag find(String name)
    {
      synchronized(flags)
      {
	for (int i = 0; i < flags.length; i++)
	{
	  if ((flags[i] != null)
	      && (flags[i].name.equalsIgnoreCase(name)))
	    return flags[i];
	}
	throw new Error("No isolate.trace flag with name " +name);
      }
    }
  }
  
  static final Flag NONE        = new Flag("NONE",        0x00000000L);
  static final Flag OWLINK_WAIT = new Flag("OWLINK_WAIT", 0x00000001L);
  static final Flag OWLINK_SEND = new Flag("OWLINK_SEND", 0x00000002L|OWLINK_WAIT.mask);
  static final Flag OWLINK_RECV = new Flag("OWLINK_RECV", 0x00000004L|OWLINK_WAIT.mask);
  static final Flag OWLINK_OP   = new Flag("OWLINK_OP",   0x00000008L);
  static final Flag OWLINK      = new Flag("OWLINK",      OWLINK_OP.mask|OWLINK_SEND.mask|OWLINK_RECV.mask);
  static final Flag ISO_START   = new Flag("ISO_START",   0x00000010L);
  static final Flag ISO_CONTEXT = new Flag("ISO_CONTEXT", 0x00000010L);
  static final Flag JANOS_EVENTS= new Flag("JANOS_EVENTS",0x00000020L);
  static final Flag ISO_EVENTS  = new Flag("ISO_EVENTS",  0x00000040L);
  // IMD == IsolateMessageDispatcher
  static final Flag IMD_RUN     = new Flag("IMD_RUN",     0x00000080L);
  static final Flag IMD_MAINT   = new Flag("IMD_MAINT",   0x00000100L);
  static final Flag AGGSTATE    = new Flag("AGGSTATE",    0x00000200L);
  static final Flag ISO_STARTFD = new Flag("ISO_STARTFD", 0x00000400L);
  static final Flag EVLINK      = new Flag("EVLINK",      0x00000800L);
  static final Flag ISO_STATE   = new Flag("ISO_STATE",   0x00001000L);
  static final Flag AGGSTATEREAD= new Flag("AGGSTATEREAD",0x00002000L);
  static final Flag ALL         = new Flag("ALL",         0xFFFFFFFFL);

  private static long currentTraceMask;
  private static final String traceInitProp = "isolate.trace";
  private static final String traceInitOut = "isolate.trace.sysout";
  private static final Object iolock = traceInitProp;

  static
  {
    String propVal = System.getProperty(traceInitProp, "NONE");
    StringTokenizer propWalk = new StringTokenizer(propVal, ",| ");
    currentTraceMask |= 0;
    while(propWalk.hasMoreTokens())
    {
      String flag = propWalk.nextToken();
      currentTraceMask |= Flag.find(flag).mask;
    }
  }

  static boolean flagSet(Flag f)
  {
    return (f.mask & currentTraceMask) != 0;
  }

  static void write(String s)
  {
    synchronized(iolock)
    {
      VMInstance.println(s);
    }
  }

  static void write(String pfx, String s)
  {
    synchronized(iolock)
    {
      VMInstance.println(pfx+ ": " +s);
    }
  }

  static void write(Throwable t)
  {
    synchronized(iolock)
    {
      VMInstance.println(t);
    }
  }

  static void write(String pfx, Throwable t)
  {
    synchronized(iolock)
    {
      VMInstance.print(pfx+ ": ");
      VMInstance.println(t);
    }
  }

  static void t(Flag f, String s)
  {
    if (flagSet(f))
      write(f.name, s);
  }

  static void t(Flag f, Throwable t)
  {
    if (flagSet(f))
      write(f.name, t);
  }
}
