/*
 * LinkSerializable.java
 *
 * Copyright (c) 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * Permission to use, copy, modify, distribute, and sell this file for
 * any purpose is hereby granted without fee, provided that the above
 * copyright notice is maintained in any derived source code.
 */

import java.lang.isolate.Isolate;
import java.lang.isolate.Link;
import java.lang.isolate.IsolateMessage;
import java.lang.isolate.ClosedLinkException;
import java.io.Serializable;
import java.io.ObjectOutputStream;
import java.io.ObjectInputStream;
import java.io.IOException;

/**
 * Test sending of serializable objects across a link.  Test that
 * serialization occurs at send time (not message-construct time), and
 * test that deserialization occurs once per received object.
 */
public class LinkSerializable
{
  public static class InnerSerializable
  implements java.io.Serializable
  {
    int x = 11;
    
    public String toString()
    {
      return "InnerSerializable[" +x+ "]";
    }
  }
  
  public static class SerializationSideEffect
  implements java.io.Serializable
  {
    private static int wObjCt = 0;
    private static int rObjCt = 0;

    int x = 11;

    public String toString()
    {
      return "SerializationSideEffect[wct=" +wObjCt+ "; rct=" +rObjCt+ "]";
    }
    
    private void writeObject(ObjectOutputStream stream)
      throws IOException
    {
      synchronized(SerializationSideEffect.class)
      {
	wObjCt++;
      }
      stream.writeInt(this.x);
    }

    private void readObject(ObjectInputStream stream)
      throws IOException
    {
      synchronized(SerializationSideEffect.class)
      {
	rObjCt++;
      }
      this.x = stream.readInt();
    }
    
  }
  
  private static void delay(int millis)
  {
    try
    {
      Thread.sleep(millis);
    }
    catch (InterruptedException ie)
    {
    }
  }

  public static void main(String args[])
    throws Throwable
  {
    if( args.length > 0 )
    {
      new InnerSerializable();

      // Child
      Thread.currentThread().setName("LinkSerializable Child Thread");

      Link l = Isolate.currentIsolateStartMessages()[0].getLink();
      while (l.isOpen())
      {
	try
	{
	  IsolateMessage im = l.receive();
	  Object o1 = im.getSerializable();
	  Object o2 = im.getSerializable(); // two deserializations?  I think not.
	  System.out.println("got: " +o1.toString());
	}
	catch (ClosedLinkException cle)
	{
	  // okay
	}
      }
      System.exit(0);
    }
    else
    {
      // Parent
      Thread.currentThread().setName("LinkSerializable Main Thread");

      Isolate iso = new Isolate("LinkSerializable", new String[] { "fooey!" });
      Link l = Link.newLink(Isolate.currentIsolate(), iso);
      iso.start(new IsolateMessage[] { IsolateMessage.newLinkMessage(l) });

      l.send(IsolateMessage.newSerializableMessage(new InnerSerializable()));
      Object obj = new SerializationSideEffect();
      IsolateMessage serMsg1 = IsolateMessage.newSerializableMessage(obj);
      IsolateMessage serMsg2 = IsolateMessage.newSerializableMessage(obj);
      l.send(serMsg1);
      l.send(serMsg1);
      l.send(serMsg1);
      l.send(serMsg2);
      l.send(serMsg2);
      l.send(serMsg2);

      // Cleanup
      l.close();
      while (!iso.hasTerminated())
	delay(100);

      System.out.println("sender obj: " +obj);
      System.exit(11);
    }
  }
}

// Loader: edu.utah.janosvm.kit.loader.Loader

/* Expected Output:
got: InnerSerializable[11]
got: SerializationSideEffect[wct=0; rct=1]
got: SerializationSideEffect[wct=0; rct=2]
got: SerializationSideEffect[wct=0; rct=3]
got: SerializationSideEffect[wct=0; rct=4]
got: SerializationSideEffect[wct=0; rct=5]
got: SerializationSideEffect[wct=0; rct=6]
sender obj: SerializationSideEffect[wct=6; rct=0]
*/
