/*
 * SameIsoLinkCheck.java
 *
 * Copyright (c) 2002, 2003 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * Permission to use, copy, modify, distribute, and sell this file for
 * any purpose is hereby granted without fee, provided that the above
 * copyright notice is maintained in any derived source code.
 */

import java.lang.isolate.Link;
import java.lang.isolate.Isolate;
import java.lang.isolate.IsolateMessage;
import java.lang.isolate.ClosedLinkException;
import java.io.Serializable;

/**
 * XXX todo check that each object is sent/received exactly once.
 */
public class SameIsoLinkCheck
{
  static boolean verbose = false;

  static final Object ioLock = new Object();
  
  private static void openCheck(Link l)
  {
  }

  private static void closedCheck(Link l)
  {
  }
  

  public static void main(String[] args)
    throws InterruptedException, ClosedLinkException
  {
    Isolate me = Isolate.currentIsolate();
    
    Link l = Link.newLink(me,me);

    openCheck(l);

    if( verbose )
      System.err.println("Link is: " +l);

    synchronized( ioLock )
    {
      System.err.println("Trying sender first.");
    }
    Thread t1 = new Sender(l, 100);
    Thread t2 = new Receiver(l);

    openCheck(l);
    t1.start();
    t2.start();

    openCheck(l);
    t1.join();
    t2.join();

    openCheck(l);

    synchronized( ioLock )
    {
      System.err.println("Trying receiver first.");
    }
    t1 = new Receiver(l);
    t2 = new Sender(l, 200);

    openCheck(l);
    t1.start();
    t2.start();

    openCheck(l);
    t1.join();
    t2.join();

    int n = 16;
    synchronized( ioLock )
    {
      System.err.println("Trying " +n+ " senders and " +n+ " receivers.");
    }
    Thread[] st = new Thread[n];
    Thread[] rt = new Thread[n];
    for (int i = 0; i < n; i++)
    {
      st[i] = new Sender(l, 1000 + i*100);
      rt[i] = new Receiver(l);
    }
    for (int i = 0; i < n; i++)
    {
      st[i].start();
      rt[i].start();
    }
    for (int i = 0; i < n; i++)
    {
      st[i].join();
      rt[i].join();
    }

    synchronized( ioLock )
    {
      System.err.println("Send tests complete.");
    }
    
    openCheck(l);
    l.close();
    closedCheck(l);

    synchronized( ioLock )
    {
      System.err.println("Link checks done.");
    }
  }

  public static class Message
  implements Serializable
  {
    private final int i;

    Message(int i)
    {
      this.i = i;
    }

    public String toString()
    {
      return "Message[" +i+ "]";
    }
  }


  public static class Sender
    extends Thread
  {
    private Link link;
    private int i;

    Sender(Link l, int i)
    {
      this.link = l;
      this.i = i;
    }

    public void run()
    {
      if( verbose )
	System.out.println("Sender (" +this+ ") starting.");
      
      try
      {
	this.link.send(IsolateMessage.newSerializableMessage(new Message(this.i++)));
      }
      catch (Exception e)
      {
	System.err.println("Sender except: " +e);
	e.printStackTrace(System.err);
      }

      openCheck(this.link);

      synchronized( ioLock )
      {
	System.err.println("Sender done.");
      }
    }
  }
  
  public static class Receiver
    extends Thread
  {
    private Link link;
    
    Receiver(Link l)
    {
      this.link = l;
    }
    
    public void run()
    {
      Object m = null;

      if( verbose )
	System.out.println("Receiver (" +this+ ") starting.");

      try
      {
	IsolateMessage lm = this.link.receive();
	m = lm.getSerializable();
      }
      catch (Exception e)
      {
	System.err.println("Receiver except: " +e);
	e.printStackTrace(System.err);
      }
      synchronized( ioLock )
      {
	System.err.println("Receiver got: " +m);
      }

      openCheck(this.link);
    }
  }
}

// Loader: edu.utah.janosvm.kit.loader.Loader

// Sort Output

/* Expected Output:
Link checks done.
Receiver got: Message[1000]
Receiver got: Message[100]
Receiver got: Message[1100]
Receiver got: Message[1200]
Receiver got: Message[1300]
Receiver got: Message[1400]
Receiver got: Message[1500]
Receiver got: Message[1600]
Receiver got: Message[1700]
Receiver got: Message[1800]
Receiver got: Message[1900]
Receiver got: Message[2000]
Receiver got: Message[200]
Receiver got: Message[2100]
Receiver got: Message[2200]
Receiver got: Message[2300]
Receiver got: Message[2400]
Receiver got: Message[2500]
Send tests complete.
Sender done.
Sender done.
Sender done.
Sender done.
Sender done.
Sender done.
Sender done.
Sender done.
Sender done.
Sender done.
Sender done.
Sender done.
Sender done.
Sender done.
Sender done.
Sender done.
Sender done.
Sender done.
Trying 16 senders and 16 receivers.
Trying receiver first.
Trying sender first.
*/
